/* ------------------------------------------------------------------------- */
/* tftp.c                                                                    */
/*                                                                           */
/* A simple tftp client for busybox.                                         */
/* Tries to follow RFC1350.                                                  */
/* Only "octet" mode supported.                                              */
/* Optional blocksize negotiation (RFC2347 + RFC2348)                        */
/*                                                                           */
/* Copyright (C) 2001 Magnus Damm <damm@opensource.se>                       */
/*                                                                           */
/* Parts of the code based on:                                               */
/*                                                                           */
/* atftp:  Copyright (C) 2000 Jean-Pierre Lefebvre <helix@step.polymtl.ca>   */
/*                        and Remi Lefebvre <remi@debian.org>                */
/*                                                                           */
/* utftp:  Copyright (C) 1999 Uwe Ohse <uwe@ohse.de>                         */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU          */
/* General Public License for more details.                                  */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   */
/*                                                                           */
/* ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>

#include "busybox.h"

//#define CONFIG_FEATURE_TFTP_DEBUG

#define TFTP_BLOCKSIZE_DEFAULT 512 /* according to RFC 1350, don't change */
#define TFTP_TIMEOUT 50000              /* 50ms, in microseconds */
#define TFTP_MAXTIMEOUT 999000          /* about 1 second, in microseconds */
#define TFTP_NUM_RETRIES 12             /* number of backed-off retries */

/* opcodes we support */

#define TFTP_RRQ   1
#define TFTP_WRQ   2
#define TFTP_DATA  3
#define TFTP_ACK   4
#define TFTP_ERROR 5
#define TFTP_OACK  6

static const char *tftp_bb_error_msg[] = {
	"Undefined error",
	"File not found",
	"Access violation",
	"Disk full or allocation error",
	"Illegal TFTP operation",
	"Unknown transfer ID",
	"File already exists",
	"No such user"
};
static const char *tftp_gs_option_str[] = {
     "grandstream_MODEL",
     /*
     "grandstream_NAT",
     */
     "grandstream_ID",
     "grandstream_REV_BOOT",
     "grandstream_REV_CORE",
     "grandstream_REV_BASE",
     "grandstream_REV_PROG"
};
static const int tftp_gs_option = sizeof(tftp_gs_option_str)/sizeof(char *);

static void pipe_quit(int sig);
static jmp_buf	jmp_env;

const int tftp_cmd_get = 1;
const int tftp_cmd_put = 2;

#ifdef CONFIG_FEATURE_TFTP_BLOCKSIZE

static int tftp_blocksize_check(int blocksize, int bufsize)
{
        /* Check if the blocksize is valid:
	 * RFC2348 says between 8 and 65464,
	 * but our implementation makes it impossible
	 * to use blocksizes smaller than 22 octets.
	 */

        if ((bufsize && (blocksize > bufsize)) ||
	    (blocksize < 8) || (blocksize > 65464)) {
	        bb_error_msg("bad blocksize");
	        return 0;
	}

	return blocksize;
}

static char *tftp_option_get(char *buf, int len, char *option)
{
        int opt_val = 0;
	int opt_found = 0;
	int k;

	while (len > 0) {

	        /* Make sure the options are terminated correctly */

	        for (k = 0; k < len; k++) {
		        if (buf[k] == '\0') {
			        break;
			}
		}

		if (k >= len) {
		        break;
		}

		if (opt_val == 0) {
			if (strcasecmp(buf, option) == 0) {
			        opt_found = 1;
			}
		}
		else {
		        if (opt_found) {
				return buf;
			}
		}

		k++;

		buf += k;
		len -= k;

		opt_val ^= 1;
	}

	return NULL;
}

#endif

static inline int tftp(const int cmd, const struct hostent *host,
	const char *remotefile, int localfd, unsigned short port, int tftp_bufsize)
{
	struct sockaddr_in sa;
	struct sockaddr_in from;
	socklen_t fromlen;

	const int cmd_get = cmd & tftp_cmd_get;
	const int cmd_put = cmd & tftp_cmd_put;

	struct timeval tv;
	fd_set rfds;
	int socketfd;
	int len;
	int send_len;
#ifdef	CONFIG_FEATURE_TFTP_BLOCKSIZE
    int want_option_ack = 0;
#endif
	int finished = 0;
	uint16_t opcode;
	uint16_t block_nr = 1;
	uint16_t recv_blk;
    int retries, waittime;
	char *cp;

	unsigned org_port;

	/* Can't use RESERVE_CONFIG_BUFFER here since the allocation
	 * size varies meaning BUFFERS_GO_ON_STACK would fail */
	/* We must keep the transmit and receive buffers seperate */
	/* In case we rcv a garbage pkt and we need to rexmit the last pkt */
	char *xbuf = xmalloc(tftp_bufsize += 4);
	char *rbuf = xmalloc(tftp_bufsize);

	org_port = port;

	if ((socketfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		bb_perror_msg("socket");
		return EXIT_FAILURE;
	}

	len = sizeof(sa);

	memset(&sa, 0, len);
	bind(socketfd, (struct sockaddr *)&sa, len);

	sa.sin_family = host->h_addrtype;
	sa.sin_port = port;
	memcpy(&sa.sin_addr, (struct in_addr *) host->h_addr,
		   sizeof(sa.sin_addr));

	/* build opcode */
	opcode = TFTP_WRQ;
	if (cmd_get) {
		opcode = TFTP_RRQ;
	}
	cp = xbuf + 2;
	/* add filename and mode */
	/* fill in packet if the filename fits into xbuf */
	len = strlen(remotefile) + 1;
	if (2 + len + sizeof("octet") >= tftp_bufsize) {
		bb_error_msg("remote filename is too long");
		goto ret;
	}
	strcpy(cp, remotefile);
	cp += len;
	/* add "mode" part of the package */
	strcpy(cp, "octet");
	cp += sizeof("octet");

#if defined(CONFIG_FEATURE_TFTP_BLOCKSIZE)
	len = tftp_bufsize - 4;	/* data block size */
	if (len != TFTP_BLOCKSIZE_DEFAULT) {
		/* rfc2348 says that 65464 is a max allowed value */
		if ((&xbuf[tftp_bufsize - 1] - cp) < sizeof("blksize NNNNN")) {
			bb_error_msg("remote filename is too long");
			goto ret;
		}
		/* add "blksize", <nul>, blocksize */
		strcpy(cp, "blksize");
		cp += sizeof("blksize");
		cp += snprintf(cp, 6, "%d", len) + 1;
		want_option_ack = 1;
	}
#endif
    {
        /* SXIE: option extension: 
         * grandstream_MODEL,
         * grandstream_NAT, (not implemented -- sxie )
         * grandstream_ID,
         * grandstream_REV_BOOT
         * grandstream_REV_CORE
         * grandstream_REV_BASE
         * grandstream_REV_PROG
         */ 
        {
            int ii;
            int gs_option_str_len;
            char *gs_option_value;

            for ( ii = 0; ii < tftp_gs_option; ii ++)
            {
                gs_option_str_len = strlen(tftp_gs_option_str[ii]) + 1;
                gs_option_value = getenv(tftp_gs_option_str[ii]);
                if (ii == 0 && gs_option_value == NULL)
                    break; // do nothing if no grandstream_MODEL was set
                memcpy(cp, tftp_gs_option_str[ii], gs_option_str_len);
                cp += gs_option_str_len;

                cp += sprintf(cp, "%s", gs_option_value ? gs_option_value : "unknown") + 1;
            }
        }
    }


	/* First packet is built, so skip packet generation */

	if (setjmp(jmp_env)) goto tftp_cancelled;
	signal(SIGPIPE, pipe_quit);

	goto send_pkt;

	while (1) {
		/* Build ACK or DATA */
		cp = xbuf + 2;
		*((uint16_t*)cp) = htons(block_nr);
		cp += 2;
		block_nr++;
		opcode = TFTP_ACK;
		if (cmd_put) {
			opcode = TFTP_DATA;
			len = bb_full_read(localfd, cp, tftp_bufsize - 4);
			if (len < 0) {
				bb_perror_msg("read error");
				goto ret;
			}
			if (len != (tftp_bufsize - 4)) {
				finished = 1;
			}
			cp += len;
		}
 send_pkt:
		/* Send packet */
		*((uint16_t*)xbuf) = htons(opcode); /* fill in opcode part */
		while (1) {
			send_len = cp - xbuf;
			/* nb: need to preserve send_len value in code below
			 * for potential resend! */

            retries = TFTP_NUM_RETRIES; /* re-initialize */
            waittime = TFTP_TIMEOUT;
 send_again:
#if ENABLE_DEBUG_TFTP
			fprintf(stderr, "sending %u bytes\n", send_len);
			for (cp = xbuf; cp < &xbuf[send_len]; cp++)
				fprintf(stderr, "%02x ", (unsigned char) *cp);
			fprintf(stderr, "\n");
#endif
			if (sendto(socketfd, xbuf, send_len, 0,
					(struct sockaddr *) &sa, sizeof(sa)) < 0) {
				bb_perror_msg("send");
				len = -1;
				break;
			}
			/* Was it final ACK? then exit */
			if (finished && (opcode == TFTP_ACK))
				goto ret;

			/* Receive packet */
 recv_again:
            tv.tv_sec = 0;
            tv.tv_usec = waittime;
			FD_ZERO(&rfds);
			FD_SET(socketfd, &rfds);
			switch (select(socketfd + 1, &rfds, NULL, NULL, &tv)) {
				unsigned from_port;
			case 1:
			    fromlen = sizeof(from);
			    memset(&from, 0, sizeof(from));
				len = recvfrom(socketfd, rbuf, tftp_bufsize, 0,
						(struct sockaddr *) &from, &fromlen);
				if (len < 0) {
					bb_perror_msg("recvfrom");
					goto ret;
				}
				from_port = from.sin_port;
				if (port == org_port) {
					/* Our first query went to port 69
					 * but reply will come from different one.
					 * Remember and use this new port */
					port = from_port;
                    sa.sin_port = from_port;
				}
				if (port != from_port)
					goto recv_again;
				goto recvd_good;
			case 0:
                retries--;

                if (retries == 0) {
                    bb_error_msg("last timeout");
					goto ret;
				}
#if 0
                if (waittime > TFTP_TIMEOUT * 2 )   /* ignore the first couple of timeout with small wait time */
				    bb_error_msg("last timeout" + 5);
#endif

                /* exponential backoff with limit */
                waittime += waittime/2;
                if (waittime > TFTP_MAXTIMEOUT) {
                    waittime = TFTP_MAXTIMEOUT;
                }

				goto send_again; /* resend last sent pkt */
			default:
				bb_perror_msg("select");
				goto ret;
			}
		} /* while we don't see recv packet with correct port# */

		/* Process recv'ed packet */
 recvd_good:
		opcode = ntohs( ((uint16_t*)rbuf)[0] );
		recv_blk = ntohs( ((uint16_t*)rbuf)[1] );

#if ENABLE_DEBUG_TFTP
		fprintf(stderr, "received %d bytes: %04x %04x\n", len, opcode, recv_blk);
#endif

		if (opcode == TFTP_ERROR) {
			static const char *const errcode_str[] = {
				"",
				"file not found",
				"access violation",
				"disk full",
				"illegal TFTP operation",
				"unknown transfer id",
				"file already exists",
				"no such user",
			};
			enum { NUM_ERRCODE = sizeof(errcode_str) / sizeof(errcode_str[0]) };

			const char *msg = "";

			if (rbuf[4] != '\0') {
				msg = &rbuf[4];
				rbuf[tftp_bufsize - 1] = '\0';
			} else if (recv_blk < NUM_ERRCODE) {
				msg = errcode_str[recv_blk];
			}
			bb_error_msg("server error: (%u) %s", recv_blk, msg);
			goto ret;
		}

#if defined(CONFIG_FEATURE_TFTP_BLOCKSIZE)
		if (want_option_ack) {
			want_option_ack = 0;

			if (opcode == TFTP_OACK) {
				/* server seems to support options */
				char *res;

				res = tftp_option_get(&rbuf[2], len - 2, "blksize");
				if (res) {
					int blksize = atoi(res);
					if (!tftp_blocksize_check(blksize, tftp_bufsize - 4)) {
						bb_error_msg("server proposes bad blksize %d, exiting", blksize);
						// FIXME: must also send ERROR 8 to server...
						goto ret;
					}
#if ENABLE_DEBUG_TFTP
					fprintf(stderr, "using blksize %u\n",
							blksize);
#endif
					tftp_bufsize = blksize + 4;
					block_nr = 0; // TODO: explain why???
					continue;
				}
				/* rfc2347:
				 * "An option not acknowledged by the server
				 *  must be ignored by the client and server
				 *  as if it were never requested." */
			}

			bb_error_msg("blksize is not supported by server"
						" - reverting to 512");
			tftp_bufsize = TFTP_BLOCKSIZE_DEFAULT + 4;
		}
#endif
		/* block_nr is already advanced to next block# we expect
		 * to get / block# we are about to send next time */

		if (cmd_get && (opcode == TFTP_DATA)) {
			if (recv_blk == block_nr) {
				len = bb_full_write(localfd, &rbuf[4], len - 4);
				if (len < 0) {
					bb_perror_msg("write_error");
					goto ret;
				}
				if (len != (tftp_bufsize - 4)) {
					finished = 1;
				}
				continue; /* send ACK */
			}
			if (recv_blk == (block_nr - 1)) {
				/* Server lost our TFTP_ACK.  Resend it */
				block_nr = recv_blk;
			    printf("server lost ack, reduce block_nr to %d\n", block_nr);
				continue;
			} 
		}

		if (cmd_put && (opcode == TFTP_ACK)) {
			/* did server ACK our last DATA pkt? */
			if (recv_blk == (uint16_t) (block_nr - 1)) {
				if (finished)
					goto ret;
				continue; /* send next block */
			}
		}
		/* Awww... recv'd packet is not recognized! */
		goto recv_again;
		/* why recv_again? - rfc1123 says:
		 * "The sender (i.e., the side originating the DATA packets)
		 *  must never resend the current DATA packet on receipt
		 *  of a duplicate ACK".
		 * DATA pkts are resent ONLY on timeout.
		 * Thus "goto send_again" will ba a bad mistake above.
		 * See:
		 * http://en.wikipedia.org/wiki/Sorcerer's_Apprentice_Syndrome
		 */
	}
 tftp_cancelled:
    /* handle SIGPIPE */
    *((uint16_t*)xbuf) = htons(TFTP_ERROR); /* fill in opcode part */
    *((uint16_t*)xbuf + 1) = htons(6); /* error string: file exist */
    {
        const char *err_msg = "Cancelled";
        int err_msg_len = strlen(err_msg);
        char *msg = &xbuf[4];
        strncpy(msg, err_msg, err_msg_len);
        msg[err_msg_len] = 0;

        if (sendto(socketfd, xbuf, 4 + err_msg_len + 1, 0,
                (struct sockaddr *) &sa, sizeof(sa)) < 0) {
            bb_perror_msg("send error");
        }
    }
#ifdef CONFIG_FEATURE_CLEAN_UP
    {
		close(socketfd);
		free(xbuf);
		free(rbuf);
	}
#endif
    return 2;

 ret:
#ifdef CONFIG_FEATURE_CLEAN_UP
    {
		close(socketfd);
		free(xbuf);
		free(rbuf);
	}
#endif
	return finished == 0; /* returns 1 on failure */

}

int tftp_main(int argc, char **argv)
{
	struct hostent *host = NULL;
	const char *localfile = NULL;
	const char *remotefile = NULL;
	int port;
	int cmd = 0;
	int fd = -1;
	int flags = 0;
	int opt;
	int result;
	int blocksize = TFTP_BLOCKSIZE_DEFAULT;

	/* figure out what to pass to getopt */

#ifdef CONFIG_FEATURE_TFTP_BLOCKSIZE
#define BS "b:"
#else
#define BS
#endif

#ifdef CONFIG_FEATURE_TFTP_GET
#define GET "g"
#else
#define GET
#endif

#ifdef CONFIG_FEATURE_TFTP_PUT
#define PUT "p"
#else
#define PUT
#endif

	while ((opt = getopt(argc, argv, BS GET PUT "l:r:")) != -1) {
		switch (opt) {
#ifdef CONFIG_FEATURE_TFTP_BLOCKSIZE
		case 'b':
			blocksize = atoi(optarg);
			if (!tftp_blocksize_check(blocksize, 0)) {
                                return EXIT_FAILURE;
			}
			break;
#endif
#ifdef CONFIG_FEATURE_TFTP_GET
		case 'g':
			cmd = tftp_cmd_get;
			flags = O_WRONLY | O_CREAT | O_TRUNC;
			break;
#endif
#ifdef CONFIG_FEATURE_TFTP_PUT
		case 'p':
			cmd = tftp_cmd_put;
			flags = O_RDONLY;
			break;
#endif
		case 'l':
			localfile = optarg;
			break;
		case 'r':
			remotefile = optarg;
			break;
		}
	}

	if ((cmd == 0) || (optind == argc)) {
		bb_show_usage();
	}
	if(localfile && strcmp(localfile, "-") == 0) {
	    fd = fileno((cmd==tftp_cmd_get)? stdout : stdin);
	}
	if(localfile == NULL)
	    localfile = remotefile;
	if(remotefile == NULL)
	    remotefile = localfile;
	if (fd==-1) {
	    fd = open(localfile, flags, 0644);
	}
	if (fd < 0) {
		bb_perror_msg_and_die("local file");
	}

	host = xgethostbyname(argv[optind]);
	port = bb_lookup_port(argv[optind + 1], "udp", 69);

#ifdef CONFIG_FEATURE_TFTP_DEBUG
	fprintf(stderr, "using server \"%s\", remotefile \"%s\", "
		"localfile \"%s\".\n",
		inet_ntoa(*((struct in_addr *) host->h_addr)),
		remotefile, localfile);
#endif

	result = tftp(cmd, host, remotefile, fd, port, blocksize);

#ifdef CONFIG_FEATURE_CLEAN_UP
	if (!(fd == STDOUT_FILENO || fd == STDIN_FILENO)) {
	    close(fd);
	}
#endif
	return(result);
}

static void pipe_quit(int sig){
	signal(SIGPIPE, SIG_IGN);
    //fprintf(stderr, "pipe broken detected, send out tftp_error\n");
	longjmp(jmp_env, 1);
}
