/* dhcpc.c
 *
 * udhcp DHCP client
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/time.h>
#include <sys/file.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "dhcpd.h"
#include "dhcpc.h"
#include "options.h"
#include "clientpacket.h"
#include "clientsocket.h"
#include "script.h"
#include "socket.h"
#include "common.h"
#include "signalpipe.h"
#include "arpping.h"

//defined for nvram variable storing device serial number populated
//by gs_cpe
#define DEVICE_SERIAL_NO ":device_serial_no"

#ifdef GXV
#define  DEVICE_PRODUCT_CLASS "/proc/gxvboard/dev_info/dev_alias"
#define DEVICE_MAC "/proc/gxvboard/dev_info/dev_mac"
#elif defined (GXW4200)
#define  DEVICE_PRODUCT_CLASS "/proc/gsboard/dev_info/dev_alias"
#define DEVICE_MAC "/proc/gsboard/dev_info/dev_mac"
#else
#error "product model not defined, check build setting."
#endif
//pen is 3561(0xde9) for the Broadband forum from the list on http://www.iana.org/assignments/enterprise-numbers
#define PEN 0xde9 

static int previous_state;
static int state;
static unsigned long requested_ip; /* = 0 */
static unsigned long server_addr;
static unsigned long prev_timeout;
static int fd = -1;

#define LISTEN_NONE 0
#define LISTEN_KERNEL 1
#define LISTEN_RAW 2
static int listen_mode;

struct client_config_t client_config = {
	/* Default options. */
	abort_if_no_lease: 0,
	foreground: 0,
	quit_after_lease: 0,
	release_on_quit: 0,
	background_if_no_lease: 0,
	interface: "eth0",
	pidfile: NULL,
	script: DEFAULT_SCRIPT,
	clientid: NULL,
	vendorclass: NULL,
	hostname: NULL,
	ifindex: 0,
	arp: "\0\0\0\0\0\0",		/* appease gcc-3.0 */
};

#ifndef IN_BUSYBOX
static void __attribute__ ((noreturn)) show_usage(void)
{
	printf(
"Usage: udhcpc [OPTIONS]\n\n"
"  -c, --clientid=CLIENTID         Client identifier\n"
"  -V, --vendorclass=CLASSID       Set vendor class identifier\n"
"  -H, --hostname=HOSTNAME         Client hostname\n"
"  -h                              Alias for -H\n"
"  -f, --foreground                Do not fork after getting lease\n"
"  -b, --background                Fork to background if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -i, --interface=INTERFACE       Interface to use (default: eth0)\n"
"  -n, --now                       Exit with failure if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -p, --pidfile=file              Store process ID of daemon in file\n"
"  -q, --quit                      Quit after obtaining lease\n"
"  -R, --release                   Release IP on quit\n"
"  -r, --request=IP                IP address to request (default: none)\n"
"  -s, --script=file               Run file at dhcp events (default:\n"
"                                  " DEFAULT_SCRIPT ")\n"
"  -v, --version                   Display version\n"
	);
	exit(0);
}
#else
#define show_usage bb_show_usage
extern void show_usage(void) __attribute__ ((noreturn));
#endif

/* ---- Unaligned access ------------------------------------ */

/* parameter is supposed to be an uint32_t* ptr */
#if defined(i386) || defined(__x86_64__) /* + other arches? */
#define get_unaligned_u32p(u32p) (*(u32p))
#else
/* performs reasonably well (gcc usually inlines memcpy here) */
#define get_unaligned_u32p(u32p) ({ uint32_t __t; memcpy(&__t, (u32p), 4); __t; })
#endif

unsigned long long monotonic_us(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000000ULL + tv.tv_usec;
}

unsigned monotonic_sec(void)
{
    return time(NULL);
}

/* just a little helper */
static void change_listen_mode(int new_mode)
{
	DEBUG(LOG_INFO, "entering %s listen mode",
		new_mode ? (new_mode == 1 ? "kernel" : "raw") : "none");
	if (fd >= 0) close(fd);
	fd = -1;
	listen_mode = new_mode;
}


/* perform a renew */
static void perform_renew(void)
{
	LOG(LOG_INFO, "Performing a DHCP renew");
	switch (state) {
	case BOUND:
		change_listen_mode(LISTEN_KERNEL);
	case RENEWING:
	case REBINDING:
		state = RENEW_REQUESTED;
		break;
	case RENEW_REQUESTED: /* impatient are we? fine, square 1 */
	case REQUESTING:
	case RELEASED:
		change_listen_mode(LISTEN_RAW);
		state = INIT_SELECTING;
		break;
	case INIT_SELECTING:
		break;
	}
}


/* perform an inform*/
static void perform_inform(void)
{
	/* send inform packet */
	if (state == BOUND || state == RENEWING || state == REBINDING) {
		change_listen_mode(LISTEN_KERNEL);
		previous_state = state ;
		state = INFORM;
		LOG( LOG_INFO, "Entering inform state");
	}
}


/* perform a release */
static void perform_release(void)
{
	char buffer[16];
	struct in_addr temp_addr;

	/* send release packet */
	if (state == BOUND || state == RENEWING || state == REBINDING) {
		temp_addr.s_addr = server_addr;
		sprintf(buffer, "%s", inet_ntoa(temp_addr));
		temp_addr.s_addr = requested_ip;
		LOG(LOG_INFO, "Unicasting a release of %s to %s",
				inet_ntoa(temp_addr), buffer);
		send_release(server_addr, requested_ip); /* unicast */
		run_script(NULL, "deconfig");
	}
	LOG(LOG_INFO, "Entering released state");

	change_listen_mode(LISTEN_NONE);
	state = RELEASED;
}

/*mac has to have at least a length of 18 chars*/
static void get_mac_address( char *mac )
{
	FILE *fp ;
	int i , colon = 0 ;

	fp = fopen( DEVICE_MAC, "r" ) ;
	fread( mac, 17, 1, fp ) ;
	fclose( fp ) ;
	for( i = 0 ; i < 17 ; i++ )
	{
		if( mac[i+colon] == ':' )
			colon++ ;
		if( ( i + colon ) >= 17 )
			break ;
		mac[i] = mac[i+colon] ;
	}
	mac[i] = '\0' ;
}

/*gets nvram value*/
static size_t get_nvram_value( const char *var, char *buffer, int size )
{
	FILE* fp ;
	char cmd[256] ;
	size_t bytes_read = 0 ;
	
	snprintf( cmd, 256, "nvram get %s", var ) ;
	cmd[255] = '\0' ;
	fp = popen( cmd, "r" ) ;
	if( !fp )
		return 0 ;
	bytes_read = fread( buffer, 1, size, fp ) ;
	pclose (fp) ;
	if(( bytes_read == 0 ) )
		return 0 ;
	if( bytes_read == size )
		buffer[bytes_read-1] = '\0' ;
	else
	{               
		/*remove new line from the end if it exists*/
		if( buffer[bytes_read - 1] == '\n' )
		{
			buffer[bytes_read-1] = '\0' ;
			bytes_read-- ;
		}
		else	
			buffer[bytes_read] = '\0' ;
	}
	return bytes_read ;
}

/*set device serial no got from nvram variable set by gs_cpe*/
static int set_device_serial_no( void )
{
	int len = 0 ;
	char sn[128], *sn_ptr, mac[18] ;

	get_mac_address( mac ) ;
	memset( ( void * )sn, 0, 128 ) ;
    get_nvram_value( DEVICE_SERIAL_NO, sn, 128 ) ;
	if( !strlen( sn ) )
		sn_ptr = mac ;
	else
		sn_ptr = sn ;
	if( !strcmp( client_config.Device.SerialNumber, sn_ptr ) )
	{
		return 0 ;
	}
//fprintf( stderr, "%s:%d\n", __FILE__, __LINE__ ) ;		
    len = strlen( sn_ptr ) ;
	strcpy( client_config.Device.SerialNumber,  sn_ptr ) ;
	client_config.Device.sn_len = len ;
	return 1 ;
}

/*build the option 125 string to send to server*/
static void build_option_125_string( void )
{
	int len = 0 ;

	client_config.Device.data_len = client_config.Device.oui_len + client_config.Device.sn_len + client_config.Device.pc_len + 6 ;
	client_config.Device.opt_len = client_config.Device.data_len + sizeof( client_config.Device.pen ) + 1 ;
	if ( client_config.Device.string )
		free( client_config.Device.string ) ;
	client_config.Device.string = xmalloc( client_config.Device.opt_len + 2 ) ;
	client_config.Device.string[OPT_CODE] = client_config.Device.opt_code ;
	client_config.Device.string[OPT_LEN] = client_config.Device.opt_len ;
	len += 2 ;
    memcpy( &client_config.Device.string[len], &client_config.Device.pen, sizeof( client_config.Device.pen ) ) ;
	len += sizeof( client_config.Device.pen ) ;
    client_config.Device.string[len] = client_config.Device.data_len ;
	len += sizeof( client_config.Device.data_len ) ;
	client_config.Device.string[len++] = 1 ;
	client_config.Device.string[len++] = client_config.Device.oui_len ;
	strncpy((char *) &client_config.Device.string[len], client_config.Device.ManufacturerOUI, client_config.Device.oui_len ) ;  
	len += client_config.Device.oui_len ;
	client_config.Device.string[len++] = 2 ;
	client_config.Device.string[len++] = client_config.Device.sn_len ;
	strncpy((char *) &client_config.Device.string[len], client_config.Device.SerialNumber, client_config.Device.sn_len ) ;  
	len += client_config.Device.sn_len ;
	client_config.Device.string[len++] = 3 ;
	client_config.Device.string[len++] = client_config.Device.pc_len ;
	strncpy((char *) &client_config.Device.string[len], client_config.Device.ProductClass, client_config.Device.pc_len ) ;  
}

/*Initially populate the device info structure  */
static void populate_device_info( void )
{
	int flag = 0 ;
	int len = 0 ;
	FILE *fp ;
	char mac[18] ;
	int i ;

	client_config.Device.opt_code = 125 ;
	get_mac_address( mac ) ;
	if ( !client_config.Device.ManufacturerOUI )
	{   
		char oui[7] ;

		for( i = 0 ; i < 6 ; i++ )
			oui[i] = mac[i] ;
		oui[6] = '\0' ;
		len = strlen( oui ) ;
		client_config.Device.ManufacturerOUI = xmalloc( len + 1 ) ;
		memcpy( client_config.Device.ManufacturerOUI, oui, len ) ;
		client_config.Device.oui_len = len ;
		flag = 1 ;
	}
	flag = set_device_serial_no( ) ;
	if ( !client_config.Device.ProductClass )
	{                   
		char pc[128] ;
		memset(  ( void * )pc, 0, 128 ) ; 
		fp = fopen( DEVICE_PRODUCT_CLASS, "r") ;
		fread( pc, 128, 1, fp ) ;
		fclose( fp ) ;
        len = strlen( pc ) ;
		client_config.Device.ProductClass = xmalloc( len + 1 ) ;
		memcpy( client_config.Device.ProductClass, pc, len ) ;
		client_config.Device.pc_len = len ;
		flag = 1 ;
	}
	client_config.Device.pen = PEN ;
	client_config.Device.pen = htonl( client_config.Device.pen ) ;
	client_config.Device.data_len = client_config.Device.oui_len + client_config.Device.sn_len + client_config.Device.pc_len + 6 ;
	client_config.Device.opt_len = client_config.Device.data_len + sizeof( client_config.Device.pen ) + 1 ;
	if( !flag )
		return ;
		build_option_125_string() ;
}


static void client_background(void)
{
	background(client_config.pidfile);
	client_config.foreground = 1; /* Do not fork again. */
	client_config.background_if_no_lease = 0;
}


#ifdef COMBINED_BINARY
int udhcpc_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	uint8_t *temp, *message;
	unsigned long xid = 0;
	unsigned long lease_seconds;
	fd_set rfds;
	int retval;
	struct timeval tv;
    int timeout;    /* must be signed */
    unsigned already_waited_sec;
    int tryagain_timeout = 20;
    int discover_wait_timeout = 40;
    int discover_timeout = 3;
    int discover_retries = 5;
    int packet_num; /* = 0 */
	int c, len;
	struct dhcpMessage packet;
	struct in_addr temp_addr;
	int max_fd;
	int sig;
	int prev_mode = LISTEN_NONE ;


	static const struct option arg_options[] = {
		{"clientid",	required_argument,	0, 'c'},
        {"vendorclass", required_argument,  0, 'V'},
		{"foreground",	no_argument,		0, 'f'},
		{"background",	no_argument,		0, 'b'},
		{"hostname",	required_argument,	0, 'H'},
		{"hostname",    required_argument,      0, 'h'},
		{"interface",	required_argument,	0, 'i'},
		{"now", 	no_argument,		0, 'n'},
		{"pidfile",	required_argument,	0, 'p'},
		{"quit",	no_argument,		0, 'q'},
		{"release",	no_argument,		0, 'R'},
		{"request",	required_argument,	0, 'r'},
		{"script",	required_argument,	0, 's'},
		{"version",	no_argument,		0, 'v'},
		{"request-option", required_argument, 0, 'O'},
		{0, 0, 0, 0}
	};

	/* get options */
	while (1) {
		int option_index = 0;
		c = getopt_long(argc, argv, "c:V:fbH:h:i:np:qRr:s:vmO:", arg_options, &option_index);
		if (c == -1) break;

		switch (c) {
        case 'm':
            minpkt = 1;
            break;
		case 'c':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			if (client_config.clientid) free(client_config.clientid);
			client_config.clientid = xmalloc(len + 2);
			client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
			client_config.clientid[OPT_LEN] = len;
			client_config.clientid[OPT_DATA] = '\0';
			strncpy((char *)client_config.clientid + OPT_DATA, optarg, len);
			break;
        case 'V':
            len = strlen(optarg) > 255 ? 255 : strlen(optarg);
            if (client_config.vendorclass) free(client_config.vendorclass);
            client_config.vendorclass = xmalloc(len + 2);
            client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
            client_config.vendorclass[OPT_LEN] = len;
            strncpy((char *)client_config.vendorclass + OPT_DATA, optarg, len);
            break;
		case 'f':
			client_config.foreground = 1;
			break;
		case 'b':
			client_config.background_if_no_lease = 1;
			break;
		case 'h':
		case 'H':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			if (client_config.hostname) free(client_config.hostname);
			client_config.hostname = xmalloc(len + 2);
			client_config.hostname[OPT_CODE] = DHCP_HOST_NAME;
			client_config.hostname[OPT_LEN] = len;
			strncpy((char *)client_config.hostname + 2, optarg, len);
			break;
		case 'i':
			client_config.interface =  optarg;
			break;
		case 'n':
			client_config.abort_if_no_lease = 1;
			break;
		case 'p':
			client_config.pidfile = optarg;
			break;
		case 'q':
			client_config.quit_after_lease = 1;
			break;
		case 'R':
			client_config.release_on_quit = 1;
			break;
		case 'r':
			requested_ip = inet_addr(optarg);
			break;
		case 's':
			client_config.script = optarg;
			break;
		case 'v':
			printf("udhcpcd, version %s\n\n", VERSION);
			return 0;
			break;
		case 'O':
			{
				int n =	get_dhcp_option_index( optarg ) ;
				//printf( "%s:%d %d\n", __FILE__, __LINE__, n ) ;
				if( n < 0 )
					break ;
				n = dhcp_options[n].code;
				//printf( "%s:%d %d\n", __FILE__, __LINE__, n ) ;
				client_config.opt_mask[n >> 3] |= 1 << (n & 7);
				break;
			}
		default:
			show_usage();
		}
	}

	/* Start the log, sanitize fd's, and write a pid file */
	start_log_and_pid("udhcpc", client_config.pidfile);

	if (read_interface(client_config.interface, &client_config.ifindex,
			   NULL, client_config.arp) < 0)
		return 1;

	if (!client_config.clientid) {
		client_config.clientid = xmalloc(6 + 3);
		client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
		client_config.clientid[OPT_LEN] = 7;
		client_config.clientid[OPT_DATA] = 1;
		memcpy(client_config.clientid + 3, client_config.arp, 6);
	}

    if (!client_config.vendorclass) {
#ifdef GXV
        client_config.vendorclass = xmalloc(sizeof("gxv31xx") + 2);
        client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
        client_config.vendorclass[OPT_LEN] = sizeof("gxv31xx") - 1;
        client_config.vendorclass[OPT_DATA] = 1;
        memcpy(&client_config.vendorclass[OPT_DATA], "gxv31xx", sizeof("gxv31xx") - 1);
#elif defined (GXW4200)
        client_config.vendorclass = xmalloc(sizeof("gxw42xx") + 2);
        client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
        client_config.vendorclass[OPT_LEN] = sizeof("gxw42xx") - 1;
        client_config.vendorclass[OPT_DATA] = 1;
        memcpy(&client_config.vendorclass[OPT_DATA], "gxw42xx", sizeof("gxw42xx") - 1);
#elif define (HT5XX)
        client_config.vendorclass = xmalloc(sizeof("ht5xx") + 2);
        client_config.vendorclass[OPT_CODE] = DHCP_VENDOR;
        client_config.vendorclass[OPT_LEN] = sizeof("ht5xx") - 1;
        client_config.vendorclass[OPT_DATA] = 1;
        memcpy(&client_config.vendorclass[OPT_DATA], "ht5xx", sizeof("ht5xx") - 1);
#endif
    }
	/*populate device info structure used in option 125*/
	populate_device_info() ;
    
	/* setup the signal pipe */
	udhcp_sp_setup();

	state = INIT_SELECTING;
	run_script(NULL, "deconfig");
	change_listen_mode(LISTEN_RAW);
    timeout = 0;
    packet_num = 0;
    already_waited_sec = 0;

	for (;;) {
        unsigned timestamp_before_wait;

		tv.tv_sec = timeout - already_waited_sec;
		tv.tv_usec = 0;

		if (listen_mode != LISTEN_NONE && fd < 0) {
			if (listen_mode == LISTEN_KERNEL)
				fd = listen_socket(INADDR_ANY, CLIENT_PORT, client_config.interface);
			else
				fd = raw_socket(client_config.ifindex);
			if (fd < 0) {
				LOG(LOG_ERR, "FATAL: couldn't listen on socket, %m");
				return 0;
			}
		}
		max_fd = udhcp_sp_fd_set(&rfds, fd);

		if (tv.tv_sec > 0) {
            timestamp_before_wait = (unsigned)monotonic_sec();
			DEBUG(LOG_INFO, "Waiting on select..., timeout=%d", tv.tv_sec);
			retval = select(max_fd + 1, &rfds, NULL, NULL, &tv);
		} else retval = 0; /* If we already timed out, fall through */

		if (retval == 0) {
			/* We will restart wait afresh in any case */
            already_waited_sec = 0;

			switch (state) {
			case INIT_SELECTING:
				if (!discover_retries || packet_num < discover_retries) {
					if (packet_num == 0)
						xid = random_xid();

					/*set_device_serial_no if it changed- option 125 */
					if( set_device_serial_no() )
						build_option_125_string() ;
					/* send discover packet */
					send_discover(xid, requested_ip); /* broadcast */

                    if (packet_num == discover_retries - 1)
                        timeout = discover_wait_timeout;
                    else
                        timeout = discover_timeout;
					packet_num++;
                    continue;
				} 
                run_script(NULL, "leasefail");
                if (client_config.background_if_no_lease) {
                    LOG(LOG_INFO, "No lease, forking to background.");
                    client_background();
                    /* do not background again! */
                    client_config.background_if_no_lease = 0;
                } else if (client_config.abort_if_no_lease) {
                    LOG(LOG_INFO, "No lease, failing.");
                    return 1;
                }
                /* wait to try again */
                packet_num = 0;
                timeout = tryagain_timeout;
				continue;
			case RENEW_REQUESTED:
			case REQUESTING:
                if (!discover_retries || packet_num < discover_retries) {
					/*set_device_serial_no if it changed- option 125 */
					if( set_device_serial_no() )
						build_option_125_string() ;
					/* send request packet */
					if (state == RENEW_REQUESTED)
						send_renew(xid, server_addr, requested_ip); /* unicast */
					else send_selecting(xid, server_addr, requested_ip); /* broadcast */

                    if (packet_num == discover_retries - 1)
                        timeout = discover_wait_timeout;
                    else
                        timeout = discover_timeout;
					packet_num++;
                    continue;
				} 
                /* timed out, go back to init state */
                if (state == RENEW_REQUESTED) run_script(NULL, "deconfig");
                change_listen_mode(LISTEN_RAW);
                state = INIT_SELECTING;
                timeout = 0;
                packet_num = 0;
				continue;
			case BOUND:
				/* Lease is starting to run out, time to enter renewing state */
				change_listen_mode(LISTEN_KERNEL);
				DEBUG(LOG_INFO, "Entering renew state");
				state = RENEWING;
				/* fall right through */
			case RENEWING:
                if (timeout > 60) {
					/*set_device_serial_no if it changed- option 125 */
					if( set_device_serial_no() )
						build_option_125_string() ;
                    /* send a request packet */
                    send_renew(xid, server_addr, requested_ip); /* unicast */
                    timeout >>= 1;
                    continue;
                }
                /* Timed out, enter rebinding state */
                DEBUG(LOG_INFO, "Entering rebinding state");
                state = REBINDING;
                /* fall right through */
			case REBINDING:
                /* Lease is *really* about to run out,
                 * try to find DHCP server using broadcast */
                if (timeout > 0) {
					/*set_device_serial_no if it changed- option 125 */
					if( set_device_serial_no() )
						build_option_125_string() ;
                    /* send a request packet */
                    send_renew(xid, 0, requested_ip); /* broadcast */
                    timeout >>= 1;
                    continue;
                }
                /* Timed out, enter init state */
				LOG(LOG_INFO, "Lease lost, entering init state");
                run_script(NULL, "deconfig");
                change_listen_mode(LISTEN_RAW);
                state = INIT_SELECTING;
                /*timeout = 0; - already is */
                packet_num = 0;
                continue;
			case INFORM:
			case_INFORM:
                if (!discover_retries || packet_num < discover_retries) {
					/*set_device_serial_no if it changed- option 125 */
					if( set_device_serial_no() )
						build_option_125_string() ;
					send_inform_client(xid, server_addr, requested_ip);
					timeout = discover_timeout;
					packet_num++ ;
                    continue;
				} 
                /*servers could be non-authoritative or offline*/
                /*set to the previous state*/
                timeout = prev_timeout ;
                change_listen_mode( prev_mode ) ;
                state = previous_state ;
                packet_num = 0 ;
				continue ;
            /* case RELEASED: */
			}
            /* yah, I know, *you* say it would never happen */
            timeout = INT_MAX;
            continue; /* back to main loop */
		} 

        /* select() didn't timeout, something did happen. */
        /* Is is a packet? */
        if (listen_mode != LISTEN_NONE && FD_ISSET(fd, &rfds)) {
			/* a packet is ready, read it */

			if (listen_mode == LISTEN_KERNEL)
				len = get_packet(&packet, fd);
			else len = get_raw_packet(&packet, fd);

			if (len == -1 && errno != EINTR) {
				DEBUG(LOG_INFO, "error on read, %m, reopening socket");
                sleep(discover_timeout); /* 3 seconds by default */
				change_listen_mode(listen_mode); /* just close and reopen */
			}
            /* If this packet will turn out to be unrelated/bogus,
             * we will go back and wait for next one.
             * Be sure timeout is properly decreased. */
            already_waited_sec += (unsigned)monotonic_sec() - timestamp_before_wait;

			if (len < 0) continue;

			if (packet.xid != xid) {
				DEBUG(LOG_INFO, "Ignoring XID %lx (our xid is %lx)",
					(unsigned long) packet.xid, xid);
				continue;
			}

			if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
				continue;
			}

			switch (state) {
			case INIT_SELECTING:
				/* Must be a DHCPOFFER to one of our xid's */
				if (*message == DHCPOFFER) {
	                LOG(LOG_DEBUG, "DHCP OFFER received");
					if ((temp = get_option(&packet, DHCP_SERVER_ID))) {
                        server_addr = get_unaligned_u32p((uint32_t*)temp);

						xid = packet.xid;
						requested_ip = packet.yiaddr;

						/* enter requesting state */
						state = REQUESTING;
						timeout = 0;
						packet_num = 0;
                        already_waited_sec = 0;

					} else {
						DEBUG(LOG_ERR, "No server ID in message");
                        continue;
					}
				}
				continue;
			case RENEW_REQUESTED:
			case REQUESTING:
			case RENEWING:
			case REBINDING:
	            LOG(LOG_DEBUG, "RENEW/REQUEST/REBIND state, received %d", *message);
				if (*message == DHCPACK) {
	                LOG(LOG_DEBUG, "DHCP ACK received");
					if (!(temp = get_option(&packet, DHCP_LEASE_TIME))) {
						LOG(LOG_ERR, "No lease time with ACK, using 1 hour lease");
						lease_seconds = 60 * 60;
					} else {
						memcpy(&lease_seconds, temp, 4);
						lease_seconds = ntohl(lease_seconds);
                        lease_seconds &= 0x0fffffff; /* paranoia: must not be prone to overflows */
                        if (lease_seconds < 10)
                            lease_seconds = 10;
					}

					temp_addr.s_addr = packet.yiaddr;
/* RFC 2131 3.1 paragraph 5:
 * "The client receives the DHCPACK message with configuration
 * parameters. The client SHOULD perform a final check on the
 * parameters (e.g., ARP for allocated network address), and notes
 * the duration of the lease specified in the DHCPACK message. At this
 * point, the client is configured. If the client detects that the
 * address is already in use (e.g., through the use of ARP),
 * the client MUST send a DHCPDECLINE message to the server and restarts
 * the configuration process..." */
					//if( !arpping( packet.yiaddr, packet.yiaddr, client_config.arp, client_config.interface ) ) SXIE:
                    //seems linux kernel does not like gratuitous arp here 
					if( !arpping( packet.yiaddr, (uint32_t)0, client_config.arp, client_config.interface ) )
					{
						LOG(LOG_INFO, "Address %s already allocated( got an ARP reply), declining",
							inet_ntoa(temp_addr));

						send_decline( xid, server_addr, packet.yiaddr ) ;

						if (state != REQUESTING)
							run_script(NULL, "deconfig");

						state = INIT_SELECTING;
						timeout = tryagain_timeout;
						requested_ip = 0;
						packet_num = 0;
						change_listen_mode(LISTEN_RAW);
						//sleep(10); /* avoid excessive network traffic */
                        already_waited_sec = 0;
                        continue; /* back to main loop */
					}
                    
                    LOG(LOG_INFO, "Lease of %s obtained, lease time %ld",
                        inet_ntoa(temp_addr), lease_seconds);
                    timeout = lease_seconds / 2;
                    requested_ip = packet.yiaddr;
                    run_script(&packet, ((state == RENEWING || state == REBINDING) ? "renew" : "bound"));
                    state = BOUND;
                    change_listen_mode(LISTEN_NONE);
                    if (client_config.quit_after_lease) {
                        if (client_config.release_on_quit) 
                            perform_release();
                        return 0;
                    }
                    if (!client_config.foreground)
                        client_background();

                    already_waited_sec = 0;
                    continue; /* back to main loop */

				} else if (*message == DHCPNAK) {
					/* return to init state */
					LOG(LOG_INFO, "Received DHCP NAK");
					run_script(&packet, "nak");
					if (state != REQUESTING)
						run_script(NULL, "deconfig");

					change_listen_mode(LISTEN_RAW);
					sleep(3); /* avoid excessive network traffic */
					state = INIT_SELECTING;
					requested_ip = 0;
					packet_num = 0;
                    timeout = 0;
                    already_waited_sec = 0;
				}
				break;
			case INFORM:
				LOG(LOG_DEBUG, "INFORM state");
				if (*message == DHCPACK) {
                    LOG(LOG_DEBUG, "DHCPACK received\n");
					run_script(&packet, "inform" );
					/*now set to the previous state*/
					state = previous_state ;
					timeout = prev_timeout ;
					packet_num = 0;
					change_listen_mode(prev_mode);
                    already_waited_sec = 0; //SXIE: Add now, not quite sure
				}
				else if (*message == DHCPNAK) {
					/* return to init state */
					LOG( LOG_INFO, "Received DHCP NAK");
					run_script(&packet, "nak");
					if (state != REQUESTING)
						run_script(NULL, "deconfig");
					change_listen_mode(LISTEN_RAW);
					sleep(3); /* avoid excessive network traffic */
					state = INIT_SELECTING;
					requested_ip = 0;
					timeout = 0;
					packet_num = 0;
                    already_waited_sec = 0;
				}
				continue;
			/* case BOUND, RELEASED: - ignore all packets */
			}
            continue;
		} 

        if ((sig = udhcp_sp_read(&rfds))) {
			switch (sig) {
			case SIGUSR1:
				perform_renew();
                packet_num = 0;
                /* Kill any timeouts because the user wants this to hurry along */
                timeout = 0;
				break;
			case SIGUSR2:
				if ( state == INFORM )
					goto case_INFORM ;
				prev_mode = listen_mode ;
				perform_inform() ;
				packet_num = 0;
				prev_timeout = timeout ;
				timeout = 0;
				continue;
			case SIGTERM:
				LOG( LOG_INFO, "Received SIGTERM");
				perform_release();
                
				return 0 ;
			}
		} else if (retval == -1 && errno == EINTR) {
			/* a signal was caught */
		} else {
			/* An error occured */
			DEBUG(LOG_ERR, "Error on select");
		}

	} // for (;;) mainloop
	return 0;
}
