/* vi: set sw=4 ts=4: */
/*
 * Mini syslogd implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Copyright (C) 2000 by Karl M. Hegbloom <karlheg@debian.org>
 *
 * "circular buffer" Copyright (C) 2001 by Gennady Feldman <gfeldman@gena01.com>
 *
 * Maintainer: Gennady Feldman <gfeldman@gena01.com> as of Mar 12, 2001
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */

/*
 * Done in syslogd_and_logger.c:
#include "libbb.h"
#define SYSLOG_NAMES
#define SYSLOG_NAMES_CONST
#include <syslog.h>
*/

#include <paths.h>
#include <sys/un.h>
#include <sys/uio.h>

#if ENABLE_FEATURE_REMOTE_LOG
#include <netinet/in.h>
#endif

#if ENABLE_FEATURE_IPC_SYSLOG
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#endif

#include <net/if.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <arpa/inet.h>

#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <nvram.h>
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
#include <dlfcn.h>
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */


#define SYSLOG_PORT_TLS         6514
#define SYSLOG_PORT_UDP         514
#define SSL_CONNECT_TIME_OUT    10
#define DELAY_MULTIPLYER        30
#define ENABLE_BB_MSG           1

#define PVALUE_LOAD_CA_CERT         "8502"
#define PVALUE_VERIFY_SERVER        "8463"
#define PVALUE_DISABLE_WEAK_CIPHER  "8536"
#define PVALUE_WAN_DEVICE           "wan_if"
#define TLS_CA_DIR                  "/etc/pki/tls/"
#define TLS_CA_FILE                 "/tmp/ssl/ca_certs.pem"
#define DEVICE_PRODUCT_CLASS        "/proc/gsboard/dev_info/dev_alias"
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
#define LIBSSL                      "/usr/lib/libssl.so"
#define LIBCRYPTO                   "/usr/lib/libcrypto.so"
#define LIBPTHREAD                  "/lib/libpthread.so"
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
typedef int (*SSL_library_init_t)(void);
typedef int (*SSL_set_fd_t)(SSL *s, int fd);
typedef int (*SSL_connect_t)(SSL *ssl);
typedef SSL *(*SSL_new_t)(SSL_CTX *);
typedef int (*SSL_shutdown_t)(SSL *);
typedef void (*SSL_free_t)(SSL *);
typedef int (*SSL_read_t)(SSL *, void *, int);
typedef int (*SSL_write_t)(SSL *, const void *, int);
typedef int (*SSL_get_error_t)(const SSL *, int);
typedef void (*SSL_CTX_free_t)(SSL_CTX *);
typedef SSL_CTX *(*SSL_CTX_new_t)(const SSL_METHOD *);
typedef const SSL_METHOD *(*SSLv23_method_t)(void);
typedef int (*SSL_CTX_load_verify_locations_t)(SSL_CTX *, const char *, const char *);
typedef void (*SSL_CTX_set_verify_t)(SSL_CTX *, int, int (*) (int, X509_STORE_CTX *));
typedef long (*SSL_CTX_ctrl_t)(SSL_CTX *ctx, int cmd, long larg, void *parg);
typedef void (*OpenSSL_add_all_algorithms_t)(void);
typedef int (*pthread_mutexattr_init_t)(pthread_mutexattr_t *);
typedef int (*pthread_mutexattr_destroy_t)(pthread_mutexattr_t *);
typedef int (*pthread_mutexattr_settype_t)(pthread_mutexattr_t *, int);
typedef int (*pthread_mutex_lock_t)(pthread_mutex_t *);
typedef int (*pthread_mutex_unlock_t)(pthread_mutex_t *);
typedef int (*pthread_mutex_init_t)(pthread_mutex_t *, const pthread_mutexattr_t *);
typedef int (*pthread_mutex_destroy_t)(pthread_mutex_t *);
typedef int (*pthread_create_t)(pthread_t *, const pthread_attr_t *, void *(*) (void *), void *);

static void *ssl_handle = NULL;
static void *crypto_handle = NULL;
static void *pthread_handle = NULL;
static SSL_library_init_t SSL_library_init_p = NULL;
static SSL_set_fd_t SSL_set_fd_p = NULL;
static SSL_connect_t SSL_connect_p = NULL;
static SSL_new_t SSL_new_p = NULL;
static SSL_shutdown_t SSL_shutdown_p = NULL;
static SSL_free_t SSL_free_p = NULL;
static SSL_read_t SSL_read_p = NULL;
static SSL_write_t SSL_write_p = NULL;
static SSL_get_error_t SSL_get_error_p = NULL;
static SSL_CTX_free_t SSL_CTX_free_p = NULL;
static SSL_CTX_new_t SSL_CTX_new_p = NULL;
static SSLv23_method_t SSLv23_method_p = NULL;
static SSL_CTX_load_verify_locations_t SSL_CTX_load_verify_locations_p = NULL;
static SSL_CTX_set_verify_t SSL_CTX_set_verify_p = NULL;
static SSL_CTX_ctrl_t SSL_CTX_ctrl_p = NULL;
#define SSL_CTX_set_cb_cipher_get_disabled_p(ctx,op) \
        if ( SSL_CTX_ctrl_p ) SSL_CTX_ctrl_p((ctx),SSL_CTRL_SET_CB_CIPHER_GET_DISABLED,(op),NULL)
#define SSL_CTX_rebuild_cipher_list_p(ctx) \
        if ( SSL_CTX_ctrl_p ) SSL_CTX_ctrl_p((ctx),SSL_CTRL_REBUILD_CIPHER_LIST,0,NULL)
static OpenSSL_add_all_algorithms_t OpenSSL_add_all_algorithms_p = NULL;
static pthread_mutexattr_init_t pthread_mutexattr_init_p = NULL;
static pthread_mutexattr_destroy_t pthread_mutexattr_destroy_p = NULL;
static pthread_mutexattr_settype_t pthread_mutexattr_settype_p = NULL;
static pthread_mutex_lock_t pthread_mutex_lock_p = NULL;
static pthread_mutex_unlock_t pthread_mutex_unlock_p = NULL;
static pthread_mutex_init_t pthread_mutex_init_p = NULL;
static pthread_mutex_destroy_t pthread_mutex_destroy_p = NULL;
static pthread_create_t pthread_create_p = NULL;
#else
static SSL_library_init_t SSL_library_init_p = SSL_library_init;
static SSL_set_fd_t SSL_set_fd_p = SSL_set_fd;
static SSL_connect_t SSL_connect_p = SSL_connect;
static SSL_new_t SSL_new_p = SSL_new;
static SSL_shutdown_t SSL_shutdown_p = SSL_shutdown;
static SSL_free_t SSL_free_p = SSL_free;
static SSL_read_t SSL_read_p = SSL_read;
static SSL_write_t SSL_write_p = SSL_write;
static SSL_get_error_t SSL_get_error_p = SSL_get_error;
static SSL_CTX_free_t SSL_CTX_free_p = SSL_CTX_free;
static SSL_CTX_new_t SSL_CTX_new_p = SSL_CTX_new;
static SSLv23_method_t SSLv23_method_p = SSLv23_method;
static SSL_CTX_load_verify_locations_t SSL_CTX_load_verify_locations_p = SSL_CTX_load_verify_locations;
static SSL_CTX_set_verify_t SSL_CTX_set_verify_p = SSL_CTX_set_verify;
#define SSL_CTX_set_cb_cipher_get_disabled_p(ctx, op) SSL_CTX_set_cb_cipher_get_disabled(ctx, op)
#define SSL_CTX_rebuild_cipher_list_p(ctx) SSL_CTX_rebuild_cipher_list(ctx)
#define OpenSSL_add_all_algorithms_p() OpenSSL_add_all_algorithms()
static pthread_mutexattr_init_t pthread_mutexattr_init_p = pthread_mutexattr_init;
static pthread_mutexattr_destroy_t pthread_mutexattr_destroy_p = pthread_mutexattr_destroy;
static pthread_mutexattr_settype_t pthread_mutexattr_settype_p = pthread_mutexattr_settype;
static pthread_mutex_lock_t pthread_mutex_lock_p = pthread_mutex_lock;
static pthread_mutex_unlock_t pthread_mutex_unlock_p = pthread_mutex_unlock;
static pthread_mutex_init_t pthread_mutex_init_p = pthread_mutex_init;
static pthread_mutex_destroy_t pthread_mutex_destroy_p = pthread_mutex_destroy;
static pthread_create_t pthread_create_p = pthread_create;
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

#define DEBUG 0

/* MARK code is not very useful, is bloat, and broken:
 * can deadlock if alarmed to make MARK while writing to IPC buffer
 * (semaphores are down but do_mark routine tries to down them again) */
#undef SYSLOGD_MARK

/* Write locking does not seem to be useful either */
#undef SYSLOGD_WRLOCK

enum {
	MAX_READ = 256,
	DNS_WAIT_SEC = 2 * 60,
};

/* Semaphore operation structures */
struct shbuf_ds {
	int32_t size;   /* size of data - 1 */
	int32_t tail;   /* end of message list */
	char data[1];   /* data/messages */
};

/* Allows us to have smaller initializer. Ugly. */
#define GLOBALS \
	const char *logFilePath;                \
	int logFD;                              \
	/* interval between marks in seconds */ \
	/*int markInterval;*/                   \
	/* level of messages to be logged */    \
	int logLevel;                           \
	char * filter_prefix;                   \
	int filter_prefix_len;                  \
	/*bug 10817 Vonage */                   \
	int user;                               \
	/* bug 141499 Support SSL/TLS */        \
	int tls;                                \
	int tls_con_status;                     \
	int tls_retry;                          \
	SSL_CTX* tls_ctx;                       \
	SSL* tls_ssl;                           \
	pthread_mutex_t cleanup_lock;           \
	pthread_t tls_disconnect_thread;        \
	pthread_t tls_reconnect_thread;         \
	pthread_t tls_linkstatus_thread;        \
IF_FEATURE_ROTATE_LOGFILE( \
	/* max size of file before rotation */  \
	unsigned logFileSize;                   \
	/* number of rotated message files */   \
	unsigned logFileRotate;                 \
	unsigned curFileSize;                   \
	smallint isRegular;                     \
) \
IF_FEATURE_REMOTE_LOG( \
	/* udp socket for remote logging */     \
	int remoteFD;                           \
	len_and_sockaddr* remoteAddr;           \
) \
IF_FEATURE_IPC_SYSLOG( \
	int shmid; /* ipc shared memory id */   \
	int s_semid; /* ipc semaphore id */     \
	int shm_size;                           \
	struct sembuf SMwup[1];                 \
	struct sembuf SMwdn[3];                 \
)

struct init_globals {
	GLOBALS
};

struct globals {
	GLOBALS

#if ENABLE_FEATURE_REMOTE_LOG
	unsigned last_dns_resolve;
	char *remoteAddrStr;
#endif

#if ENABLE_FEATURE_IPC_SYSLOG
	struct shbuf_ds *shbuf;
#endif
	time_t last_log_time;
	/* localhost's name. We print only first 64 chars */
	char *hostname;

	/* We recv into recvbuf... */
	char recvbuf[MAX_READ * (1 + ENABLE_FEATURE_SYSLOGD_DUP)];
	/* ...then copy to parsebuf, escaping control chars */
	/* (can grow x2 max) */
	char parsebuf[MAX_READ*2];
	/* ...then sprintf into printbuf, adding timestamp (15 chars),
	 * host (64), fac.prio (20) to the message */
	/* (growth by: 15 + 64 + 20 + delims = ~110) */
	char printbuf[MAX_READ*2 + 128];
};

static const struct init_globals init_data = {
	.logFilePath = "/var/log/messages",
	.logFD = -1,
#ifdef SYSLOGD_MARK
	.markInterval = 20 * 60,
#endif
	.logLevel = 8,
#if ENABLE_FEATURE_ROTATE_LOGFILE
	.logFileSize = 200 * 1024,
	.logFileRotate = 1,
#endif
#if ENABLE_FEATURE_REMOTE_LOG
	.remoteFD = -1,
#endif
#if ENABLE_FEATURE_IPC_SYSLOG
	.shmid = -1,
	.s_semid = -1,
	.shm_size = ((CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE)*1024), // default shm size
	.SMwup = { {1, -1, IPC_NOWAIT} },
	.SMwdn = { {0, 0}, {1, 0}, {1, +1} },
#endif
	.user = 0, //Bug 10817 Vonage
	.tls = 0, /* Bug 141499 */
	.tls_con_status = 0,
	.tls_retry = 0,
	.tls_ctx = NULL,
	.tls_ssl = NULL,
};

#define G (*ptr_to_globals)
#define INIT_G() do { \
	SET_PTR_TO_GLOBALS(memcpy(xzalloc(sizeof(G)), &init_data, sizeof(init_data))); \
} while (0)


/* Options */
enum {
	OPTBIT_mark = 0, // -m
	OPTBIT_nofork, // -n
	OPTBIT_outfile, // -O
	OPTBIT_loglevel, // -l
	OPTBIT_small, // -S
	OPTBIT_user, // -U  Bug 10817 Vonage
	OPTBIT_TLS, // -T  Bug 141499
	IF_FEATURE_ROTATE_LOGFILE(OPTBIT_filesize   ,)	// -s
	IF_FEATURE_ROTATE_LOGFILE(OPTBIT_rotatecnt  ,)	// -b
	IF_FEATURE_REMOTE_LOG(    OPTBIT_remotelog  ,)	// -R
	IF_FEATURE_REMOTE_LOG(    OPTBIT_locallog   ,)	// -L
	IF_FEATURE_IPC_SYSLOG(    OPTBIT_circularlog,)	// -C
	IF_FEATURE_SYSLOGD_DUP(   OPTBIT_dup        ,)	// -D

	OPT_mark        = 1 << OPTBIT_mark    ,
	OPT_nofork      = 1 << OPTBIT_nofork  ,
	OPT_outfile     = 1 << OPTBIT_outfile ,
	OPT_loglevel    = 1 << OPTBIT_loglevel,
	OPT_small       = 1 << OPTBIT_small   ,
	OPT_user        = 1 << OPTBIT_user   ,
	OPT_TLS         = 1 << OPTBIT_TLS   ,
	OPT_filesize    = IF_FEATURE_ROTATE_LOGFILE((1 << OPTBIT_filesize   )) + 0,
	OPT_rotatecnt   = IF_FEATURE_ROTATE_LOGFILE((1 << OPTBIT_rotatecnt  )) + 0,
	OPT_remotelog   = IF_FEATURE_REMOTE_LOG(    (1 << OPTBIT_remotelog  )) + 0,
	OPT_locallog    = IF_FEATURE_REMOTE_LOG(    (1 << OPTBIT_locallog   )) + 0,
	OPT_circularlog = IF_FEATURE_IPC_SYSLOG(    (1 << OPTBIT_circularlog)) + 0,
	OPT_dup         = IF_FEATURE_SYSLOGD_DUP(   (1 << OPTBIT_dup        )) + 0,
};
#define OPTION_STR "m:nO:l:SUT" \
	IF_FEATURE_ROTATE_LOGFILE("s:" ) \
	IF_FEATURE_ROTATE_LOGFILE("b:" ) \
	IF_FEATURE_REMOTE_LOG(    "R:" ) \
	IF_FEATURE_REMOTE_LOG(    "L"  ) \
	IF_FEATURE_IPC_SYSLOG(    "C::") \
	IF_FEATURE_SYSLOGD_DUP(   "D"  )
#define OPTION_DECL *opt_m, *opt_l \
	IF_FEATURE_ROTATE_LOGFILE(,*opt_s) \
	IF_FEATURE_ROTATE_LOGFILE(,*opt_b) \
	IF_FEATURE_IPC_SYSLOG(    ,*opt_C = NULL)
#define OPTION_PARAM &opt_m, &G.logFilePath, &opt_l \
	IF_FEATURE_ROTATE_LOGFILE(,&opt_s) \
	IF_FEATURE_ROTATE_LOGFILE(,&opt_b) \
	IF_FEATURE_REMOTE_LOG(    ,&G.remoteAddrStr) \
	IF_FEATURE_IPC_SYSLOG(    ,&opt_C)

#if ENABLE_BB_MSG
#define print_msg(s, ...) bb_error_msg(s, ##__VA_ARGS__);
#else
static void print_msg(const char *s, ...)
{
#define MSG_BUF_LEN 8192
	char buf[MSG_BUF_LEN] = {0};
	va_list p;

	va_start(p, s);
	vsnprintf(buf, MSG_BUF_LEN - 1, s, p);
	va_end(p);
	fprintf(stdout, "%s", buf);
}
#endif

/* circular buffer variables/structures */
#if ENABLE_FEATURE_IPC_SYSLOG

#if CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE < 4
#error Sorry, you must set the syslogd buffer size to at least 4KB.
#error Please check CONFIG_FEATURE_IPC_SYSLOG_BUFFER_SIZE
#endif

/* our shared key (syslogd.c and logread.c must be in sync) */
enum { KEY_ID = 0x414e4547 }; /* "GENA" */

static void ipcsyslog_cleanup(void)
{
	if (G.shmid != -1) {
		shmdt(G.shbuf);
	}
	if (G.shmid != -1) {
		shmctl(G.shmid, IPC_RMID, NULL);
	}
	if (G.s_semid != -1) {
		semctl(G.s_semid, 0, IPC_RMID, 0);
	}
}

static void ipcsyslog_init(void)
{
	if (DEBUG)
		printf("shmget(%x, %d,...)\n", (int)KEY_ID, G.shm_size);

	G.shmid = shmget(KEY_ID, G.shm_size, IPC_CREAT | 0644);
	if (G.shmid == -1) {
		bb_perror_msg_and_die("shmget");
	}

	G.shbuf = shmat(G.shmid, NULL, 0);
	if (G.shbuf == (void*) -1L) { /* shmat has bizarre error return */
		bb_perror_msg_and_die("shmat");
	}

	memset(G.shbuf, 0, G.shm_size);
	G.shbuf->size = G.shm_size - offsetof(struct shbuf_ds, data) - 1;
	/*G.shbuf->tail = 0;*/

	// we'll trust the OS to set initial semval to 0 (let's hope)
	G.s_semid = semget(KEY_ID, 2, IPC_CREAT | IPC_EXCL | 1023);
	if (G.s_semid == -1) {
		if (errno == EEXIST) {
			G.s_semid = semget(KEY_ID, 2, 0);
			if (G.s_semid != -1)
				return;
		}
		bb_perror_msg_and_die("semget");
	}
}

/* Write message to shared mem buffer */
static void log_to_shmem(const char *msg, int len)
{
	int old_tail, new_tail;

	if (semop(G.s_semid, G.SMwdn, 3) == -1) {
		bb_perror_msg_and_die("SMwdn");
	}

	/* Circular Buffer Algorithm:
	 * --------------------------
	 * tail == position where to store next syslog message.
	 * tail's max value is (shbuf->size - 1)
	 * Last byte of buffer is never used and remains NUL.
	 */
	len++; /* length with NUL included */
 again:
	old_tail = G.shbuf->tail;
	new_tail = old_tail + len;
	if (new_tail < G.shbuf->size) {
		/* store message, set new tail */
		memcpy(G.shbuf->data + old_tail, msg, len);
		G.shbuf->tail = new_tail;
	} else {
		/* k == available buffer space ahead of old tail */
		int k = G.shbuf->size - old_tail;
		/* copy what fits to the end of buffer, and repeat */
		memcpy(G.shbuf->data + old_tail, msg, k);
		msg += k;
		len -= k;
		G.shbuf->tail = 0;
		goto again;
	}
	if (semop(G.s_semid, G.SMwup, 1) == -1) {
		bb_perror_msg_and_die("SMwup");
	}
	if (DEBUG)
		printf("tail:%d\n", G.shbuf->tail);
}
#else
void ipcsyslog_cleanup(void);
void ipcsyslog_init(void);
void log_to_shmem(const char *msg);
#endif /* FEATURE_IPC_SYSLOG */


/* Print a message to the log file. */
static void log_locally(time_t now, char *msg)
{
#ifdef SYSLOGD_WRLOCK
	struct flock fl;
#endif
	int len = strlen(msg);

#if ENABLE_FEATURE_IPC_SYSLOG
	if ((option_mask32 & OPT_circularlog) && G.shbuf) {
		log_to_shmem(msg, len);
		return;
	}
#endif
	if (G.logFD >= 0) {
		/* Reopen log file every second. This allows admin
		 * to delete the file and not worry about restarting us.
		 * This costs almost nothing since it happens
		 * _at most_ once a second.
		 */
		if (!now)
			now = time(NULL);
		if (G.last_log_time != now) {
			G.last_log_time = now;
			close(G.logFD);
			goto reopen;
		}
	} else {
 reopen:
		G.logFD = open(G.logFilePath, O_WRONLY | O_CREAT
					| O_NOCTTY | O_APPEND | O_NONBLOCK,
					0666);
		if (G.logFD < 0) {
			/* cannot open logfile? - print to /dev/console then */
			int fd = device_open(DEV_CONSOLE, O_WRONLY | O_NOCTTY | O_NONBLOCK);
			if (fd < 0)
				fd = 2; /* then stderr, dammit */
			full_write(fd, msg, len);
			if (fd != 2)
				close(fd);
			return;
		}
#if ENABLE_FEATURE_ROTATE_LOGFILE
		{
			struct stat statf;
			G.isRegular = (fstat(G.logFD, &statf) == 0 && S_ISREG(statf.st_mode));
			/* bug (mostly harmless): can wrap around if file > 4gb */
			G.curFileSize = statf.st_size;
		}
#endif
	}

#ifdef SYSLOGD_WRLOCK
	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 1;
	fl.l_type = F_WRLCK;
	fcntl(G.logFD, F_SETLKW, &fl);
#endif

#if ENABLE_FEATURE_ROTATE_LOGFILE
	if (G.logFileSize && G.isRegular && G.curFileSize > G.logFileSize) {
		if (G.logFileRotate) { /* always 0..99 */
			int i = strlen(G.logFilePath) + 3 + 1;
			char oldFile[i];
			char newFile[i];
			i = G.logFileRotate - 1;
			/* rename: f.8 -> f.9; f.7 -> f.8; ... */
			while (1) {
				sprintf(newFile, "%s.%d", G.logFilePath, i);
				if (i == 0) break;
				sprintf(oldFile, "%s.%d", G.logFilePath, --i);
				/* ignore errors - file might be missing */
				rename(oldFile, newFile);
			}
			/* newFile == "f.0" now */
			rename(G.logFilePath, newFile);
#ifdef SYSLOGD_WRLOCK
			fl.l_type = F_UNLCK;
			fcntl(G.logFD, F_SETLKW, &fl);
#endif
			close(G.logFD);
			goto reopen;
		}
		ftruncate(G.logFD, 0);
	}
	G.curFileSize +=
#endif
			full_write(G.logFD, msg, len);
#ifdef SYSLOGD_WRLOCK
	fl.l_type = F_UNLCK;
	fcntl(G.logFD, F_SETLKW, &fl);
#endif
}

static void parse_fac_prio_20(int pri, char *res20)
{
	const CODE *c_pri, *c_fac;

	if (pri != 0) {
		c_fac = facilitynames;
		while (c_fac->c_name) {
			if (c_fac->c_val != (LOG_FAC(pri) << 3)) {
				c_fac++;
				continue;
			}
			/* facility is found, look for prio */
			c_pri = prioritynames;
			while (c_pri->c_name) {
				if (c_pri->c_val != LOG_PRI(pri)) {
					c_pri++;
					continue;
				}
				snprintf(res20, 20, "%s.%s",
						c_fac->c_name, c_pri->c_name);
				return;
			}
			/* prio not found, bail out */
			break;
		}
		snprintf(res20, 20, "<%d>", pri);
	}
}

/* len parameter is used only for "is there a timestamp?" check.
 * NB: some callers cheat and supply len==0 when they know
 * that there is no timestamp, short-circuiting the test. */
static void timestamp_and_log(int pri, char *msg, int len)
{
	char *timestamp;
	time_t now;

	/* Jan 18 00:11:22 msg... */
	/* 01234567890123456 */
	if (len < 16 || msg[3] != ' ' || msg[6] != ' '
	 || msg[9] != ':' || msg[12] != ':' || msg[15] != ' '
	) {
		time(&now);
		timestamp = ctime(&now) + 4; /* skip day of week */
	} else {
		now = 0;
		timestamp = msg;
		msg += 16;
	}
	timestamp[15] = '\0';

	if (option_mask32 & OPT_small)
		sprintf(G.printbuf, "%s %s\n", timestamp, msg);
	else {
		char res[20];
		parse_fac_prio_20(pri, res);
		sprintf(G.printbuf, "%s %.64s %s %s\n", timestamp, G.hostname, res, msg);
	}

	/* Log message locally (to file or shared mem) */
	log_locally(now, G.printbuf);
}

static void timestamp_and_log_internal(const char *msg)
{
	/* -L, or no -R */
	if (ENABLE_FEATURE_REMOTE_LOG && !(option_mask32 & OPT_locallog))
		return;
	timestamp_and_log(LOG_SYSLOG | LOG_INFO, (char*)msg, 0);
}

/* tmpbuf[len] is a NUL byte (set by caller), but there can be other,
 * embedded NULs. Split messages on each of these NULs, parse prio,
 * escape control chars and log each locally. */
static void split_escape_and_log(char *tmpbuf, int len)
{
	char *p = tmpbuf;

	tmpbuf += len;
	while (p < tmpbuf) {
		char c;
		char *q = G.parsebuf;
		int pri = (LOG_USER | LOG_NOTICE);

		if (*p == '<') {
			/* Parse the magic priority number */
			pri = bb_strtou(p + 1, &p, 10);
			if (*p == '>')
				p++;
			if (pri & ~(LOG_FACMASK | LOG_PRIMASK))
				pri = (LOG_USER | LOG_NOTICE);
		}

		while ((c = *p++)) {
			if (c == '\n')
				c = ' ';
			if (!(c & ~0x1f) && c != '\t') {
				*q++ = '^';
				c += '@'; /* ^@, ^A, ^B... */
			}
			*q++ = c;
		}
		*q = '\0';

		/* Now log it */
		if (LOG_PRI(pri) < G.logLevel)
			timestamp_and_log(pri, G.parsebuf, q - G.parsebuf);
	}
}

#ifdef SYSLOGD_MARK
static void do_mark(int sig)
{
	if (G.markInterval) {
		timestamp_and_log_internal("-- MARK --");
		alarm(G.markInterval);
	}
}
#endif

/* Don't inline: prevent struct sockaddr_un to take up space on stack
 * permanently */
static NOINLINE int create_socket(void)
{
	struct sockaddr_un sunx;
	int sock_fd;
	char *dev_log_name;

	memset(&sunx, 0, sizeof(sunx));
	sunx.sun_family = AF_UNIX;

	/* Unlink old /dev/log or object it points to. */
	/* (if it exists, bind will fail) */
	strcpy(sunx.sun_path, "/dev/log");
	/*
	dev_log_name = xmalloc_follow_symlinks("/dev/log");
	if (dev_log_name) {
		safe_strncpy(sunx.sun_path, dev_log_name, sizeof(sunx.sun_path));
		free(dev_log_name);
	}
	*/
	unlink(sunx.sun_path);

	sock_fd = xsocket(AF_UNIX, SOCK_DGRAM, 0);
	xbind(sock_fd, (struct sockaddr *) &sunx, sizeof(sunx));
	chmod("/dev/log", 0666);

	return sock_fd;
}

#if ENABLE_FEATURE_REMOTE_LOG
static int try_to_resolve_remote(void)
{
	int port, socktype;
	if ( G.tls )
	{
		port = SYSLOG_PORT_TLS;
		socktype = SOCK_STREAM;
	}
	else
	{
		port = SYSLOG_PORT_UDP;
		socktype = SOCK_DGRAM;
	}
	if (!G.remoteAddr) {
		unsigned now = monotonic_sec();

		/* Don't resolve name too often - DNS timeouts can be big */
		if ((now - G.last_dns_resolve) < DNS_WAIT_SEC)
			return -1;
		G.last_dns_resolve = now;
		G.remoteAddr = host2sockaddr(G.remoteAddrStr, port);
		if (!G.remoteAddr)
			return -1;
	}
	return socket(G.remoteAddr->u.sa.sa_family, socktype, 0);
}
#endif

/* Bug 136852/140696 */
enum WEAK_CIPHER_LEVEL
{
    WCL_ENABLE = 0,
    WCL_ENC_DES_3DES_RC4,
    WCL_ENC_SEED,
    WCL_ENC_ALL, /* Disable all weak enc. */
    WCL_MAC_MD5, 
    WCL_MAC_ALL, /* Disable all weak mac. */
    WCL_SSL_SSLV2_SSLV3,
    WCL_SSL_ALL, /* Disable all weak ssl. */
    WCL_DISABLE_ALL, /* Disable all weak cipher. */
    /* All other options can only be added following WCL_DISABLE_ALL. */
};

static void
getDisabledCipherMask(unsigned long *mkey, unsigned long *auth, unsigned long *enc, unsigned long *mac, unsigned long *ssl)
{
	char *pDisWeakCipher = nvram_get(PVALUE_DISABLE_WEAK_CIPHER);
	int nDisWeakCipher = 0;
	if ( pDisWeakCipher )
	{
		nDisWeakCipher = atoi(pDisWeakCipher);
	}

	switch ( nDisWeakCipher )
	{
	case WCL_ENC_DES_3DES_RC4:
	{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
		}
		break;
	case WCL_ENC_SEED:
		{
			*enc |= SSL_SEED;
		}
		break;
	case WCL_ENC_ALL:
		{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
			*enc |= SSL_SEED;
		}
		break;
	case WCL_MAC_MD5:
		{
			*mac |= SSL_MD5;
		}
		break;
	case WCL_MAC_ALL:
		{
			*mac |= SSL_MD5;
		}
		break;
	case WCL_SSL_SSLV2_SSLV3:
		{
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_SSL_ALL:
		{
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_DISABLE_ALL:
		{
			*enc |= SSL_DES;
			*enc |= SSL_3DES;
			*enc |= SSL_RC4;
			*enc |= SSL_SEED;
			*mac |= SSL_MD5;
			*ssl |= SSL_SSLV2;
			*ssl |= SSL_SSLV3;
		}
		break;
	case WCL_ENABLE:
	default:
		{
			return;
		}
		break;
	}
}

static void tls_cleanup( void )
{
	print_msg( "%s start\n", __FUNCTION__ );
	pthread_mutex_lock_p( &G.cleanup_lock );
	print_msg( "%s locked\n", __FUNCTION__ );

	if ( G.tls_ssl != NULL )
	{
		if ( G.tls_con_status )
		{
			print_msg( "%s TLS shutdown\n", __FUNCTION__ );
			SSL_shutdown_p( G.tls_ssl );
		}
		SSL_free_p( G.tls_ssl );
	}

	if ( G.tls_ctx != NULL )
	{
		SSL_CTX_free_p( G.tls_ctx );
	}

	if ( -1 != G.remoteFD )
	{
		close( G.remoteFD );
		G.remoteFD = -1;
	}
	G.tls_ssl = NULL;
	G.tls_ctx = NULL;
	G.tls_con_status = 0;

	pthread_mutex_unlock_p(&G.cleanup_lock);
	print_msg( " %s done\n", __FUNCTION__ );
}

void *tls_disconnect( void* dummy )
{
	char buf[ MAX_READ ];

	print_msg( "%s start monitor\n", __FUNCTION__ );
	while ( !bb_got_signal )
	{
		pthread_mutex_lock_p( &G.cleanup_lock );
		if ( G.tls_con_status && G.tls_ssl != NULL )
		{
			int read_res = SSL_read_p( G.tls_ssl, buf, MAX_READ );

			if ( read_res == 0 ) // tls disconnect case
			{
				print_msg( "%s TLS disconnected\n", __FUNCTION__ );
				tls_cleanup();
			}
			else if ( read_res < 0 )
			{
				int ssl_error = SSL_get_error_p (G.tls_ssl, read_res);

				switch ( ssl_error )
				{
					case SSL_ERROR_WANT_READ:
					case SSL_ERROR_WANT_WRITE:
						break;
					default:
						print_msg( "%s TLS disconnect error, %d\n", __FUNCTION__, ssl_error );
						tls_cleanup();
						break;
				}
			}
		}
		pthread_mutex_unlock_p(&G.cleanup_lock);

		sleep( 3 );
	}
	return NULL;
}

static int get_link_status( void )
{
	struct ifreq ifr;
	struct ethtool_value edata;
	int fd;
	int rc;
	char buf[ MAX_READ ] = {0};

	if ( ( fd = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
	{
		return 0;
	}

	memset( &ifr, 0, sizeof( ifr ) );
	edata.cmd = ETHTOOL_GLINK;

	nvram_get_safe( PVALUE_WAN_DEVICE, buf, MAX_READ );
	if ( buf[0] == 0 )
	{
		strcpy(buf, "eth0");
	}
	strncpy( ifr.ifr_name, buf, sizeof( ifr.ifr_name ) - 1 );
	ifr.ifr_data = ( char* )&edata;

	rc = ioctl( fd, SIOCETHTOOL, &ifr );
	close( fd );
	if ( rc == -1 )
	{
		return 0;
	}
	return edata.data != 0 ? 1 : 0;
}

void *tls_linkstatus( void* dummy )
{
	print_msg( "%s start monitor\n", __FUNCTION__ );

	while ( !bb_got_signal )
	{
		if ( G.tls_con_status && get_link_status() == 0 )
		{
			print_msg( "%s link down\n", __FUNCTION__ );
			tls_cleanup();
		}
		sleep( 1 );
	}
	return NULL;
}

static int tls_connect( void )
{
	char ip_addr[ INET_ADDRSTRLEN + INET6_ADDRSTRLEN ];
	char buf[MAX_READ] = {0};
	int flags;
	fd_set set;
	int width;
	struct timeval tv;
	int con_ret;
	void *addr;

	print_msg("%s start\n", __FUNCTION__);

	if ( G.remoteAddrStr )
	{
		if ( -1 == G.remoteFD )
		{
			G.remoteFD = try_to_resolve_remote();
			if ( -1 == G.remoteFD )
			{
				return -1;
			}
		}
	}
	else
	{
		return -1;
	}

	if ( NULL == G.tls_ctx )
	{
		G.tls_ctx = SSL_CTX_new_p( SSLv23_method_p() );
	}
	if ( NULL == G.tls_ctx )
	{
		return -1;
	}
	/* Bug 136852 */
	SSL_CTX_set_cb_cipher_get_disabled_p(G.tls_ctx, (long)getDisabledCipherMask);
	SSL_CTX_rebuild_cipher_list_p(G.tls_ctx);

	/* enforce CA verification in TLS mode
	 *              rfc5425
	 * 4.2.1.  Certificate-Based Authentication
	 *
	 * Both syslog transport sender (TLS client) and syslog transport
	 * receiver (TLS server) MUST implement certificate-based
	 * authentication.
	 */
	nvram_get_safe( PVALUE_VERIFY_SERVER, buf, sizeof( buf ) );
	if ( buf[0] == '1' )
	{
		buf[0] = 0;
		nvram_get_safe( PVALUE_LOAD_CA_CERT, buf, sizeof( buf ) );
		if ( SSL_CTX_load_verify_locations_p( G.tls_ctx, TLS_CA_FILE, NULL ) < 1 )
		{
			print_msg( "%s load CA failed\n", __FUNCTION__ );
			goto connect_failed;
			/* SSL_CTX_set_default_verify_paths( G.tls_ctx ); */
		}
		SSL_CTX_set_verify_p( G.tls_ctx, SSL_VERIFY_PEER, NULL );
	}

	if ( connect( G.remoteFD, &G.remoteAddr->u.sa, G.remoteAddr->len ) == -1 )
	{
		switch ( G.remoteAddr->u.sa.sa_family )
		{
		case AF_INET:
			addr = &((struct sockaddr_in*)&G.remoteAddr->u.sa)->sin_addr;
			break;
		case AF_INET6:
			addr = &((struct sockaddr_in6*)&G.remoteAddr->u.sa)->sin6_addr;
			break;
		default:
			print_msg( "%s unsupported sa family %d\n", __FUNCTION__, G.remoteAddr->u.sa.sa_family );
			goto connect_failed;
			break;
		}
		inet_ntop(G.remoteAddr->u.sa.sa_family, addr, ip_addr, INET_ADDRSTRLEN + INET6_ADDRSTRLEN);
		print_msg( "%s sock %x Cannot connect to host %s [%s] on port %d, err %d - %s\n",
					__FUNCTION__,
					G.remoteFD,
					G.remoteAddrStr,
					ip_addr,
					ntohs( get_nport( &G.remoteAddr->u.sa ) ),
					errno, strerror(errno) );
		goto connect_failed;
	}
	if ( NULL == G.tls_ssl )
	{
		G.tls_ssl = SSL_new_p( G.tls_ctx );
	}
	if ( NULL == G.tls_ssl )
	{
		goto connect_failed;
	}

	SSL_set_fd_p( G.tls_ssl, G.remoteFD );

#if 1
	if ( ( flags = fcntl( G.remoteFD, F_GETFL ) ) >= 0 )
	{
		if ( ( flags & O_NONBLOCK ) == 0 )
		{
			fcntl( G.remoteFD, F_SETFL, flags | O_NONBLOCK );
		}
	}
#endif

	while( (con_ret = SSL_connect_p( G.tls_ssl )) != 1 )
	{
		FD_ZERO(&set);
		FD_SET( G.remoteFD, &set );
		width = G.remoteFD;

		tv.tv_sec = SSL_CONNECT_TIME_OUT;
		tv.tv_usec = 0;

		switch( SSL_get_error_p( G.tls_ssl, con_ret ) )
		{
		case SSL_ERROR_WANT_READ:
			print_msg( "%s SSL connection get error SSL_ERROR_WANT_READ\n", __FUNCTION__ );
			if ( select( width + 1, &set, NULL, NULL, &tv ) <= 0 )
			{
				goto connect_failed;
			}
			break;

		case SSL_ERROR_WANT_WRITE:
			print_msg( "%s SSL connection get error SSL_ERROR_WANT_WRITE\n", __FUNCTION__ );
			break;

		default:
			goto connect_failed;
			break;
		}
	}

	print_msg( "%s TLS connected\n", __FUNCTION__ );
	G.tls_con_status = 1;
	G.tls_retry = 0;
	return 0;

connect_failed:
	tls_cleanup();
	return -1;
}

void *tls_reconnect( void* dummy )
{
	print_msg( "%s start monitor\n", __FUNCTION__ );
	while ( !bb_got_signal )
	{
		if ( !G.tls_con_status && tls_connect() < 0 )
		{
			print_msg( "TLS connect failed retry %d wait %d\n", G.tls_retry, G.tls_retry * DELAY_MULTIPLYER );
			sleep( G.tls_retry * DELAY_MULTIPLYER );
			G.tls_retry++;
		}
		sleep( 1 );
	}
	return NULL;
}

static void do_syslogd(void) NORETURN;
static void do_syslogd(void)
{
	int fd;
	int sock_fd;
#if ENABLE_FEATURE_SYSLOGD_DUP
	int last_sz = -1;
	char *last_buf;
	char *recvbuf = G.recvbuf;
#else
#define recvbuf (G.recvbuf)
#endif                              


	if( G.user )
	{
		char *log_mask = getenv( "syslog_id" ) ;
		if( log_mask )
		{
			G.filter_prefix = strdup( log_mask ) ;
			G.filter_prefix_len = strlen( G.filter_prefix ) ;
			goto filter_prefix_done;
		}
		else
		{
			/* Read filter prefix */
			fd = open( DEVICE_PRODUCT_CLASS, O_RDONLY );
			if ( fd < 0 )
			{
				goto filter_prefix_read_fail;
			}
			else
			{
				char log_msg[] = "syslogd: loaded filter prefix                ";

				G.filter_prefix = ( char * )malloc( 16 );
				memset(G.filter_prefix, 0, 16);
				if ( read( fd, G.filter_prefix, 16 ) <= 0 )
				{
					goto filter_prefix_read_fail;
				}
				G.filter_prefix_len = strlen( G.filter_prefix );
				sprintf( log_msg + 31, "%s", G.filter_prefix );
				timestamp_and_log_internal( log_msg );
				goto filter_prefix_done;
			}
		}
	}
	else
	{
		/* Read filter prefix */
		fd = open( DEVICE_PRODUCT_CLASS, O_RDONLY );
		if ( fd < 0 )
		{
			goto filter_prefix_read_fail;
		}
		else
		{
			char log_msg[] = "syslogd: loaded filter prefix                ";

			G.filter_prefix = ( char * )malloc( 16 );
			memset(G.filter_prefix, 0, 16);
			if ( read( fd, G.filter_prefix, 16 ) <= 0 )
			{
				goto filter_prefix_read_fail;
			}
			G.filter_prefix_len = strlen( G.filter_prefix );
			sprintf( log_msg + 31, "%s", G.filter_prefix );
			timestamp_and_log_internal( log_msg );
			goto filter_prefix_done;
		}
	}

filter_prefix_read_fail:
	timestamp_and_log_internal("syslogd: falling back to default filter prefix GXW42XX");
	G.filter_prefix = "GXW42XX";
	G.filter_prefix_len = 7;
filter_prefix_done:

	/* Set up signal handlers (so that they interrupt read()) */
	signal_no_SA_RESTART_empty_mask(SIGTERM, record_signo);
	signal_no_SA_RESTART_empty_mask(SIGINT, record_signo);
	//signal_no_SA_RESTART_empty_mask(SIGQUIT, record_signo);
	signal(SIGHUP, SIG_IGN);
#ifdef SYSLOGD_MARK
	signal(SIGALRM, do_mark);
	alarm(G.markInterval);
#endif
	sock_fd = create_socket();

	if (ENABLE_FEATURE_IPC_SYSLOG && (option_mask32 & OPT_circularlog)) {
		ipcsyslog_init();
	}

	timestamp_and_log_internal("syslogd started: BusyBox v" BB_VER);

	while (!bb_got_signal) {
		ssize_t sz;

#if ENABLE_FEATURE_SYSLOGD_DUP
		last_buf = recvbuf;
		if (recvbuf == G.recvbuf)
			recvbuf = G.recvbuf + MAX_READ;
		else
			recvbuf = G.recvbuf;
#endif
 read_again:
		sz = read(sock_fd, recvbuf, MAX_READ - 1);
		if (sz < 0) {
			if (!bb_got_signal)
				print_msg("read from /dev/log");
			break;
		}

		/* Drop trailing '\n' and NULs (typically there is one NUL) */
		while (1) {
			if (sz == 0)
				goto read_again;
			/* man 3 syslog says: "A trailing newline is added when needed".
			 * However, neither glibc nor uclibc do this:
			 * syslog(prio, "test")   sends "test\0" to /dev/log,
			 * syslog(prio, "test\n") sends "test\n\0".
			 * IOW: newline is passed verbatim!
			 * I take it to mean that it's syslogd's job
			 * to make those look identical in the log files. */
			if (recvbuf[sz-1] != '\0' && G.user )
				break;
			if (recvbuf[sz-1] != '\0' && recvbuf[sz-1] != '\n')
				break;
			sz--;
		}
#if ENABLE_FEATURE_SYSLOGD_DUP
		if ((option_mask32 & OPT_dup) && (sz == last_sz))
			if (memcmp(last_buf, recvbuf, sz) == 0)
				continue;
		last_sz = sz;
#endif
#if ENABLE_FEATURE_REMOTE_LOG
		/* We are not modifying log messages in any way before send */
		/* Remote site cannot trust _us_ anyway and need to do validation again */
		if (G.remoteAddrStr) {
			if (-1 == G.remoteFD) {
				G.remoteFD = try_to_resolve_remote();
				if (-1 == G.remoteFD)
					goto no_luck;
			}
			if ( G.tls )
			{
				if ( !G.tls_con_status )
				{
					goto no_luck;
				}
			}
			/* Bug 141499 */
			/* Stock syslogd sends it '\n'-terminated over network, mimic that */
			recvbuf[ sz++ ] = '\n';
			/* send message to remote logger, ignore possible error */
			/* TODO: on some errors, close and set G.remoteFD to -1
			 * so that DNS resolution and connect is retried? */
			if( G.user && strstr( recvbuf, G.filter_prefix ) )
			{
				char * tmpbuf = recvbuf ;
				if( recvbuf[0] == ':' && recvbuf[1] == '<' )
				{
					tmpbuf= &recvbuf[1] ;
					sz-- ;
				}
				if ( G.tls )
				{
					SSL_write_p( G.tls_ssl, tmpbuf, sz );
				}
				else
				{
					sendto(G.remoteFD, tmpbuf, sz, MSG_DONTWAIT,
						    &G.remoteAddr->u.sa, G.remoteAddr->len);
				}
			}
			else if ( strncmp( recvbuf, G.filter_prefix, G.filter_prefix_len ) == 0 )
			{
				if ( G.tls )
				{
					SSL_write_p( G.tls_ssl, recvbuf, sz );
				}
				else
				{
					sendto(G.remoteFD, recvbuf, sz, MSG_DONTWAIT,
						    &G.remoteAddr->u.sa, G.remoteAddr->len);
				}
			}
 no_luck: ;
		}
#endif
		if (!ENABLE_FEATURE_REMOTE_LOG || (option_mask32 & OPT_locallog)) {
			recvbuf[sz] = '\0'; /* ensure it *is* NUL terminated */
			split_escape_and_log(recvbuf, sz);
		}
	} /* while (!bb_got_signal) */

	timestamp_and_log_internal("syslogd exiting");
	puts("syslogd exiting");
	if (ENABLE_FEATURE_IPC_SYSLOG)
		ipcsyslog_cleanup();
	if ( G.tls )
	{
		tls_cleanup();
		pthread_mutex_destroy(&G.cleanup_lock);
	}
	kill_myself_with_sig(bb_got_signal);
#undef recvbuf
}

#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
static int open_libssl(void)
{
	char *error;

	ssl_handle = dlopen(LIBSSL, RTLD_LAZY);
	if (!ssl_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}
	crypto_handle = dlopen(LIBCRYPTO, RTLD_LAZY);
	if (!crypto_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}
	pthread_handle = dlopen(LIBPTHREAD, RTLD_LAZY);
	if (!pthread_handle) {
		print_msg("%s\n", dlerror());
		goto HANDLE_ERROR;
	}

	SSL_library_init_p = (SSL_library_init_t)dlsym(ssl_handle, "SSL_library_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_set_fd_p = (SSL_set_fd_t)dlsym(ssl_handle, "SSL_set_fd");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_connect_p = (SSL_connect_t)dlsym(ssl_handle, "SSL_connect");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_new_p = (SSL_new_t)dlsym(ssl_handle, "SSL_new");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_shutdown_p = (SSL_new_t)dlsym(ssl_handle, "SSL_shutdown");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_free_p = (SSL_free_t)dlsym(ssl_handle, "SSL_free");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_read_p = (SSL_read_t)dlsym(ssl_handle, "SSL_read");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_write_p = (SSL_write_t)dlsym(ssl_handle, "SSL_write");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_get_error_p = (SSL_get_error_t)dlsym(ssl_handle, "SSL_get_error");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_free_p = (SSL_CTX_free_t)dlsym(ssl_handle, "SSL_CTX_free");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_new_p = (SSL_new_t)dlsym(ssl_handle, "SSL_CTX_new");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSLv23_method_p = (SSLv23_method_t)dlsym(ssl_handle, "SSLv23_method");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_load_verify_locations_p =
		(SSL_CTX_load_verify_locations_t)dlsym(ssl_handle, "SSL_CTX_load_verify_locations");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_set_verify_p = (SSL_CTX_set_verify_t)dlsym(ssl_handle, "SSL_CTX_set_verify");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	SSL_CTX_ctrl_p = (SSL_CTX_ctrl_t)dlsym(ssl_handle, "SSL_CTX_ctrl");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
#ifdef OPENSSL_LOAD_CONF
	OpenSSL_add_all_algorithms_p =
		(OpenSSL_add_all_algorithms_t)dlsym(crypto_handle, "OPENSSL_add_all_algorithms_conf");
#else
	OpenSSL_add_all_algorithms_p =
		(OpenSSL_add_all_algorithms_t)dlsym(crypto_handle, "OPENSSL_add_all_algorithms_noconf");
#endif
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_init_p = (pthread_mutexattr_init_t)dlsym(pthread_handle, "pthread_mutexattr_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_destroy_p = (pthread_mutexattr_destroy_t)dlsym(pthread_handle, "pthread_mutexattr_destroy");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutexattr_settype_p = (pthread_mutexattr_settype_t)dlsym(pthread_handle, "pthread_mutexattr_settype");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_lock_p = (pthread_mutex_lock_t)dlsym(pthread_handle, "pthread_mutex_lock");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_unlock_p = (pthread_mutex_unlock_t)dlsym(pthread_handle, "pthread_mutex_unlock");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_init_p = (pthread_mutex_init_t)dlsym(pthread_handle, "pthread_mutex_init");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_mutex_destroy_p = (pthread_mutex_destroy_t)dlsym(pthread_handle, "pthread_mutex_destroy");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}
	pthread_create_p = (pthread_create_t)dlsym(pthread_handle, "pthread_create");
	if ((error = dlerror())) {
		print_msg("%s %d err:%s\n", __func__, __LINE__, error);
		goto HANDLE_ERROR;
	}

	return 0;
HANDLE_ERROR:
	if ( ssl_handle )
	{
		dlclose(ssl_handle);
		ssl_handle = NULL;
	}
	if ( crypto_handle )
	{
		dlclose(crypto_handle);
		crypto_handle = NULL;
	}
	if ( pthread_handle )
	{
		dlclose(pthread_handle);
		pthread_handle = NULL;
	}
	return -1;
}

static void close_libssl(void)
{
	if ( ssl_handle )
	{
		dlclose(ssl_handle);
		ssl_handle = NULL;
	}
	if ( crypto_handle )
	{
		dlclose(crypto_handle);
		crypto_handle = NULL;
	}
	if ( pthread_handle )
	{
		dlclose(pthread_handle);
		pthread_handle = NULL;
	}
}
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */

int syslogd_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
int syslogd_main(int argc UNUSED_PARAM, char **argv)
{
	char OPTION_DECL;
	int opts;
	pthread_mutexattr_t attr;

	INIT_G();
#if ENABLE_FEATURE_REMOTE_LOG
	G.last_dns_resolve = monotonic_sec() - DNS_WAIT_SEC - 1;
#endif

	/* do normal option parsing */
	opt_complementary = "=0"; /* no non-option params */
	opts = getopt32(argv, OPTION_STR, OPTION_PARAM);
#ifdef SYSLOGD_MARK
	if (opts & OPT_mark) // -m
		G.markInterval = xatou_range(opt_m, 0, INT_MAX/60) * 60;
#endif
	//if (opts & OPT_nofork) // -n
	//if (opts & OPT_outfile) // -O
	if (opts & OPT_loglevel) // -l
		G.logLevel = xatou_range(opt_l, 1, 8);
	//if (opts & OPT_small) // -S
	if (opts & OPT_user) // -U  Bug 10817 Vonage
		G.user = 1 ;
#if ENABLE_FEATURE_ROTATE_LOGFILE
	if (opts & OPT_filesize) // -s
		G.logFileSize = xatou_range(opt_s, 0, INT_MAX/1024) * 1024;
	if (opts & OPT_rotatecnt) // -b
		G.logFileRotate = xatou_range(opt_b, 0, 99);
#endif
#if ENABLE_FEATURE_IPC_SYSLOG
	if (opt_C) // -Cn
		G.shm_size = xatoul_range(opt_C, 4, INT_MAX/1024) * 1024;
#endif
	if (opts & OPT_TLS) // -T  Bug 141499
	{
		G.tls = 1;
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
		if ( open_libssl() )
		{
			exit(-1);
		}
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
	}

	/* If they have not specified remote logging, then log locally */
	if (ENABLE_FEATURE_REMOTE_LOG && !(opts & OPT_remotelog)) // -R
		option_mask32 |= OPT_locallog;

	/* Store away localhost's name before the fork */
	G.hostname = safe_gethostname();
	*strchrnul(G.hostname, '.') = '\0';

	if (!(opts & OPT_nofork)) {
		bb_daemonize_or_rexec(DAEMON_CHDIR_ROOT, argv);
	}
	umask(0);
	write_pidfile("/var/run/syslogd.pid");

	/* bug 141499 */
	if ( G.tls )
	{
		if ( SSL_library_init_p() < 0 )
		{
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
			close_libssl();
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
			bb_perror_msg_and_die( "Could not initialize the OpenSSL library!\n" );
		}
		/* tls connection */
		OpenSSL_add_all_algorithms_p();
		//ERR_load_BIO_strings();
		//ERR_load_crypto_strings();
		//SSL_load_error_strings();

		pthread_mutexattr_init_p( &attr );
		pthread_mutexattr_settype_p( &attr, PTHREAD_MUTEX_RECURSIVE );
		pthread_mutex_init_p( &G.cleanup_lock, &attr );
		pthread_mutexattr_destroy_p( &attr );

		pthread_create_p( &G.tls_disconnect_thread, NULL, tls_disconnect, NULL );
		pthread_create_p( &G.tls_reconnect_thread, NULL, tls_reconnect, NULL );
		pthread_create_p( &G.tls_linkstatus_thread, NULL, tls_linkstatus, NULL );
	}

	do_syslogd();
#if ENABLE_LOAD_LIB_DYNAMIC_SUPPORT
	close_libssl();
#endif /* ENABLE_LOAD_LIB_DYNAMIC_SUPPORT */
	/* return EXIT_SUCCESS; */
}

/* Clean up. Needed because we are included from syslogd_and_logger.c */
#undef DEBUG
#undef SYSLOGD_MARK
#undef SYSLOGD_WRLOCK
#undef G
#undef GLOBALS
#undef INIT_G
#undef OPTION_STR
#undef OPTION_DECL
#undef OPTION_PARAM
