/*
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *   Copyright (C) 2007 John Crispin <blogic@openwrt.org>
 */

#include <linux/autoconf.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/mtd/physmap.h>
#include <linux/kernel.h>
#include <linux/reboot.h>
#include <linux/platform_device.h>
#include <linux/leds.h>
#include <linux/etherdevice.h>
#include <linux/reboot.h>
#include <linux/time.h>
#include <linux/io.h>
#include <linux/i2c-gpio.h>
#ifdef CONFIG_GENERIC_GPIO
#include <linux/gpio.h>
#endif
#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/ifxmips/ifxmips.h>

#define MAX_BOARD_NAME_LEN		32
#define MAX_IFXMIPS_DEVS		9

#define SYSTEM_DANUBE_CHIPID1	0x129
#define SYSTEM_DANUBE_CHIPID2	0x12B

#define SYSTEM_TWINPASS_CHIPID	0x12D
#define SYSTEM_SVIP_CHIPID       0x169
#define SYSTEM_SVIPx_CHIPID      0x160

enum {
	EASY50712,
	EASY4010,
	ARV4519,
	EASY336,
	EASY33016,
	EASY3201EVS
};

#ifdef CONFIG_PCI
extern int ifxmips_pci_external_clock;
#endif

static u16 chiprev;
static int cmdline_mac;
char board_name[MAX_BOARD_NAME_LEN + 1] = { 0 };

struct ifxmips_system_type {
	u16	partnum;
	char name[32];
};

struct ifxmips_board {
	int type;
	char name[32];
	unsigned int system_type;
	struct platform_device **devs;
	struct resource reset_resource;
	struct resource gpiodev_resource;
#ifdef CONFIG_LEDS_GPIO
	struct gpio_led *ifxmips_leds;
	struct gpio_led *gpio_leds;
#endif
	int pci_external_clock;
	int num_devs;
};

DEFINE_SPINLOCK(ebu_lock);
EXPORT_SYMBOL_GPL(ebu_lock);

static unsigned char ifxmips_ethaddr[6];
static int ifxmips_brn;

#ifdef CONFIG_LEDS_GPIO
static struct gpio_led_platform_data ifxmips_led_data;

static struct platform_device ifxmips_led = {
	.id = 0,
	.name = "ifxmips_led",
	.dev = {
		.platform_data = (void *) &ifxmips_led_data,
	}
};
#endif

#ifdef CONFIG_GENERIC_GPIO
static struct platform_device ifxmips_gpio = {
	.id = 0,
	.name = "ifxmips_gpio",
	.num_resources = 1,
};
#endif

#if defined (CONFIG_I2C) && defined (CONFIG_IFX_BOARD_GS_GXW)
static struct i2c_gpio_platform_data i2c_gpio_adapter_data_a = {     
    .sda_pin = 31,   //P1.P11  
    .scl_pin = 33,   //P1.P13     
    .udelay = 5, //100KHz     
    .timeout = 100,     
    .sda_is_open_drain = 1,     
    .scl_is_open_drain = 1,
    .scl_is_output_only = 1  
};     

static struct platform_device ifxmips_i2c_gpio_a = {     
    .id = 1,     
    .name = "ifxmips_i2c_gpio_a",     
    .dev = {     
        .platform_data = &i2c_gpio_adapter_data_a,     
        .release = NULL, //mxs_nop_release,     
    },     
 }; 
 
static struct i2c_gpio_platform_data i2c_gpio_adapter_data_b = {     
    .sda_pin = 30,   //P1.P10 
    .scl_pin = 32,   //P1.P12     
    .udelay = 5, //100KHz     
    .timeout = 100,     
    .sda_is_open_drain = 1,     
    .scl_is_open_drain = 1,
    .scl_is_output_only = 1  
};     

static struct platform_device ifxmips_i2c_gpio_b = {     
    .id = 2,     
    .name = "ifxmips_i2c_gpio_b",     
    .dev = {     
        .platform_data = &i2c_gpio_adapter_data_b,     
        .release = NULL, //mxs_nop_release,     
    },     
 };    
#endif

static struct platform_device ifxmips_mii = {
	.id = 0,
	.name = "ifxmips_mii0",
	.dev = {
		.platform_data = ifxmips_ethaddr,
	}
};

#if defined( CONFIG_IFX_BOARD_GS_GXW )
static struct platform_device gxw42xx_kpad_device = {
    .name               = "gxw42xx_keypad",
    .id                 = 0,
    .dev                = {
        .release = NULL,
        }
};
#endif

#ifndef IFX_EASY33016
static struct platform_device svip_ve = {
	.id = 0,
	.name = "ifxmips_svip_ve",
};
#endif

static struct platform_device ifxmips_wdt = {
	.id = 0,
	.name = "ifxmips_wdt",
};

#ifdef CONFIG_MTD_IFXMIPS
static struct resource ifxmips_mtd_resource = {
	.start  = IFXMIPS_FLASH_START,
	.end    = IFXMIPS_FLASH_START + IFXMIPS_FLASH_MAX - 1,
	.flags  = IORESOURCE_MEM,
};

static struct platform_device ifxmips_mtd = {
	.id = 0,
	.name = "ifxmips_mtd",
	.num_resources  = 1,
	.resource   = &ifxmips_mtd_resource,
};
#endif

static struct platform_device ifxmips_gpio_dev = {
	.name     = "GPIODEV",
	.id     = -1,
	.num_resources = 1
};

#ifdef CONFIG_LEDS_GPIO
static struct gpio_led arv4519_gpio_leds[] = {
	{ .name = "ifx:green:power", .gpio = 3, .active_low = 1, },
	{ .name = "ifx:red:power", .gpio = 7, .active_low = 1, },
	{ .name = "ifx:green:adsl", .gpio = 4, .active_low = 1, },
	{ .name = "ifx:green:internet", .gpio = 5, .active_low = 1, },
	{ .name = "ifx:red:internet", .gpio = 8, .active_low = 1, },
	{ .name = "ifx:green:wlan", .gpio = 6, .active_low = 1, },
	{ .name = "ifx:green:usb", .gpio = 19, .active_low = 1, },
};

static struct gpio_led_platform_data ifxmips_gpio_led_data;

static struct platform_device ifxmips_gpio_leds = {
	.name = "leds-gpio",
	.id = -1,
	.dev = {
		.platform_data = (void *) &ifxmips_gpio_led_data,
	}
};
#endif

struct platform_device *easy50712_devs[] = {
#ifdef CONFIG_LEDS_GPIO
	&ifxmips_led,
#endif
#ifdef CONFIG_GENERIC_GPIO
   &ifxmips_gpio,
#endif
   &ifxmips_mii,
#ifdef CONFIG_MTD_IFXMIPS
   &ifxmips_mtd,
#endif
   &ifxmips_wdt, &ifxmips_gpio_dev
};

struct platform_device *easy4010_devs[] = {
#ifdef CONFIG_LEDS_GPIO
	&ifxmips_led,
#endif
#ifdef CONFIG_GENERIC_GPIO
   &ifxmips_gpio,
#endif
	&ifxmips_mii, &ifxmips_wdt, &ifxmips_gpio_dev
};

struct platform_device *arv5419_devs[] = {
#ifdef CONFIG_GENERIC_GPIO
	&ifxmips_gpio,
#endif
	&ifxmips_mii,
#ifdef CONFIG_MTD_IFXMIPS
   &ifxmips_mtd,
#endif
   &ifxmips_wdt,
#ifdef CONFIG_LEDS_GPIO
	&ifxmips_gpio_leds,
#endif
};

#ifdef CONFIG_SVIP
struct platform_device *svip_devs[] = {
#ifdef CONFIG_GENERIC_GPIO
   &ifxmips_gpio,
#endif
#ifdef CONFIG_I2C
   &ifxmips_i2c_gpio_a,
   &ifxmips_i2c_gpio_b,
#endif
   &ifxmips_mii,
#if defined(CONFIG_IFX_EASY336) || defined(CONFIG_IFX_BOARD_GS_GXW)
   &svip_ve,
#endif
#ifdef CONFIG_MTD_IFXMIPS
   &ifxmips_mtd,
#endif
   &ifxmips_wdt, 
   &ifxmips_gpio_dev,
   &gxw42xx_kpad_device
};
#endif

#ifdef CONFIG_LEDS_GPIO
static struct gpio_led easy50712_leds[] = {
	{ .name = "ifx:green:test0", .gpio = 0,},
	{ .name = "ifx:green:test1", .gpio = 1,},
	{ .name = "ifx:green:test2", .gpio = 2,},
	{ .name = "ifx:green:test3", .gpio = 3,},
};

static struct gpio_led easy4010_leds[] = {
	{ .name = "ifx:green:test0", .gpio = 0,},
	{ .name = "ifx:green:test1", .gpio = 1,},
	{ .name = "ifx:green:test2", .gpio = 2,},
	{ .name = "ifx:green:test3", .gpio = 3,},
};
#endif

static struct ifxmips_board boards[] = {
	{
		.type = EASY50712,
		.name = "EASY50712",
		.system_type = SYSTEM_DANUBE_CHIPID1,
		.devs = easy50712_devs,
#ifdef CONFIG_LEDS_GPIO
		.reset_resource = {.name = "reset", .start = 1, .end = 15,},
		.gpiodev_resource = { .name = "gpio",
			.start = (1 << 0) | (1 << 1),
			.end = (1 << 0) | (1 << 1)},
		.ifxmips_leds = easy50712_leds,
#endif
	}, {
		.type = EASY4010,
		.name = "EASY4010",
		.system_type = SYSTEM_TWINPASS_CHIPID,
		.devs = easy4010_devs,
#ifdef CONFIG_LEDS_GPIO
		.reset_resource = {.name = "reset", .start = 1, .end = 15},
		.gpiodev_resource = { .name = "gpio",
			.start = (1 << 0) | (1 << 1),
			.end = (1 << 0) | (1 << 1)},
		.ifxmips_leds = easy4010_leds,
#endif
	}, {
		.type = ARV4519,
		.name = "ARV4519",
		.system_type = SYSTEM_DANUBE_CHIPID2,
		.devs = arv5419_devs,
#ifdef CONFIG_LEDS_GPIO
		.reset_resource = {.name = "reset", .start = 1, .end = 14},
		.pci_external_clock = 1,
		.gpio_leds = arv4519_gpio_leds,
#endif
	},
#ifdef CONFIG_SVIP
#ifdef CONFIG_IFX_BOARD_GS_GXW
   {
		.type = EASY336,
		.name = "EASY336",
		.system_type = SYSTEM_SVIP_CHIPID,
		.devs = svip_devs,
		.gpiodev_resource = {
         .name = "gpio",
         .start = 0 |
#ifdef CONFIG_IFX_MUX_P0_12
            (1 << 12) |
#endif /* CONFIG_IFX_MUX_P0_12 */
#ifdef CONFIG_IFX_MUX_P0_13
            (1 << 13) |
#endif /* CONFIG_IFX_MUX_P0_13 */
            0,
			.end = 0 |
#ifdef CONFIG_IFX_MUX_P0_12
            (1 << 12) |
#endif /* CONFIG_IFX_MUX_P0_12 */
#ifdef CONFIG_IFX_MUX_P0_13
            (1 << 13) |
#endif /* CONFIG_IFX_MUX_P0_13 */
            0},
	},
#endif /* CONFIG_IFX_BOARD_GS_GXW */

#ifdef CONFIG_IFX_EASY336
   {
		.type = EASY336,
		.name = "EASY336",
		.system_type = SYSTEM_SVIP_CHIPID,
		.devs = svip_devs,
		.gpiodev_resource = {
         .name = "gpio",
         .start = 0 |
#ifdef CONFIG_IFX_MUX_P0_12
            (1 << 12) |
#endif /* CONFIG_IFX_MUX_P0_12 */
#ifdef CONFIG_IFX_MUX_P0_13
            (1 << 13) |
#endif /* CONFIG_IFX_MUX_P0_13 */
            0,
			.end = 0 |
#ifdef CONFIG_IFX_MUX_P0_12
            (1 << 12) |
#endif /* CONFIG_IFX_MUX_P0_12 */
#ifdef CONFIG_IFX_MUX_P0_13
            (1 << 13) |
#endif /* CONFIG_IFX_MUX_P0_13 */
            0},
	},
#endif /* CONFIG_IFX_EASY336 */
#ifdef CONFIG_IFX_EASY33016
   {
		.type = EASY33016,
		.name = "EASY33016",
		.system_type = SYSTEM_SVIP_CHIPID,
		.devs = svip_devs,
		.gpiodev_resource = {
         .name = "gpio",
         .start = 0 |
#ifdef CONFIG_IFX_MUX_P0_9
            (1 << 9) |
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
            (1 << 10) |
#endif /* CONFIG_IFX_MUX_P0_10 */
            0,
			.end = 0 |
#ifdef CONFIG_IFX_MUX_P0_9
            (1 << 9) |
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
            (1 << 10) |
#endif /* CONFIG_IFX_MUX_P0_10 */
            0},
	},
#endif /* CONFIG_IFX_EASY33016 */
#ifdef CONFIG_IFX_EASY3201EVS
   {
		.type = EASY3201EVS,
		.name = "EASY3201EVS",
		.system_type = SYSTEM_SVIP_CHIPID,
		.devs = svip_devs,
		.gpiodev_resource = {
         .name = "gpio",
         .start = 0 |
#ifdef CONFIG_IFX_MUX_P0_9
            (1 << 9) |
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
            (1 << 10) |
#endif /* CONFIG_IFX_MUX_P0_10 */
            0,
			.end = 0 |
#ifdef CONFIG_IFX_MUX_P0_9
            (1 << 9) |
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
            (1 << 10) |
#endif /* CONFIG_IFX_MUX_P0_10 */
            0},
	},
#endif /* CONFIG_IFX_EASY3201EVS */
#endif /* CONFIG_SVIP */
};

static struct ifxmips_system_type system_type_list[] = {
	{
		.partnum = 0x129,
		.name = "Danube-S",
	}, {
		.partnum = 0x12B,
		.name = "Danube",
	}, {
		.partnum = 0x12C,
		.name = "Vinax-VE",
	}, {
		.partnum = 0x12D,
		.name = "Twinpass-VE",
	}, {
		.partnum = 0x12E,
		.name = "Twinpass-E",
	}, {
		.partnum = 0x150,
		.name = "Amazon-S",
	}, {
		.partnum = 0x151,
		.name = "Vinax-E",
	}, {
		.partnum = 0x152,
		.name = "Amazon-SE",
	}, {
		.partnum = 0x153,
		.name = "Amazon-SE",
	}, {
		.partnum = 0x169,
		.name = "SVIP",
	}, {
		.partnum = 0x16C,
		.name = "ARX188",
	}, {
		.partnum = 0x16D,
		.name = "ARX168",
	}, {
		.partnum = 0x16F,
		.name = "ARX182",
	}, {
		.partnum = 0x170,
		.name = "GRX188",
	}, {
		.partnum = 0x171,
		.name = "GRX168",
	},
};

static inline u16 get_chip_partnum(void)
{
#ifndef CONFIG_SVIP
	return (ifxmips_r32(IFXMIPS_MPS_CHIPID) & 0x0FFFF000) >> 12;
#else
   return (ifxmips_r32(IFXMIPS_STATUS_CHIPID) & 0x0FFFF000) >> 12;
#endif
}

const char *get_system_type(void)
{
	int i;
	u16 partnum = get_chip_partnum();

	for (i = 0; i < ARRAY_SIZE(system_type_list); i++)
		if (system_type_list[i].partnum == partnum)
			return system_type_list[i].name;

	return BOARD_SYSTEM_TYPE;
}

static int __init ifxmips_set_board_type(char *str)
{
	str = strchr(str, '=');
	if (!str)
		goto out;
	str++;
	if (strlen(str) > MAX_BOARD_NAME_LEN)
		goto out;
	strncpy(board_name, str, MAX_BOARD_NAME_LEN);
	printk(KERN_INFO "bootloader told us, that this is a %s board\n",
		board_name);
out:
	return 1;
}
__setup("ifxmips_board", ifxmips_set_board_type);

static int __init ifxmips_set_ethaddr(char *str)
{
#define IS_HEX(x) \
	(((x >= '0' && x <= '9') || (x >= 'a' && x <= 'f') \
		|| (x >= 'A' && x <= 'F')) ? (1) : (0))
	int i;
	str = strchr(str, '=');
	if (!str)
		goto out;
	str++;
	if (strlen(str) != 17)
		goto out;
	for (i = 0; i < 6; i++) {
		if (!IS_HEX(str[3 * i]) || !IS_HEX(str[(3 * i) + 1]))
			goto out;
		if ((i != 5) && (str[(3 * i) + 2] != ':'))
			goto out;
		ifxmips_ethaddr[i] = simple_strtoul(&str[3 * i], NULL, 16);
	}
	if (is_valid_ether_addr(ifxmips_ethaddr))
		cmdline_mac = 1;
out:
	return 1;
}
__setup("ethaddr", ifxmips_set_ethaddr);

int ifxmips_replace_ethaddr(char *mac_bytes)
{
    memcpy(ifxmips_ethaddr, mac_bytes, 6);
}

int ifxmips_find_brn_block(void)
{
	unsigned char temp[8];
	memcpy_fromio(temp,
		(void *)KSEG1ADDR(IFXMIPS_FLASH_START + 0x800000 - 0x10000), 8);
	if (memcmp(temp, "BRN-BOOT", 8) == 0) {
		if (!cmdline_mac)
			memcpy_fromio(ifxmips_ethaddr,
				(void *)KSEG1ADDR(IFXMIPS_FLASH_START +
					0x800000 - 0x10000 + 0x16), 6);
		if (is_valid_ether_addr(ifxmips_ethaddr))
			cmdline_mac = 1;
		return 1;
	} else {
		return 0;
	}
}

int ifxmips_has_brn_block(void)
{
	return ifxmips_brn;
}
EXPORT_SYMBOL(ifxmips_has_brn_block);

struct ifxmips_board *ifxmips_find_board(void)
{
	int i;
	if (!*board_name)
		return 0;
	for (i = 0; i < ARRAY_SIZE(boards); i++)
		if ((boards[i].system_type == chiprev) &&
		    (!strcmp(boards[i].name, board_name)))
			return &boards[i];
	return 0;
}

int __init ifxmips_init_devices(void)
{
	struct ifxmips_board *board = ifxmips_find_board();

	chiprev = get_chip_partnum();
	ifxmips_brn = ifxmips_find_brn_block();

	if (!cmdline_mac)
		random_ether_addr(ifxmips_ethaddr);

	if (!board) {
		switch (chiprev) {
		case SYSTEM_DANUBE_CHIPID1:
		case SYSTEM_DANUBE_CHIPID2:
		default:
			board = &boards[0];
#ifdef CONFIG_SVIP
            printk("SVIP CHIP ID IS INVALID (0x%04X)\n", chiprev);
#endif
			break;
		case SYSTEM_TWINPASS_CHIPID:
			board = &boards[1];
			break;
#ifdef CONFIG_SVIP
		case SYSTEM_SVIP_CHIPID:
		case SYSTEM_SVIPx_CHIPID:
			board = &boards[3];
			break;
#endif
		}
	}

#ifdef CONFIG_GENERIC_GPIO
	switch (board->type) {
	case EASY50712:
		board->num_devs = ARRAY_SIZE(easy50712_devs);
#ifdef CONFIG_LEDS_GPIO
		ifxmips_led_data.num_leds = ARRAY_SIZE(easy50712_leds);
#endif
		break;
	case EASY4010:
		board->num_devs = ARRAY_SIZE(easy4010_devs);
#ifdef CONFIG_LEDS_GPIO
		ifxmips_led_data.num_leds = ARRAY_SIZE(easy4010_leds);
#endif
		break;
	case ARV4519:
		gpio_set_value(3, 0);
		gpio_set_value(4, 0);
		gpio_set_value(5, 0);
		gpio_set_value(6, 0);
		gpio_set_value(7, 1);
		gpio_set_value(8, 1);
		gpio_set_value(19, 0);
		board->num_devs = ARRAY_SIZE(arv5419_devs);
#ifdef CONFIG_LEDS_GPIO
		ifxmips_gpio_led_data.num_leds = ARRAY_SIZE(arv4519_gpio_leds);
#endif
		break;
#ifdef CONFIG_SVIP
	case EASY336:
#ifdef CONFIG_IFX_MUX_P0_12
		gpio_set_value(12, 1);
#endif /* CONFIG_IFX_MUX_P0_12 */
#ifdef CONFIG_IFX_MUX_P0_13
		gpio_set_value(13, 1);
#endif /* CONFIG_IFX_MUX_P0_13 */
		board->num_devs = ARRAY_SIZE(svip_devs);
		break;
	case EASY33016:
#ifdef CONFIG_IFX_MUX_P0_9
		gpio_set_value(9, 1);
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
		gpio_set_value(10, 1);
#endif /* CONFIG_IFX_MUX_P0_10 */
		board->num_devs = ARRAY_SIZE(svip_devs);
		break;
	case EASY3201EVS:
#ifdef CONFIG_IFX_MUX_P0_9
		gpio_set_value(9, 1);
#endif /* CONFIG_IFX_MUX_P0_9 */
#ifdef CONFIG_IFX_MUX_P0_10
		gpio_set_value(10, 1);
#endif /* CONFIG_IFX_MUX_P0_10 */
		board->num_devs = ARRAY_SIZE(svip_devs);
		break;
#endif /* CONFIG_SVIP */
	}
#ifdef CONFIG_LEDS_GPIO
	ifxmips_gpio_led_data.leds = board->gpio_leds;
	ifxmips_led_data.leds = board->ifxmips_leds;
#endif

	ifxmips_gpio.resource = &board->reset_resource;
	ifxmips_gpio_dev.resource = &board->gpiodev_resource;
#endif

	printk(KERN_INFO "%s: adding %d devs\n",
		__func__, board->num_devs);

#ifdef CONFIG_PCI
	if (board->pci_external_clock)
		ifxmips_pci_external_clock = 1;
#endif

	printk(KERN_INFO "using board definition %s\n", board->name);
	return platform_add_devices(board->devs, board->num_devs);
}

arch_initcall(ifxmips_init_devices);
