/******************************************************************************
     Copyright (c) 2007, Infineon Technologies.  All rights reserved.

                               No Warranty
   Because the program is licensed free of charge, there is no warranty for
   the program, to the extent permitted by applicable law.  Except when
   otherwise stated in writing the copyright holders and/or other parties
   provide the program "as is" without warranty of any kind, either
   expressed or implied, including, but not limited to, the implied
   warranties of merchantability and fitness for a particular purpose. The
   entire risk as to the quality and performance of the program is with
   you.  should the program prove defective, you assume the cost of all
   necessary servicing, repair or correction.

   In no event unless required by applicable law or agreed to in writing
   will any copyright holder, or any other party who may modify and/or
   redistribute the program as permitted above, be liable to you for
   damages, including any general, special, incidental or consequential
   damages arising out of the use or inability to use the program
   (including but not limited to loss of data or data being rendered
   inaccurate or losses sustained by you or third parties or a failure of
   the program to operate with any other programs), even if such holder or
   other party has been advised of the possibility of such damages.
 ******************************************************************************
   Module      : switchip_setup.c
   Date        : 2007-11-09
   Description : Basic setup of embedded ethernet switch "SwitchIP"
   Remarks: andreas.schmidt@infineon.com

 *****************************************************************************/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/delay.h>
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
/* applicable to Linux 2.4 version of SVIP kernel */
#include <linux/tqueue.h>
#else
#include <linux/workqueue.h>
#endif
#include <linux/time.h>

#include <asm/ifx/ifx_reg_base.h>
#include <asm/ifx/ifx_reg_es.h>
#include <asm/ifx/ifx_reg_sys1.h>
#include <asm/ifx/ifx_reg_dma.h>
#include <linux/proc_fs.h>
#include <gs_def.h>

#if 0
#define DMA_RX_CHANNEL_0_SELECT      0x00000000
#define DMA_TX_CHANNEL_1_SELECT      0x00000001
#endif

#define SINGLE_PORT

/* PHY Organizationally Unique Identifier (OUI) */
#define PHY_OUI_PMC           0x00E004
#define PHY_OUI_VITESSE       0x008083
#define PHY_OUI_REALTEK       0x00E04C
#define PHY_OUI_DEFAULT       0xFFFFFF

#define RTL8211F_INER_LINK_STATUS 0x0010
#define RTL8211F_INSR		0x1d
#define RTL8211F_PAGE_SELECT	0x1f
#define RTL8211F_TX_DELAY	0x100
#define RTL8211F_LED_REG	0x10
#define RTL8211F_LED_EEE	0x11
#define RTL8211F_GBCR		0x09
#define RTL8211F_BMCR		0x00
#define RTL8211F_PHYCR2		0x19

unsigned short switchip_phy_read(unsigned int phyaddr, unsigned int regaddr);
void switchip_phy_write(unsigned int phyaddr, unsigned int regaddr, unsigned short data);
void switchchip_adjust_timeskew(int phy, unsigned char phy_tx, unsigned char phy_rx,
                                           int sw, unsigned char sw_tx, unsigned char sw_rx);

extern struct gsboard_hw_sw_limit *board_hw_sw_ver;

static int phy_address[2] = {0,1};
static u32 gs_phy_oui;
static u32 phy_oui;
#define CONFIG_IFX_MDIO_POLLING
#ifdef CONFIG_IFX_MDIO_POLLING
static void switchip_mdio_poll_init (void);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void switchip_mdio_poll_schedule (void);
static void _switchip_mdio_poll(void *arg);
static void switchip_mdio_speed_delay_poll_schedule (void);
static void _switchip_mdio_speed_delay_poll(void *arg);
#else
static void _switchip_mdio_poll(struct work_struct *work);
static void _switchip_mdio_speed_delay_poll(struct work_struct *work);
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static struct tq_struct mdio_poll_task;
static struct timer_list mdio_poll_timer;
static struct tq_struct mdio_speed_delay_poll_task;
static struct timer_list mdio_speed_delay_poll_timer;
#else
/* struct workqueue_struct mdio_poll_task; */
static struct workqueue_struct *mdio_poll_workqueue;
static struct workqueue_struct *mdio_speed_delay_poll_workqueue;
DECLARE_DELAYED_WORK(mdio_poll_work, _switchip_mdio_poll);
DECLARE_DELAYED_WORK(mdio_speed_delay_poll_work, _switchip_mdio_speed_delay_poll);
#endif
int old_link_status[2] = {-1,-1};
int net_linked = 0;
int old_phy_pause[2] = {-1,-1};
int old_phy_speed[2] = {-1,-1};
int old_phy_duplex[2] = {-1,-1};
int phy_speed_changed[2] = {0,0};

/**
 * Autonegotiation check.
 * This funtion checks for link changes. If a link change has occured it will
 * update certain switch registers.
 */
static void _switchip_check_phy_status(int port)
{
   int new_link_status;
   unsigned short reg1;

   reg1 = switchip_phy_read(phy_address[port], 1);
   if ((reg1 == 0xFFFF) || (reg1 == 0x0000))
   {
      return; /* no PHY connected */
   }

   new_link_status = reg1 & 4;
   if (old_link_status[port] ^ new_link_status) /* link status change */
   {
      printk("\n-----------------------------------------\nNetwork port %d status changed\n", port);
      if (!new_link_status)
      {
         if (port == 0)
            IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P0_CTL_REG, 0);
         else
            IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P1_CTL_REG, 0);

         /* read again; link bit is latched low! */
         reg1 = switchip_phy_read(phy_address[port], 1);
         new_link_status = reg1 & 4;
      }

      if (new_link_status)
      {
         unsigned short reg0, reg4, reg5, reg9, reg10;
         int phy_pause, phy_speed, phy_duplex;
         int aneg_enable, aneg_cmpt;

         reg0 = switchip_phy_read(phy_address[port], 0);
         reg4 = switchip_phy_read(phy_address[port], 4);
         aneg_enable = reg0 & 0x1000;
         aneg_cmpt = reg1 & 0x20;

         printk("Auto-Negotiation: %s, Completed: %s\n", aneg_enable? "Yes": "No", aneg_cmpt? "Yes":"No");

         if (aneg_enable && aneg_cmpt)
         {
            reg5 = switchip_phy_read(phy_address[port], 5);
            switch (phy_oui)
            {
#if defined(CONFIG_IFX_EASY336) || defined (CONFIG_IFX_BOARD_GS_GXW)
               case PHY_OUI_PMC:
                  /* PMC Sierra supports 1Gigabit FD, only. On successfull
                     auto-negotiation, we are sure this is what the LP can. */
                  phy_pause = ((reg4 & reg5) & 0x0080) >> 7;
                  phy_speed = PHY_SPEED_1000M;
                  phy_duplex = PHY_DUPLEX_FULL;
                  break;
#endif
               case PHY_OUI_VITESSE:
               case PHY_OUI_DEFAULT:
                  reg9 = switchip_phy_read(phy_address[port], 9);
                  reg10 = switchip_phy_read(phy_address[port], 10);

                  /* Check if advertise and partner agree on pause */
                  phy_pause = ((reg4 & reg5) & 0x0400) >> 10;

                  /* Find the best mode both partners support */
                  /* Priority: 1GB-FD, 1GB-HD, 100MB-FD, 100MB-HD, 10MB-FD, 10MB-HD */
                  phy_speed = ((((reg9<<2) & reg10) & 0x0c00) >> 6) |
                              (((reg4 & reg5) & 0x01e0) >> 5);

                  if (phy_speed >= 0x0020) {
                     phy_speed = PHY_SPEED_1000M;
                     phy_duplex = PHY_DUPLEX_FULL;
                  }
                  else if (phy_speed >= 0x0010) {
                     phy_speed = PHY_SPEED_1000M;
                     phy_duplex = PHY_DUPLEX_HALF;
                  }
                  else if (phy_speed >= 0x0008) {
                     phy_speed = PHY_SPEED_100M;
                     phy_duplex = PHY_DUPLEX_FULL;
                  }
                  else if (phy_speed >= 0x0004) {
                     phy_speed = PHY_SPEED_100M;
                     phy_duplex = PHY_DUPLEX_HALF;
                  }
                  else if (phy_speed >= 0x0002) {
                     phy_speed = PHY_SPEED_10M;
                     phy_duplex = PHY_DUPLEX_FULL;
                  }
                  else {
                     phy_speed = PHY_SPEED_10M;
                     phy_duplex = PHY_DUPLEX_HALF;
                  }
                  break;
               default:
                  phy_pause = (reg4 & 0x0400) >> 10;
                  phy_speed = (reg0 & 0x40 ? PHY_SPEED_1000M : (reg0 >> 13)&1);
                  phy_duplex = (reg0 >> 8)&1;
                  break;
            }
         }
         else /* parallel detection or fixed speed */
         {
            phy_pause = (reg4 & 0x0400) >> 10;
            phy_speed = (reg0 & 0x40 ? PHY_SPEED_1000M : (reg0 >> 13)&1);
            phy_duplex = (reg0 >> 8)&1;
         }
         printk("NETWORK: ON, %s Mbps %s duplex\n",
                 (phy_speed == PHY_SPEED_1000M)? "1000" : ((phy_speed == PHY_SPEED_100M)? "100" : "10"),
                 phy_duplex? "Full" : "Half");

         if ( old_phy_speed[port] != phy_speed )
         {
            phy_speed_changed[port] = 1;
         }
         old_phy_speed[port] = phy_speed;
         old_phy_duplex[port] = phy_duplex;
         old_phy_pause[port] = phy_pause;

         if (port == 0)
         {
            IFX_ES_RGMII_CTL_REG_P0SPD_SET(*IFX_ES_RGMII_CTL_REG, phy_speed);
            IFX_ES_RGMII_CTL_REG_P0DUP_SET(*IFX_ES_RGMII_CTL_REG, phy_duplex);
            IFX_ES_RGMII_CTL_REG_P0FCE_SET(*IFX_ES_RGMII_CTL_REG, 0 /*phy_pause */);
            IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P0_CTL_REG, 1);
         }
         else
         {
            IFX_ES_RGMII_CTL_REG_P1SPD_SET(*IFX_ES_RGMII_CTL_REG, phy_speed);
            IFX_ES_RGMII_CTL_REG_P1DUP_SET(*IFX_ES_RGMII_CTL_REG, phy_duplex);
            IFX_ES_RGMII_CTL_REG_P1FCE_SET(*IFX_ES_RGMII_CTL_REG, phy_pause);
            IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P1_CTL_REG, 1);
         }
      }
      else
      {
         printk("NETWORK: OFF, (Was %s Mbps %s duplex, 0x%08X)\n",
                 (old_phy_speed[port] == PHY_SPEED_1000M)? "1000" : ((old_phy_speed[port] == 1)? "100" : "10"),
                 old_phy_duplex[port]? "Full" : "Half", old_phy_pause[port]);
      }
      printk("-----------------------------------------\n\n");

      {
        extern void gxw_kpad_notify_link(int); 
        net_linked = new_link_status? 1: 0;
        gxw_kpad_notify_link(net_linked? 0: 1);    // in state machine we use 0 for connected
      }
   }
   old_link_status[port] = new_link_status;
}

static void _switchip_check_phy_speed_status(int port)
{
   if ( phy_speed_changed[port] )
   {
      if ( board_hw_sw_ver )
      {
         switch ( old_phy_speed[port] )
         {
         case PHY_SPEED_1000M:
            {
               switchchip_adjust_timeskew( board_hw_sw_ver->phy.enable,
                  board_hw_sw_ver->phy.s1000.tx, board_hw_sw_ver->phy.s1000.rx,
                  board_hw_sw_ver->sw.enable,
                  board_hw_sw_ver->sw.s1000.tx, board_hw_sw_ver->sw.s1000.rx);
            }
            break;
         case PHY_SPEED_100M:
            {
               switchchip_adjust_timeskew( board_hw_sw_ver->phy.enable,
                  board_hw_sw_ver->phy.s100.tx, board_hw_sw_ver->phy.s100.rx,
                  board_hw_sw_ver->sw.enable,
                  board_hw_sw_ver->sw.s100.tx, board_hw_sw_ver->sw.s100.rx);
            }
            break;
         case PHY_SPEED_10M:
         default:
            {
               switchchip_adjust_timeskew( board_hw_sw_ver->phy.enable,
                  board_hw_sw_ver->phy.s10.tx, board_hw_sw_ver->phy.s10.rx,
                  board_hw_sw_ver->sw.enable,
                  board_hw_sw_ver->sw.s10.tx, board_hw_sw_ver->sw.s10.rx);
            }
            break;
         }
         phy_speed_changed[port] = 0;
      }
   }
}

int switchchip_link_status(void)
{
    return (net_linked? 0: 1);  // in state machine we use 0 for connected
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void _switchip_mdio_poll(void *arg)
#else
static void _switchip_mdio_poll(struct work_struct *work)
#endif
{
   if (IFX_ES_SW_GCTL0_REG_SE_GET(*IFX_ES_SW_GCTL0_REG) == 1)
   {
      _switchip_check_phy_status(0);
      _switchip_check_phy_status(1);
   }

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   switchip_mdio_poll_schedule ();
#else
   queue_delayed_work(mdio_poll_workqueue, &mdio_poll_work, HZ/2);
#endif
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void _switchip_mdio_speed_delay_poll(void *arg)
#else
static void _switchip_mdio_speed_delay_poll(struct work_struct *work)
#endif
{
   /* only need to check port 0 */
   _switchip_check_phy_speed_status(0);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   switchip_mdio_speed_delay_poll_schedule ();
#else
   queue_delayed_work(mdio_speed_delay_poll_workqueue, &mdio_speed_delay_poll_work, HZ/2);
#endif
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void switchip_mdio_poll_timer_call_back (unsigned long arg)
{
   /* do the operation in process context, not in interrupt context */
   schedule_task(&mdio_poll_task);
}
static void switchip_mdio_speed_delay_poll_timer_call_back (unsigned long arg)
{
   /* do the operation in process context, not in interrupt context */
   schedule_task(&mdio_speed_delay_poll_task);
}
#endif

static void switchip_mdio_poll_init (void)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   init_timer(&mdio_poll_timer);
   mdio_poll_timer.function = switchip_mdio_poll_timer_call_back;
   mdio_poll_timer.data = (unsigned long)NULL;

   memset(&mdio_poll_task, 0, sizeof(struct tq_struct));
   mdio_poll_task.routine = _switchip_mdio_poll;
   mdio_poll_task.data = NULL;

   switchip_mdio_poll_schedule ();
   if ( gs_phy_oui != PHY_OUI_REALTEK ) {
      init_timer(&mdio_speed_delay_poll_timer);
      mdio_speed_delay_poll_timer.function = switchip_mdio_speed_delay_poll_timer_call_back;
      mdio_speed_delay_poll_timer.data = (unsigned long)NULL;
      memset(&mdio_speed_delay_poll_task, 0, sizeof(struct tq_struct));
      mdio_speed_delay_poll_task.routine = _switchip_mdio_speed_delay_poll;
      mdio_speed_delay_poll_task.data = NULL;
      switchip_mdio_speed_delay_poll_schedule ();
   }
#else
   mdio_poll_workqueue = create_workqueue("SVIP MDIP poll");
   INIT_DELAYED_WORK(&mdio_poll_work, _switchip_mdio_poll);
   queue_delayed_work(mdio_poll_workqueue, &mdio_poll_work, HZ/2);
   if ( gs_phy_oui != PHY_OUI_REALTEK ) {
      mdio_speed_delay_poll_workqueue = create_workqueue("SVIP MDIP SPEED poll");
      INIT_DELAYED_WORK(&mdio_speed_delay_poll_work, _switchip_mdio_speed_delay_poll);
      queue_delayed_work(mdio_speed_delay_poll_workqueue, &mdio_speed_delay_poll_work, HZ/2);
   }

#endif

}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void switchip_mdio_poll_schedule (void)
{
   if (timer_pending(&mdio_poll_timer))
   {
      if (mod_timer(&mdio_poll_timer, jiffies + HZ/2))
         return;
   }

   mdio_poll_timer.expires = jiffies + HZ/2;
   add_timer(&mdio_poll_timer);
}

static void switchip_mdio_speed_delay_poll_schedule (void)
{
   if (timer_pending(&mdio_speed_delay_poll_timer))
   {
      if (mod_timer(&mdio_speed_delay_poll_timer, jiffies + HZ/2))
         return;
   }

   mdio_speed_delay_poll_timer.expires = jiffies + HZ/2;
   add_timer(&mdio_speed_delay_poll_timer);
}
#endif
#endif /* CONFIG_IFX_MDIO_POLLING */

void switchchip_adjust_timeskew(int phy, unsigned char phy_tx, unsigned char phy_rx,
                                           int sw, unsigned char sw_tx, unsigned char sw_rx)
{
   unsigned short reg17, reg0, value;
   unsigned char tx, rx, change;

   if ( phy )
   {
      change = 0;
      reg0 = switchip_phy_read(phy_address[0], 0x0);
      reg17 = switchip_phy_read(phy_address[0], 0x17);
      tx = (reg17 >> 8) & 0x7;
      rx = (reg17 >> 12) & 0x7;
      if ( ( /*phy_tx >= 0 &&*/ phy_tx <= 7 ) && ( tx ^ phy_tx ) )
      {
         reg17 &= ~(7 << 8);
         reg17 |= phy_tx << 8;
         printk("[PHY] TxDelay: %d/10 ns\n", phy_tx * 5);
         change = 1;
      }
      if ( ( /*phy_rx >= 0 &&*/ phy_rx <= 7 ) && ( rx ^ phy_rx ) )
      {
         reg17 &= ~(7 << 12);
         reg17 |= phy_rx << 12;
         printk("[PHY] RxDelay: %d/10 ns\n", phy_rx * 5);
         change = 1;
      }
      if ( change )
      {
         switchip_phy_write(phy_address[0], 0x17, reg17);
         if ( reg0 & (1 << 12) )
         {
            reg0 |= (1 << 9); /* restart auto negotiation to apply the change */
            switchip_phy_write(phy_address[0], 0, reg0);
         }
      }
   }
   if ( sw )
   {
      change = 0;
      tx = IFX_ES_RGMII_CTL_REG_P0TDLY_GET(*IFX_ES_RGMII_CTL_REG);
      rx = IFX_ES_RGMII_CTL_REG_P0RDLY_GET(*IFX_ES_RGMII_CTL_REG);
      if ( ( /*sw_tx >= 0 &&*/ sw_tx <= 3 ) && ( sw_tx ^ tx ) )
      {
         IFX_ES_RGMII_CTL_REG_P0TDLY_SET(*IFX_ES_RGMII_CTL_REG, sw_tx);
         change = 1;
      }
      if ( ( /*sw_rx >= 0 &&*/ sw_rx <= 3 ) && ( sw_rx ^ rx ) )
      {
         IFX_ES_RGMII_CTL_REG_P0RDLY_SET(*IFX_ES_RGMII_CTL_REG, sw_rx);
         change = 1;
      }
      if ( change )
      {
         value = IFX_ES_RGMII_CTL_REG_P0RDLY_GET(*IFX_ES_RGMII_CTL_REG);
         printk("[SW P0] Ctrl RxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
         value = IFX_ES_RGMII_CTL_REG_P0TDLY_GET(*IFX_ES_RGMII_CTL_REG);
         printk("[SW P0] Ctrl TxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
      }
   }
}

unsigned short switchip_phy_read(unsigned int phyaddr, unsigned int regaddr)
{
   /* TODO: protect MDIO access with semaphore */
   *IFX_ES_MDIO_CTL_REG = IFX_ES_MDIO_CTL_REG_MBUSY
                        | IFX_ES_MDIO_CTL_REG_OP_VAL(2) /* read operation */
                        | IFX_ES_MDIO_CTL_REG_PHYAD_VAL(phyaddr)
                        | IFX_ES_MDIO_CTL_REG_REGAD_VAL(regaddr);
   while (*IFX_ES_MDIO_CTL_REG & IFX_ES_MDIO_CTL_REG_MBUSY);

   return (*IFX_ES_MDIO_DATA_REG & 0xFFFF);
}

void switchip_phy_write(unsigned int phyaddr, unsigned int regaddr, unsigned short data)
{
   /* TODO: protect MDIO access with semaphore */
   *IFX_ES_MDIO_CTL_REG = IFX_ES_MDIO_CTL_REG_WD_VAL(data)
                        | IFX_ES_MDIO_CTL_REG_MBUSY
                        | IFX_ES_MDIO_CTL_REG_OP_VAL(1) /* write operation */
                        | IFX_ES_MDIO_CTL_REG_PHYAD_VAL(phyaddr)
                        | IFX_ES_MDIO_CTL_REG_REGAD_VAL(regaddr);
   while (*IFX_ES_MDIO_CTL_REG & IFX_ES_MDIO_CTL_REG_MBUSY);

   return;
}

const static u32 switch_reset_offset_000[] = {
/*b8000000:*/ 0xffffffff, 0x00000001, 0x00000001, 0x00000003,
/*b8000010:*/ 0x04070001, 0x04070001, 0x04070001, 0xffffffff,
/*b8000020:*/ 0x00001be8, 0x00001be8, 0x00001be8, 0xffffffff,
/*b8000030:*/ 0x00000000, 0x00000000, 0x00080004, 0x00020001,
/*b8000040:*/ 0x00000000, 0x00000000, 0x00080004, 0x00020001,
/*b8000050:*/ 0x00000000, 0x00000000, 0x00080004, 0x00020001,
/*b8000060:*/ 0x00000000, 0x00000000, 0x00081000, 0x001f7777,
/*b8000070:*/ 0x00000000, 0x00000000, 0x0c00ac2b, 0x0000fa50,
/*b8000080:*/ 0x00001000, 0x00001800, 0x00000000, 0x00000000,
/*b8000090:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b80000a0:*/ 0x00000000, 0x00000050, 0x00000010, 0x00000000,
/*b80000b0:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b80000c0:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b80000d0:*/ 0xffffffff, 0x00000000, 0x00000000
};
const static u32 switch_reset_offset_100[] = {
/*b8000100:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000110:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000120:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000130:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000140:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000150:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000160:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000170:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000180:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b8000190:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b80001a0:*/ 0x00000000, 0x00000000, 0x00000000, 0x00000000,
/*b80001b0:*/ 0x00000000, 0x00000000
};

/*
 * Switch Reset.
 */
void switchip_reset(void)
{
   volatile unsigned int *reg;
   volatile unsigned int rdreg;
   int i;

   *IFX_SYS1_CLKENR = IFX_SYS1_CLKENR_ETHSW;
   asm("sync");
#if 0
   *IFX_SYS1_CLKENR = IFX_SYS1_CLKENR_DMA;
   asm("sync");
#endif

   /* disable P0 */
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P0_CTL_REG, 1);
   /* disable P1 */
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P1_CTL_REG, 1);
   /* disable P2 */
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P2_CTL_REG, 1);

   /**************************************
    * BEGIN: Procedure to clear MAC table
    **************************************/
   for (i=0; i<3; i++)
   {
      int result;

      /* check if access engine is available */
      while (*IFX_ES_ADR_TB_ST2_REG & IFX_ES_ADR_TB_ST2_REG_BUSY);

      /* initialise to first address */
      *IFX_ES_ADR_TB_CTL2_REG = IFX_ES_ADR_TB_CTL2_REG_CMD_VAL(3)
                              | IFX_ES_ADR_TB_CTL2_REG_AC_VAL(0);

      /* wait while busy */
      while (*IFX_ES_ADR_TB_ST2_REG & IFX_ES_ADR_TB_ST2_REG_BUSY);

      /* setup the portmap */
      IFX_ES_ADR_TB_CTL1_REG_PMAP_SET(*IFX_ES_ADR_TB_CTL1_REG, 1<<i);

      do
      {
         /* search for addresses by port */
         *IFX_ES_ADR_TB_CTL2_REG = IFX_ES_ADR_TB_CTL2_REG_CMD_VAL(2)
                                 | IFX_ES_ADR_TB_CTL2_REG_AC_VAL(9);

         /* wait while busy */
         while (*IFX_ES_ADR_TB_ST2_REG & IFX_ES_ADR_TB_ST2_REG_BUSY);

         if ((result = IFX_ES_ADR_TB_ST2_REG_RSLT_GET(*IFX_ES_ADR_TB_ST2_REG)) == 0x101)
         {
            printk("%s, cmd error\n", __FUNCTION__);
            return;
         }
         /* if Command OK, address found... */
         if (result == 0)
         {
            unsigned char mac[6];

            mac[5] = (*IFX_ES_ADR_TB_ST0_REG >> 0) & 0xff;
            mac[4] = (*IFX_ES_ADR_TB_ST0_REG >> 8) & 0xff;
            mac[3] = (*IFX_ES_ADR_TB_ST0_REG >> 16) & 0xff;
            mac[2] = (*IFX_ES_ADR_TB_ST0_REG >> 24) & 0xff;
            mac[1] = (*IFX_ES_ADR_TB_ST1_REG >> 0) & 0xff;
            mac[0] = (*IFX_ES_ADR_TB_ST1_REG >> 8) & 0xff;

            /* setup address */
            *IFX_ES_ADR_TB_CTL0_REG =
                           (mac[5] << 0) |
                           (mac[4] << 8) |
                           (mac[3] << 16) |
                           (mac[2] << 24);
            *IFX_ES_ADR_TB_CTL1_REG =
                           IFX_ES_ADR_TB_CTL1_REG_PMAP_VAL(1<<i) |
                           IFX_ES_ADR_TB_CTL1_REG_FID_VAL(0) |
                           (mac[0] << 8) |
                           (mac[1] << 0);
            /* erase address */
            *IFX_ES_ADR_TB_CTL2_REG =
                           IFX_ES_ADR_TB_CTL2_REG_CMD_VAL(1) |
                           IFX_ES_ADR_TB_CTL2_REG_AC_VAL(15);

            /* wait, while busy */
            while (*IFX_ES_ADR_TB_ST2_REG & IFX_ES_ADR_TB_ST2_REG_BUSY);
         }
      } while (result == 0);
   }
   /**************************************
    * END: Procedure to clear MAC table
    **************************************/

   /* reset RMON counters */
   *IFX_ES_RMON_CTL_REG = IFX_ES_RMON_CTL_REG_BAS | IFX_ES_RMON_CTL_REG_CAC_VAL(3);

   /* bring all registers to reset state */
   reg = IFX_ES_PS_REG;
   for (i=0; i<(sizeof(switch_reset_offset_000)/sizeof(unsigned int)); i++)
   {
      if ((reg == IFX_ES_PS_REG) ||
          (reg >= IFX_ES_ADR_TB_CTL0_REG && reg <= IFX_ES_ADR_TB_ST2_REG))
         continue;

      if (switch_reset_offset_000[i] != 0xFFFFFFFF)
      {
         /* write reset value to register */
         *reg = switch_reset_offset_000[i];
         /* read register value back */
         rdreg = *reg;
         if (reg == IFX_ES_SW_GCTL1_REG)
            rdreg &= ~IFX_ES_SW_GCTL1_REG_BISTDN;
         /* compare read value with written one */
         if (rdreg != switch_reset_offset_000[i])
         {
            printk("%s,%d: reg %08x mismatch [has:%08x, expect:%08x]\n", __FUNCTION__, __LINE__,
                  (unsigned int)reg, rdreg, switch_reset_offset_000[i]);
         }
      }
      reg++;
   }

   reg = IFX_ES_VLAN_FLT0_REG;
   for (i=0; i<(sizeof(switch_reset_offset_100)/sizeof(unsigned int)); i++)
   {
      *reg = switch_reset_offset_100[i];
      rdreg = *reg;
      if (rdreg != switch_reset_offset_100[i])
      {
         printk("%s,%d: reg %08x mismatch [has:%08x, expect:%08x]\n", __FUNCTION__, __LINE__,
               (unsigned int)reg, rdreg, switch_reset_offset_100[i]);
      }
      reg++;
   }

#if 0
   /* Reset DMA channels used by the SWITCH */

	/* select channel, Tx */
	*IFX_DMA_CS = DMA_TX_CHANNEL_1_SELECT;
	/* reset Tx channel */
	IFX_DMA_CCTRL_RST_SET(*IFX_DMA_CCTRL, 1);
	/* wait until Tx channel reset completed */
	while (IFX_DMA_CCTRL_RST_GET(*IFX_DMA_CCTRL) == 1);
	/* shut off Tx channel */
   IFX_DMA_CCTRL_ON_OFF_SET(*IFX_DMA_CCTRL, 0);

	/* select channel, Rx */
	*IFX_DMA_CS = DMA_RX_CHANNEL_0_SELECT;
	/* reset Rx channel */
	IFX_DMA_CCTRL_RST_SET(*IFX_DMA_CCTRL, 1);
	/* wait until Rx channel reset completed */
	while (IFX_DMA_CCTRL_RST_GET(*IFX_DMA_CCTRL) == 1);
	/* shut off Rx channel */
   IFX_DMA_CCTRL_ON_OFF_SET(*IFX_DMA_CCTRL, 0);
#endif
}

/*   Set SVIP switch port2 mac address 
   * To resolve kernel packet lost problem in recovery image.
   * Equal to "switch_control mac_entry_add xxxxxxxxxxxx 4" 
     code from ltq.SP61/drv_ethsw-0.7.7/src/ifx_switchip.c */
void switchip_set_mac_table_entry(char * fact_mac)
{
    int i;
    unsigned char mac [6];
    memcpy(mac, fact_mac, 6);
    /* check if access engine is available */
    while(*IFX_ES_ADR_TB_ST2_REG & IFX_ES_ADR_TB_ST2_REG_BUSY);

    *IFX_ES_ADR_TB_CTL0_REG = (mac[2] << 24)
                           | (mac[3] << 16)
                           | (mac[4] << 8)
                           | (mac[5] << 0);
    *IFX_ES_ADR_TB_CTL1_REG = IFX_ES_ADR_TB_CTL1_REG_PMAP_VAL(1<<2)
        //| IFX_ES_ADR_TB_CTL1_REG_FID_VAL(mac_table_entry->nFid)
        | IFX_ES_ADR_TB_CTL1_REG_FID_VAL(0)
        | (mac[0] << 8)
        | (mac[1] << 0);
    *IFX_ES_ADR_TB_CTL2_REG = IFX_ES_ADR_TB_CTL2_REG_CMD_VAL(0)  /* enter address */
            | IFX_ES_ADR_TB_CTL2_REG_AC_VAL(7)   /* create new address */
            | IFX_ES_ADR_TB_CTL2_REG_INFOT;     /* static entry */
    /*      | IFX_ES_ADR_TB_CTL2_REG_ITAT_VAL(4 |
                              mac_table_entry->nPriority */

    /* wait for command completion */
    while((*IFX_ES_ADR_TB_ST2_REG) & IFX_ES_ADR_TB_ST2_REG_BUSY);

}

static u32 get_phy_oui(unsigned char phy_addr)
{
   u32 oui;
   int i, bit, byte, shift, w;
   u16 reg_id[2];

   /* read PHY identifier registers 1 and 2 */
   reg_id[0] = switchip_phy_read(phy_addr, 2);
   reg_id[1] = switchip_phy_read(phy_addr, 3);

   oui = 0;
   w=1;
   shift=7;
   byte=1;
   for (i=0, bit=10; i<=21; i++, bit++)
   {
      oui |= ((reg_id[w] & (1<<bit)) ? 1 : 0) << shift;
      if (!(shift%8))
      {
         byte++;
         if (byte==2)
         {
            shift=15;
         }
         else
            shift=21;
      }
      else
         shift--;
      if (w==1 && bit==15)
      {
         bit=-1;
         w=0;
      }
   }
   return oui;
}

#define MII_IFDIR "driver/mii"
struct proc_dir_entry *mii_proc_ifdir = NULL;
struct proc_dir_entry *phy_ifdir = NULL;
struct proc_dir_entry *sw_ifdir = NULL;

static int phy_if_proc_read (
   char *page, char **start, off_t off, int count,
   int *eof, void *data)
{
   char *p = page;
   int phyaddr = phy_address[0];
   int  len = 0;
   const int max = 16, total_standard_regs = 32, total_extend_regs = 15;
   const int total = total_standard_regs + total_extend_regs, extend_offs = 16;
   int i = 0, offs = 0 , size, extend = 0;
   unsigned short val[max];

   p += sprintf(p, "[PHY] SWPORT: %d, OUI: %x\n", 0, phy_oui);
   while ( offs < total )
   {
      memset(val, 0xcc, max);
      size = (total - offs) > max ? max : (total - offs);
      if ( offs >= total_standard_regs ) {
         extend = 1;
      } else {
         extend = 0;
      }
      i = 0;
      if ( extend ) {
         switchip_phy_write(phyaddr, 31, 1);
      }
      while ( i < size ) {
         val[i] = switchip_phy_read(phyaddr, extend ? (extend_offs + i) : (offs + i));
         i++;
      }
      if ( extend ) {
         switchip_phy_write(phyaddr, 31, 0);
      }
      p += sprintf(p, "[PHY] %sREG (%02x): ", extend ? "E" : "", offs);
      offs += size;
      i = 0;
      while ( i < size ) {
         p += sprintf(p, "%04x,", val[i]);
         i++;
      }
      p += sprintf(p, "%s", "\n");
   }

   len = (p - page);
   if (len <= off+count)
           *eof = 1;
   *start = page + off;
   len -= off;
   if (len>count)
           len = count;
   if (len<0)
           len = 0;

   return (len);
}

static int phy_if_proc_write (
   struct file *file, const char *buffer, unsigned long count,
   void *data)
{
   char proc_str[128] = {'\0'};
   char *proc_strp = proc_str;
   char *reg_addrptr = NULL, *reg_valptr = NULL, *reg_extendptr = NULL;
   unsigned int regaddr, val, extend = 0;
   int phyaddr = phy_address[0];

   if ((count > sizeof(proc_str) - 1))
      return (-EINVAL);

   if (copy_from_user(proc_str, buffer, count))
      return (-EFAULT);

   proc_str[count] = '\0';

   reg_addrptr = strsep(&proc_strp, " ");
   if (reg_addrptr) {
      regaddr = simple_strtoul(reg_addrptr, NULL, 16);
   } else {
     printk("Please specify register address\n");
     return (-EINVAL);
  }

   reg_valptr = strsep(&proc_strp, " ");
   if (reg_valptr) {
      val = simple_strtoul(reg_valptr, NULL, 16) & 0xFFFF;
   } else {
      printk("Please specify register value\n");
      return (-EINVAL);
   }

   reg_extendptr = strsep(&proc_strp, " ");
   if (reg_extendptr) {
      extend = simple_strtoul(reg_extendptr, NULL, 0);
   }

   if ( extend )
   {
      if ( regaddr < 16 || regaddr > 30 )
      {
         printk("Invalid register address, valid extend address range: [16, 30]\n");
         return (-EINVAL);
      }
   }
   else
   {
      if ( regaddr < 0 || regaddr > 31 )
      {
         printk("Invalid register address, valid address range: [0, 31]\n");
         return (-EINVAL);
      }
   }

   if ( extend )
   {
      switchip_phy_write(phyaddr, 31, 1);
      switchip_phy_write(phyaddr, regaddr, val);
      switchip_phy_write(phyaddr, 31, 0);
   }
   else
   {
      switchip_phy_write(phyaddr, regaddr, val);
   }
   printk("[PHY%d] [%sREG%d] VAL: %x\n", phyaddr, extend ? "E" : "", regaddr, val);

   return (count);
}

static int sw_if_proc_read (
   char *page, char **start, off_t off, int count,
   int *eof, void *data)
{
   char *p = page;
   int len = 0, value = 0;

   value = IFX_ES_RGMII_CTL_REG_P0SPD_GET(*IFX_ES_RGMII_CTL_REG);
   p += sprintf(p, "[SW P0] Ctrl Speed:%x, %s\n", value, value == 0 ? "10M" : (value == 1 ? "100M" : (value == 2 ? "1000M" : "Unknown")));
   value = IFX_ES_RGMII_CTL_REG_P0DUP_GET(*IFX_ES_RGMII_CTL_REG);
   p += sprintf(p, "[SW P0] Ctrl Duplex:%x, %s\n", value, value == 0 ? "Half duplex" : "Full duplex");
   value = IFX_ES_RGMII_CTL_REG_P0RDLY_GET(*IFX_ES_RGMII_CTL_REG);
   p += sprintf(p, "[SW P0] Ctrl RxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
   value = IFX_ES_RGMII_CTL_REG_P0TDLY_GET(*IFX_ES_RGMII_CTL_REG);
   p += sprintf(p, "[SW P0] Ctrl TxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
   value = IFX_ES_RGMII_CTL_REG_P0FCE_GET(*IFX_ES_RGMII_CTL_REG);
   p += sprintf(p, "[SW P0] Ctrl FlowControl:%x, %s\n", value, value == 0 ? "disabled" : "enabled");

   value = IFX_ES_PS_REG_P0SHS_GET(*IFX_ES_PS_REG);
   p += sprintf(p, "[SW P0] Status High Speed:%x\n", value);
   value = IFX_ES_PS_REG_P0SS_GET(*IFX_ES_PS_REG);
   p += sprintf(p, "[SW P0] Status Speed:%x\n", value);
   value = IFX_ES_PS_REG_P0DS_GET(*IFX_ES_PS_REG);
   p += sprintf(p, "[SW P0] Status Duplex:%x\n", value);
   value = IFX_ES_PS_REG_P0LS_GET(*IFX_ES_PS_REG);
   p += sprintf(p, "[SW P0] Status Link status:%x\n", value);
   value = IFX_ES_PS_REG_P0FCS_GET(*IFX_ES_PS_REG);
   p += sprintf(p, "[SW P0] Status FlowControl:%x\n", value);

   value = IFX_ES_SW_GCTL0_REG_DPWECH_GET(*IFX_ES_SW_GCTL0_REG);
   p += sprintf(p, "[SW CTL0] Drop Packet When Excessive Collision Happen:%x\n", value);
   value = IFX_ES_SW_GCTL0_REG_ICRCCD_GET(*IFX_ES_SW_GCTL0_REG);
   p += sprintf(p, "[SW CTL0] CRC Check Disable:%x\n", value);
   value = IFX_ES_SW_GCTL0_REG_MPL_GET(*IFX_ES_SW_GCTL0_REG);
   p += sprintf(p, "[SW CTL0] Max Packet Length:%x\n", value);

   value = IFX_ES_SW_GCTL1_REG_EDSTX_GET(*IFX_ES_SW_GCTL1_REG);
   p += sprintf(p, "[SW CTL1] Enable drop scheme of TX and RX:%x\n", value);

   value = IFX_ES_STRM_CTL_REG_STORM_10_TH_GET(*IFX_ES_STRM_CTL_REG);
   p += sprintf(p, "[SW STORM] Storm 10M Threshold:%x\n", value);
   value = IFX_ES_STRM_CTL_REG_STORM_100_TH_GET(*IFX_ES_STRM_CTL_REG);
   p += sprintf(p, "[SW STORM] Storm 100M Threshold:%x\n", value);

   len = (p - page);
   if (len <= off+count)
           *eof = 1;
   *start = page + off;
   len -= off;
   if (len>count)
           len = count;
   if (len<0)
           len = 0;

   return (len);
}

static int sw_if_proc_write (
   struct file *file, const char *buffer, unsigned long count,
   void *data)
{
   char proc_str[128] = {'\0'}, opt_str[128];
   char *proc_strp = proc_str;
   char *optptr = NULL, *valptr = NULL;
   int value;

   if ((count > sizeof(proc_str) - 1))
      return (-EINVAL);

   if (copy_from_user(proc_str, buffer, count))
      return (-EFAULT);

   proc_str[count] = '\0';

   optptr = strsep(&proc_strp, " ");
   if (optptr) {
      memcpy(opt_str, optptr, strlen(optptr) > 128 ? 128 : strlen(optptr));
   } else {
      printk("Please spefify setting option\n");
      return (-EINVAL);
   }

   valptr = strsep(&proc_strp, " ");
   if (valptr) {
      value = simple_strtoul(valptr, NULL, 0);
   } else {
      printk("Please specify register value\n");
      return (-EINVAL);
   }

   if ( !strncasecmp("CRC", optptr, strlen("CRC")) )
   {
      if ( value )
      {
         IFX_ES_SW_GCTL0_REG_ICRCCD_SET(*IFX_ES_SW_GCTL0_REG, 1);
      }
      else
      {
         IFX_ES_SW_GCTL0_REG_ICRCCD_SET(*IFX_ES_SW_GCTL0_REG, 0);
      }
      value = IFX_ES_SW_GCTL0_REG_ICRCCD_GET(*IFX_ES_SW_GCTL0_REG);
      printk("[SW CTL0] CRC Check Disable:%x\n", value);
   }
   else if ( !strncasecmp("FlowCtrl", optptr, strlen("FlowCtrl")) )
   {
      if ( value )
      {
         IFX_ES_RGMII_CTL_REG_P0FCE_SET(*IFX_ES_RGMII_CTL_REG, 1);
      }
      else
      {
         IFX_ES_RGMII_CTL_REG_P0FCE_SET(*IFX_ES_RGMII_CTL_REG, 0);
      }
      value = IFX_ES_RGMII_CTL_REG_P0FCE_GET(*IFX_ES_RGMII_CTL_REG);
      printk("[SW P0] Ctrl FlowControl:%x, %s\n", value, value == 0 ? "disabled" : "enabled");
      value = IFX_ES_PS_REG_P0FCS_GET(*IFX_ES_PS_REG);
      printk("[SW P0] Status FlowControl:%x\n", value);
   }
   else if ( !strncasecmp("Drop", optptr, strlen("Drop")) )
   {
      if ( value )
      {
         IFX_ES_SW_GCTL1_REG_EDSTX_SET(*IFX_ES_SW_GCTL1_REG, 1);
      }
      else
      {
         IFX_ES_SW_GCTL1_REG_EDSTX_SET(*IFX_ES_SW_GCTL1_REG, 0);
      }
      value = IFX_ES_SW_GCTL1_REG_EDSTX_GET(*IFX_ES_SW_GCTL1_REG);
      printk("[SW CTL1] Enable drop scheme of TX and RX:%x\n", value);
   }
   else  if ( !strncasecmp("RxDelay", optptr, strlen("RxDelay")) )
   {
      if ( value >= 0 && value <= 3 )
      {
         IFX_ES_RGMII_CTL_REG_P0RDLY_SET(*IFX_ES_RGMII_CTL_REG, value);
      }
      else
      {
         printk("Invalid delay, valid delay range: [0, 3]\n");
         return (-EINVAL);
      }
      value = IFX_ES_RGMII_CTL_REG_P0RDLY_GET(*IFX_ES_RGMII_CTL_REG);
      printk("[SW P0] Ctrl RxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
   }
   else  if ( !strncasecmp("TxDelay", optptr, strlen("TxDelay")) )
   {
      if ( value >= 0 && value <= 3 )
      {
         IFX_ES_RGMII_CTL_REG_P0TDLY_SET(*IFX_ES_RGMII_CTL_REG, value);
      }
      else
      {
         printk("Invalid delay, valid delay range: [0, 3]\n");
         return (-EINVAL);
      }
      value = IFX_ES_RGMII_CTL_REG_P0TDLY_GET(*IFX_ES_RGMII_CTL_REG);
      printk("[SW P0] Ctrl TxDelay:%x, %s\n", value, value == 0 ? "0ns" : ( value == 1 ? "1.5ns" : (value == 2 ? "1.75ns" : "2ns")));
   }
   else
   {
      printk("Invalid setting option name, valid option name is\n");
      printk("[CRC, FlowCtrl, Drop, RxDelay, TxDelay]\n");
      return (-EINVAL);
   }

   return (count);
}

static int mii_proc_install_if(void)
{
   int ret = 0;

   printk("mii: using mii proc fs\n");
   mii_proc_ifdir = proc_mkdir(MII_IFDIR, NULL);
   if ( mii_proc_ifdir )
   {
      phy_ifdir = create_proc_entry("phy", S_IFREG|S_IRUGO, mii_proc_ifdir);
      if ( phy_ifdir )
      {
         phy_ifdir->read_proc = phy_if_proc_read;
         phy_ifdir->write_proc = phy_if_proc_write;
      }
      else
      {
         printk("mii: create proc entry %s/phy failed\n", MII_IFDIR);
      }
      sw_ifdir = create_proc_entry("sw", S_IFREG|S_IRUGO, mii_proc_ifdir);
      if ( sw_ifdir )
      {
         sw_ifdir->read_proc = sw_if_proc_read;
         sw_ifdir->write_proc = sw_if_proc_write;
      }
      else
      {
         printk("mii: create proc entry %s/sw failed\n", MII_IFDIR);
      }
   }
   else
   {
      printk("mii: create proc entry dir %s failed\n", MII_IFDIR);
      ret = -ENOMEM;
      remove_proc_entry(MII_IFDIR, NULL);
   }

   return ret;
}

static int __attribute__((unused)) mii_proc_remove_if(void)
{
   if (mii_proc_ifdir)
   {
      if ( phy_ifdir )
      {
         remove_proc_entry(MII_IFDIR "/phy", NULL);
         phy_ifdir = NULL;
      }
      if ( sw_ifdir )
      {
         remove_proc_entry(MII_IFDIR "/sw", NULL);
         sw_ifdir = NULL;
      }
      remove_proc_entry(MII_IFDIR, NULL);
      mii_proc_ifdir = NULL;
   }
   return 0;
}

/*
 * Switch Initialization.
 */
int switchip_init(void)
{
#if 0
   int i;
#endif
   int eth_port, phy_present = 0;
   u16 reg, mode;

   *IFX_SYS1_CLKENR = IFX_SYS1_CLKENR_ETHSW;
   asm("sync");

   /* Enable Switch, if not already done so */
   if (IFX_ES_SW_GCTL0_REG_SE_GET(*IFX_ES_SW_GCTL0_REG) == 0)
   {
      IFX_ES_SW_GCTL0_REG_SE_SET(*IFX_ES_SW_GCTL0_REG, 1);
   }
   /* Wait for completion of MBIST */
   while (IFX_ES_SW_GCTL1_REG_BISTDN_GET(*IFX_ES_SW_GCTL1_REG) == 0);

   switchip_reset();

   mode = IFX_ES_RGMII_CTL_REG_IS_GET(*IFX_ES_RGMII_CTL_REG);
   eth_port = (mode == 2 ? 1 : 0);
   printk("mode=%d, port=%d\n", mode, eth_port);

#ifndef CONFIG_IFX_SVIP_SW_MIRRORING
   /* Set the primary port(port toward backplane) as sniffer port,
      changing from P2 which is the reset setting */
   IFX_ES_SW_GCTL0_REG_SNIFFPN_SET(*IFX_ES_SW_GCTL0_REG, eth_port);
#else
   /* applicable to GMII mode only */
   if (mode == 2)
   {
      /* Mirror received packets having errors, as well */
      IFX_ES_SW_GCTL0_REG_MRA_SET(*IFX_ES_SW_GCTL0_REG, 1);
      /* Mirror packets received on P1 */
      IFX_ES_P0_CTL_REG_IPMO_SET(*IFX_ES_P1_CTL_REG, 1);
   }
#endif

#ifdef CONFIG_IFX_MDIO_POLLING
   /* Point MDIO state machine to invalid PHY addresses 8 and 9 */
   IFX_ES_SW_GCTL0_REG_PHYBA_SET(*IFX_ES_SW_GCTL0_REG, 1);
#endif

   /* Add CRC for packets from DMA to PMAC.
      Remove CRC for packets from PMAC to DMA. */
   *IFX_ES_PMAC_HD_CTL = IFX_ES_PMAC_HD_CTL_RC |
      IFX_ES_PMAC_HD_CTL_AC;

   gs_phy_oui = get_phy_oui(0);
   phy_oui = gs_phy_oui;
   switch (phy_oui)
   {
#if defined(CONFIG_IFX_EASY336) || defined (CONFIG_IFX_BOARD_GS_GXW)
      case PHY_OUI_PMC:
         phy_address[0] = (mode == 2 ? -1 : 2);
         phy_address[1] = (mode == 2 ? 2 : -1);
         break;
#endif
      case PHY_OUI_VITESSE:
      case PHY_OUI_REALTEK:
      default:
         phy_oui = PHY_OUI_DEFAULT;
         phy_address[0] = (mode == 2 ? 1 : 0);
         phy_address[1] = (mode == 2 ? 0 : 1);
         break;
   }
   printk("OUI: %x - %x\n", gs_phy_oui, phy_oui);

   /****** PORT 0 *****/
   reg = switchip_phy_read(phy_address[0], 1);
   if ((reg != 0x0000) && (reg != 0xffff)) /* PHY connected? */
   {
      phy_present |= 1;
      /* Set Rx- and TxDelay in case of RGMII */
      switch (mode)
      {
      case 0 : /* *RGMII,RGMII */
      case 2 : /* RGMII,*GMII */
         /* program clock delay in PHY, not in SVIP */
         IFX_ES_RGMII_CTL_REG_P0RDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         IFX_ES_RGMII_CTL_REG_P0TDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         if ( ( gs_phy_oui != PHY_OUI_REALTEK ) &&
            (phy_oui == PHY_OUI_VITESSE ||
             phy_oui == PHY_OUI_DEFAULT)
             )
         {
            switchip_phy_write(phy_address[0], 31, 0x0001);
            switchip_phy_write(phy_address[0], 28, 0xA000);
            switchip_phy_write(phy_address[0], 31, 0x0000);
         }
      default:
         break;
      }
#ifdef CONFIG_IFX_MDIO_POLLING
      if (phy_oui == PHY_OUI_VITESSE ||
          phy_oui == PHY_OUI_DEFAULT)
      {
         /* Program PHY advertisements and restart auto-negotiation */
         switchip_phy_write(phy_address[0], 4, 0x05E1);
         switchip_phy_write(phy_address[0], 9, 0x0300);
         switchip_phy_write(phy_address[0], 0, 0x3300);
      }
      else
      {
         reg = switchip_phy_read(phy_address[1], 0);
         reg |= 0x1000; /* auto-negotiation enable */
         switchip_phy_write(phy_address[1], 0, reg);
         reg |= 0x0200; /* auto-negotiation restart */
         switchip_phy_write(phy_address[1], 0, reg);
      }
#if 0
      printk("Waiting for auto-negotiation to complete on P0... ");
      {
      int i=0;
      do {
         msleep(200);
         reg = switchip_phy_read(phy_address[0], 1);
         if (reg & (1<<5))
            break;
      } while (++i < 20);
      }
      if (reg & (1<<5))
         printk("done\n");
      else
         printk("failed\n");
#endif
#endif
   }
   else
   {
      /* Force SWITCH link with highest capability:
       * 100M FD for MII
       * 1G FD for GMII/RGMII
       */
      switch (mode)
      {
      case 1 : /* *MII,MII */
      case 3 : /* *MII,RGMII */
         IFX_ES_RGMII_CTL_REG_P0SPD_SET(*IFX_ES_RGMII_CTL_REG, 1);
         IFX_ES_RGMII_CTL_REG_P0DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
         break;
      case 0 : /* *RGMII,RGMII */
      case 2 : /* RGMII,*GMII */
         IFX_ES_RGMII_CTL_REG_P0SPD_SET(*IFX_ES_RGMII_CTL_REG, 2);
         IFX_ES_RGMII_CTL_REG_P0DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
         IFX_ES_RGMII_CTL_REG_P0RDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         IFX_ES_RGMII_CTL_REG_P0TDLY_SET(*IFX_ES_RGMII_CTL_REG, 2);
         break;
      }
      IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P0_CTL_REG, 1);
   }

   /****** PORT 1 *****/
   reg = switchip_phy_read(phy_address[1], 1);
   if ((reg != 0x0000) && (reg != 0xffff)) /* PHY connected? */
   {
      phy_present |= 2;
      /* Set Rx- and TxDelay in case of RGMII */
      switch (mode)
      {
      case 0 : /* *RGMII,RGMII */
      case 3 : /* *MII,RGMII */
         /* program clock delay in PHY, not in SVIP */
         IFX_ES_RGMII_CTL_REG_P1RDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         IFX_ES_RGMII_CTL_REG_P1TDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         if ( ( gs_phy_oui != PHY_OUI_REALTEK ) &&
             (phy_oui == PHY_OUI_VITESSE ||
             phy_oui == PHY_OUI_DEFAULT)
             )
         {
            switchip_phy_write(phy_address[1], 31, 0x0001);
            switchip_phy_write(phy_address[1], 28, 0xA000);
            switchip_phy_write(phy_address[1], 31, 0x0000);
         }
         break;
      case 2 : /* RGMII,*GMII */
         IFX_ES_RGMII_CTL_REG_P1SPD_SET(*IFX_ES_RGMII_CTL_REG, 2);
         IFX_ES_RGMII_CTL_REG_P1DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
#if defined(CONFIG_IFX_EASY336) || defined (CONFIG_IFX_BOARD_GS_GXW)
         if (phy_oui == PHY_OUI_PMC)
         {
            switchip_phy_write(phy_address[1], 24, 0x0510);
            switchip_phy_write(phy_address[1], 17, 0xA38C);
            switchip_phy_write(phy_address[1], 17, 0xA384);
         }
#endif
         break;
      default:
         break;
      }
#ifdef CONFIG_IFX_MDIO_POLLING
      /* Program PHY advertisements and restart auto-negotiation */
      if (phy_oui == PHY_OUI_VITESSE ||
          phy_oui == PHY_OUI_DEFAULT)
      {
         switchip_phy_write(phy_address[1], 4, 0x05E1);
         switchip_phy_write(phy_address[1], 9, 0x0300);
         switchip_phy_write(phy_address[1], 0, 0x3300);
      }
      else
      {
         reg = switchip_phy_read(phy_address[1], 0);
         reg |= 0x1000; /* auto-negotiation enable */
         switchip_phy_write(phy_address[1], 0, reg);
         reg |= 0x0200; /* auto-negotiation restart */
         switchip_phy_write(phy_address[1], 0, reg);
      }
#if 0
      printk("Waiting for auto-negotiation to complete on P1... ");
      i=0;
      do {
         msleep(200);
         reg = switchip_phy_read(phy_address[1], 1);
         if (reg & (1<<5))
            break;
      } while (++i < 20);
      if (reg & (1<<5))
         printk("done\n");
      else
         printk("failed\n");
#endif
#endif
   }
   else
   {
      /* Force SWITCH link with highest capability:
       * 100M FD for MII
       * 1G FD for GMII/RGMII
       */
      switch (mode)
      {
      case 1 : /* *MII,MII */
         IFX_ES_RGMII_CTL_REG_P1SPD_SET(*IFX_ES_RGMII_CTL_REG, 1);
         IFX_ES_RGMII_CTL_REG_P1DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
         break;
      case 0 : /* *RGMII,RGMII */
      case 3 : /* *MII,RGMII */
         IFX_ES_RGMII_CTL_REG_P1SPD_SET(*IFX_ES_RGMII_CTL_REG, 2);
         IFX_ES_RGMII_CTL_REG_P1DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
         IFX_ES_RGMII_CTL_REG_P1RDLY_SET(*IFX_ES_RGMII_CTL_REG, 0);
         IFX_ES_RGMII_CTL_REG_P1TDLY_SET(*IFX_ES_RGMII_CTL_REG, 2);
         break;
      case 2 : /* RGMII,*GMII */
         IFX_ES_RGMII_CTL_REG_P1SPD_SET(*IFX_ES_RGMII_CTL_REG, 2);
         IFX_ES_RGMII_CTL_REG_P1DUP_SET(*IFX_ES_RGMII_CTL_REG, 1);
         break;
      }
      IFX_ES_P0_CTL_REG_FLP_SET(*IFX_ES_P1_CTL_REG, 1);
   }

   /*
   * Allow unknown unicast/multicast and broadcasts
   * on all ports.
   */
   IFX_ES_SW_GCTL1_REG_UP_SET(*IFX_ES_SW_GCTL1_REG, 1);
   IFX_ES_SW_GCTL1_REG_BP_SET(*IFX_ES_SW_GCTL1_REG, 7);
   IFX_ES_SW_GCTL1_REG_MP_SET(*IFX_ES_SW_GCTL1_REG, 7);
   IFX_ES_SW_GCTL1_REG_RP_SET(*IFX_ES_SW_GCTL1_REG, 7);

   /* user port priority as QoS */
   /* set priority of port towards external network to lowest */
   /* set priority of port towards private network to highest */
   if (eth_port == 0)
   {
      IFX_ES_P0_VLAN_REG_PP_SET(*IFX_ES_P0_VLAN_REG, 0);
      IFX_ES_P0_VLAN_REG_PP_SET(*IFX_ES_P1_VLAN_REG, 2);
   }
   else
   {
      IFX_ES_P0_VLAN_REG_PP_SET(*IFX_ES_P0_VLAN_REG, 2);
      IFX_ES_P0_VLAN_REG_PP_SET(*IFX_ES_P1_VLAN_REG, 0);
   }
   *IFX_ES_P0_VLAN_REG |= IFX_ES_P0_VLAN_REG_PPE;
   *IFX_ES_P1_VLAN_REG |= IFX_ES_P0_VLAN_REG_PPE;
   /* set port 2 priority to medium high */
   IFX_ES_P0_VLAN_REG_PP_SET(*IFX_ES_P2_VLAN_REG, 1);
   *IFX_ES_P2_VLAN_REG |= IFX_ES_P0_VLAN_REG_PPE;

   /* Enable LAN port(s) */
#ifdef SINGLE_PORT
   if (eth_port == 0)
   {
      IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P0_CTL_REG, 0);
   }
   else
   {
      IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P1_CTL_REG, 0);
   }
#else
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P0_CTL_REG, 0);
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P1_CTL_REG, 0);
#endif
   /* Enable CPU Port (Forwarding State) */
   IFX_ES_P0_CTL_REG_SPS_SET(*IFX_ES_P2_CTL_REG, 0);

   if ( gs_phy_oui == PHY_OUI_REALTEK )
   {
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0xd04);
      /*
       * Link: on at 10/100/1000Mbps
       * Act: blink and on 10/100/1000Mbps
       */
      switchip_phy_write(phy_address[eth_port], RTL8211F_LED_REG, 0x6d60);
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0x0);
      /* Disable EEE */
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0xa43);
      reg = switchip_phy_read(phy_address[eth_port], RTL8211F_PHYCR2);
      reg &= ~(1 << 5 );
      switchip_phy_write(phy_address[eth_port], RTL8211F_PHYCR2, reg);
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0x0);
      /* Disable EEE LED */
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0xd04);
      switchip_phy_write(phy_address[eth_port], RTL8211F_LED_EEE, 0x0);
      switchip_phy_write(phy_address[eth_port], RTL8211F_PAGE_SELECT, 0x0);
   }
   else
   {
      /* Set the LED1H, LED LINK, offset 1F.01E4 */
      switchip_phy_write( phy_address[eth_port], 13, 00 << 14 | 0x1F ); /* Write Function field to 00 */
      switchip_phy_write( phy_address[eth_port], 14, 0x01E4); /* Address of LED1H */
      switchip_phy_write( phy_address[eth_port], 13, 01 << 14 | 0x1F); /* Write Function field to 01 */
      switchip_phy_write( phy_address[eth_port], 14, 0x42); /* 100M fast blink, 1000M ON */

      /* Set the LED1L, LED LINK, offset 1F.01E5 */
      switchip_phy_write( phy_address[eth_port], 13, 00 << 14 | 0x1F ); /* Write Function field to 00 */
      switchip_phy_write( phy_address[eth_port], 14, 0x01E5); /* Address of LED1L */
      switchip_phy_write( phy_address[eth_port], 13, 01 << 14 | 0x1F); /* Write Function field to 01 */
      switchip_phy_write( phy_address[eth_port], 14, 0x10); /* 10M slow blink, 1000M ON */
   }
   mii_proc_install_if();


#ifdef CONFIG_IFX_MDIO_POLLING
   if (phy_present)
      switchip_mdio_poll_init();
#endif
   printk("switchip initialized\n");
   return 0;
}

__initcall(switchip_init);

EXPORT_SYMBOL(switchip_phy_read);
EXPORT_SYMBOL(switchip_phy_write);
EXPORT_SYMBOL(switchip_reset);
EXPORT_SYMBOL(switchip_init);

