/*
 * Bitbanging I2C bus driver using the GPIO API
 *
 * Copyright (C) 2007 Atmel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/i2c-gpio.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>

#include <asm/gpio.h>
#define LOW_UNIT_ADDR 0x22
#define HIGH_UNIT_ADDR 0x23
#define OUTPUT_PORT0 0x04
#define CONF_PORT0 0x0c

#define LEDPORT_TO_UNIT(x) (((x-1) / 24))
#define PORT_TO_PORT(x) ((x >> 3))   // equal to x/8
#define PORT_TO_PIN(x) ((x & 7))     // equal to x%8
#define LEDPORT_TO_PORT(x) ((23- (x-1)%24))
#define UNIT_TO_ADDR(x) (HIGH_UNIT_ADDR - (x & 1)) 

static struct timer_list port_timer;
static struct timer_list port_setting_timer;
static int first_time = 1;

static int set_port_value_in_ram (int number, int status);
static int set_port_value_all (void);

extern int i2c_gpio_flag;

#define GlobalControlRegister 0x11
#define OutputPort0 0x02
#define OutputPort1 0x03

struct tca6424_unit 
{
    char iic;
    char addr;
    char port_vals[4]; // 1 bit for one port
};

struct aw9523_unit
{
    char iic;
    char addr;
    char port_vals[2]; // 1 bit for one port
};

/* to store the i2c-gpio ports output value in ram */
static struct tca6424_unit unit_a[] = 
{
    {'A', 0x23, {0x84, 0xff, 0xff, 0xff} }, //0x84 is addr cmd
    {'A', 0x22, {0x84, 0xff, 0xff, 0xff} },
};

static struct tca6424_unit unit_b[] = 
{
    {'B', 0x23, {0x84, 0xff, 0xff, 0xff} },
    {'B', 0x22, {0x84, 0xff, 0xff, 0xff} },
};

static struct aw9523_unit aw9523_unit[] =
{
	{'B', 0x58, {OutputPort0, 0xff} },
	{'B', 0x58, {OutputPort1, 0xff} },

	{'B', 0x59, {OutputPort0, 0xff} },
	{'B', 0x59, {OutputPort1, 0xff} },

	{'B', 0x5a, {OutputPort0, 0xff} },
	{'B', 0x5a, {OutputPort1, 0xff} },
};

static struct aw9523_unit prev_aw9523_unit[] =
{
	{'B', 0x58, {OutputPort0, 0xff} },
	{'B', 0x58, {OutputPort1, 0xff} },

	{'B', 0x59, {OutputPort0, 0xff} },
	{'B', 0x59, {OutputPort1, 0xff} },

	{'B', 0x5a, {OutputPort0, 0xff} },
	{'B', 0x5a, {OutputPort1, 0xff} },
};

static char prev_unita0 [4] = {0x84, 0xff, 0xff, 0xff};
static char prev_unita1 [4] = {0x84, 0xff, 0xff, 0xff};
static char prev_unitb0 [4] = {0x84, 0xff, 0xff, 0xff};
static char prev_unitb1 [4] = {0x84, 0xff, 0xff, 0xff};
/* Toggle SDA by changing the direction of the pin */
static void i2c_gpio_setsda_dir(void *data, int state)
{
	struct i2c_gpio_platform_data *pdata = data;

	if (state)
		gpio_direction_input(pdata->sda_pin);
	else
		gpio_direction_output(pdata->sda_pin, 0);
}

/*
 * Toggle SDA by changing the output value of the pin. This is only
 * valid for pins configured as open drain (i.e. setting the value
 * high effectively turns off the output driver.)
 */
static void i2c_gpio_setsda_val(void *data, int state)
{
	struct i2c_gpio_platform_data *pdata = data;

    gpio_direction_output(pdata->sda_pin, state);
	//gpio_set_value(pdata->sda_pin, state);
}

/* Toggle SCL by changing the direction of the pin. */
static void i2c_gpio_setscl_dir(void *data, int state)
{
	struct i2c_gpio_platform_data *pdata = data;

	if (state)
		gpio_direction_input(pdata->scl_pin);
	else
		gpio_direction_output(pdata->scl_pin, 0);
}

/*
 * Toggle SCL by changing the output value of the pin. This is used
 * for pins that are configured as open drain and for output-only
 * pins. The latter case will break the i2c protocol, but it will
 * often work in practice.
 */
static void i2c_gpio_setscl_val(void *data, int state)
{
	struct i2c_gpio_platform_data *pdata = data;

	gpio_set_value(pdata->scl_pin, state);
}

static int i2c_gpio_getsda(void *data)
{
	struct i2c_gpio_platform_data *pdata = data;
    
    gpio_direction_input(pdata->sda_pin);
	return gpio_get_value(pdata->sda_pin);
}

static int i2c_gpio_getscl(void *data)
{
	struct i2c_gpio_platform_data *pdata = data;

	return gpio_get_value(pdata->scl_pin);
}

static int set_port_value_in_ram (int number, int status)
{
    int unit, unit_port, unit_pin, port;
    struct tca6424_unit * unitP;

    unit = LEDPORT_TO_UNIT(number);
    unit_port = LEDPORT_TO_PORT(number);
    unit_pin = PORT_TO_PIN(unit_port);
    port = PORT_TO_PORT(unit_port); 
     
    if (unit == 0)
        unitP = &unit_a[0];
    else if (unit == 1) 
        unitP = &unit_a[1];
    else if (unit ==2) 
        unitP = &unit_b[0];
    else 
        unitP = &unit_b[1];
    
    if (status) // on  
    {
        unitP->port_vals[port+1] &= ~(1<< unit_pin);    
    }
    else 
    {
        unitP->port_vals[port+1] |= (1<< unit_pin);    
    }
    return 1;
}

static int set_aw9523_port_value_in_ram (int number, int status)
{
	int unit_port, unit_pin;
	unit_port = (number - 49)/8;
	unit_pin = (number - 49)%8;
	//printk("%d %d\n", unit_port, unit_pin);

	if (status) // on
	{
	    aw9523_unit[unit_port].port_vals[1] &= ~(1<< unit_pin);
	}
	else
	{
	    aw9523_unit[unit_port].port_vals[1] |= (1<< unit_pin);
	}

	return 1;
}

/* which actually send commands to four tca6424a */
static int set_port_value_all (void)
{
    int ret;
    struct i2c_adapter *i2cadap_a;
    struct i2c_adapter *i2cadap_b;

    struct i2c_msg msg_a[] =
    {
        { 0x23,  0, 4, unit_a[0].port_vals },
        { 0x22,  0, 4, unit_a[1].port_vals },
    };
    struct i2c_msg msg_b[] = 
    {
        { 0x23,  0, 4, unit_b[0].port_vals },
        { 0x22,  0, 4, unit_b[1].port_vals },
    };

    i2cadap_a = i2c_get_adapter(1);
    i2cadap_b = i2c_get_adapter(2);

    if (i2cadap_a)
    {
        if (*(u32*)prev_unita0 != *(u32*)unit_a[0].port_vals)
            ret = i2c_transfer(i2cadap_a, &msg_a[0], 1);

        if (*(u32*)prev_unita1 != *(u32*)unit_a[1].port_vals)
            ret = i2c_transfer(i2cadap_a, &msg_a[1], 1);
    }    
    if (i2cadap_b)
    {
        if (*(u32*)prev_unitb0 != *(u32*)unit_b[0].port_vals)
            ret = i2c_transfer(i2cadap_b, &msg_b[0], 1);
        
        if (*(u32*)prev_unitb1 != *(u32*)unit_b[1].port_vals)
            ret = i2c_transfer(i2cadap_b, &msg_b[1], 1);
    }

    memcpy(prev_unita0, unit_a[0].port_vals, 4);
    memcpy(prev_unita1, unit_a[1].port_vals, 4);
    memcpy(prev_unitb0, unit_b[0].port_vals, 4);
    memcpy(prev_unitb1, unit_b[1].port_vals, 4);

    return 1;
}

/* which actually send commands to four aw9523 */
static int set_port_value_aw9523_all (void)
{
	struct i2c_msg msg = { 0x58,  0, 2 };
	int i = 0;
	struct i2c_adapter *i2cadap_b;
	i2cadap_b = i2c_get_adapter(2);

	if ( i2cadap_b )
	{
		for(i=0; i<6; i++)
		{
			if (*(u16*)prev_aw9523_unit[i].port_vals != *(u16*)aw9523_unit[i].port_vals)
			{
				switch(i/2)
				{
					case 0:msg.addr = 0x58;break;
					case 1:msg.addr = 0x59;break;
					case 2:msg.addr = 0x5a;break;
					default:break;
				}
				msg.buf = aw9523_unit[i].port_vals;
				i2c_transfer(i2cadap_b, &msg, 1);
			}
		}
	}

	memcpy(prev_aw9523_unit, aw9523_unit, sizeof(aw9523_unit));
	return 1;
}

static void port_setting_timer_fun (unsigned long data)
{
	if(i2c_gpio_flag == 0)
	{
		set_port_value_all(); //set i2c-gpio port in real world
	}
	else
	{
		set_port_value_aw9523_all();
	}
	first_time = 1; //reset
}

/* interfact to set i2c-gpio port */
int ifxmips_i2c_gpio_set (int number, int status)
{
	if(i2c_gpio_flag == 0)
	{
	    set_port_value_in_ram(number, status);
	}
	else
	{
		set_aw9523_port_value_in_ram(number, status);
	}
    if (first_time)
    {
        first_time = 0;
        if (timer_pending(&port_setting_timer))
        {
            if (mod_timer(&port_setting_timer, jiffies + HZ/20))
            {
                return 1;
            }
        }
        //wait 50ms to set together
        port_setting_timer.expires = jiffies + HZ/20; //50ms
        add_timer(&port_setting_timer);
    }
    return 1;
}

static void port_timer_fun_all (unsigned long data)
{
    int i;
    static int on_or_off =0;
    
    port_timer.expires = jiffies + HZ/5;
    add_timer(&port_timer);

    //mod_timer(&led_pattern_timer, jiffies + 500 * HZ / 1000);
    if (on_or_off )
    {
        for(i=0; i<97; i++)
        {   
            set_port_value_in_ram( i, 0);
        }
        on_or_off = 0;
    }
    else
    {
        for(i=0; i<97; i++)
        {   
            set_port_value_in_ram( i, 1);
        }
        on_or_off = 1;
    }
    set_port_value_all();
}

static int __devinit i2c_gpio_probe(struct platform_device *pdev)
{
	struct i2c_gpio_platform_data *pdata;
	struct i2c_algo_bit_data *bit_data;
	struct i2c_adapter *adap;
	int ret;

	char buf1[4] = {0x8c, 0x0, 0x0, 0x0};
	char buf2[4] = {0x84, 0xff, 0xff, 0xff};
	char buf3[4] = {0x06, 0xff, 0xff, 0xff};

	char buf4[2] = {GlobalControlRegister, 0x10}; //set PO port to push-pull
	char buf5[2] = {OutputPort0, 0xff};
	char buf6[2] = {OutputPort1, 0xff};
    struct i2c_msg msg[10] =
    {
        { 0x23,  0, 4,  buf1 },
        { 0x23,  0, 4,  buf2 },
        { 0x22,  0, 4,  buf1 },
        { 0x22,  0, 4,  buf2 }
        //{ 0x22,  1, 2,  buf3 }, 
    };

	if(i2c_gpio_flag == 1)// 1=aw9523
	{
		msg[0].addr = 0x58; msg[0].flags = 0; msg[0].len = 2; msg[0].buf = buf4;
		msg[1].addr = 0x59; msg[1].flags = 0; msg[1].len = 2; msg[1].buf = buf4;
		msg[2].addr = 0x5a; msg[2].flags = 0; msg[2].len = 2; msg[2].buf = buf4;

		msg[3].addr = 0x58; msg[3].flags = 0; msg[3].len = 2; msg[3].buf = buf5;
		msg[4].addr = 0x59; msg[4].flags = 0; msg[4].len = 2; msg[4].buf = buf5;
		msg[5].addr = 0x5a; msg[5].flags = 0; msg[5].len = 2; msg[5].buf = buf5;

		msg[6].addr = 0x58; msg[6].flags = 0; msg[6].len = 2; msg[6].buf = buf6;
		msg[7].addr = 0x59; msg[7].flags = 0; msg[7].len = 2; msg[7].buf = buf6;
		msg[8].addr = 0x5a; msg[8].flags = 0; msg[8].len = 2; msg[8].buf = buf6;
	}
	pdata = pdev->dev.platform_data;
	if (!pdata)
		return -ENXIO;

	ret = -ENOMEM;
	adap = kzalloc(sizeof(struct i2c_adapter), GFP_KERNEL);
	if (!adap)
		goto err_alloc_adap;
	bit_data = kzalloc(sizeof(struct i2c_algo_bit_data), GFP_KERNEL);
	if (!bit_data)
		goto err_alloc_bit_data;

	ret = gpio_request(pdata->sda_pin, "sda");
	if (ret)
		goto err_request_sda;
	ret = gpio_request(pdata->scl_pin, "scl");
	if (ret)
		goto err_request_scl;

	if (pdata->sda_is_open_drain) {
		gpio_direction_output(pdata->sda_pin, 1);
        //gpio_set_open_drain(pdata->sda_pin);
		bit_data->setsda = i2c_gpio_setsda_val;
	} else {
		gpio_direction_input(pdata->sda_pin);
		bit_data->setsda = i2c_gpio_setsda_dir;
	}

	if (pdata->scl_is_open_drain || pdata->scl_is_output_only) {
		gpio_direction_output(pdata->scl_pin, 1);
		bit_data->setscl = i2c_gpio_setscl_val;
	} else {
		gpio_direction_input(pdata->scl_pin);
		bit_data->setscl = i2c_gpio_setscl_dir;
	}

	if (!pdata->scl_is_output_only)
		bit_data->getscl = i2c_gpio_getscl;
	bit_data->getsda = i2c_gpio_getsda;

	if (pdata->udelay)
		bit_data->udelay = pdata->udelay;
	else if (pdata->scl_is_output_only)
		bit_data->udelay = 50;			/* 10 kHz */
	else
		bit_data->udelay = 5;			/* 100 kHz */

	if (pdata->timeout)
		bit_data->timeout = pdata->timeout;
	else
		bit_data->timeout = HZ / 10;		/* 100 ms */

	bit_data->data = pdata;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "i2c-gpio%d", pdev->id);
	adap->algo_data = bit_data;
	adap->class = I2C_CLASS_HWMON | I2C_CLASS_SPD;
	adap->dev.parent = &pdev->dev;

	/*
	 * If "dev->id" is negative we consider it as zero.
	 * The reason to do so is to avoid sysfs names that only make
	 * sense when there are multiple adapters.
	 */
	adap->nr = (pdev->id != -1) ? pdev->id : 0;
	ret = i2c_bit_add_numbered_bus(adap);
	if (ret)
		goto err_add_bus;

	platform_set_drvdata(pdev, adap);

	if(i2c_gpio_flag == 0)// 0=tca6424 ;1=aw9523
	{
	    /* set tca6424a all ports output */
	    ret = i2c_transfer(adap, msg, 4);
	    /* if cannot find receiver, this device doesn't exist */
	    if (ret < 0)
	       goto err_no_such_device;
	}
	else
	{
	    /* set aw9523 all port0 push_pull */
	    ret = i2c_transfer(adap, msg, 9);
	    /* if cannot find receiver, this device doesn't exist */
	    if (ret < 0)  
	       goto err_no_such_device; 
	}
	dev_info(&pdev->dev, "using pins %u (SDA) and %u (SCL%s)\n",
		 pdata->sda_pin, pdata->scl_pin,
		 pdata->scl_is_output_only
		 ? ", no clock stretching" : "");
   
        
#if 0 
    if (pdev->id ==1 )
        return 0;
	init_timer(&port_timer);
    port_timer.function = port_timer_fun_all;
    port_timer.expires = jiffies + HZ/5;
    add_timer(&port_timer);
#endif
    return 0;

err_no_such_device:
    printk("error no such device, %s\n", pdev->name);
	i2c_del_adapter(adap);
err_add_bus:
	gpio_free(pdata->scl_pin);
err_request_scl:
	gpio_free(pdata->sda_pin);
err_request_sda:
	kfree(bit_data);
err_alloc_bit_data:
	kfree(adap);
err_alloc_adap:
	return ret;
}

static int __devexit i2c_gpio_remove(struct platform_device *pdev)
{
	struct i2c_gpio_platform_data *pdata;
	struct i2c_adapter *adap;

	adap = platform_get_drvdata(pdev);
	pdata = pdev->dev.platform_data;

	i2c_del_adapter(adap);
	gpio_free(pdata->scl_pin);
	gpio_free(pdata->sda_pin);
	kfree(adap->algo_data);
	kfree(adap);

	return 0;
}

static struct platform_driver i2c_gpio_driver_a = {
	.driver		= {
		.name	= "ifxmips_i2c_gpio_a",
		.owner	= THIS_MODULE,
	},
	.probe		= i2c_gpio_probe,
	.remove		= __devexit_p(i2c_gpio_remove),
};

static struct platform_driver i2c_gpio_driver_b = {
	.driver		= {
		.name	= "ifxmips_i2c_gpio_b",
		.owner	= THIS_MODULE,
	},
	.probe		= i2c_gpio_probe,
	.remove		= __devexit_p(i2c_gpio_remove),
};

static int __init i2c_gpio_init(void)
{
	int ret;

	ret = platform_driver_register(&i2c_gpio_driver_a);
	if (ret)
		printk(KERN_ERR "i2c-gpio-a: probe failed: %d\n", ret);
    
    ret = platform_driver_register(&i2c_gpio_driver_b);
	if (ret)
		printk(KERN_ERR "i2c-gpio-b: probe failed: %d\n", ret);
	init_timer(&port_setting_timer);
	port_setting_timer.function = port_setting_timer_fun;

	return ret;
}
module_init(i2c_gpio_init);

static void __exit i2c_gpio_exit(void)
{
	if (timer_pending(&port_setting_timer))
	{
		del_timer_sync(&port_setting_timer);
	}
	platform_driver_unregister(&i2c_gpio_driver_a);
	platform_driver_unregister(&i2c_gpio_driver_b);
}
module_exit(i2c_gpio_exit);



MODULE_AUTHOR("Haavard Skinnemoen <hskinnemoen@atmel.com>");
MODULE_DESCRIPTION("Platform-independent bitbanging I2C driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:i2c-gpio");
