/* 
 * Copyright (C) 2012 Grandstream Network, Inc.
 * By Shaoyun Xie <sxie@grandstream.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* gxw42xx_keypad.c file */
/*
 * this driver monitor two gpio keys and one fake pin, link for link up/down events */

/* include Linux files */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/device.h>
#include <linux/cdev.h>		/* Used for struct cdev */
#include <linux/interrupt.h>    /* For IRQ_HANDLED and irqreturn_t */
#include <linux/semaphore.h>
#include <linux/poll.h>
#include <linux/vmalloc.h>
#include <linux/platform_device.h>
#include <linux/irq.h>	
#include <linux/sched.h>	
#include <linux/workqueue.h>
#include <asm/uaccess.h>	/* for VERIFY_READ/VERIFY_WRITE/ copy_from_user */
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/mach-ifxmips/gpio.h>
#include <asm/ifxmips/ifxmips_irq.h>

#undef KEY_SM_DEBUG
//#define KEY_SM_DEBUG

#define FAN_FAIL_INTERRUPT_SUPPORT

#define DRIVERNAME  "gxw-keypad"

#define PIN_DEBOUNCE_TIME       10  //ms
#define LINK_DEBOUNCE_TIME      1200
#define LCD_BACKLIT_DEALY       1200    // second

#define KEY_EXINT_MODE      2   /* Interrupt on falling edge */
#define KEY_EXINT12         12  /* EXINT3 for reset button*/
#define KEY_EXINT13         13  /* EXINT3 for reset button*/
#define IFXMIPS_KEY_DOWN_IRQ    (160+8+24)
#define IFXMIPS_KEY_ENTER_IRQ   (160+8+25)

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
#define FAN_DEBOUNCE_TIME       100  //ms
#define IFXMIPS_FAN_OFF_IRQ    (160+8+16)
#define IFXMIPS_FAN1_OFF_IRQ    (160+8+23)
#define IFXMIPS_FAN2_OFF_IRQ    (160+8+26)
#define IFXMIPS_FAN3_OFF_IRQ    (160+8+27)
#define IFXMIPS_FAN4_OFF_IRQ    (160+8+17)

#ifndef CONFIG_GXW42XX_BUILD_48_96
/* gxw4232/24/16 version 2.3 */
#define GPIO_FAN_DET       3    //P0.P3
#define GPIO_FAN_CTRL      34   //P1.P14
#else
/* gxw4248/96 version 1.1 */
#define GPIO_FAN_DET       6    //P0.P6
#define GPIO_FAN_CTRL      28   //P1.P8
#endif

#endif

/* for lcd backlight pin, only 4248 v1.1 is different */
extern short get_led_gpioset();
#define GPIO_LCDBACKLIT     ((get_led_gpioset() != 9 && get_led_gpioset() != 13)?38:27)
#define GPIO_DOWN           37
#define GPIO_ENTER          36

#define PIN_VAL_LINK  0
#define PIN_VAL_DOWN  0 //when down key pressed ,pin val is low
#define PIN_VAL_ENTER 0

/*
 * Keypad event definition
 */
#define EVENT_LINK                      0x4000
#define EVENT_NOTIFY                    0x2000
#define EVENT_KEY                       0x1000
#define EVENT_LINK_ON                   (EVENT_LINK     | 0x01)
#define EVENT_LINK_OFF                  (EVENT_LINK     | 0x02)
#define EVENT_KEY_DOWN                  (EVENT_KEY      | 0x01)
#define EVENT_KEY_ENTER                 (EVENT_KEY      | 0x02)
#define EVENT_NOTIFY_RESET1             (EVENT_NOTIFY	| 0x01)

typedef struct {
    unsigned long    event_id;
    unsigned long    timestamp;        /* in ms */
} GXW_INPUT_EVENT, *GXW_INPUT_EVENT_P;

typedef struct {
    struct list_head node;
    GXW_INPUT_EVENT  gs_event;
} GXW_LIST_INPUT_EVENT, *GXW_LIST_INPUT_EVENT_P;

typedef struct {
    unsigned long    param;                /* for input/output */
    unsigned long    reserved;
} GXW_KP_IOCTL_PARAM, *GXW_KP_IOCTL_PARAM_P;

/* IOCTL command */
typedef enum
{
    GXW_KP_IOCTL_START = 0,
    GXW_KP_IOCTL_END
} GXW_KP_IOCTL_CMD;

/** state machine events */ 
typedef enum
{  
    GXW_KP_EVT_TIMEOUT,
    GXW_KP_EVT_ON,
    GXW_KP_EVT_OFF,

} GXW_KP_EVT_t;

typedef enum
{
    GXW_KP_STATE_PRE_ON = 0,
    GXW_KP_STATE_PRE_OFF,
    GXW_KP_STATE_ON,
    GXW_KP_STATE_OFF,
} GXW_KP_KEY_SM_STATE_t;

#define GXW_KP_TOTAL_STATE                ((int)GXW_KP_STATE_OFF + 1)

struct gxw_kp {
    struct input_dev *input;
    struct timer_list timer;
    int irq;
    unsigned int rows;
    unsigned int cols;
    unsigned long delay;
    unsigned int debounce;
};

struct gxw_kpad_platform_data {
    int rows;
    int cols;
    int *keymap;
    unsigned int keymapsize;
    unsigned int rep:1;
    unsigned long delay;
    unsigned int dbounce:1;
    unsigned int row_gpios[5];
    unsigned int col_gpios[6];
};

#define KEY(col, row, val) (((col) << 28) | ((row) << 24) | (val))

static int gxw42xx_keymap[] = {
    KEY(0, 0, KEY_DOWN),
    KEY(0, 1, KEY_ENTER),
#ifdef FAN_FAIL_INTERRUPT_SUPPORT
    KEY(0, 1, KEY_FAN_FAIL_1),
    KEY(0, 1, KEY_FAN_FAIL_2),
    KEY(0, 1, KEY_FAN_FAIL_3),
    KEY(0, 1, KEY_FAN_FAIL_4),
#endif
};


static struct gxw_kpad_platform_data gxw42xx_platform_data = {
    .rows       = 2, /* we have no actual keypad for gxw */
    .cols       = 1,    
    .keymap     = gxw42xx_keymap,
    .keymapsize = 2,
    .rep        = 0,
    .delay      = 10,
    .dbounce    = 0,
    .row_gpios  = {},
    .col_gpios  = {},
};

typedef  void (*generate_event_t)(GXW_KP_EVT_t event);

typedef struct {
    char name[32];
    GXW_KP_KEY_SM_STATE_t state;
    unsigned debounce_time;
    generate_event_t generate_event;
    struct timer_list timer;
#ifdef FAN_FAIL_INTERRUPT_SUPPORT
    int gpio;
    int exint;
    int irq;
#endif
} GXW_KP_SM_t;

static struct timer_list lcdbacklit_timer;

void link_generate_event(GXW_KP_EVT_t event);
void down_generate_event(GXW_KP_EVT_t event);
void enter_generate_event(GXW_KP_EVT_t event);

GXW_KP_SM_t gxw_kpad_sm[] = {
    {   .name = "network_link",
        .generate_event = (generate_event_t)link_generate_event,
        .debounce_time  = LINK_DEBOUNCE_TIME,
    },
    {   .name = "key_down",
        .generate_event = (generate_event_t)down_generate_event,
        .debounce_time  = PIN_DEBOUNCE_TIME,
    },
    {   .name = "key_okay",
        .generate_event = (generate_event_t)enter_generate_event,
        .debounce_time  = PIN_DEBOUNCE_TIME,
    },
};
#define gxw_kpad_sm_link gxw_kpad_sm[0]
#define gxw_kpad_sm_down gxw_kpad_sm[1]
#define gxw_kpad_sm_enter gxw_kpad_sm[2]

#define GXW_KP_SM_COUNT     (sizeof(gxw_kpad_sm)/sizeof(GXW_KP_SM_t))

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
static struct timer_list fan_timer;
static int timer_lock = 0;
void fan_generate_event(GXW_KP_EVT_t event);

/*  Only gxw4232/24/16 hardware verion >=2.3,  gxw4248/96 hardware version >= 1.1 
 *  support fan failure interruption 
 */
#ifndef CONFIG_GXW42XX_BUILD_48_96
GXW_KP_SM_t gxw_fan_sm[] = {
    {   .name = "fan_off_1",
        .generate_event = (generate_event_t)fan_generate_event,
        .debounce_time  = FAN_DEBOUNCE_TIME,
        .gpio = 11, //P0.P11 
        .exint = 4, //EXINT4
        .irq = IFXMIPS_FAN_OFF_IRQ,
    },
};
#else 
GXW_KP_SM_t gxw_fan_sm[] = {
    {   .name = "fan_off_1",
        .generate_event = (generate_event_t)fan_generate_event,
        .debounce_time  = FAN_DEBOUNCE_TIME,
        .gpio = 38,  //P1.P18
        .exint = 11, //EXINT11
        .irq = IFXMIPS_FAN1_OFF_IRQ, 
    },
    {   .name = "fan_off_2",
        .generate_event = (generate_event_t)fan_generate_event,
        .debounce_time  = FAN_DEBOUNCE_TIME,
        .gpio = 35,  //P1.P15
        .exint = 14, //EXINT14  
        .irq = IFXMIPS_FAN2_OFF_IRQ, 
    },
    {   .name = "fan_off_3",
        .generate_event = (generate_event_t)fan_generate_event,
        .debounce_time  = FAN_DEBOUNCE_TIME,
        .gpio = 39,  //P1.P19
        .exint = 15, //EXINT15
        .irq = IFXMIPS_FAN3_OFF_IRQ, 
    },
    {   .name = "fan_off_4",
        .generate_event = (generate_event_t)fan_generate_event,
        .debounce_time  = FAN_DEBOUNCE_TIME,
        .gpio = 12, //P0.P12
        .exint = 5, //EXINT5
        .irq = IFXMIPS_FAN4_OFF_IRQ, 
    },
};

#endif
#define GXW_FAN_SM_COUNT     (sizeof(gxw_fan_sm)/sizeof(GXW_KP_SM_t))
#endif /* FAN_FAIL_INTERRUPT_SUPPORT */

typedef void (*GXW_KP_STATE_HANDLER) (GXW_KP_SM_t *pState, GXW_KP_EVT_t evt);

static void  gxw_sh_PreON(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt);
static void  gxw_sh_PreOFF(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt);
static void  gxw_sh_ON(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt);
static void  gxw_sh_OFF(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt);

static void gxw_lcd_timed_on(void);
static unsigned int lcd_backlight = 0;

int gxw42xx_kp_key_generate_event (unsigned long data);

GXW_KP_STATE_HANDLER  gKeyStates[GXW_KP_TOTAL_STATE] = {
                        gxw_sh_PreON,
                        gxw_sh_PreOFF,
                        gxw_sh_ON,
                        gxw_sh_OFF,                      
                        };

/* global variable of type cdev to register driver to the kernel */
static struct cdev cdev;
static dev_t dev;
static struct class *gxw_kpad_class = NULL;

//static struct input_dev gxw42xx_input_dev;
static struct gxw_kp gxw42xx_kp;
static int *keymap;

static DEFINE_MUTEX(kp_enable_mutex);
static int kp_enable = 1;
static int link_notify_ready = 0;
static int ate_testmode = 0;
static int ate_teststate = 0;

LIST_HEAD(gxw42xx_kp_gs_event_list);
DECLARE_WAIT_QUEUE_HEAD(gxw42xx_kp_gs_event_queue);

struct workqueue_struct *ate_workqueue;
struct work_struct ate_work;

static DEFINE_SPINLOCK(gxw42xx_kp_gs_event_lock);
int    kpad_is_open = 0;

static void atelcd_work(struct work_struct *work)
{
    extern void kgui_lcdtest( int mode, int width, int height);

    printk("LCD test mode %d\n", ate_teststate);
    kgui_lcdtest( ate_teststate, 128, 32 );
    if ( ++ate_teststate > 1 )
        ate_teststate = 0;
}

void link_generate_event(GXW_KP_EVT_t event)
{
    /* Turn on the LCD backlight in case of a cable plug/unplug */
    gxw_lcd_timed_on();

    if (event == GXW_KP_EVT_ON)
        gxw42xx_kp_key_generate_event(EVENT_LINK_ON);
    else
        gxw42xx_kp_key_generate_event(EVENT_LINK_OFF);
}

void down_generate_event(GXW_KP_EVT_t event)
{
    struct gxw_kp *gxw_kp = &gxw42xx_kp;
    unsigned int temp_lcd_backlight = lcd_backlight;

//    printk("generate key_down event: %d\n", event);
    gxw_lcd_timed_on();
    /* If the LCD backlight was off untill now then ignore the first button press*/
    if ( 0 == temp_lcd_backlight)
        return;

    /* ate test mode */
    if (GXW_KP_EVT_ON == event )
    {
        if ( gxw_kpad_sm_enter.state == GXW_KP_STATE_ON )
        {
            /* both down and enter are pressed, enter/exit test mode */
            ate_testmode = !ate_testmode;

            if ( ate_testmode )
                printk("enter LCD test mode\n");
            else
                printk("quit LCD test mode\n");
            ate_teststate = 0;

            /* turn off lcd backlight */
            gpio_set_value(GPIO_LCDBACKLIT, 0);
        }
        if ( ate_testmode)
        {
            queue_work(ate_workqueue, &ate_work);
        }

        input_report_key(gxw_kp->input, KEY_DOWN, 1);
    }
    else
        input_report_key(gxw_kp->input, KEY_DOWN, 0);

}

void enter_generate_event(GXW_KP_EVT_t event)
{
    struct gxw_kp *gxw_kp = &gxw42xx_kp;
    unsigned int temp_lcd_backlight = lcd_backlight;

//    printk("generate key_okay event: %d\n", event);
    gxw_lcd_timed_on();
    /* If the LCD backlight was off untill now then ignore the first button press*/
    if ( 0 == temp_lcd_backlight)
        return;

    if (GXW_KP_EVT_ON == event )
    {
        input_report_key(gxw_kp->input, KEY_ENTER, 1);
    }
    else
        input_report_key(gxw_kp->input, KEY_ENTER, 0);
}

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
void fan_generate_event(GXW_KP_EVT_t event)
{
    struct gxw_kp *gxw_kp = &gxw42xx_kp;
    unsigned int temp_lcd_backlight = lcd_backlight;

//    printk("generate FAN event: %d\n", event);
    gxw_lcd_timed_on();
    /* If the LCD backlight was off untill now then ignore the first button press*/
    if ( 0 == temp_lcd_backlight)
        return;

    if (GXW_KP_EVT_OFF == event )
    {
        input_report_key(gxw_kp->input, KEY_FAN_FAIL_1, 1);
    }
    else
        input_report_key(gxw_kp->input, KEY_FAN_FAIL_1, 0);
}
#endif

static void gxw_kp_sm_timer(unsigned long data)
{
    GXW_KP_EVT_t evt = GXW_KP_EVT_TIMEOUT;
    GXW_KP_SM_t *kp_sm = (GXW_KP_SM_t *)data;

#ifdef KEY_SM_DEBUG
    printk("%s timer expired\n", kp_sm->name);
#endif
    gKeyStates[kp_sm->state](kp_sm, evt);
}

static void gxw_lcdbacklit_timer(unsigned long data)
{
    /* turn off lcd backlight */
    gpio_set_value(GPIO_LCDBACKLIT, 0);
    lcd_backlight = 0;
}

static void gxw_lcd_timed_on(void)
{
    mod_timer(&lcdbacklit_timer, jiffies + LCD_BACKLIT_DEALY * HZ);
    /* turn on lcd backlight */
    gpio_set_value(GPIO_LCDBACKLIT, 1);
    lcd_backlight = 1;
}

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
static void fan_enable_int(void)
{
    del_timer(&fan_timer);

    timer_lock = 0;
 //   printk("FAN: re-enable interrupts ts=%ld\n", jiffies * 1000 / HZ);   
#ifndef CONFIG_GXW42XX_BUILD_48_96
    enable_irq(IFXMIPS_FAN_OFF_IRQ);
#else
    enable_irq(IFXMIPS_FAN1_OFF_IRQ);
    enable_irq(IFXMIPS_FAN2_OFF_IRQ);
    enable_irq(IFXMIPS_FAN3_OFF_IRQ);
    enable_irq(IFXMIPS_FAN4_OFF_IRQ);
#endif
}

static irqreturn_t gxw_fan_off_interrupt(int irq, void *dev_id)
{
    /* down pin toggled */
    GXW_KP_EVT_t evt;
    GXW_KP_SM_t * ptr_fan_sm = NULL;
    int pinval;
    int fan_ctrl = 0, fan_det = 0;
    
    /* Use lock to make sure we don't disable fan off irq many times */
    if(0 == timer_lock)
    {
#ifndef CONFIG_GXW42XX_BUILD_48_96
        disable_irq_nosync(IFXMIPS_FAN_OFF_IRQ);
#else 
        disable_irq_nosync(IFXMIPS_FAN1_OFF_IRQ);
        disable_irq_nosync(IFXMIPS_FAN2_OFF_IRQ);
        disable_irq_nosync(IFXMIPS_FAN3_OFF_IRQ);
        disable_irq_nosync(IFXMIPS_FAN4_OFF_IRQ);
#endif
    }
    ptr_fan_sm = (GXW_KP_SM_t *) dev_id;
        
    if ( ptr_fan_sm == NULL )
    {
       add_timer(&fan_timer);
       printk("dev_id is NULL, fan interrupt handler returns\n");
       return IRQ_HANDLED;
    }
    
    fan_ctrl = gpio_get_value(GPIO_FAN_CTRL);
    fan_det = gpio_get_value(GPIO_FAN_DET);

    /* In case of interrupt when the fan is not working simply ignore it */
    if ( fan_ctrl == 0 && fan_det == 1 )
    {
#ifdef KEY_SM_DEBUG
        printk("FAN: SKIP: fan_ctrl = %d, fan_det = %d\n", fan_ctrl, fan_det );
#endif
        fan_timer.expires = jiffies + 1000 * HZ / 1000; //FAN_DEBOUNCE_TIME;
    }
    /* We got interrupt while the fan was ON indicating a actual failure */
    else if ( fan_ctrl == 1 || fan_det == 0 )
    {
#ifdef KEY_SM_DEBUG
        printk("FAN: fan_ctrl = %d, fan_det = %d\n", fan_ctrl, fan_det );
#endif
        pinval = gpio_get_value(ptr_fan_sm->gpio);
        evt = (pinval == 0?  GXW_KP_EVT_ON : GXW_KP_EVT_OFF);

#ifdef KEY_SM_DEBUG
        printk("%s, %s\n", ptr_fan_sm->name, pinval == 0?  "EVT_ON" : "EVT_OFF");
#endif

        if (gKeyStates[ptr_fan_sm->state])
            gKeyStates[ptr_fan_sm->state](ptr_fan_sm, evt);

        fan_timer.expires = jiffies + 100 * HZ / 1000; //FAN_DEBOUNCE_TIME;
    }

    /* We start a timer to re-enable the FAN failure IRQ because of spurious triggeres
       caused by the fan circuitry. timer_lock ensure we only have one fan timer enabled */
    if( 0 == timer_lock)
    { 
        add_timer(&fan_timer);
        timer_lock = 1;
    }
    
    return IRQ_HANDLED;
}
#endif /* FAN_FAIL_INTERRUPT_SUPPORT */

static irqreturn_t gxw_kpad_down_interrupt(int irq, void *dev_id)
{
    /* down pin toggled */
    GXW_KP_EVT_t evt;
    int pinval;

    disable_irq_nosync(IFXMIPS_KEY_DOWN_IRQ);

    pinval = gpio_get_value(GPIO_DOWN);
    evt = (pinval == 0?  GXW_KP_EVT_ON : GXW_KP_EVT_OFF);
#ifdef KEY_SM_DEBUG
    printk("%s, %s\n", gxw_kpad_sm_down.name, pinval == 0?  "EVT_ON" : "EVT_OFF");
#endif

    if (gKeyStates[gxw_kpad_sm_down.state])
        gKeyStates[gxw_kpad_sm_down.state](&gxw_kpad_sm_down, evt);

    enable_irq(IFXMIPS_KEY_DOWN_IRQ);

    return IRQ_HANDLED;
}

static irqreturn_t gxw_kpad_enter_interrupt(int irq, void *dev_id)
{
    /* enter pin toggled */
    GXW_KP_EVT_t evt;
    int pinval;

    disable_irq_nosync(IFXMIPS_KEY_ENTER_IRQ);

    pinval = gpio_get_value(GPIO_ENTER);
    evt = (pinval == 0?  GXW_KP_EVT_ON : GXW_KP_EVT_OFF);
#ifdef KEY_SM_DEBUG
    printk("%s, %s\n", gxw_kpad_sm_down.name, pinval == 0?  "EVT_ON" : "EVT_OFF");
#endif

    if (gKeyStates[gxw_kpad_sm_enter.state])
        gKeyStates[gxw_kpad_sm_enter.state](&gxw_kpad_sm_enter, evt);

    enable_irq(IFXMIPS_KEY_ENTER_IRQ);

    return IRQ_HANDLED;
}

void gxw_kpad_notify_link(int link)
{
    /* enter pin toggled */
    GXW_KP_EVT_t evt;
    int pinval = link;

    if (link_notify_ready == 0)     // we have not initialized yet.
        return;

    evt = (pinval == 0?  GXW_KP_EVT_ON : GXW_KP_EVT_OFF);
#ifdef KEY_SM_DEBUG
    printk("%s, %s\n", gxw_kpad_sm_link.name, pinval == 0?  "EVT_ON" : "EVT_OFF");
#endif

    if (gKeyStates[gxw_kpad_sm_link.state])
        gKeyStates[gxw_kpad_sm_link.state](&gxw_kpad_sm_link, evt);
}


/*
static inline int gxw_kpad_find_key(int col, int row)
{
   int i, key;

   key = KEY(col, row, 0);
   for (i = 0; keymap[i] != 0; i++)
      if ((keymap[i] & 0xff000000) == key)
         return keymap[i] & 0x00ffffff;
   return -1;
}
*/

static ssize_t gxw_kp_enable_show(struct device *dev,
                   struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", kp_enable);
}

static ssize_t gxw_kp_enable_store(struct device *dev, struct device_attribute *attr,
                    const char *buf, size_t count)
{
    int state;

    if (sscanf(buf, "%u", &state) != 1)
        return -EINVAL;

    if ((state != 1) && (state != 0))
        return -EINVAL;

    mutex_lock(&kp_enable_mutex);
    if (state != kp_enable) {
        kp_enable = state;
    }
    mutex_unlock(&kp_enable_mutex);

    return strnlen(buf, count);
}

static DEVICE_ATTR(enable, S_IRUGO | S_IWUSR, gxw_kp_enable_show, gxw_kp_enable_store);

#ifdef CONFIG_PM
static int gxw_kp_suspend(struct platform_device *dev, pm_message_t state)
{
   /* Nothing yet */

   return 0;
}

static int gxw_kp_resume(struct platform_device *dev)
{
   /* Nothing yet */

   return 0;
}
#else
#define gxw_kp_suspend   NULL
#define gxw_kp_resume   NULL
#endif

static int gxw_kpad_start(struct platform_device *pdev)
{
    struct gxw_kp *gxw_kp = &gxw42xx_kp;
    struct input_dev *input_dev; // = &gxw42xx_input_dev;
    struct gxw_kpad_platform_data *pdata =  &gxw42xx_platform_data;
    int i;
    int err;

    input_dev = input_allocate_device();
    gxw_kp->input = input_dev;

    keymap = pdata->keymap;

    if (pdata->rep)
       set_bit(EV_REP, input_dev->evbit);

    if (pdata->delay)
       gxw_kp->delay = pdata->delay;

    printk("Initialize keypad for gxw42xx.\n");

    err = device_create_file(&pdev->dev, &dev_attr_enable);
    if (err < 0)
        goto err2;

    /* setup input device */
    set_bit(EV_KEY, input_dev->evbit);
    for (i = 0; keymap[i] != 0; i++)
       set_bit(keymap[i] & KEY_MAX, input_dev->keybit);
    input_dev->name = DRIVERNAME;
    input_dev->phys = "gxw-keypad/input0";
	input_dev->dev.parent = &pdev->dev;

    input_dev->id.bustype = BUS_HOST;
    input_dev->id.vendor = 0x0001;
    input_dev->id.product = 0x0001;
    input_dev->id.version = 0x0001;

    input_dev->keycode = keymap;
    input_dev->keycodesize = sizeof(unsigned int);
    input_dev->keycodemax = pdata->keymapsize;

    err = input_register_device(gxw_kp->input);
    if (err)
        goto err2;

    return 0;
err2:
    return -EINVAL;
}

/* Functions */
int gxw_kpad_open(struct inode *inode, struct file *filp)
{
    kpad_is_open = 1;
	return 0;
}

int gxw_kpad_release(struct inode *inode, struct file *filp)
{
    kpad_is_open = 0;
	return 0;
}

int gxw42xx_kp_key_generate_event (unsigned long data)
{
    GXW_LIST_INPUT_EVENT    *event = NULL;
    unsigned long           flags = 0;

    /* drop event on the floor if no one's listening */
    if ((data & EVENT_NOTIFY) == 0)
        printk("report event 0x%lx\n", data);
    if (!kpad_is_open)
        return 0;

    event = kmalloc(sizeof(*event), GFP_ATOMIC);
    if (!event)
        return -ENOMEM;

    event->gs_event.event_id = data;
    event->gs_event.timestamp = jiffies * 1000/HZ;

    spin_lock_irqsave(&gxw42xx_kp_gs_event_lock, flags);
    list_add_tail(&event->node, &gxw42xx_kp_gs_event_list);
    spin_unlock_irqrestore(&gxw42xx_kp_gs_event_lock, flags);

    wake_up_interruptible(&gxw42xx_kp_gs_event_queue);

    return  0;
}

int gxw42xx_kp_key_receive_event (GXW_INPUT_EVENT *event)
{   
    GXW_LIST_INPUT_EVENT    *entry = NULL;
    unsigned long           flags = 0;

    DECLARE_WAITQUEUE(wait, current);
    
    if (!event)
        return  -EINVAL;
        
    if (list_empty(&gxw42xx_kp_gs_event_list)) {

        set_current_state(TASK_INTERRUPTIBLE);
        add_wait_queue(&gxw42xx_kp_gs_event_queue, &wait);

        if (list_empty(&gxw42xx_kp_gs_event_list))
            schedule();

        remove_wait_queue(&gxw42xx_kp_gs_event_queue, &wait);
        set_current_state(TASK_RUNNING);

        if (signal_pending(current))
            return -ERESTARTSYS;
    }

    spin_lock_irqsave(&gxw42xx_kp_gs_event_lock, flags);
    entry = list_entry(gxw42xx_kp_gs_event_list.next, GXW_LIST_INPUT_EVENT, node);
    if (entry)
        list_del(&entry->node);
    spin_unlock_irqrestore(&gxw42xx_kp_gs_event_lock, flags);

    if (!entry)
        return -ENODEV;

    memcpy(event, &entry->gs_event, sizeof(*event));

    kfree(entry);

    return 0;
}

ssize_t gxw_kpad_read(struct file *fp, char __user *buf, size_t len, loff_t *off)
{
    GXW_INPUT_EVENT event;
    int result;

    memset(&event, 0, sizeof(event));

    if ((fp->f_flags & O_NONBLOCK)
        && (list_empty(&gxw42xx_kp_gs_event_list)))
        return  -EAGAIN;

    result = gxw42xx_kp_key_receive_event(&event);
    if (result) {
        return -EIO;
    }
    
    if (len < sizeof(event))
        return -EIO;
    else
        len = sizeof(event);

    if (copy_to_user(buf, &event, len))
        return -EFAULT; 

    return len;
}

int gxw_kpad_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    //GXW_KP_IOCTL_PARAM ioctl_param;
    //GXW_KP_IOCTL_PARAM_P p_ioctl_param = &ioctl_param;
    
    switch ((GXW_KP_IOCTL_CMD)cmd)
    {
	default:
            printk(" Un-supported IO control !!!");
            return -EINVAL;
    }

	return 0;
}


static void  gxw_sh_PreON(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt)
{
   switch (evt)
   {
        case GXW_KP_EVT_TIMEOUT:
#ifdef KEY_SM_DEBUG
            printk("%s_PreON: timout received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            pState->state = GXW_KP_STATE_ON;
            /* report enter on event */
            pState->generate_event(GXW_KP_EVT_ON);
            break;
        case GXW_KP_EVT_OFF:
#ifdef KEY_SM_DEBUG
            printk("%s_PreON: off irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            /* back to off state and cancel the timer */
            pState->state = GXW_KP_STATE_OFF;
            del_timer(&pState->timer);
            break;
        case GXW_KP_EVT_ON:
#ifdef KEY_SM_DEBUG
            printk("%s_PreON: on irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            break;
        default:
            break;
    }
}
static void  gxw_sh_PreOFF(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt)
{
   switch (evt)
   {
        case GXW_KP_EVT_TIMEOUT:
#ifdef KEY_SM_DEBUG
            printk("%s_PreOFF: timout received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            pState->state = GXW_KP_STATE_OFF;
            /* report enter off event */
            pState->generate_event(GXW_KP_EVT_OFF);
            break;
        case GXW_KP_EVT_ON:
#ifdef KEY_SM_DEBUG
            printk("%s_PreOFF: on irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            /* back to on state and cancel the timer */
            pState->state = GXW_KP_STATE_ON;
            del_timer(&pState->timer);
            break;
        case GXW_KP_EVT_OFF:
#ifdef KEY_SM_DEBUG
            printk("%s_PreOFF: off irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            break;
        default:
            break;
    }
}

static void  gxw_sh_ON(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt)
{
    switch (evt)
    {
        case GXW_KP_EVT_OFF:
#ifdef KEY_SM_DEBUG
            printk("%s_ON: off irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            pState->state = GXW_KP_STATE_PRE_OFF;
            /* start debounce timer */
#ifdef KEY_SM_DEBUG
            printk("%s_ON: start debounce timer, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            mod_timer(&pState->timer, jiffies + pState->debounce_time * HZ / 1000);
            break;
        case GXW_KP_EVT_TIMEOUT:
            break;
        case GXW_KP_EVT_ON:
#ifdef KEY_SM_DEBUG
            printk("%s_ON: on irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            break;
        default:
            /* do nothing */
            break;
    }
}

static void  gxw_sh_OFF(GXW_KP_SM_t *pState, GXW_KP_EVT_t evt)
{
   switch (evt)
   {
        case GXW_KP_EVT_ON:
#ifdef KEY_SM_DEBUG
            printk("%s_OFF: off irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            pState->state = GXW_KP_STATE_PRE_ON;
            /* start debounce timer */
#ifdef KEY_SM_DEBUG
            printk("%s_OFF: start debounce timer, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            if (pState->debounce_time)
                mod_timer(&pState->timer, jiffies + pState->debounce_time * HZ / 1000);
            break;
        case GXW_KP_EVT_TIMEOUT:
            break;
        case GXW_KP_EVT_OFF:
#ifdef KEY_SM_DEBUG
            printk("%s_OFF: off irq received, ts=%ld\n", pState->name, jiffies * 1000 / HZ);
#endif
            break;
        default:
            /* do nothing */
            break;
    }
}


static int gxw42xx_kp_sm_init(void)
{
    int ii;
    int pinval = 0;

    /* initial states */
    if (gpio_request(GPIO_DOWN, DRIVERNAME) < 0) {
         printk(KERN_ERR "Failed to request"
                "GPIO%d for down state\n",
                GPIO_DOWN);
    }
    if (gpio_request(GPIO_ENTER, DRIVERNAME) < 0) {
         printk(KERN_ERR "Failed to request"
                "GPIO%d for enter state\n",
                GPIO_ENTER);
    }
    //gpio_direction_input(GPIO_DOWN);
    //gpio_direction_input(GPIO_ENTER);
    ifxmips_port_set_dir_in(1, 17);
    ifxmips_port_set_dir_in(1, 16);

    /* initialize state machines */
    for (ii = 0; ii < GXW_KP_SM_COUNT; ii++ )
    {
        printk("initilize State Machine [%s]\n", gxw_kpad_sm[ii].name);
        setup_timer(&gxw_kpad_sm[ii].timer, gxw_kp_sm_timer, (unsigned long)&gxw_kpad_sm[ii]);
    }

    {
        extern int switchchip_link_status(void);

        pinval = switchchip_link_status();
        gxw_kpad_sm_link.state = (pinval == 0?  GXW_KP_STATE_ON : GXW_KP_STATE_OFF);
    }

    pinval = gpio_get_value(GPIO_DOWN);
    gxw_kpad_sm_down.state = (pinval == 0?  GXW_KP_STATE_ON : GXW_KP_STATE_OFF);

    pinval = gpio_get_value(GPIO_ENTER);
    gxw_kpad_sm_enter.state = (pinval == 0?  GXW_KP_STATE_ON : GXW_KP_STATE_OFF);

    /* register gpio interrupt for the link,down and enter */
    if (request_irq(IFXMIPS_KEY_DOWN_IRQ, gxw_kpad_down_interrupt, 
                0, "gxw_kpgxw-keypad-downkey", 0) < 0)
    {
        printk("request down irq failed\n");
        return -1;
    }
    if (ifx_enable_external_int(KEY_EXINT12, KEY_EXINT_MODE) != 0)
        printk("error, enable EXINT%d failed\n", KEY_EXINT12);

    if (request_irq(IFXMIPS_KEY_ENTER_IRQ, gxw_kpad_enter_interrupt, 
                0, "gxw_kpgxw-keypad-enterkey", 0) < 0)
    {
        printk("request enter irq failed\n");
        return -1;
    }
    if (ifx_enable_external_int(KEY_EXINT13, KEY_EXINT_MODE) != 0)
        printk("error, enable EXINT%d failed\n", KEY_EXINT13);

    return 0;
}

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
/* Modified version of ifx_enable_external_int in interrupt.c */
static int ifx_enable_external_int_no_mask(int port, int pin, u32 mode)
{   
   if ((mode < 0) || (mode > 6))
      return -EINVAL;
    
   ifxmips_port_clear_exintcr0 (port, pin);
   ifxmips_port_clear_exintcr1 (port, pin);
   ifxmips_port_clear_irncfg (port, pin);

   if (mode & 0x1)
      ifxmips_port_set_exintcr0 (port, pin);
   if (mode & 0x2)
      ifxmips_port_set_exintcr1 (port, pin);
   if (mode & 0x4)
      ifxmips_port_set_irncfg (port, pin);
    
   return ifxmips_port_set_irnen (port, pin);
}

static int gxw42xx_fan_sm_init(void)
{
    int ii;
    int pinval = 0;
    
    /* for gxw4232/24/16 version <= 2.2, we don't need fan interruption support */
    /* for gxw4248/96 version <=1.0, we don't have fan interruption support */ 
    if ( get_led_gpioset() < 6 )
        return 0; 
   
#ifndef CONFIG_GXW42XX_BUILD_48_96
/* for gxw4232/24/16 version 2.3 */
    /*  We have only one fan on 4232 */
    ii = 0 ;
    /* Gpio request and direction set aready have been done in nand_gs.c */ 
    /* Already request in proc_fs init */
    printk("initilize FAN State Machine [%s]\n", gxw_fan_sm[ii].name);
    setup_timer(&gxw_fan_sm[ii].timer, gxw_kp_sm_timer, (unsigned long)&gxw_fan_sm[ii]);

    /* initialize state machines */
    pinval = gpio_get_value(gxw_fan_sm[ii].gpio);
    gxw_fan_sm[ii].state = (pinval == 0?  GXW_KP_STATE_ON : GXW_KP_STATE_OFF);

    /* register gpio interrupt for the fan failure controls */
    if (request_irq(gxw_fan_sm[ii].irq, gxw_fan_off_interrupt, 
                            IRQF_SHARED, "gxw_kpgxw-fan_off", &gxw_fan_sm[ii]) < 0)
    {
            printk("request fan irq failed\n");
            return -1;
    }

    /* Make sure interrupt won't come in before we start fan timer */
    disable_irq_nosync(gxw_fan_sm[ii].irq);

    if (ifx_enable_external_int(gxw_fan_sm[ii].exint, 6) != 0)
        printk("error, enable EXINT%d failed\n", gxw_fan_sm[ii].exint);

#else
/* for gxw4248/96 version 1.1 */
    for (ii = 0; ii < GXW_FAN_SM_COUNT; ii++ )
    {
        /* Gpio request and direction set aready have been done in nand_gs.c */ 
        /* EXINT ports used to be set in mux.c. We set them here because we don't want to change kernel config file */
        ifxmips_port_set_altsel0 (GPIO_TO_PORT(gxw_fan_sm[ii].gpio), GPIO_TO_GPIO(gxw_fan_sm[ii].gpio));
        ifxmips_port_clear_altsel1 (GPIO_TO_PORT(gxw_fan_sm[ii].gpio), GPIO_TO_GPIO(gxw_fan_sm[ii].gpio));
       
        /* Already request in proc_fs init */
        printk("initilize FAN State Machine [%s]\n", gxw_fan_sm[ii].name);
        setup_timer(&gxw_fan_sm[ii].timer, gxw_kp_sm_timer, (unsigned long)&gxw_fan_sm[ii]);

        /* initialize state machines */
        pinval = gpio_get_value(gxw_fan_sm[ii].gpio);
        gxw_fan_sm[ii].state = (pinval == 0?  GXW_KP_STATE_ON : GXW_KP_STATE_OFF);
       
        /* register gpio interrupt for the fan failure controls */
        if (request_irq(gxw_fan_sm[ii].irq, gxw_fan_off_interrupt, 
                IRQF_SHARED, "gxw_kpgxw-fan_off", &gxw_fan_sm[ii]) < 0)
        {
            printk("request fan irq failed\n");
            return -1;
        }
        /* Make sure interrupt won't come in before we start fan timer */
        disable_irq_nosync(gxw_fan_sm[ii].irq);

        /* Use modified version ifx_enable_external_int, because we don't want to change kernel configure file */
        if (ifx_enable_external_int_no_mask(GPIO_TO_PORT(gxw_fan_sm[ii].gpio), GPIO_TO_GPIO(gxw_fan_sm[ii].gpio), 6) != 0)
            printk("error, enable EXINT%d failed\n", gxw_fan_sm[ii].exint);
    
    }
#endif /* CONFIG_GXW42XX_BUILD_48_96 */
   
    /* Init a timer that controlls the interrupt re-enabling */
    init_timer(&fan_timer);
    fan_timer.function = fan_enable_int;
    fan_timer.expires = jiffies + 1000 * HZ / 1000; //FAN_DEBOUNCE_TIME;
    add_timer(&fan_timer);
    return 0;
}
#endif /* FAN_FAIL_INTERRUPT_SUPPORT */

unsigned int gxw_kpad_poll (struct file *fp, struct poll_table_struct *wait)
{
    poll_wait(fp, &gxw42xx_kp_gs_event_queue, wait);
    if (!list_empty(&gxw42xx_kp_gs_event_list))
        return POLLIN | POLLRDNORM;
    return 0;
}

/*
int gxw_kpad_get_key_state(int *link, int *down)
{
    if (link)
    {
        if (kp_key_state.state == GXW_KP_STATE_LINKON || kp_key_state.state == GXW_KP_STATE_PRE_LINKOFF)
            *link = 1;
        else
            *link = 0;
    }

    if (down)
    {
        if (kp_key_state.state == GXW_KP_STATE_CAMON || kp_key_state.state == GXW_KP_STATE_PRE_CAMOFF)
            *down = 0;
        else
            *down = 1;
    }

    return 0;
}
*/

/* global variable of type file_operations containing function 
pointers of file operations */
static struct file_operations gxw_kpad_fops = {
	.owner = THIS_MODULE,
	.open = gxw_kpad_open,
	.read = gxw_kpad_read,
	.poll = gxw_kpad_poll,
	.release = gxw_kpad_release,
	.ioctl = gxw_kpad_ioctl,
};

#if 0
static char *gxw_kp_devnode(struct device *dev, mode_t *mode)
{
    return kasprintf(GFP_KERNEL, "gxw-keypad/input0");
}
#endif

static int __devinit gxw_kp_probe(struct platform_device *pdev)
{
	dev_t devid;
    int result;

    /* Register the driver in the kernel */
    /* dynmically get the major number for the driver using 
       alloc_chrdev_region function */
    result = alloc_chrdev_region(&devid, 0, 1, DRIVERNAME);

    /* if it fails return error */
    if (result < 0) {
        printk("GXW42xx_KeyPad: Module intialization \
                failed. could not register character device\n");
        return -ENODEV;
    } else {
    }

    /* initialize cdev with file operations */
    cdev_init(&cdev, &gxw_kpad_fops);

    cdev.owner = THIS_MODULE;
    cdev.ops = &gxw_kpad_fops;

    /* add cdev to the kernel */
    result = cdev_add(&cdev, devid, 1);

    if (result) {
        unregister_chrdev_region(devid, 1);
        printk("GXW42xx_KeyPad: Error adding \
        GXW42xx_KeyPad .. error no:%d\n", result);
        return -EINVAL;
    }

    gxw_kpad_class = class_create(THIS_MODULE, "gxw42xx_keypad");
    if (IS_ERR(gxw_kpad_class)) {
        printk("gxw_kpad_init: error in creating device class\n");
        unregister_chrdev_region(dev, 1);
        cdev_del(&cdev);
        return -EIO;
    }
    //gxw_kpad_class->devnode = gxw_kp_devnode;
    device_create(gxw_kpad_class, NULL, devid, NULL, "gxw42xx_keypad");

    gxw_kpad_start(pdev);

    gxw42xx_kp_sm_init();

#ifdef FAN_FAIL_INTERRUPT_SUPPORT
    gxw42xx_fan_sm_init();
#endif

    setup_timer(&lcdbacklit_timer, gxw_lcdbacklit_timer, 0);
    gxw_lcd_timed_on();

    INIT_WORK(&ate_work, atelcd_work);
    ate_workqueue = create_singlethread_workqueue("lcd_ate");

    link_notify_ready = 1;

    return 0;
}

static int __devexit gxw_kp_remove(struct platform_device *pdev)
{
    struct gxw_kp *gxw_kp = &gxw42xx_kp;

    /* unregister everything */
    input_unregister_device(gxw_kp->input);

    /* remove major number allocated to this driver */
    unregister_chrdev_region(dev, 1);

    kfree(gxw_kp);

    return 0;
}

static struct platform_driver gxw_kp_driver = {
    .probe      = gxw_kp_probe,
    .remove     = __devexit_p(gxw_kp_remove),
    .suspend    = gxw_kp_suspend,
    .resume     = gxw_kp_resume,
    .driver     = {
        .name   = "gxw42xx_keypad",
        .owner  = THIS_MODULE,
    },
};

static int __init gxw_kp_init(void)
{
    //printk(KERN_INFO "GXW42XX Keypad Driver\n");
    return platform_driver_register(&gxw_kp_driver);
}

static void __exit gxw_kp_exit(void)
{
    platform_driver_unregister(&gxw_kp_driver);
}

EXPORT_SYMBOL(gxw_kpad_notify_link);

module_init(gxw_kp_init)
module_exit(gxw_kp_exit)
MODULE_LICENSE("GPL");
