/*
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 *  Description:
 *    device driver of NAND controller of IFXMIPS chips
 *  Author:
 *    Samuels Xu Liang
 *  Created:
 *    6 Sep 2005
 *  History & Modification Tag:
 *  ___________________________________________________________________________
 *  |  Tag   |                  Comments                   | Modifier & Time  |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.0  | First version of this driver and the tag is | Samuels Xu Liang |
 *  |        | implied.                                    |    6 Jul 2005    |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.1  | Second version of this driver for INCAIP2   | Winder Sung      |
 *  |        |                                             |    3 Nov 2005    |
 *  |--------+---------------------------------------------+------------------|
 *  |  S1.0  | Adapted to Linux 2.6.28 with new control    | INgo Rah         |
 *  |        | API. Inserted NOPs for SVIP                 |    10 May 2010   |
 *  ---------------------------------------------------------------------------
 */

/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#ifdef LINUX_2_4
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/delay.h>
#else
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <linux/gpio.h>
#include <linux/io.h>
#endif 

/*
 *  Chip Specific Head File
 */
#include <asm/ifx/ifx_reg_base.h>
#include <asm/ifx/ifx_reg_ebu.h>

#if defined (CONFIG_INCAIP2) || defined (CONFIG_SVIP)
#define IFX_NAND_ADDR_REGION_BASE IFX_EBU_SEG1_BASE /* NAND flash in region0 (ebu_address_region  = 0) */
#endif


/*
 * ####################################
 *     Parameters to Configure NAND
 * ####################################
 */
static volatile u8 *ebu_base_address = (volatile u8*)(IFX_NAND_ADDR_REGION_BASE);

#ifdef MODULE
    static int  ebu_address_region  = 1;    /*  number of address region (0-4)      */
    static int  ebu_config          = 0;    /*  1 to config EBU, else to do nothing */

    MODULE_PARM(ebu_address_region, "i");
    MODULE_PARM_DESC(ebu_address_region, "Number of address region (0-4).");
    MODULE_PARM(ebu_config, "i");
    MODULE_PARM_DESC(ebu_config, "Set 1 to config EBU, else to do nothing.");
#else
    static int ebu_address_region    = 0;
    static int ebu_config            = 0;
#endif

#ifndef CONFIG_MTD_CMDLINE_PARTS
#define NUM_PARTITIONS                  6
static struct mtd_partition partition_info[NUM_PARTITIONS] = {
	{
		.name   = "2nd Level Loader",
		.offset = 0x0000000,
		.size   = 0x0004000
	},
	{
		.name   = "U-Boot",
		.offset = 0x0004000,
		.size   = 0x004C000
	},
        {
		.name   = "U-Boot Environment",
		.offset = 0x0050000,
		.size   = 0x0010000
	},
	{
		.name   = "Linux Kernel",
		.offset = 0x0060000,
		.size   = 0x01A0000
	},
	{
		.name   = "rootfs",
		.offset = 0x0200000,
		.size   = 0x0400000
	},
	{
		.name   = "spare",
		.offset = 0x0600000,
		.size   = 0x1A00000
	}
};
#endif

/*
 * ####################################
 *              Definition
 * ####################################
 */

#undef IFX_NAND_DEBUG_ON

/*
 *  Bits Operation
 */
#define GET_BITS(x, msb, lsb)           (((x) & ((1 << ((msb) + 1)) - 1)) >> (lsb))
#define SET_BITS(x, msb, lsb, value)    (((x) & ~(((1 << ((msb) + 1)) - 1) ^ ((1 << (lsb)) - 1))) | (((value) & ((1 << (1 + (msb) - (lsb))) - 1)) << (lsb)))

/*
 *  NAND IO Port Mapping
 */
#define NAND_CMD_DECODER_ALE(x)         ((x) ? (1 << 2) : 0)
#define NAND_CMD_DECODER_CLE(x)         ((x) ? (1 << 3) : 0)
#define NAND_CMD_DECODER_CS(x)          ((x) ? (1 << 4) : 0)
#define NAND_CMD_DECODER_SE(x)          ((x) ? (1 << 5) : 0)
#define NAND_CMD_DECODER_WP(x)          ((x) ? (1 << 6) : 0)
#define NAND_CMD_DECODER_PRE(x)         ((x) ? (1 << 7) : 0)
#define NAND_CLE_BIT	(1 << 3)
#define NAND_ALE_BIT	(1 << 2)

#define IFX_NAND_CMD                 ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(1) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_WRITE               ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(1) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))

#define IFX_NAND_WRITE_OVER          ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))


/*
 *  NAND Flash Device RD/BY State Register
 */
#define NAND_WAIT_WR_C                  (*IFX_EBU_NAND_WAIT & (1 << 3))
#define NAND_WAIT_RD                    ((*IFX_EBU_NAND_WAIT & 0x01) == 0x01)

/*
 *  NAND Flash ECC Register
 */
#define NAND_ECC0_B2                    GET_BITS(*IFX_EBU_NAND_ECC0, 23, 16)
#define NAND_ECC0_B1                    GET_BITS(*IFX_EBU_NAND_ECC0, 15, 8)
#define NAND_ECC0_B0                    GET_BITS(*IFX_EBU_NAND_ECC0, 7, 0)

#define NAND_ECC0_CLEAR()               do { *IFX_EBU_NAND_ECC0 = 0; } while (0)

/* Additional NOPs to let the FW still handle interrupts, 
 * otherwise the NAND accesss to the EBU would block the other cores to 
 * long when they also access the EBU */
#define IFX_NAND_NOPS 150 

/*
 * ####################################
 * Preparation of Debug trick
 * ####################################
 */

/*
 * ####################################
 *             Declaration
 * ####################################
 */

/*
 *  MTD Device Operations to Replace Default Routines
 */
#ifdef LINUX_2_4
static void ifx_hwcontrol(struct mtd_info *mtd, int cmd);
#else
static void ifx_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl);
#endif
static int ifx_dev_ready(struct mtd_info *);
static int ifx_calculate_ecc(struct mtd_info *, const u_char *, u_char *);
static void ifx_enable_hwecc(struct mtd_info *, int);

/*
 *  Init Functions
 */
static inline void check_parameter(void);
static inline void config_ebu(void);
static inline void config_nand(void);

/*
 * ####################################
 *            Local Variable
 * ####################################
 */

static struct mtd_info mtd = {0};
static struct nand_chip chip = {0};
static struct mtd_info mtd_256ecc = {0};
static struct nand_chip chip_256ecc = {0};
#ifdef LINUX_2_4
static u_char data_buf[512 + 16];
#endif

static int f_mtd_partitions = 0;
static int f_ecc_write = 0;


/*
 * ####################################
 *           Global Variable
 * ####################################
 */


/*
 * ####################################
 *            Local Function
 * ####################################
 */

#if defined (CONFIG_SVIP)
 /**
 * ifx_nand_write_buf - write buffer to chip
 * @mtd:	MTD device structure
 * @buf:	data buffer
 * @len:	number of bytes to write
 *
 * Write function for 8bit buswith.
 * This function inserts some NOPs to let the FW still handle interrupts, 
 * otherwise the NAND accesss to the EBU would block the other cores to 
 * long when they also access the EBU
 */
static void ifx_nand_write_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	int i, j;
	struct nand_chip *this = mtd->priv;

	for (i=0; i<len; i++)
   {
		writeb(buf[i], this->IO_ADDR_W);
      for (j=0; j<IFX_NAND_NOPS; j++)
         asm("nop");
   }
}

/**
 * ifx_nand_read_buf - read chip data into buffer 
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 * Read function for 8bit buswith.
 * This function inserts some NOPs to let the FW still handle interrupts, 
 * otherwise the NAND accesss to the EBU would block the other cores to 
 * long when they also access the EBU
 */
static void ifx_nand_read_buf(struct mtd_info *mtd, u_char *buf, int len)
{
	int i, j;
	struct nand_chip *this = mtd->priv;

	for (i=0; i<len; i++)
   {
		buf[i] = readb(this->IO_ADDR_R);
      for (j=0; j<IFX_NAND_NOPS; j++)
         asm("nop");
#ifdef IFX_NAND_DEBUG_ON      
      printk ("NAND read %02X\n", buf[i]);
#endif      
   }         
}
#endif /* defined (CONFIG_SVIP) */

#ifdef IFX_NAND_DEBUG_ON      
/**
 * nand_read_byte - [DEFAULT] read one byte from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 8bit buswith
 */
static uint8_t ifx_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
   uint8_t val;
	
   val = readb(chip->IO_ADDR_R);
   printk ("NAND readb %02X\n", val);

   return val;
}

/**
 * nand_write_byte - [DEFAULT] write one byte to the chip
 * @mtd:	MTD device structure
 * @byte:	pointer to data byte to write
 *
 * Default write function for 8it buswith
 */
static void ifx_write_byte(struct mtd_info *mtd, u_char byte)
{
	struct nand_chip *this = mtd->priv;
   static int cnt = 0;
	
   writeb(byte, this->IO_ADDR_W);
   if (cnt ++ < 25)
      printk ("NAND writeb %02X\n", byte);
}
#endif      

/* Hardware control function */
#ifdef LINUX_2_4
static void ifx_hwcontrol (struct mtd_info *mtd, int cmd)
{
	struct nand_chip *this = (struct nand_chip *) mtd->priv;

	switch (cmd) {
		case NAND_CTL_SETCLE:
			/**(unsigned long*)*/(unsigned long) this->IO_ADDR_W |= NAND_CLE_BIT;
			break;
		case NAND_CTL_CLRCLE:
			/**(unsigned long*)*/(unsigned long) this->IO_ADDR_W &= ~NAND_CLE_BIT;
			break;
		case NAND_CTL_SETALE:
			/**(unsigned long*)*/(unsigned long) this->IO_ADDR_W |= NAND_ALE_BIT;
			break;
		case NAND_CTL_CLRALE:
			/**(unsigned long*)*/(unsigned long) this->IO_ADDR_W &= ~NAND_ALE_BIT;
			break;
		case NAND_CTL_SETNCE:
			*IFX_EBU_NAND_CON |= IFX_EBU_NAND_CON_NANDMODE_VAL(1);
			break;
		case NAND_CTL_CLRNCE:
			*IFX_EBU_NAND_CON &= ~IFX_EBU_NAND_CON_NANDMODE_VAL(1);
			break;
    }
        printk ("NAND ctrl = %08X\n",  cmd);
}
#else
/*
 *	hardware specific access to control-lines
 *
 *	NAND_NCE: bit 0 -> CON register
 *	NAND_CLE: bit 1 -> bit 3
 *	NAND_ALE: bit 2 == bit 2
 */
static void ifx_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	 struct nand_chip *this = mtd->priv;

    if (ctrl & NAND_CTRL_CHANGE) {
         unsigned long adr;
         /* comming here means to change either the enable state or the address for
            controlling ALE or CLE */
        
        /* NAND_NCE: Select the chip by setting nCE to low. This is done in CON register */
        if (ctrl & NAND_NCE)
            *IFX_EBU_NAND_CON |= IFX_EBU_NAND_CON_NANDMODE_VAL(1);
        else
            *IFX_EBU_NAND_CON &= ~IFX_EBU_NAND_CON_NANDMODE_VAL(1);
        /* The addressing of CLE or ALE is done via different addresses.
           We are now changing the address depending on the given action 
           IFXMIPSs NAND_CLE_BIT = (1 << 3), NAND_CLE = 0x02 
           NAND_ALE_BIT = (1 << 2) = NAND_ALE (0x04) */
        adr = (unsigned long)this->IO_ADDR_W;
        adr &= ~(NAND_CLE_BIT | NAND_ALE_BIT);
        adr |= (ctrl & NAND_CLE) << 2 | (ctrl & NAND_ALE);
        this->IO_ADDR_W = (void __iomem *)adr;
    }
	
	if (cmd != NAND_CMD_NONE)
   {
      /* send a command */
		writeb(cmd, this->IO_ADDR_W);
#ifdef IFX_NAND_DEBUG_ON      
      printk ("NAND writeb %02X\n",  cmd);
#endif      
   }
}
#endif

/*
 *  Description:
 *    Check if the MTD device is ready to accesss.
 *  Input:
 *    mtd --- struct mtd_info *, MTD device to check.
 *  Output:
 *    none
 */
static int ifx_dev_ready(struct mtd_info *mtd)
{
    return NAND_WAIT_RD;
}

/*
 *  Description:
 *    Caculate ECC of a page. For read, ECC is calculated by hardware, however,
 *    for write, a standard function implemented in source file "nand_ecc.c" is
 *    called to produce the ECC.
 *  Input:
 *    mtd      --- struct mtd_info *, MTD device to do ECC calculation.
 *    dat      --- const u_char *, page data.
 *    ecc_code --- u_char *, buffer to put ECC.
 *  Output:
 *    none
 */
static int ifx_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
    if ( f_ecc_write )
        nand_calculate_ecc(mtd, dat, ecc_code);
    else
    {
        ecc_code[0] = NAND_ECC0_B0;
        ecc_code[1] = NAND_ECC0_B1;
        ecc_code[2] = NAND_ECC0_B2;
        *IFX_EBU_NAND_CON &= ~IFX_EBU_NAND_CON_ECC_ON_VAL(1);
    }
    return 0;
}


void ifx_nand_waitcmd(void)
{
    while(!NAND_WAIT_WR_C);
}

/*
 *  Description:
 *    Send signal to prepare for ECC calculation. For read, need to enable
 *    ECC hardware.
 *  Input:
 *    mtd  --- struct mtd_info *, MTD device to enable hardware ECC calculation.
 *    mode --- u_char *, signal ECC calculation for read or write.
 *  Output:
 *    none
 */
static void ifx_enable_hwecc(struct mtd_info *mtd, int mode)
{
    switch ( mode )
    {
    case NAND_ECC_READ:
        f_ecc_write = 0;
        NAND_ECC0_CLEAR();
        *IFX_EBU_NAND_CON |= IFX_EBU_NAND_CON_ECC_ON_VAL(1);
        break;
    case NAND_ECC_WRITE:
        f_ecc_write = 1;
        break;
    }
}

/*
 *  Description:
 *    Check and adjust parameters passed in by command "insmod".
 *  Input:
 *    none
 *  Output:
 *    none
 */
#ifdef MODULE
    static inline void check_parameter(void)
    {
        if ( ebu_address_region < 0 || ebu_address_region > 4 )
            ebu_address_region = 1;
        else if ( ebu_address_region == 4 )
            ebu_address_region = 3;

        if ( ebu_config != 1 )
            ebu_config = 0;

        switch ( ebu_address_region )
        {
        case 0:
            ebu_base_address = (volatile u8*)(IFX_EBU_SEG1_BASE); break;
        case 1:
            ebu_base_address = (volatile u8*)(IFX_EBU_SEG2_BASE); break;
        case 2:
            ebu_base_address = (volatile u8*)(IFX_EBU_SEG3_BASE); break;
        case 3:
            ebu_base_address = (volatile u8*)(IFX_EBU_SEG4_BASE); break;
        }
    }
#else
    static inline void check_parameter(void) {}
#endif

/*
 *  Description:
 *    Config EBU region.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static inline void config_ebu(void)
{
    IFX_EBU_ADDR_SEL_0[ebu_address_region] = IFX_EBU_ADDR_SEL_0_BASE_VAL(CPHYSADDR(IFX_NAND_ADDR_REGION_BASE) >> 12)
                         | IFX_EBU_ADDR_SEL_0_MASK_VAL(15)
                         | IFX_EBU_ADDR_SEL_0_MRME_VAL(0)
                         | IFX_EBU_ADDR_SEL_0_REGEN_VAL(1);

#if defined (CONFIG_SVIP)
    IFX_EBU_CON_0[ebu_address_region] = IFX_EBU_CON_0_WRDIS_VAL(0) | IFX_EBU_CON_0_ADSWP_VAL(1) | IFX_EBU_CON_0_AGEN_VAL(0x00)
                    | IFX_EBU_CON_0_SETUP_VAL(1) | IFX_EBU_CON_0_WAIT_VAL(0x00) | IFX_EBU_CON_0_WINV_VAL(0)
                    | IFX_EBU_CON_0_PW_VAL(0x00) | IFX_EBU_CON_0_ALEC_VAL(0) | IFX_EBU_CON_0_BCGEN_VAL(0x01)
                    | IFX_EBU_CON_0_WAITWRC_VAL(1) | IFX_EBU_CON_0_WAITRDC_VAL(1) | IFX_EBU_CON_0_HOLDC_VAL(1)
                    | IFX_EBU_CON_0_RECOVC_VAL(0) | IFX_EBU_CON_0_CMULT_VAL(0x01);
#elif defined (CONFIG_INCAIP2)
    IFX_EBU_CON_0[ebu_address_region] = IFX_EBU_CON_0_WRDIS_VAL(0) | IFX_EBU_CON_0_ADSWP_VAL(1) | IFX_EBU_CON_0_AGEN_VAL(0x00)
                    | IFX_EBU_CON_0_SETUP_VAL(1) | IFX_EBU_CON_0_WAIT_VAL(0x00) | IFX_EBU_CON_0_WINV_VAL(0)
                    | IFX_EBU_CON_0_PW_VAL(0x00) | IFX_EBU_CON_0_ALEC_VAL(0) | IFX_EBU_CON_0_BCGEN_VAL(0x01)
                    | IFX_EBU_CON_0_WAITWRC_VAL(2) | IFX_EBU_CON_0_WAITRDC_VAL(2) | IFX_EBU_CON_0_HOLDC_VAL(1)
                    | IFX_EBU_CON_0_RECOVC_VAL(1) | IFX_EBU_CON_0_CMULT_VAL(0x01);
#endif

}

/*
 *  Description:
 *    Config NAND controller.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static inline void config_nand(void)
{
    /*
     *  ECC disabled
     *  CLE, ALE and CS are pulse, all other signal are latches based
     *  CLE and ALE are active high, PRE, WP, SE and CS/CE are active low
     *  OUT_CS_S is disabled
     *  NAND mode is disabled
     */
    *IFX_EBU_NAND_CON = IFX_EBU_NAND_CON_ECC_ON_VAL(0)
                     | IFX_EBU_NAND_CON_LAT_EN_VAL(0x38)
                     | IFX_EBU_NAND_CON_OUT_CS_S_VAL(ebu_address_region)
                     | IFX_EBU_NAND_CON_IN_CS_S_VAL(ebu_address_region)
                     | IFX_EBU_NAND_CON_PRE_P_VAL(1) | IFX_EBU_NAND_CON_WP_P_VAL(1)
                     | IFX_EBU_NAND_CON_SE_P_VAL(1) | IFX_EBU_NAND_CON_CS_P_VAL(1)
                     | IFX_EBU_NAND_CON_CLE_P_VAL(0) | IFX_EBU_NAND_CON_ALE_P_VAL(0)
                     | IFX_EBU_NAND_CON_CSMUX_E_VAL(0)
                     | IFX_EBU_NAND_CON_NANDMODE_VAL(0);
}


/*
 * ####################################
 *           Global Function
 * ####################################
 */


/*
 * ####################################
 *           Init/Cleanup API
 * ####################################
 */

/*
 *  Description:
 *    Register MTD partitions or device.
 *  Input:
 *    none
 *  Output:
 *    0    --- successful
 *    else --- failure, usually it is negative value of error code
 */
int __init ifx_nand_init(void)
{
    struct nand_chip *this;
    /*
     *  Configure hardware
     */
    check_parameter();

    if ( ebu_config )
        config_ebu();

    config_nand();
    /* Get pointer to private data */
    this = (struct nand_chip *)&chip;

    /* insert callbacks */
    this->IO_ADDR_R      = (void *)IFX_NAND_READ;
    this->IO_ADDR_W      = (void *)IFX_NAND_WRITE;
#ifdef LINUX_2_4
    this->hwcontrol      = ifx_hwcontrol;
    this->calculate_ecc  = ifx_calculate_ecc;
    this->correct_data   = nand_correct_data;
    this->eccmode        = NAND_ECC_SOFT;
    this->enable_hwecc   = ifx_enable_hwecc;
    this->data_buf       = data_buf;
#else    
    this->cmd_ctrl       = ifx_hwcontrol;
    this->ecc.calculate  = ifx_calculate_ecc;
    this->ecc.mode       = NAND_ECC_SOFT;
#ifdef CONFIG_MTD_NAND_ECC_BCH
    /* we may use BCH aogrithm for large page nand.
     * please check nand_base.c nand_get_flash_type
     */
#endif
    this->ecc.hwctl      = ifx_enable_hwecc;
#endif
    this->dev_ready      = ifx_dev_ready;
#if defined (CONFIG_SVIP)
    /* to insert some NOPs to let the FW still handle interrupts, otherwise
       the NAND accesss to the EBU would block the other cores to long when they
       also access the EBU */
    this->write_buf      = ifx_nand_write_buf;
    this->read_buf       = ifx_nand_read_buf;
#endif /* defined (CONFIG_SVIP) */    
#ifdef IFX_NAND_DEBUG_ON      
    this->read_byte = ifx_read_byte;
#endif      

    mtd.priv = &chip;
    mtd_256ecc.priv = &chip_256ecc;

    /*
     *  Scan to find existance of the device
     */
    if ( nand_scan_impl(&mtd, &mtd_256ecc, 1) )
    {
        return -ENXIO;
    }

    /*
     *  Register MTD partitions or device
     */
#if 0
#ifdef CONFIG_MTD_CMDLINE_PARTS
	{
		int mtd_parts_nb = 0;
		struct mtd_partition *mtd_parts = NULL;
		static const char *part_probes[] = { "cmdlinepart", NULL, };

		mtd.name = "ifx-nand";
		mtd_parts_nb = parse_mtd_partitions(&mtd, part_probes, &mtd_parts, 0);
		if ( mtd_parts_nb > 0 )
		{
			add_mtd_partitions(&mtd, mtd_parts, mtd_parts_nb);
			f_mtd_partitions = 1;
		}
		else
			add_mtd_device(&mtd);
	}
#else
#if defined(NUM_PARTITIONS) && NUM_PARTITIONS > 0
    add_mtd_partitions(&mtd, partition_info, NUM_PARTITIONS);
    f_mtd_partitions = 1;
#else
    add_mtd_device(&mtd);
#endif
#endif
#else

#ifdef CONFIG_MTD_PARTITIONS
#define mtd_has_partitions()    (1)
#else
#define mtd_has_partitions()    (0)
#endif

    if (mtd_has_partitions())
    {
        extern int gs_add_nandpartition(struct mtd_info *mtd, struct mtd_info *mtd_256ecc, int index);
        if ((mtd.writesize == 2048) && (mtd.size > 120*1024*1024)) {
            gs_add_nandpartition(&mtd, &mtd_256ecc, 1);
        }
        else
        {
            gs_add_nandpartition(&mtd, &mtd_256ecc, 0);
        }
    }
    else
        add_mtd_device(&mtd);

    /* Read out gxv31xx factory data */
    {
        extern void gs_init(struct mtd_info *mtd, struct mtd_info *mtd_256ecc);
        gs_init(&mtd, &mtd_256ecc);
    }
#endif

    return 0;
}

/*
 *  Description:
 *    Deregister MTD partitions or device.
 *  Input:
 *    none
 *  Output:
 *    none
 */
void __exit ifx_nand_exit(void)
{
    if ( f_mtd_partitions )
        del_mtd_partitions(&mtd);
    else
        del_mtd_device(&mtd);
}

module_init(ifx_nand_init);
module_exit(ifx_nand_exit);
