/*
 *  drivers/mtd/nand/nand_gs.c
 *
 *   This is nand partition definition for Grandstream GXW4200
 *
 *  Author: Shaoyun Xie <sxie@grandstream.com>
 *
 *  Copyright 2012 Grandstream Networks, Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 */

#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/mach-ifxmips/gpio.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <gs_def.h>
//#include "board-gxw4200.h"

struct gs_procfs_entry
{
    char * name;
    read_proc_t * read_proc;
    write_proc_t * write_proc;
    void ( *init )( void );
    int value;
};

#define PWM_ON  1
#define PWM_OFF 0

#define HW_SECURITY_TABLE_VERSION        2

#define GS_IMAGE_HEADER_OFFSET          0x4000      //0x200 seems will mess up the bad block remapping in squashfs
#define GS_IMAGE_OFFSET(address)        (address + GS_IMAGE_HEADER_OFFSET)

#define GS_IMAGE_HEADER_OFFSET_NLP      0x20000      
#define GS_IMAGE_OFFSET_NLP(address)        (address + GS_IMAGE_HEADER_OFFSET_NLP)

#define GS_IMAGE_ADJUST(a)              (a - GS_IMAGE_HEADER_OFFSET)
#define GS_IMAGE_ADJUST_NLP(a)              (a - GS_IMAGE_HEADER_OFFSET_NLP)

#define GS_ARRAY_SIZE(a)                (sizeof(a) / sizeof(a[0]))
#define GS_SW_VER(a, b, c, d)           ((((a) & 0xFF) << 24) | ((b) & 0xFF) << 16 | ((c) & 0xFF) << 8 | ((d) & 0xFF))
#define GS_SW_VER_HILONG_HIWORD(a)      ((a >> 24) & 0xFF)
#define GS_SW_VER_HILONG_LOWORD(a)      ((a >> 16) & 0xFF)
#define GS_SW_VER_LOLONG_HIWORD(a)      ((a >> 8) & 0xFF)
#define GS_SW_VER_LOLONG_LOWORD(a)      (a & 0xFF)


/* external values exported on /proc */
#ifdef CONFIG_GXW42XX_ATE_BUILD
static int ate_flag = 1;
#endif
#ifdef CONFIG_GXW42XX_RECOVERY_BUILD
static int recovery_flag = 1;
#endif

int i2c_gpio_flag = 0; //0=tca6424 ;1=aw9523

extern int wdt_feed;
extern int feed_wdt(void);
extern int net_linked;

extern int nand_recover_bad(struct mtd_info *mtd, unsigned long block_addr);
extern int ifxmips_replace_ethaddr(char *mac_bytes);
extern void switchip_set_mac_table_entry(char * fact_mac);

#ifdef CONFIG_GXW42XX_BUILD_48_96
extern int ifxmips_i2c_gpio_set(unsigned int number, int state);
#else
#define ifxmips_i2c_gpio_set(a, b) 
#endif

int soft_pwm_tick(void);
static int soft_pwm_active = 0;
static int led_tickonoff = 0;
static int led_tickonoff_saved = 0;

static char hw_feature_modelbase[129];
static char hw_feature_modelstr[129];
static char hw_feature_modelalias[129];
static char hw_feature_vendor[129];
static char hw_feature_vendor_fullname[129];
static char hw_feature_fwkey[33];
static unsigned short hw_feature_oem_id = 0;

static unsigned int require_pem;
static unsigned int require_sn;
static unsigned int require_admin_pwd;
static unsigned int require_xml_key;
static unsigned int require_hash;

static sec_info_t sec_info;

unsigned short rtp_port_inspect = 0;

static short led_gpioset = -1;
static short lcd_model = 0;

static short led_pattern = 0;
static short led_pattern_count = 0;
static struct timer_list led_pattern_timer;

#define FW_PART_SIZE     0x00a00000
#define FW_PART_SIZE_NLP 0x01900000

#define CORE_SIZE       0x001a0000
#define CORE_SIZE_NLP   0x00300000

#define RCVR_START      0x000e0000
#define RC_PART_SIZE    0x00200000
#define RCVR_START_NLP      0x00400000
#define RC_PART_SIZE_NLP    0x00500000

#define BASE_SIZE       FW_PART_SIZE
#define PROG_SIZE       FW_PART_SIZE
#define BUFF_SIZE       FW_PART_SIZE
#define RCVR_SIZE       RC_PART_SIZE

#define BASE_SIZE_NLP    FW_PART_SIZE_NLP
#define PROG_SIZE_NLP    FW_PART_SIZE_NLP
#define BUFF_SIZE_NLP    FW_PART_SIZE_NLP
#define RCVR_SIZE_NLP    RC_PART_SIZE_NLP

#define BOOT_START 0x000a0000
#define BOOT_SIZE  0x00020000
#define BOOT_START_NLP 0x00300000
#define BOOT_SIZE_NLP  0x00080000

#define CORE_START 0x002e0000
#define CORE_START_NLP 0x00900000

#define BASE_START (CORE_START + CORE_SIZE)
#define PROG_START (BASE_START + BASE_SIZE)
#define BUFF_START (PROG_START + PROG_SIZE)
#define BASE_START_NLP (CORE_START_NLP + CORE_SIZE_NLP)
#define PROG_START_NLP (BASE_START_NLP + BASE_SIZE_NLP)
#define BUFF_START_NLP (PROG_START_NLP + PROG_SIZE_NLP)

#define DATA_START (BUFF_START + BUFF_SIZE)
#define DATA_SIZE  (0x4000000 - DATA_START)
#define DATA_START_NLP (BUFF_START_NLP + BUFF_SIZE_NLP)
#define DATA_SIZE_NLP  (0x8000000 - DATA_START_NLP)

#define BOOTIMG_SIZE    98304 /* 96KB for dual copy of boot image */

static int sec_feature_proc_read( char *page, char **start, off_t off, int count, int *eof, void *data );
struct gs_procfs_entry sec_feature_entry[] =
{
#define SEC_FEATURE_ENTRY( name, value ) { #name, NULL, NULL, NULL, value }
    SEC_FEATURE_ENTRY( private_key, SEC_FEATURE_PRIVATE_KEY ),
    SEC_FEATURE_ENTRY( certificate, SEC_FEATURE_CERTIFICATE ),
    SEC_FEATURE_ENTRY( checksum, SEC_FEATURE_CHECKSUM ),
    SEC_FEATURE_ENTRY( sn, SEC_FEATURE_SN ),
    SEC_FEATURE_ENTRY( hash, SEC_FEATURE_VONAGE_HASH ),
    SEC_FEATURE_ENTRY( password, SEC_FEATURE_VONAGE_PASSWORD ),
    SEC_FEATURE_ENTRY( xml_key, SEC_FEATURE_VONAGE_XML_KEY ),
};
#define NUM_SEC_FEATURE_ENTRIES ( sizeof ( sec_feature_entry ) / sizeof ( struct gs_procfs_entry ) )

/* 
 * SXIE: right now the mtd partition is mtd0-mtd9. If we have mtd10 then make sure some of the shell scripts 
 * needs to be changed as they assume this number is one digit right now.
 */
struct mtd_partition _gs_nand_partition_info_nand_small_page[] = {
    /* 64 MiB Nand, Page Size 512*/
    { 
        .name =     "bootloader",
        .start =    0x000a0000,
        .offset =   GS_IMAGE_OFFSET(BOOT_START),
        .size =     GS_IMAGE_ADJUST(BOOT_SIZE),
    },
    { 
        .name =     "core",
        .start =    CORE_START,
        .offset =   GS_IMAGE_OFFSET(CORE_START),
        .size =     GS_IMAGE_ADJUST(CORE_SIZE),
    },
    {
        .name =     "base",
        .start =    BASE_START,
        .offset =   GS_IMAGE_OFFSET(BASE_START),
        .size =     GS_IMAGE_ADJUST(BASE_SIZE), 
    },
    {
        .name =     "prog",
        .start =    PROG_START,
        .offset =   GS_IMAGE_OFFSET(PROG_START),
        .size =     GS_IMAGE_ADJUST(PROG_SIZE),
    },
    {
        .name =     "upgrade_buffer",
        .start =    BUFF_START,
        .offset =   GS_IMAGE_OFFSET(BUFF_START),
        .size =     GS_IMAGE_ADJUST(BUFF_SIZE),
    },
    { 
        .name =     "rcvr",
        .start =    RCVR_START,
        .offset =   GS_IMAGE_OFFSET(RCVR_START),
        .size =     GS_IMAGE_ADJUST(RCVR_SIZE),
    },
    {
        .name =     "data",
        .start =    DATA_START,
        .offset =   DATA_START,
        .size =     DATA_SIZE,
    },
    {
        .name =     "ubootenv",
        .start =    0x00008000,
        .offset =   0x00008000,
        .size =     0x00004000,
    },
    {
        .name = "factory",
        .start = 0x00004000,
        .offset = GS_FACTORY_CFG_OFFSET,
        .size = GS_FACTORY_CFG_SIZE,
    },
};

struct mtd_partition _gs_nand_partition_info_nand_large_page[] = {
     /* 128 MiB Nand, Page Size 2048*/
    { 
        .name =     "bootloader",
        .start =    0x00300000,
        .offset =   GS_IMAGE_OFFSET_NLP(BOOT_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(BOOT_SIZE_NLP),
    },
    { 
        .name =     "core",
        .start =    CORE_START_NLP,
        .offset =   GS_IMAGE_OFFSET_NLP(CORE_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(CORE_SIZE_NLP),
    },
    {
        .name =     "base",
        .start =    BASE_START_NLP,
        .offset =   GS_IMAGE_OFFSET_NLP(BASE_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(BASE_SIZE_NLP), 
    },
    {
        .name =     "prog",
        .start =    PROG_START_NLP,
        .offset =   GS_IMAGE_OFFSET_NLP(PROG_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(PROG_SIZE_NLP),
    },
    {
        .name =     "upgrade_buffer",
        .start =    BUFF_START_NLP,
        .offset =   GS_IMAGE_OFFSET_NLP(BUFF_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(BUFF_SIZE_NLP),
    },
    { 
        .name =     "rcvr",
        .start =    RCVR_START_NLP,
        .offset =   GS_IMAGE_OFFSET_NLP(RCVR_START_NLP),
        .size =     GS_IMAGE_ADJUST_NLP(RCVR_SIZE_NLP),
    },
    {
        .name =     "data",
        .start =    DATA_START_NLP,
        .offset =   DATA_START_NLP,
        .size =     DATA_SIZE_NLP,
    },
    {
        .name =     "ubootenv",
        .start =    0x00040000,
        .offset =   0x00040000,
        .size =     0x00020000,
    },
    {
        .name = "factory",
        .start = 0x00020000,
        .offset = GS_FACTORY_CFG_OFFSET_NLP,
        .size = GS_FACTORY_CFG_SIZE_NLP,
    },
};

/* default partition info */
struct mtd_partition *gs_nand_partition_info = _gs_nand_partition_info_nand_small_page;
int gs_nand_partition_count = ((sizeof(_gs_nand_partition_info_nand_small_page))/(sizeof(struct mtd_partition)));

static char factParaBuf_nsp[GS_FACTORY_CFG_SIZE];
static char factParaBuf_nlp[GS_FACTORY_CFG_SIZE_NLP];

/* default factParaBuf */
char *factParaBuf = factParaBuf_nsp;
unsigned int factory_cfg_size = GS_FACTORY_CFG_SIZE;
unsigned int factory_cfg_offset = GS_FACTORY_CFG_OFFSET;

static struct mtd_info *gs_mtd = NULL;
struct mtd_info *gs_mtd_256ecc = NULL;
int gs_nand_type = 0; /* 0 for small page size(512), 1 for large page size(2048) */

struct soft_pwm_control
{
    int pwm_value;
    int pwm_stop;
    int tick;
    int pwm_onoff;
};

#define default_pwm {0,0,0,0}

struct gsboard_gpio
{
    char *gpio_name;
    int gpio_pin;
    int gpio_dir;   /* 0 for in, 1 for out */
    int skip_init;
    int init_val;
    struct soft_pwm_control pwm;
    int extended_gpio;
};

struct gsboard_port_map
{
    int vineticDevId;
    int vineticDevCh;
    int vineticPort;
    int slicPort;
    int svip16;
};

typedef enum _Map_Proc_Entry
{
    ENTRY_MAP_TABLE_INFO = 0,
    ENTRY_MAP_TABLE_SIZE,
    ENTRY_MAP_TABLE_INDEX,
} Map_Proc_Entry;

struct gsboard_port_map *board_port_map_gs = NULL;
struct gsboard_hw_sw_limit *board_hw_sw_ver = NULL;
static int board_port_map_gs_count = 0;
static int board_port_map_gs_index = 0;
unsigned long board_port_map_proc_entry = ENTRY_MAP_TABLE_INFO; // 0: port map table; 1: port map table size
struct gsboard_gpio *board_gpio_gs = NULL;
static int board_gpio_gs_count = 0;

struct gsboard_gpio board_gpio_gxw4200_v20[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "led_power", 
        .gpio_pin  = 11,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slic_reset_1", 
        .gpio_pin  = 3,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slic_reset_2", 
        .gpio_pin  = 4,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  40, 1, 0, 1, default_pwm, 0 },
    { "led1",  41, 1, 0, 1, default_pwm, 0 },
    { "led2",  44, 1, 0, 1, default_pwm, 0 },
    { "led3",  45, 1, 0, 1, default_pwm, 0 },
    { "led4",  47, 1, 0, 1, default_pwm, 0 },
    { "led5",  48, 1, 0, 1, default_pwm, 0 },
    { "led6",  49, 1, 0, 1, default_pwm, 0 },
    { "led7",  50, 1, 0, 1, default_pwm, 0 },
    { "led8",  51, 1, 0, 1, default_pwm, 0 },
    { "led9",  68, 1, 0, 1, default_pwm, 0 },
    { "led10", 69, 1, 0, 1, default_pwm, 0 },
    { "led11", 70, 1, 0, 1, default_pwm, 0 },
    { "led12", 71, 1, 0, 1, default_pwm, 0 },
    { "led13", 79, 1, 0, 1, default_pwm, 0 },
    { "led14", 57, 1, 0, 1, default_pwm, 0 },
    { "led15", 58, 1, 0, 1, default_pwm, 0 },
    /* 16-17 not connected in v2.0 */
    { "led16", 72, 1, 0, 1, default_pwm, 0 },
    { "led17", 72, 1, 0, 1, default_pwm, 0 },
    { "led18", 72, 1, 0, 1, default_pwm, 0 },
    { "led19", 72, 1, 0, 1, default_pwm, 0 },
    { "led20", 72, 1, 0, 1, default_pwm, 0 },
    { "led21", 72, 1, 0, 1, default_pwm, 0 },
    { "led22", 72, 1, 0, 1, default_pwm, 0 },
    { "led23", 72, 1, 0, 1, default_pwm, 0 },
    { "led24", 72, 1, 0, 1, default_pwm, 0 },
    { "led25", 72, 1, 0, 1, default_pwm, 0 },
    { "led26", 72, 1, 0, 1, default_pwm, 0 },
    { "led27", 72, 1, 0, 1, default_pwm, 0 },
    { "led28", 72, 1, 0, 1, default_pwm, 0 },
    { "led29", 72, 1, 0, 1, default_pwm, 0 },
    { "led30", 72, 1, 0, 1, default_pwm, 0 },
    { "led31", 72, 1, 0, 1, default_pwm, 0 },
};

struct gsboard_gpio board_gpio_gxw4200_v21[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  40, 1, 0, 0, default_pwm, 0 },
    { "led1",  41, 1, 0, 0, default_pwm, 0 },
    { "led2",  44, 1, 0, 0, default_pwm, 0 },
    { "led3",  45, 1, 0, 0, default_pwm, 0 },
    { "led4",  47, 1, 0, 0, default_pwm, 0 },
    { "led5",  48, 1, 0, 0, default_pwm, 0 },
    { "led6",  49, 1, 0, 0, default_pwm, 0 },
    { "led7",  50, 1, 0, 0, default_pwm, 0 },
    { "led8",  51, 1, 0, 0, default_pwm, 0 },
    { "led9",  68, 1, 0, 0, default_pwm, 0 },
    { "led10", 69, 1, 0, 0, default_pwm, 0 },
    { "led11", 70, 1, 0, 0, default_pwm, 0 },
    { "led12", 71, 1, 0, 0, default_pwm, 0 },
    { "led13", 79, 1, 0, 0, default_pwm, 0 },
    { "led14", 57, 1, 0, 0, default_pwm, 0 },
    { "led15", 58, 1, 0, 0, default_pwm, 0 },

    { "led16", 72, 1, 0, 0, default_pwm, 0 },
    { "led17", 73, 1, 0, 0, default_pwm, 0 },
    { "led18", 74, 1, 0, 0, default_pwm, 0 },
    { "led19", 75, 1, 0, 0, default_pwm, 0 },
    { "led20", 54, 1, 0, 0, default_pwm, 0 },
    { "led21", 77, 1, 0, 0, default_pwm, 0 },
    { "led22", 78, 1, 0, 0, default_pwm, 0 },
    { "led23", 55, 1, 0, 0, default_pwm, 0 },
    { "led24",  4, 1, 0, 0, default_pwm, 0 },
    { "led25",  6, 1, 0, 0, default_pwm, 0 },
    { "led26",  7, 1, 0, 0, default_pwm, 0 },
    { "led27",  9, 1, 0, 0, default_pwm, 0 },
    { "led28", 13, 1, 0, 0, default_pwm, 0 },
    { "led29", 16, 1, 0, 0, default_pwm, 0 },
    { "led30", 17, 1, 0, 0, default_pwm, 0 },
    { "led31",  1, 1, 0, 0, default_pwm, 0 },
};

#ifdef CONFIG_GXW42XX_BUILD_48_96
int slave_reset_set[] = {
    12,13,3,4,5,
};
int slave_reset_set_rev_11A[] = {
    40,41,3,4,5,
};
int slave_reset_size = sizeof(slave_reset_set) / sizeof(int);
#endif

struct gsboard_gpio board_gpio_gxw4248_v10[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 9,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave1_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave2_reset", 
        .gpio_pin  = 13,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont",
        .gpio_pin  = 6,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det",
        .gpio_pin  = 28,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1",
        .gpio_pin  = 27,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_2",
        .gpio_pin  = 26,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "exgpio_reset",
        .gpio_pin  = 34,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 1,
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  49, 1, 1, 0, default_pwm, 1 },
    { "led1",  50, 1, 1, 0, default_pwm, 1 },
    { "led2",  51, 1, 1, 0, default_pwm, 1 },
    { "led3",  52, 1, 1, 0, default_pwm, 1 },
    { "led4",  53, 1, 1, 0, default_pwm, 1 },
    { "led5",  54, 1, 1, 0, default_pwm, 1 },
    { "led6",  55, 1, 1, 0, default_pwm, 1 },
    { "led7",  56, 1, 1, 0, default_pwm, 1 },
    { "led8",  57, 1, 1, 0, default_pwm, 1 },
    { "led9",  58, 1, 1, 0, default_pwm, 1 },
    { "led10", 59, 1, 1, 0, default_pwm, 1 },
    { "led11", 60, 1, 1, 0, default_pwm, 1 },
    { "led12", 61, 1, 1, 0, default_pwm, 1 },
    { "led13", 62, 1, 1, 0, default_pwm, 1 },
    { "led14", 63, 1, 1, 0, default_pwm, 1 },
    { "led15", 64, 1, 1, 0, default_pwm, 1 },

    { "led16", 65, 1, 1, 0, default_pwm, 1 },
    { "led17", 66, 1, 1, 0, default_pwm, 1 },
    { "led18", 67, 1, 1, 0, default_pwm, 1 },
    { "led19", 68, 1, 1, 0, default_pwm, 1 },
    { "led20", 69, 1, 1, 0, default_pwm, 1 },
    { "led21", 70, 1, 1, 0, default_pwm, 1 },
    { "led22", 71, 1, 1, 0, default_pwm, 1 },
    { "led23", 72, 1, 1, 0, default_pwm, 1 },
    { "led24", 73, 1, 1, 0, default_pwm, 1 },
    { "led25", 74, 1, 1, 0, default_pwm, 1 },
    { "led26", 75, 1, 1, 0, default_pwm, 1 },
    { "led27", 76, 1, 1, 0, default_pwm, 1 },
    { "led28", 77, 1, 1, 0, default_pwm, 1 },
    { "led29", 78, 1, 1, 0, default_pwm, 1 },
    { "led30", 79, 1, 1, 0, default_pwm, 1 },
    { "led31", 80, 1, 1, 0, default_pwm, 1 },

    { "led32", 81, 1, 1, 0, default_pwm, 1 },
    { "led33", 82, 1, 1, 0, default_pwm, 1 },
    { "led34", 83, 1, 1, 0, default_pwm, 1 },
    { "led35", 84, 1, 1, 0, default_pwm, 1 },
    { "led36", 85, 1, 1, 0, default_pwm, 1 },
    { "led37", 86, 1, 1, 0, default_pwm, 1 },
    { "led38", 87, 1, 1, 0, default_pwm, 1 },
    { "led39", 88, 1, 1, 0, default_pwm, 1 },
    { "led40", 89, 1, 1, 0, default_pwm, 1 },
    { "led41", 90, 1, 1, 0, default_pwm, 1 },
    { "led42", 91, 1, 1, 0, default_pwm, 1 },
    { "led43", 92, 1, 1, 0, default_pwm, 1 },
    { "led44", 93, 1, 1, 0, default_pwm, 1 },
    { "led45", 94, 1, 1, 0, default_pwm, 1 },
    { "led46", 95, 1, 1, 0, default_pwm, 1 },
    { "led47", 96, 1, 1, 0, default_pwm, 1 },
};


struct gsboard_gpio board_gpio_gxw4248_v11[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 9,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave1_reset", 
        .gpio_pin  = 40,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave2_reset", 
        .gpio_pin  = 41,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 27,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont",
        .gpio_pin  = 28,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det",
        .gpio_pin  = 6,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1",
        .gpio_pin  = 39,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_2",
        .gpio_pin  = 12,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "vbat_det",
        .gpio_pin  = 13,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "vbat_ctrl",
        .gpio_pin  = 1,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "exgpio_reset",
        .gpio_pin  = 34,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 1,
        .pwm = default_pwm
    },


    /* phone LEDs */
    { "led0",  49, 1, 1, 0, default_pwm, 1 },
    { "led1",  50, 1, 1, 0, default_pwm, 1 },
    { "led2",  51, 1, 1, 0, default_pwm, 1 },
    { "led3",  52, 1, 1, 0, default_pwm, 1 },
    { "led4",  53, 1, 1, 0, default_pwm, 1 },
    { "led5",  54, 1, 1, 0, default_pwm, 1 },
    { "led6",  55, 1, 1, 0, default_pwm, 1 },
    { "led7",  56, 1, 1, 0, default_pwm, 1 },
    { "led8",  57, 1, 1, 0, default_pwm, 1 },
    { "led9",  58, 1, 1, 0, default_pwm, 1 },
    { "led10", 59, 1, 1, 0, default_pwm, 1 },
    { "led11", 60, 1, 1, 0, default_pwm, 1 },
    { "led12", 61, 1, 1, 0, default_pwm, 1 },
    { "led13", 62, 1, 1, 0, default_pwm, 1 },
    { "led14", 63, 1, 1, 0, default_pwm, 1 },
    { "led15", 64, 1, 1, 0, default_pwm, 1 },

    { "led16", 65, 1, 1, 0, default_pwm, 1 },
    { "led17", 66, 1, 1, 0, default_pwm, 1 },
    { "led18", 67, 1, 1, 0, default_pwm, 1 },
    { "led19", 68, 1, 1, 0, default_pwm, 1 },
    { "led20", 69, 1, 1, 0, default_pwm, 1 },
    { "led21", 70, 1, 1, 0, default_pwm, 1 },
    { "led22", 71, 1, 1, 0, default_pwm, 1 },
    { "led23", 72, 1, 1, 0, default_pwm, 1 },
    { "led24", 73, 1, 1, 0, default_pwm, 1 },
    { "led25", 74, 1, 1, 0, default_pwm, 1 },
    { "led26", 75, 1, 1, 0, default_pwm, 1 },
    { "led27", 76, 1, 1, 0, default_pwm, 1 },
    { "led28", 77, 1, 1, 0, default_pwm, 1 },
    { "led29", 78, 1, 1, 0, default_pwm, 1 },
    { "led30", 79, 1, 1, 0, default_pwm, 1 },
    { "led31", 80, 1, 1, 0, default_pwm, 1 },

    { "led32", 81, 1, 1, 0, default_pwm, 1 },
    { "led33", 82, 1, 1, 0, default_pwm, 1 },
    { "led34", 83, 1, 1, 0, default_pwm, 1 },
    { "led35", 84, 1, 1, 0, default_pwm, 1 },
    { "led36", 85, 1, 1, 0, default_pwm, 1 },
    { "led37", 86, 1, 1, 0, default_pwm, 1 },
    { "led38", 87, 1, 1, 0, default_pwm, 1 },
    { "led39", 88, 1, 1, 0, default_pwm, 1 },
    { "led40", 89, 1, 1, 0, default_pwm, 1 },
    { "led41", 90, 1, 1, 0, default_pwm, 1 },
    { "led42", 91, 1, 1, 0, default_pwm, 1 },
    { "led43", 92, 1, 1, 0, default_pwm, 1 },
    { "led44", 93, 1, 1, 0, default_pwm, 1 },
    { "led45", 94, 1, 1, 0, default_pwm, 1 },
    { "led46", 95, 1, 1, 0, default_pwm, 1 },
    { "led47", 96, 1, 1, 0, default_pwm, 1 },
};

struct gsboard_gpio board_gpio_gxw4248_v15[] = {
	{
		.gpio_name = "board_reset", 
		.gpio_pin  = 10,
		.gpio_dir  = 0, 
		.skip_init = 0, 
		.init_val  = 1, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "phy_reset", 
		.gpio_pin  = 9,
		.gpio_dir  = 1, 
		.skip_init = 1, 
		.init_val  = 1, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "slave1_reset", 
		.gpio_pin  = 40,
		.gpio_dir  = 1, 
		.skip_init = 0, 
		.init_val  = 0, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "slave2_reset", 
		.gpio_pin  = 41,
		.gpio_dir  = 1, 
		.skip_init = 0, 
		.init_val  = 0, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "lcd_backlight", 
		.gpio_pin  = 27,
		.gpio_dir  = 1, 
		.skip_init = 0, 
		.init_val  = 1, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "key_down", 
		.gpio_pin  = 37,
		.gpio_dir  = 0, 
		.skip_init = 0, 
		.init_val  = 0, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "key_enter", 
		.gpio_pin  = 36,
		.gpio_dir  = 0, 
		.skip_init = 0, 
		.init_val  = 0, 
		.pwm = default_pwm
	},
	{
		.gpio_name = "fan_cont",
		.gpio_pin  = 28,
		.gpio_dir  = 1,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "fan_det",
		.gpio_pin  = 6,
		.gpio_dir  = 0,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "fan_off_1",
		.gpio_pin  = 39,
		.gpio_dir  = 0,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "fan_off_2",
		.gpio_pin  = 12,
		.gpio_dir  = 0,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "vbat_det",
		.gpio_pin  = 13,
		.gpio_dir  = 0,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "vbat_ctrl",
		.gpio_pin  = 1,
		.gpio_dir  = 1,
		.skip_init = 0,
		.init_val  = 0,
		.pwm = default_pwm
	},
	{
		.gpio_name = "exgpio_reset",
		.gpio_pin  = 34,
		.gpio_dir  = 1,
		.skip_init = 0,
		.init_val  = 1,
		.pwm = default_pwm
	},


	/* phone LEDs */
	{ "led0",  49+7, 1, 1, 0, default_pwm, 1 },
	{ "led1",  49+6, 1, 1, 0, default_pwm, 1 },
	{ "led2",  49+5, 1, 1, 0, default_pwm, 1 },
	{ "led3",  49+4, 1, 1, 0, default_pwm, 1 },
	{ "led4",  49+3, 1, 1, 0, default_pwm, 1 },
	{ "led5",  49+2, 1, 1, 0, default_pwm, 1 },
	{ "led6",  49+1, 1, 1, 0, default_pwm, 1 },
	{ "led7",  49+0, 1, 1, 0, default_pwm, 1 },
	{ "led8",  49+11, 1, 1, 0, default_pwm, 1 },
	{ "led9",  49+10, 1, 1, 0, default_pwm, 1 },
	{ "led10", 49+9, 1, 1, 0, default_pwm, 1 },
	{ "led11", 49+8, 1, 1, 0, default_pwm, 1 },
	{ "led12", 49+15, 1, 1, 0, default_pwm, 1 },
	{ "led13", 49+14, 1, 1, 0, default_pwm, 1 },
	{ "led14", 49+13, 1, 1, 0, default_pwm, 1 },
	{ "led15", 49+12, 1, 1, 0, default_pwm, 1 },

	{ "led16", 64+8, 1, 1, 0, default_pwm, 1 },
	{ "led17", 64+7, 1, 1, 0, default_pwm, 1 },
	{ "led18", 64+6, 1, 1, 0, default_pwm, 1 },
	{ "led19", 64+5, 1, 1, 0, default_pwm, 1 },
	{ "led20", 64+4, 1, 1, 0, default_pwm, 1 },
	{ "led21", 64+3, 1, 1, 0, default_pwm, 1 },
	{ "led22", 64+2, 1, 1, 0, default_pwm, 1 },
	{ "led23", 64+1, 1, 1, 0, default_pwm, 1 },
	{ "led24", 64+12, 1, 1, 0, default_pwm, 1 },
	{ "led25", 64+11, 1, 1, 0, default_pwm, 1 },
	{ "led26", 64+10, 1, 1, 0, default_pwm, 1 },
	{ "led27", 64+9, 1, 1, 0, default_pwm, 1 },
	{ "led28", 64+16, 1, 1, 0, default_pwm, 1 },
	{ "led29", 64+15, 1, 1, 0, default_pwm, 1 },
	{ "led30", 64+14, 1, 1, 0, default_pwm, 1 },
	{ "led31", 64+13, 1, 1, 0, default_pwm, 1 },

	{ "led32", 80+8, 1, 1, 0, default_pwm, 1 },
	{ "led33", 80+7, 1, 1, 0, default_pwm, 1 },
	{ "led34", 80+6, 1, 1, 0, default_pwm, 1 },
	{ "led35", 80+5, 1, 1, 0, default_pwm, 1 },
	{ "led36", 80+4, 1, 1, 0, default_pwm, 1 },
	{ "led37", 80+3, 1, 1, 0, default_pwm, 1 },
	{ "led38", 80+2, 1, 1, 0, default_pwm, 1 },
	{ "led39", 80+1, 1, 1, 0, default_pwm, 1 },
	{ "led40", 80+12, 1, 1, 0, default_pwm, 1 },
	{ "led41", 80+11, 1, 1, 0, default_pwm, 1 },
	{ "led42", 80+10, 1, 1, 0, default_pwm, 1 },
	{ "led43", 80+9, 1, 1, 0, default_pwm, 1 },
	{ "led44", 80+16, 1, 1, 0, default_pwm, 1 },
	{ "led45", 80+15, 1, 1, 0, default_pwm, 1 },
	{ "led46", 80+14, 1, 1, 0, default_pwm, 1 },
	{ "led47", 80+13, 1, 1, 0, default_pwm, 1 },
};


struct gsboard_gpio board_gpio_gxw4296_v10[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 9,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave1_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave2_reset", 
        .gpio_pin  = 13,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave3_reset", 
        .gpio_pin  = 3,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave4_reset", 
        .gpio_pin  = 4,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave5_reset", 
        .gpio_pin  = 5,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont", 
        .gpio_pin  = 6,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det", 
        .gpio_pin  = 28,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1", 
        .gpio_pin  = 27,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_2", 
        .gpio_pin  = 26,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_3", 
        .gpio_pin  = 25,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_4", 
        .gpio_pin  = 29,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "exgpio_reset",
        .gpio_pin  = 34,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 1,
        .pwm = default_pwm
    },


    /* phone LEDs */
    { "led0",   1, 1, 1, 0, default_pwm, 1 },
    { "led1",   2, 1, 1, 0, default_pwm, 1 },
    { "led2",   3, 1, 1, 0, default_pwm, 1 },
    { "led3",   4, 1, 1, 0, default_pwm, 1 },
    { "led4",   5, 1, 1, 0, default_pwm, 1 },
    { "led5",   6, 1, 1, 0, default_pwm, 1 },
    { "led6",   7, 1, 1, 0, default_pwm, 1 },
    { "led7",   8, 1, 1, 0, default_pwm, 1 },
    { "led8",   9, 1, 1, 0, default_pwm, 1 },
    { "led9",  10, 1, 1, 0, default_pwm, 1 },
    { "led10", 11, 1, 1, 0, default_pwm, 1 },
    { "led11", 12, 1, 1, 0, default_pwm, 1 },
    { "led12", 13, 1, 1, 0, default_pwm, 1 },
    { "led13", 14, 1, 1, 0, default_pwm, 1 },
    { "led14", 15, 1, 1, 0, default_pwm, 1 },
    { "led15", 16, 1, 1, 0, default_pwm, 1 },

    { "led16", 17, 1, 1, 0, default_pwm, 1 },
    { "led17", 18, 1, 1, 0, default_pwm, 1 },
    { "led18", 19, 1, 1, 0, default_pwm, 1 },
    { "led19", 20, 1, 1, 0, default_pwm, 1 },
    { "led20", 21, 1, 1, 0, default_pwm, 1 },
    { "led21", 22, 1, 1, 0, default_pwm, 1 },
    { "led22", 23, 1, 1, 0, default_pwm, 1 },
    { "led23", 24, 1, 1, 0, default_pwm, 1 },
    { "led24", 25, 1, 1, 0, default_pwm, 1 },
    { "led25", 26, 1, 1, 0, default_pwm, 1 },
    { "led26", 27, 1, 1, 0, default_pwm, 1 },
    { "led27", 28, 1, 1, 0, default_pwm, 1 },
    { "led28", 29, 1, 1, 0, default_pwm, 1 },
    { "led29", 30, 1, 1, 0, default_pwm, 1 },
    { "led30", 31, 1, 1, 0, default_pwm, 1 },
    { "led31", 32, 1, 1, 0, default_pwm, 1 },

    { "led32", 33, 1, 1, 0, default_pwm, 1 },
    { "led33", 34, 1, 1, 0, default_pwm, 1 },
    { "led34", 35, 1, 1, 0, default_pwm, 1 },
    { "led35", 36, 1, 1, 0, default_pwm, 1 },
    { "led36", 37, 1, 1, 0, default_pwm, 1 },
    { "led37", 38, 1, 1, 0, default_pwm, 1 },
    { "led38", 39, 1, 1, 0, default_pwm, 1 },
    { "led39", 40, 1, 1, 0, default_pwm, 1 },
    { "led40", 41, 1, 1, 0, default_pwm, 1 },
    { "led41", 42, 1, 1, 0, default_pwm, 1 },
    { "led42", 43, 1, 1, 0, default_pwm, 1 },
    { "led43", 44, 1, 1, 0, default_pwm, 1 },
    { "led44", 45, 1, 1, 0, default_pwm, 1 },
    { "led45", 46, 1, 1, 0, default_pwm, 1 },
    { "led46", 47, 1, 1, 0, default_pwm, 1 },
    { "led47", 48, 1, 1, 0, default_pwm, 1 },

    { "led48", 49, 1, 1, 0, default_pwm, 1 },
    { "led49", 50, 1, 1, 0, default_pwm, 1 },
    { "led50", 51, 1, 1, 0, default_pwm, 1 },
    { "led51", 52, 1, 1, 0, default_pwm, 1 },
    { "led52", 53, 1, 1, 0, default_pwm, 1 },
    { "led53", 54, 1, 1, 0, default_pwm, 1 },
    { "led54", 55, 1, 1, 0, default_pwm, 1 },
    { "led55", 56, 1, 1, 0, default_pwm, 1 },
    { "led56", 57, 1, 1, 0, default_pwm, 1 },
    { "led57", 58, 1, 1, 0, default_pwm, 1 },
    { "led58", 59, 1, 1, 0, default_pwm, 1 },
    { "led59", 60, 1, 1, 0, default_pwm, 1 },
    { "led60", 61, 1, 1, 0, default_pwm, 1 },
    { "led61", 62, 1, 1, 0, default_pwm, 1 },
    { "led62", 63, 1, 1, 0, default_pwm, 1 },
    { "led63", 64, 1, 1, 0, default_pwm, 1 },

    { "led64", 65, 1, 1, 0, default_pwm, 1 },
    { "led65", 66, 1, 1, 0, default_pwm, 1 },
    { "led66", 67, 1, 1, 0, default_pwm, 1 },
    { "led67", 68, 1, 1, 0, default_pwm, 1 },
    { "led68", 69, 1, 1, 0, default_pwm, 1 },
    { "led69", 70, 1, 1, 0, default_pwm, 1 },
    { "led70", 71, 1, 1, 0, default_pwm, 1 },
    { "led71", 72, 1, 1, 0, default_pwm, 1 },
    { "led72", 73, 1, 1, 0, default_pwm, 1 },
    { "led73", 74, 1, 1, 0, default_pwm, 1 },
    { "led74", 75, 1, 1, 0, default_pwm, 1 },
    { "led75", 76, 1, 1, 0, default_pwm, 1 },
    { "led76", 77, 1, 1, 0, default_pwm, 1 },
    { "led77", 78, 1, 1, 0, default_pwm, 1 },
    { "led78", 79, 1, 1, 0, default_pwm, 1 },
    { "led79", 80, 1, 1, 0, default_pwm, 1 },

    { "led80", 81, 1, 1, 0, default_pwm, 1 },
    { "led81", 82, 1, 1, 0, default_pwm, 1 },
    { "led82", 83, 1, 1, 0, default_pwm, 1 },
    { "led83", 84, 1, 1, 0, default_pwm, 1 },
    { "led84", 85, 1, 1, 0, default_pwm, 1 },
    { "led85", 86, 1, 1, 0, default_pwm, 1 },
    { "led86", 87, 1, 1, 0, default_pwm, 1 },
    { "led87", 88, 1, 1, 0, default_pwm, 1 },
    { "led88", 89, 1, 1, 0, default_pwm, 1 },
    { "led89", 90, 1, 1, 0, default_pwm, 1 },
    { "led90", 91, 1, 1, 0, default_pwm, 1 },
    { "led91", 92, 1, 1, 0, default_pwm, 1 },
    { "led92", 93, 1, 1, 0, default_pwm, 1 },
    { "led93", 94, 1, 1, 0, default_pwm, 1 },
    { "led94", 95, 1, 1, 0, default_pwm, 1 },
    { "led95", 96, 1, 1, 0, default_pwm, 1 },
};


struct gsboard_gpio board_gpio_gxw4232_v22[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det", 
        .gpio_pin  = 3,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1", 
        .gpio_pin  = 11,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  40, 1, 0, 0, default_pwm, 0 },
    { "led1",  41, 1, 0, 0, default_pwm, 0 },
    { "led2",  44, 1, 0, 0, default_pwm, 0 },
    { "led3",  45, 1, 0, 0, default_pwm, 0 },
    { "led4",  47, 1, 0, 0, default_pwm, 0 },
    { "led5",  48, 1, 0, 0, default_pwm, 0 },
    { "led6",  49, 1, 0, 0, default_pwm, 0 },
    { "led7",  50, 1, 0, 0, default_pwm, 0 },
    { "led8",  51, 1, 0, 0, default_pwm, 0 },
    { "led9",  68, 1, 0, 0, default_pwm, 0 },
    { "led10", 69, 1, 0, 0, default_pwm, 0 },
    { "led11", 70, 1, 0, 0, default_pwm, 0 },
    { "led12", 71, 1, 0, 0, default_pwm, 0 },
    { "led13", 79, 1, 0, 0, default_pwm, 0 },
    { "led14", 57, 1, 0, 0, default_pwm, 0 },
    { "led15", 58, 1, 0, 0, default_pwm, 0 },

    { "led16", 72, 1, 0, 0, default_pwm, 0 },
    { "led17", 73, 1, 0, 0, default_pwm, 0 },
    { "led18", 74, 1, 0, 0, default_pwm, 0 },
    { "led19", 75, 1, 0, 0, default_pwm, 0 },
    { "led20", 54, 1, 0, 0, default_pwm, 0 },
    { "led21", 77, 1, 0, 0, default_pwm, 0 },
    { "led22", 78, 1, 0, 0, default_pwm, 0 },
    { "led23", 55, 1, 0, 0, default_pwm, 0 },

    { "led24",  4, 1, 0, 0, default_pwm, 0 },
    { "led25",  6, 1, 0, 0, default_pwm, 0 },
    { "led26",  7, 1, 0, 0, default_pwm, 0 },
    { "led27",  9, 1, 0, 0, default_pwm, 0 },
    { "led28", 13, 1, 0, 0, default_pwm, 0 },
    { "led29", 16, 1, 0, 0, default_pwm, 0 },
    { "led30", 17, 1, 0, 0, default_pwm, 0 },
    { "led31",  1, 1, 0, 0, default_pwm, 0 },
};

struct gsboard_gpio board_gpio_gxw4224_v22[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det", 
        .gpio_pin  = 3,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1", 
        .gpio_pin  = 11,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  51, 1, 0, 0, default_pwm, 0 },
    { "led1",  68, 1, 0, 0, default_pwm, 0 },
    { "led2",  69, 1, 0, 0, default_pwm, 0 },
    { "led3",  70, 1, 0, 0, default_pwm, 0 },
    { "led4",  71, 1, 0, 0, default_pwm, 0 },
    { "led5",  79, 1, 0, 0, default_pwm, 0 },
    { "led6",  57, 1, 0, 0, default_pwm, 0 },
    { "led7",  58, 1, 0, 0, default_pwm, 0 },

    { "led8",  72, 1, 0, 0, default_pwm, 0 },
    { "led9",  73, 1, 0, 0, default_pwm, 0 },
    { "led10", 74, 1, 0, 0, default_pwm, 0 },
    { "led11", 75, 1, 0, 0, default_pwm, 0 },
    { "led12", 54, 1, 0, 0, default_pwm, 0 },
    { "led13", 77, 1, 0, 0, default_pwm, 0 },
    { "led14", 78, 1, 0, 0, default_pwm, 0 },
    { "led15", 55, 1, 0, 0, default_pwm, 0 },
    { "led16",  4, 1, 0, 0, default_pwm, 0 },
    { "led17",  6, 1, 0, 0, default_pwm, 0 },
    { "led18",  7, 1, 0, 0, default_pwm, 0 },
    { "led19",  9, 1, 0, 0, default_pwm, 0 },
    { "led20", 13, 1, 0, 0, default_pwm, 0 },
    { "led21", 16, 1, 0, 0, default_pwm, 0 },
    { "led22", 17, 1, 0, 0, default_pwm, 0 },
    { "led23",  1, 1, 0, 0, default_pwm, 0 },
};

struct gsboard_gpio board_gpio_gxw4216_v22[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det", 
        .gpio_pin  = 3,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1", 
        .gpio_pin  = 11,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },


    /* phone LEDs */
    { "led0",  72, 1, 0, 0, default_pwm, 0 },
    { "led1",  73, 1, 0, 0, default_pwm, 0 },
    { "led2",  74, 1, 0, 0, default_pwm, 0 },
    { "led3",  75, 1, 0, 0, default_pwm, 0 },
    { "led4",  54, 1, 0, 0, default_pwm, 0 },
    { "led5",  77, 1, 0, 0, default_pwm, 0 },
    { "led6",  78, 1, 0, 0, default_pwm, 0 },
    { "led7",  55, 1, 0, 0, default_pwm, 0 },
    { "led8",   4, 1, 0, 0, default_pwm, 0 },
    { "led9",   6, 1, 0, 0, default_pwm, 0 },
    { "led10",  7, 1, 0, 0, default_pwm, 0 },
    { "led11",  9, 1, 0, 0, default_pwm, 0 },
    { "led12", 13, 1, 0, 0, default_pwm, 0 },
    { "led13", 16, 1, 0, 0, default_pwm, 0 },
    { "led14", 17, 1, 0, 0, default_pwm, 0 },
    { "led15",  1, 1, 0, 0, default_pwm, 0 },
};


struct gsboard_gpio board_gpio_gxw4232_v23[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont",
        .gpio_pin  = 34,
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_det",
        .gpio_pin  = 3,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_off_1",
        .gpio_pin  = 11,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  40, 1, 0, 0, default_pwm, 0 },
    { "led1",  41, 1, 0, 0, default_pwm, 0 },
    { "led2",  44, 1, 0, 0, default_pwm, 0 },
    { "led3",  45, 1, 0, 0, default_pwm, 0 },
    { "led4",  47, 1, 0, 0, default_pwm, 0 },
    { "led5",  48, 1, 0, 0, default_pwm, 0 },
    { "led6",  49, 1, 0, 0, default_pwm, 0 },
    { "led7",  50, 1, 0, 0, default_pwm, 0 },
    { "led8",  51, 1, 0, 0, default_pwm, 0 },
    { "led9",  68, 1, 0, 0, default_pwm, 0 },
    { "led10", 69, 1, 0, 0, default_pwm, 0 },
    { "led11", 70, 1, 0, 0, default_pwm, 0 },
    { "led12", 71, 1, 0, 0, default_pwm, 0 },
    { "led13", 79, 1, 0, 0, default_pwm, 0 },
    { "led14", 57, 1, 0, 0, default_pwm, 0 },
    { "led15", 58, 1, 0, 0, default_pwm, 0 },

    { "led16", 72, 1, 0, 0, default_pwm, 0 },
    { "led17", 73, 1, 0, 0, default_pwm, 0 },
    { "led18", 74, 1, 0, 0, default_pwm, 0 },
    { "led19", 75, 1, 0, 0, default_pwm, 0 },
    { "led20", 54, 1, 0, 0, default_pwm, 0 },
    { "led21", 77, 1, 0, 0, default_pwm, 0 },
    { "led22", 78, 1, 0, 0, default_pwm, 0 },
    { "led23", 55, 1, 0, 0, default_pwm, 0 },

    { "led24",  4, 1, 0, 0, default_pwm, 0 },
    { "led25",  6, 1, 0, 0, default_pwm, 0 },
    { "led26",  7, 1, 0, 0, default_pwm, 0 },
    { "led27",  9, 1, 0, 0, default_pwm, 0 },
    { "led28", 13, 1, 0, 0, default_pwm, 0 },
    { "led29", 16, 1, 0, 0, default_pwm, 0 },
    { "led30", 17, 1, 0, 0, default_pwm, 0 },
    { "led31",  1, 1, 0, 0, default_pwm, 0 },
};

struct gsboard_gpio board_gpio_gxw4224_v23[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont",
        .gpio_pin  = 34, 
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },  
    {
        .gpio_name = "fan_det",
        .gpio_pin  = 3,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },  
    {
        .gpio_name = "fan_off_1",
        .gpio_pin  = 11,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  51, 1, 0, 0, default_pwm, 0 },
    { "led1",  68, 1, 0, 0, default_pwm, 0 },
    { "led2",  69, 1, 0, 0, default_pwm, 0 },
    { "led3",  70, 1, 0, 0, default_pwm, 0 },
    { "led4",  71, 1, 0, 0, default_pwm, 0 },
    { "led5",  79, 1, 0, 0, default_pwm, 0 },
    { "led6",  57, 1, 0, 0, default_pwm, 0 },
    { "led7",  58, 1, 0, 0, default_pwm, 0 },

    { "led8",  72, 1, 0, 0, default_pwm, 0 },
    { "led9",  73, 1, 0, 0, default_pwm, 0 },
    { "led10", 74, 1, 0, 0, default_pwm, 0 },
    { "led11", 75, 1, 0, 0, default_pwm, 0 },
    { "led12", 54, 1, 0, 0, default_pwm, 0 },
    { "led13", 77, 1, 0, 0, default_pwm, 0 },
    { "led14", 78, 1, 0, 0, default_pwm, 0 },
    { "led15", 55, 1, 0, 0, default_pwm, 0 },
    { "led16",  4, 1, 0, 0, default_pwm, 0 },
    { "led17",  6, 1, 0, 0, default_pwm, 0 },
    { "led18",  7, 1, 0, 0, default_pwm, 0 },
    { "led19",  9, 1, 0, 0, default_pwm, 0 },
    { "led20", 13, 1, 0, 0, default_pwm, 0 },
    { "led21", 16, 1, 0, 0, default_pwm, 0 },
    { "led22", 17, 1, 0, 0, default_pwm, 0 },
    { "led23",  1, 1, 0, 0, default_pwm, 0 },
};

struct gsboard_gpio board_gpio_gxw4216_v23[] = {
    {
        .gpio_name = "board_reset", 
        .gpio_pin  = 10,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "phy_reset", 
        .gpio_pin  = 8,
        .gpio_dir  = 1, 
        .skip_init = 1, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "slave_reset", 
        .gpio_pin  = 12,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "lcd_backlight", 
        .gpio_pin  = 38,
        .gpio_dir  = 1, 
        .skip_init = 0, 
        .init_val  = 1, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_down", 
        .gpio_pin  = 37,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "key_enter", 
        .gpio_pin  = 36,
        .gpio_dir  = 0, 
        .skip_init = 0, 
        .init_val  = 0, 
        .pwm = default_pwm
    },
    {
        .gpio_name = "fan_cont",
        .gpio_pin  = 34, 
        .gpio_dir  = 1,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },  
    {
        .gpio_name = "fan_det",
        .gpio_pin  = 3,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },  
    {
        .gpio_name = "fan_off_1",
        .gpio_pin  = 11,
        .gpio_dir  = 0,
        .skip_init = 0,
        .init_val  = 0,
        .pwm = default_pwm
    },

    /* phone LEDs */
    { "led0",  72, 1, 0, 0, default_pwm, 0 },
    { "led1",  73, 1, 0, 0, default_pwm, 0 },
    { "led2",  74, 1, 0, 0, default_pwm, 0 },
    { "led3",  75, 1, 0, 0, default_pwm, 0 },
    { "led4",  54, 1, 0, 0, default_pwm, 0 },
    { "led5",  77, 1, 0, 0, default_pwm, 0 },
    { "led6",  78, 1, 0, 0, default_pwm, 0 },
    { "led7",  55, 1, 0, 0, default_pwm, 0 },
    { "led8",   4, 1, 0, 0, default_pwm, 0 },
    { "led9",   6, 1, 0, 0, default_pwm, 0 },
    { "led10",  7, 1, 0, 0, default_pwm, 0 },
    { "led11",  9, 1, 0, 0, default_pwm, 0 },
    { "led12", 13, 1, 0, 0, default_pwm, 0 },
    { "led13", 16, 1, 0, 0, default_pwm, 0 },
    { "led14", 17, 1, 0, 0, default_pwm, 0 },
    { "led15",  1, 1, 0, 0, default_pwm, 0 },
};

struct gsboard_port_map board_port_map_gxw4216_v23[] = 
{
    { 0,  0,  0,  0, 1 },
    { 0,  1,  1,  1, 1 },
    { 0,  2,  2,  2, 1 },
    { 0,  3,  3,  3, 1 },
    { 0,  4,  4,  4, 1 },
    { 0,  5,  5,  5, 1 },
    { 0,  6,  6,  6, 1 },
    { 0,  7,  7,  7, 1 },
    { 0,  8,  8,  8, 1 },
    { 0,  9,  9,  9, 1 },
    { 0, 10, 10, 10, 1 },
    { 0, 11, 11, 11, 1 },
    { 0, 12, 12, 12, 1 },
    { 0, 13, 13, 13, 1 },
    { 0, 14, 14, 14, 1 },
    { 0, 15, 15, 15, 1 },
};

struct gsboard_port_map board_port_map_gxw4224_v23[] = 
{
    { 0,  0,  0,  0, 1 },
    { 0,  1,  1,  1, 1 },
    { 0,  2,  2,  2, 1 },
    { 0,  3,  3,  3, 1 },
    { 0,  4,  4,  4, 1 },
    { 0,  5,  5,  5, 1 },
    { 0,  6,  6,  6, 1 },
    { 0,  7,  7,  7, 1 },
    { 0,  8,  8,  8, 1 },
    { 0,  9,  9,  9, 1 },
    { 0, 10, 10, 10, 1 },
    { 0, 11, 11, 11, 1 },
    { 0, 12, 12, 12, 1 },
    { 0, 13, 13, 13, 1 },
    { 0, 14, 14, 14, 1 },
    { 0, 15, 15, 15, 1 },
    { 1,  0, 16, 16, 1 },
    { 1,  1, 17, 17, 1 },
    { 1,  2, 18, 18, 1 },
    { 1,  3, 19, 19, 1 },
    { 1,  4, 20, 20, 1 },
    { 1,  5, 21, 21, 1 },
    { 1,  6, 22, 22, 1 },
    { 1,  7, 23, 23, 1 },
};

struct gsboard_port_map board_port_map_gxw4224_v24[] = 
{
    { 0,  8,  0, 22, 1 },
    { 0,  9,  1, 23, 1 },
    { 0, 10,  2, 20, 1 },
    { 0, 11,  3, 21, 1 },
    { 0, 12,  4, 18, 1 },
    { 0, 13,  5, 19, 1 },
    { 0, 14,  6, 16, 1 },
    { 0, 15,  7, 17, 1 },
    { 1,  0,  8, 14, 1 },
    { 1,  1,  9, 15, 1 },
    { 1,  2, 10, 12, 1 },
    { 1,  3, 11, 13, 1 },
    { 1,  4, 12, 10, 1 },
    { 1,  5, 13, 11, 1 },
    { 1,  6, 14,  8, 1 },
    { 1,  7, 15,  9, 1 },
    { 1,  8, 16,  6, 1 },
    { 1,  9, 17,  7, 1 },
    { 1, 10, 18,  4, 1 },
    { 1, 11, 19,  5, 1 },
    { 1, 12, 20,  2, 1 },
    { 1, 13, 21,  3, 1 },
    { 1, 14, 22,  0, 1 },
    { 1, 15, 23,  1, 1 }, 
};

struct gsboard_port_map board_port_map_gxw4232_v23[] = 
{
    { 0,  0,  0,  0, 1 },
    { 0,  1,  1,  1, 1 },
    { 0,  2,  2,  2, 1 },
    { 0,  3,  3,  3, 1 },
    { 0,  4,  4,  4, 1 },
    { 0,  5,  5,  5, 1 },
    { 0,  6,  6,  6, 1 },
    { 0,  7,  7,  7, 1 },
    { 0,  8,  8,  8, 1 },
    { 0,  9,  9,  9, 1 },
    { 0, 10, 10, 10, 1 },
    { 0, 11, 11, 11, 1 },
    { 0, 12, 12, 12, 1 },
    { 0, 13, 13, 13, 1 },
    { 0, 14, 14, 14, 1 },
    { 0, 15, 15, 15, 1 },
    { 1,  0, 16, 16, 1 },
    { 1,  1, 17, 17, 1 },
    { 1,  2, 18, 18, 1 },
    { 1,  3, 19, 19, 1 },
    { 1,  4, 20, 20, 1 },
    { 1,  5, 21, 21, 1 },
    { 1,  6, 22, 22, 1 },
    { 1,  7, 23, 23, 1 },
    { 1,  8, 24, 24, 1 },
    { 1,  9, 25, 25, 1 },
    { 1, 10, 26, 26, 1 },
    { 1, 11, 27, 27, 1 },
    { 1, 12, 28, 28, 1 },
    { 1, 13, 29, 29, 1 },
    { 1, 14, 30, 30, 1 },
    { 1, 15, 31, 31, 1 },
};

struct gsboard_port_map board_port_map_gxw4232_v24[] = 
{
    { 0,  0,  0, 30, 1 },
    { 0,  1,  1, 31, 1 },
    { 0,  2,  2, 28, 1 },
    { 0,  3,  3, 29, 1 },
    { 0,  4,  4, 26, 1 },
    { 0,  5,  5, 27, 1 },
    { 0,  6,  6, 24, 1 },
    { 0,  7,  7, 25, 1 },
    { 0,  8,  8, 22, 1 },
    { 0,  9,  9, 23, 1 },
    { 0, 10, 10, 20, 1 },
    { 0, 11, 11, 21, 1 },
    { 0, 12, 12, 18, 1 },
    { 0, 13, 13, 19, 1 },
    { 0, 14, 14, 16, 1 },
    { 0, 15, 15, 17, 1 },
    { 1,  0, 16, 14, 1 },
    { 1,  1, 17, 15, 1 },
    { 1,  2, 18, 12, 1 },
    { 1,  3, 19, 13, 1 },
    { 1,  4, 20, 10, 1 },
    { 1,  5, 21, 11, 1 },
    { 1,  6, 22,  8, 1 },
    { 1,  7, 23,  9, 1 },
    { 1,  8, 24,  6, 1 },
    { 1,  9, 25,  7, 1 },
    { 1, 10, 26,  4, 1 },
    { 1, 11, 27,  5, 1 },
    { 1, 12, 28,  2, 1 },
    { 1, 13, 29,  3, 1 },
    { 1, 14, 30,  0, 1 },
    { 1, 15, 31,  1, 1 },
};

struct gsboard_port_map board_port_map_gxw4248_v11[] = 
{
    { 0,  0,  0, 47, 1 },
    { 0,  1,  1, 46, 1 },
    { 0,  2,  2, 45, 1 },
    { 0,  3,  3, 44, 1 },
    { 0,  4,  4, 43, 1 },
    { 0,  5,  5, 42, 1 },
    { 0,  6,  6, 41, 1 },
    { 0,  7,  7, 40, 1 },
    { 0,  8,  8, 39, 1 },
    { 0,  9,  9, 38, 1 },
    { 0, 10, 10, 37, 1 },
    { 0, 11, 11, 36, 1 },
    { 0, 12, 12, 35, 1 },
    { 0, 13, 13, 34, 1 },
    { 0, 14, 14, 33, 1 },
    { 0, 15, 15, 32, 1 },
    { 1,  0, 16, 31, 1 },
    { 1,  1, 17, 30, 1 },
    { 1,  2, 18, 29, 1 },
    { 1,  3, 19, 28, 1 },
    { 1,  4, 20, 27, 1 },
    { 1,  5, 21, 26, 1 },
    { 1,  6, 22, 25, 1 },
    { 1,  7, 23, 24, 1 },
    { 1,  8, 24, 23, 1 },
    { 1,  9, 25, 22, 1 },
    { 1, 10, 26, 21, 1 },
    { 1, 11, 27, 20, 1 },
    { 1, 12, 28, 19, 1 },
    { 1, 13, 29, 18, 1 },
    { 1, 14, 30, 17, 1 },
    { 1, 15, 31, 16, 1 },
    { 2,  0, 32, 15, 1 },
    { 2,  1, 33, 14, 1 },
    { 2,  2, 34, 13, 1 },
    { 2,  3, 35, 12, 1 },
    { 2,  4, 36, 11, 1 },
    { 2,  5, 37, 10, 1 },
    { 2,  6, 38,  9, 1 },
    { 2,  7, 39,  8, 1 },
    { 2,  8, 40,  7, 1 },
    { 2,  9, 41,  6, 1 },
    { 2, 10, 42,  5, 1 },
    { 2, 11, 43,  4, 1 },
    { 2, 12, 44,  3, 1 },
    { 2, 13, 45,  2, 1 },
    { 2, 14, 46,  1, 1 },
    { 2, 15, 47,  0, 1 },
};

struct gsboard_port_map board_port_map_gxw4248_v12[] = 
{
    { 0,  0,  0, 47, 1 },
    { 0,  1,  1, 46, 1 },
    { 0,  2,  2, 39, 1 },
    { 0,  3,  3, 38, 1 },
    { 0,  4,  4, 31, 1 },
    { 0,  5,  5, 30, 1 },
    { 0,  6,  6, 23, 1 },
    { 0,  7,  7, 22, 1 },
    { 0,  8,  8, 15, 1 },
    { 0,  9,  9, 14, 1 },
    { 0, 10, 10,  7, 1 },
    { 0, 11, 11,  6, 1 },
    { 0, 12, 12, 37, 1 },
    { 0, 13, 13, 36, 1 },
    { 0, 14, 14, 45, 1 },
    { 0, 15, 15, 44, 1 },
    { 1,  0, 16, 29, 1 },
    { 1,  1, 17, 28, 1 },
    { 1,  2, 18, 21, 1 },
    { 1,  3, 19, 20, 1 },
    { 1,  4, 20, 13, 1 },
    { 1,  5, 21, 12, 1 },
    { 1,  6, 22,  5, 1 },
    { 1,  7, 23,  4, 1 },
    { 1,  8, 24,  3, 1 },
    { 1,  9, 25,  2, 1 },
    { 1, 10, 26, 11, 1 },
    { 1, 11, 27, 10, 1 },
    { 1, 12, 28, 19, 1 },
    { 1, 13, 29, 18, 1 },
    { 1, 14, 30, 27, 1 },
    { 1, 15, 31, 26, 1 },
    { 2,  0, 32, 43, 1 },
    { 2,  1, 33, 42, 1 },
    { 2,  2, 34, 35, 1 },
    { 2,  3, 35, 34, 1 },
    { 2,  4, 36, 41, 1 },
    { 2,  5, 37, 40, 1 },
    { 2,  6, 38, 33, 1 },
    { 2,  7, 39, 32, 1 },
    { 2,  8, 40, 25, 1 },
    { 2,  9, 41, 24, 1 },
    { 2, 10, 42, 17, 1 },
    { 2, 11, 43, 16, 1 },
    { 2, 12, 44,  9, 1 },
    { 2, 13, 45,  8, 1 },
    { 2, 14, 46,  1, 1 },
    { 2, 15, 47,  0, 1 },

};

struct gsboard_hw_sw_limit board_ver_limit_gxw4216_v23 = 
{
    GS_SW_VER(1, 0, 0, 1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } }, /* PHY: default HW setting is 1.5ns */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4216_v30b = 
{
    /* 
     * Supported version is COREVER 1.0.19.3
     * Downgrade limit feature is valid on COREVER 1.0.9.2+
     */
    GS_SW_VER(1, 0, 19, 2+1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } }, /* PHY: default HW setting is 1.5ns */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4224_v23 = 
{
    GS_SW_VER(1, 0, 0, 1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } },
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4224_v24 = 
{
    /* 
     * Supported version is COREVER 1.0.5.44
     * Downgrade limit feature is valid on COREVER 1.0.9.1+
     */
    GS_SW_VER(1, 0, 9, 1+1),
    { 1, { 4, 4 }, { 4, 4 }, { 3, 3 } }, /* BH Telecom 10M/100M issue */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4224_v30b = 
{
    /* 
     * Supported version is COREVER 1.0.19.3
     * Downgrade limit feature is valid on COREVER 1.0.9.2+
     */
    GS_SW_VER(1, 0, 19, 2+1),
    { 1, { 4, 4 }, { 4, 4 }, { 3, 3 } }, /* BH Telecom 10M/100M issue */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4232_v23 = 
{
    GS_SW_VER(1, 0, 0, 1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } },
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4232_v24 = 
{
    /* 
     * Supported version is COREVER 1.0.5.44
     * Downgrade limit feature is valid on COREVER 1.0.9.1+
     */
    GS_SW_VER(1, 0, 9, 1+1),
    { 1, { 4, 4 }, { 4, 4 }, { 3, 3 } }, /* BH Telecom 10M/100M issue */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4232_v30b = 
{
    /* 
     * Supported version is COREVER 1.0.19.3
     * Downgrade limit feature is valid on COREVER 1.0.9.2+
     */
    GS_SW_VER(1, 0, 19, 2+1),
    { 1, { 4, 4 }, { 4, 4 }, { 3, 3 } }, /* BH Telecom 10M/100M issue */
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4248_v11 = 
{
    GS_SW_VER(1, 0, 0, 1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } },
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4248_v12 = 
{
    /* 
     * Supported version is COREVER 1.0.5.44
     * Downgrade limit feature is valid on COREVER 1.0.9.1+
     */
    GS_SW_VER(1, 0, 9, 1+1),
    { 0, { 3, 3 }, { 3, 3 }, { 3, 3 } },
    { 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

struct gsboard_hw_sw_limit board_ver_limit_gxw4248_v14_16 =
{
	/*
	 * Supported version is COREVER 1.0.5.44
	 * Downgrade limit feature is valid on COREVER 1.0.21.1+
	 */
	GS_SW_VER(1, 0, 21, 1+1),
	{ 0, { 3, 3 }, { 3, 3 }, { 3, 3 } },
	{ 0, { 0, 0 }, { 0, 0 }, { 0, 0 } }
};

int ascToInt(char *p)
{
    int ret;

    if (*p >= 0x30 && *p <= 0x39)
        ret = *p - 0x30;
    else if (*p >= 0x41 && *p <= 0x46)
        ret = *p - 0x37;
    else if (*p >= 0x61 && *p <= 0x66)
        ret = *p - 0x57;
    else
        ret = 0;

    return ret;
}

static void gxw_led_pattern_state_change(void)
{
    int ii, index;

    for (ii = 0; ii < led_pattern_count; ii ++)
    {
        index = board_gpio_gs_count - led_pattern_count + ii;
        if (board_gpio_gs[index].extended_gpio)
            ifxmips_i2c_gpio_set(board_gpio_gs[index].gpio_pin, led_tickonoff);
        else
            gpio_set_value(board_gpio_gs[index].gpio_pin, led_tickonoff);
    }
}

static void gxw_led_pattern_timer(unsigned long data)
{
    if (soft_pwm_tick())
        mod_timer(&led_pattern_timer, jiffies + 500 * HZ / 1000);
    else
        soft_pwm_active = 0;
}

int gs_add_nandpartition(struct mtd_info *mtd, struct mtd_info *mtd_256ecc, int index)
{
    unsigned int i, size, offset;
    struct mtd_info *mtdnand;

	if (0 == index) {
		gs_nand_partition_info = _gs_nand_partition_info_nand_small_page;
		gs_nand_partition_count = ((sizeof(_gs_nand_partition_info_nand_small_page))/(sizeof(struct mtd_partition)));
		factParaBuf = factParaBuf_nsp;
		factory_cfg_size = GS_FACTORY_CFG_SIZE;
		factory_cfg_offset = GS_FACTORY_CFG_OFFSET;
		gs_nand_type = 0;
		printk(KERN_INFO "add nand partitions for small page size\n");
	} else {
		gs_nand_partition_info = _gs_nand_partition_info_nand_large_page;
		gs_nand_partition_count = ((sizeof(_gs_nand_partition_info_nand_large_page))/(sizeof(struct mtd_partition)));
		factParaBuf = factParaBuf_nlp;
		factory_cfg_size = GS_FACTORY_CFG_SIZE_NLP;
		factory_cfg_offset = GS_FACTORY_CFG_OFFSET_NLP;
		gs_nand_type = 1;
		printk(KERN_INFO "add nand partitions for large page size\n");
	}

    /*
     * adjust offset if there is bad block on header area */
    for(i = 0; i < gs_nand_partition_count; i++)
    {
        size = gs_nand_partition_info[i].size;
        offset = gs_nand_partition_info[i].start;
        if (gs_nand_partition_info[i].start == gs_nand_partition_info[i].offset
        && strncmp(gs_nand_partition_info[i].name,"swap", strlen("swap")) != 0)//by dign, swap partition need check
            /* skip special partition without header */
            continue;
        if ( i == 0 )
        {
            mtdnand = mtd_256ecc;
        }
        else
        {
            mtdnand = mtd;
        }
        /* we need find the first good block for image header */
        while(size > mtdnand->erasesize)
        {
            if (mtdnand->block_isbad(mtdnand, offset) == 0)
                break;
            /* block is bad */
            offset += mtdnand->erasesize;
            /* if we advanced into real image area, we need adjust the partition size */
            if (offset >= gs_nand_partition_info[i].offset)
                size -= mtdnand->erasesize;
        }
        /* 
         * adjust partition offset and size based on header block position, pay attention to 
         * the special handling to make it compatible to both lp and sp nand version
         */
        gs_nand_partition_info[i].size = size; /* size shall not be zero here as 0 means the whole rest space */
        gs_nand_partition_info[i].offset = offset + (gs_nand_partition_info[i].offset - gs_nand_partition_info[i].start);
    }
    add_mtd_partitions(mtd_256ecc, gs_nand_partition_info, 1);
    add_mtd_partitions(mtd, gs_nand_partition_info + 1, gs_nand_partition_count - 1);
    return gs_nand_partition_count;
}

int gs_get_nand_type(void)
{
	return gs_nand_type;
}
EXPORT_SYMBOL(gs_get_nand_type);

int gs_internal_mtd_read(struct mtd_info *mtd, u_int32_t addr, u_int32_t len, char *buf_ptr)
{
    size_t retlen;
    int ret;

    if ( !mtd ) mtd = gs_mtd;
    ret = (*(mtd->read))(mtd, addr, len, &retlen, (unsigned char *)buf_ptr);
    if (ret || retlen != len)
    {
        printk ("mtd read failed. len=%d, ret=%d, retlen=%d\n", len, ret, retlen);
        return 1;
    }
    return 0;
}

short get_lcd_model(void)
{
    return lcd_model;
}

short get_led_gpioset(void)
{
    return led_gpioset;
}

#define LED_MAP_IS_GXW4248_REV_11A()  (get_led_gpioset() == 9 || get_led_gpioset() == 13)

static int load_fact_para(struct mtd_info *mtd, char *buf_ptr)
{
    GS_FACTORYPARA *factPara = (GS_FACTORYPARA *)buf_ptr;
    unsigned int length = 0;
    int ii,jj;
    int ret;

    memset( &sec_info, 0, sizeof ( sec_info ) );

    ret = gs_internal_mtd_read(mtd, factory_cfg_offset, factory_cfg_size, (char *)buf_ptr);
    if (ret)
    {
        printk ("Read factory data failed.\n");
        return 1;
    }

    if (fact_para_checksum_be(factPara) == factPara->checksum)
    {
        /* check feature table for logo data */
        length = (gs_endian_conv16(factPara->length[1]) << 16) | gs_endian_conv16(factPara->length[0]); // length le
        if (length > sizeof(GS_FACTORYPARA))
        {
            /* hw feature table is available */
            unsigned short *featurePtr;
            hw_feature_table_header_t * hw_table = (hw_feature_table_header_t *)factPara->hw_feature_table_word;
            unsigned short hw_feature_table_len = gs_endian_conv16(hw_table->hw_feature_table_len);

            ii = 0;
            if ( gs_endian_conv16(hw_table->hw_feature_table_version) == HW_FEATURE_TABLE_VERSION )
            {
                ii = 0;
                while (ii < (hw_feature_table_len - 3))
                {
                    unsigned short featureLength;
                    featurePtr = (unsigned short*)factPara->hw_feature_table_word + ii + sizeof(hw_feature_table_header_t)/sizeof(uint16_t);
                    printk("feature item %d\n", gs_endian_conv16(*featurePtr));
                    switch (gs_endian_conv16(*featurePtr))
                    {
                        case HW_FEATURE_MODELBASE:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 128) featureLength = 128; //max 128 bytes
                            memcpy(hw_feature_modelbase, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_MODELSTR:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 128) featureLength = 128; //max 128 bytes
                            memcpy(hw_feature_modelstr, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_MODELALIAS:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 128) featureLength = 128; //max 128 bytes
                            memcpy(hw_feature_modelalias, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_OEM_ID:
                            hw_feature_oem_id = gs_endian_conv16(*(featurePtr+2));
                            break;
                        case HW_FEATURE_LED_GPIOSET:
                            led_gpioset = gs_endian_conv16(*(featurePtr+2));
                            printk("led gpioset=%d\n", led_gpioset);
                            break;
                        case HW_FEATURE_LCD_MODEL:
                            lcd_model = gs_endian_conv16(*(featurePtr+2));
                            printk("lcd model=%d\n", lcd_model);
                            break;
                        case HW_FEATURE_VENDOR:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 128) featureLength = 128; //max 128 bytes
                            memcpy(hw_feature_vendor, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_VENDOR_FULLNAME:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 128) featureLength = 128; //max 128 bytes
                            memcpy(hw_feature_vendor_fullname, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_FWKEY:
                            featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                            if (featureLength > 32) featureLength = 32; //32 bytes long
                            memcpy(hw_feature_fwkey, (char *)(featurePtr+2), featureLength);
                            break;
                        case HW_FEATURE_REQUIRE_PEM:
                             require_pem = gs_endian_conv16(*(featurePtr+2));
                             break;
                        case HW_FEATURE_REQUIRE_SN:
                             require_sn = gs_endian_conv16(*(featurePtr+2));
                             break;
                        case HW_FEATURE_REQUIRE_ADMIN_PWD:
                             require_admin_pwd = gs_endian_conv16(*(featurePtr+2));
                             break;
                        case HW_FEATURE_REQUIRE_XML_KEY:
                             require_xml_key = gs_endian_conv16(*(featurePtr+2));
                             break;
                        case HW_FEATURE_REQUIRE_HASH:
                             require_hash = gs_endian_conv16(*(featurePtr+2));
                             break;
                    }
                    ii += gs_endian_conv16(*(featurePtr+1)) + 2;
                    continue;
                }
            }
            if ( length > ii + sizeof(GS_FACTORYPARA) )
            {
               hw_table = (hw_feature_table_header_t *)(factPara->hw_feature_table_word + ii + sizeof(hw_feature_table_header_t)/sizeof(uint16_t));
               hw_feature_table_len = gs_endian_conv16(hw_table->hw_feature_table_len);
            
               if ( gs_endian_conv16(hw_table->hw_feature_table_version) == HW_SECURITY_TABLE_VERSION )
               {
                  jj = 0;
                  while ( jj <= hw_feature_table_len - sizeof(hw_feature_table_header_t)/sizeof(uint16_t) - sizeof(hw_feature_attrib_t)/sizeof(uint16_t) )
                  {
                       unsigned short featureLength;
                       featurePtr = (unsigned short*)factPara->hw_feature_table_word + ii + jj + sizeof(hw_feature_table_header_t)/sizeof(uint16_t)*2;
                       printk("security feature item %d\n", gs_endian_conv16(*featurePtr));
                  
                       switch(gs_endian_conv16(*featurePtr))
                       {
                            case SEC_FEATURE_SN:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.sn.data = vmalloc( featureLength );
                                memcpy( sec_info.sn.data, (char *)(featurePtr+2), featureLength );
                                sec_info.sn.len = featureLength;
                                break;
                            case SEC_FEATURE_PRIVATE_KEY:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.private_key.data = vmalloc( featureLength );
                                memcpy( sec_info.private_key.data, (char *)(featurePtr+2), featureLength );
                                sec_info.private_key.len = featureLength;
                                break;
                            case SEC_FEATURE_CERTIFICATE:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.certificate.data = vmalloc( featureLength );
                                memcpy( sec_info.certificate.data, (char *)(featurePtr+2), featureLength );
                                sec_info.certificate.len = featureLength;
                                break;
                            case SEC_FEATURE_CHECKSUM:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.checksum.data = vmalloc( featureLength );
                                memcpy( sec_info.checksum.data, (char *)(featurePtr+2), featureLength );
                                sec_info.checksum.len = featureLength;
                                break;
                            case SEC_FEATURE_VONAGE_HASH:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.hash.data = vmalloc( featureLength );
                                memcpy( sec_info.hash.data, (char *)(featurePtr+2), featureLength );
                                sec_info.hash.len = featureLength;
                                break;
                            case SEC_FEATURE_VONAGE_PASSWORD:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.password.data = vmalloc( featureLength );
                                memcpy( sec_info.password.data, (char *)(featurePtr+2), featureLength);
                                sec_info.password.len = featureLength;
                                break;
                            case SEC_FEATURE_VONAGE_XML_KEY:
                                featureLength = gs_endian_conv16(*(featurePtr+1)) << 1;
                                sec_info.xml_key.data = vmalloc( featureLength );
                                memcpy( sec_info.xml_key.data, (char *)(featurePtr+2), featureLength );
                                sec_info.xml_key.len = featureLength;
                                break;
                       }
                       jj += gs_endian_conv16(*(featurePtr+1)) + 2;
                  }
               }
             }
        }
    }
    else
        printk("factory data checksum failed\n");

    return 0;
}


static void gs_mtd_erase_callback_lp (struct erase_info *instr)
{
    unsigned long delay = jiffies + 3 * HZ / 1000;

    while(time_before(jiffies, delay))
        cond_resched();

    wake_up((wait_queue_head_t *)instr->priv);
}

static void gs_mtd_erase_callback (struct erase_info *instr)
{
    wake_up((wait_queue_head_t *)instr->priv);
}

int gs_internal_mtd_erase(struct mtd_info *mtd, u_int32_t addr, u_int32_t len)
{
    struct erase_info *erase;
    int ret = 0;

    if ( !mtd ) mtd = gs_mtd;
    erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
    if (!erase)
        ret = -ENOMEM;
    else {
        wait_queue_head_t waitq;
        DECLARE_WAITQUEUE(wait, current);

        init_waitqueue_head(&waitq);

        memset (erase,0,sizeof(struct erase_info));
        erase->addr = addr;
        erase->len = len;
        erase->mtd = mtd;
        if (mtd->erasesize > 512)
            erase->callback = gs_mtd_erase_callback_lp;
       else
            erase->callback = gs_mtd_erase_callback;

        erase->priv = (unsigned long)&waitq;

        /*
          FIXME: Allow INTERRUPTIBLE. Which means
          not having the wait_queue head on the stack.
          
          If the wq_head is on the stack, and we
          leave because we got interrupted, then the
          wq_head is no longer there when the
          callback routine tries to wake us up.
        */
        ret = (*(mtd->erase))(mtd, erase);
        if (!ret) {
            set_current_state(TASK_UNINTERRUPTIBLE);
            add_wait_queue(&waitq, &wait);
            if (erase->state != MTD_ERASE_DONE &&
                erase->state != MTD_ERASE_FAILED)
                schedule();
            remove_wait_queue(&waitq, &wait);
            set_current_state(TASK_RUNNING);

            ret = (erase->state == MTD_ERASE_FAILED)?-EIO:0;
        }
        kfree(erase);
    }
    return ret;

}

int gs_internal_mtd_write(struct mtd_info *mtd, u_int32_t addr, u_int32_t len, char *buf_ptr)
{
    size_t retlen;
    int ret=0;

    if ( !mtd ) mtd = gs_mtd;
    ret = (*(mtd->write))(mtd, addr, len, &retlen, (unsigned char *)buf_ptr);

    return ret;
}

int gs_internal_mtd_recoverbad(struct mtd_info *mtd, u_int32_t addr)
{
    if ( !mtd ) mtd = gs_mtd;
    return nand_recover_bad(mtd, addr);
}

int gs_internal_mtd_checkbad(struct mtd_info *mtd, u_int32_t addr)
{
    if ( !mtd ) mtd = gs_mtd;
    return ((*(mtd->block_isbad))(mtd, addr));
}

int gs_internal_mtd_markbad(struct mtd_info *mtd, u_int32_t addr)
{
    if ( !mtd ) mtd = gs_mtd;
    return ((*(mtd->block_markbad))(mtd, addr));
}

int gs_internal_partition_info(struct mtd_info *mtd, int mtd_num, uint32_t *start, uint32_t *offset, uint32_t *size, int *erase_size)
{
	unsigned int addr = gs_nand_partition_info[mtd_num].start;
	if(mtd_num > gs_nand_partition_count || mtd_num < 0)
	{
		return -1;
	}
	if ( !mtd ) mtd = gs_mtd;
	while ((*(mtd->block_isbad))(mtd, addr))
	{
		/* try the backup block */
		addr += mtd->erasesize;
		if (addr >= gs_nand_partition_info[mtd_num].offset)
		{
			// no image header found
			return -1;
		}
    	}
	*start = gs_nand_partition_info[mtd_num].start;
	*offset = addr;
	*size = gs_nand_partition_info[mtd_num].size;
	*erase_size = mtd->erasesize;
	return  0;
}

unsigned short fact_para_checksum_be(GS_FACTORYPARA *factPara)
{
    unsigned char  *p = (unsigned char *)factPara;
    unsigned short checksum = 0;
    unsigned short checksum_word = 0;
    unsigned int length = 0;
    int i;

    // the length field is designed to be able expand the structure but since we need copy
    // the whole structure, it makes nothing except has chance to crash when length is invalid
    length = (gs_endian_conv16(factPara->length[1]) << 16) | gs_endian_conv16(factPara->length[0]); // length le

    if (length > factory_cfg_size)
        length = factory_cfg_size;

    for (i = 0; i < length; i+=2)
    {
        if ( i == offsetof(GS_FACTORYPARA,checksum))
        {
            continue;   // skip checksum itself
        }
        checksum_word = (((unsigned short int)p[i]) << 8) | ((unsigned short int)p[i+1] & 0x00ff);
        checksum += checksum_word;
    }
    return (1 + ~checksum);
}

int get_gs_fact_param_pointer(void **ptr)
{
    *ptr = (void *)(factParaBuf);
    return 0;
}

unsigned long seemem_base = 0x87ef4fc0;

static int
gsboard_seemem_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *mapped_mem;

    len = 64;

    if (seemem_base == 0)
    {
        printk("base addr is zero!\n;");
        return 0;
    }

    mapped_mem = (char *)ioremap_nocache( seemem_base, 64);

    if (mapped_mem) 
    {
        memcpy(page, mapped_mem, len);
        iounmap(mapped_mem);
    }
    else
    {
        printk("map mem failed!\n;");
        return 0;
    }

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gsboard_seemem_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    seemem_base = simple_strtoul(state_string, &endp, 16);

    return (count);
}

static int
gsboard_flag_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    char        *p = page;
    int         len = 0;
    unsigned long val = *((int*)data);

    p += sprintf(p, "%ld\n", val);

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gsboard_flag_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long       count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;
    unsigned long newValue = 0;
    int *val = (int*)data;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    newValue = simple_strtoul(state_string, &endp, 16) & 0x1;
    if (newValue == 0)
        *val = 0;
    else
        *val = 1;

    return (count);
}

#ifdef CONFIG_GXW42XX_BUILD_48_96
static int
gs_board_slavereset_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;
    unsigned long newValue = 0;
    int * ptr_slave_reset_set = slave_reset_set;
    int ii;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    newValue = simple_strtoul(state_string, &endp, 16) & 0x1;

    if ( LED_MAP_IS_GXW4248_REV_11A() )
    {
        ptr_slave_reset_set = slave_reset_set_rev_11A;
    }

    /* Note: we assume those gpios were initialized in board gpios */
    for (ii = 0; ii < slave_reset_size; ii++)
    {
        gpio_direction_output(ptr_slave_reset_set[ii], newValue);
//        printk("slave_reset[%d] = %d\n", ptr_slave_reset_set[ii], newValue);
    }

    return (count);
}
#endif

static int
gsboard_led_pattern_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long       count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;
    unsigned long newValue = 0;
    int ii, index;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    newValue = simple_strtoul(state_string, &endp, 16) & 0x1;
    if (newValue != led_pattern && led_pattern_count > 0)
    {
        switch (newValue)
        {
            case 0:
                led_pattern = 0;
                led_tickonoff = led_tickonoff_saved;
                /* for #15028. Need to differentiate led and other gpio */
                for (ii = 0; ii < led_pattern_count; ii ++)
                {
                    index = board_gpio_gs_count - led_pattern_count + ii;
                    board_gpio_gs[index].pwm.pwm_stop = 1;
                }
                break;
            case 1: // provision state
            default:
                led_pattern = 1;
                led_tickonoff_saved = led_tickonoff;
                if (!soft_pwm_active)
                {
                    mod_timer(&led_pattern_timer, jiffies + 500 * HZ / 1000);
                    soft_pwm_active = 1;
                }
                break;
        }
    }

    return (count);
}

static int
gsboard_control_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    char        *p = page;
    int         len = 0;
    unsigned long val = (unsigned long)(*((unsigned short*)data));

    p += sprintf(p, "%lu\n", val);

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gsboard_control_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;
    unsigned long newValue = 0;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    newValue = simple_strtoul(state_string, &endp, 10);

    {
        *((unsigned short*)data) = (unsigned int)newValue;
    }

    return (count);
}

static int
gs_board_factcfg_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

    len = (gs_endian_conv16(factPara->length[1]) << 16) | gs_endian_conv16(factPara->length[0]); // length le
    if (len > factory_cfg_size)
        len = factory_cfg_size;

    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        memcpy(page, factPara, len);
    }
    else
    {
        memset(page, 0x0, 86);
        len = 86;
    }

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
#if 1
    GS_FACTORYPARA* factory_para = (GS_FACTORYPARA*) factParaBuf;
    unsigned short  mac[3];

    if ((count != sizeof(mac)))
    {
        printk("invalid size\n");
        return (-EINVAL);
    }

    if (copy_from_user(mac, buffer, 6)) // 6 bytes
    {
        printk("copy error\n");
        return (-EFAULT);
    }

    /* make sure checksum is right */
    if (factory_para->checksum != fact_para_checksum_be(factory_para))
    {
        printk("checksum fail\n");
        return (-EINVAL);
    }

    printk("Request for setting MAC %04x:%04x:%04x\n", mac[0], mac[1], mac[2]);
    printk("Current MAC %04x:%04x:%04x\n", factory_para->mac[0], factory_para->mac[1], factory_para->mac[2]);

    /* Pay attemtion to endianess */
    if ( !( ( factory_para->mac[0] == 0x000b &&
            factory_para->mac[1] == 0x8200 &&
            factory_para->mac[2] == 0x0000 ) || 
            ( factory_para->mac[0] == 0xc074 &&
            factory_para->mac[1] == 0xad00 &&
            factory_para->mac[2] == 0x0000 )
          )
       )
    {
        printk("mac already set\n");
        return -EFAULT;
    }

    printk("Setting MAC %04x:%04x:%04x\n", mac[0], mac[1], mac[2]);

    factory_para->mac[0] = mac[0];
    factory_para->mac[1] = mac[1];
    factory_para->mac[2] = mac[2];

    factory_para->checksum = fact_para_checksum_be( factory_para );

#else
    /* this is for internal use only for XXU */
    GS_FACTORYPARA* factPara = vmalloc(factory_cfg_size);
    int len;

    if (factPara == NULL)
    {
        return (-ENOMEM);
    }
    if (copy_from_user(factPara, buffer, sizeof(GS_FACTORYPARA))) // 6 bytes
        return (-EFAULT);

    len = (gs_endian_conv16(factPara->length[1]) << 16) | gs_endian_conv16(factPara->length[0]); // length le
    if (len > factory_cfg_size)
        len  = factory_cfg_size;

    if (copy_from_user(factPara, buffer, len)) // 6 bytes
        return (-EFAULT);

    if (factPara->checksum != fact_para_checksum_be(factPara))
        return (-EINVAL);
    }
#endif

    /* SXIE: not needed if header is not using the entire block */
    if ( gs_internal_mtd_erase(NULL, factory_cfg_offset, gs_mtd->erasesize) < 0  )
    {
        printk("Error erasing flash\n");
        return -EFAULT;
    }
    /* Write Flash */
    if ( gs_internal_mtd_write(NULL, factory_cfg_offset, factory_cfg_size, (char *)factory_para) < 0 )
    {
        printk("Error writing flash\n");
        return -EFAULT;
    }

    return (count);
}

static int
gs_board_factcfg_parse_feature ( void )
{
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        char *pn = (char *)factPara->PartNumber;
        int Hw_rev = (pn[8] - 0x30) * 16 + (pn[9] - 0x30);
        char Hw_sub_rev = pn[10];
        printk("PN: %s\n", (char *)&factPara->PartNumber);
        if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4232_PNSTR) == 0)
        {
            if ( ( Hw_rev > 0x30 ) || ( ( Hw_rev == 0x30 ) && ( Hw_sub_rev == 'B') ) )
            {
                board_port_map_gs = board_port_map_gxw4232_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v30b;
            }
            else if ( Hw_rev >= 0x24 )
            {
                board_port_map_gs = board_port_map_gxw4232_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v24;
            }
            else
            {
                board_port_map_gs = board_port_map_gxw4232_v23;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v23);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v23;
            }
        }
        else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4224_PNSTR) == 0)
        {
            if ( ( Hw_rev > 0x30 ) || ( ( Hw_rev == 0x30 ) && ( Hw_sub_rev == 'B') ) )
            {
                board_port_map_gs = board_port_map_gxw4224_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4224_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4224_v30b;
            }
            else if ( Hw_rev >= 0x24 )
            {
                board_port_map_gs = board_port_map_gxw4224_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4224_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4224_v24;
            }
            else
            {
                board_port_map_gs = board_port_map_gxw4224_v23;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4224_v23);
                board_hw_sw_ver = &board_ver_limit_gxw4224_v23;
            }
        }
        else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4216_PNSTR) == 0)
        {
            if ( ( Hw_rev > 0x30 ) || ( ( Hw_rev == 0x30 ) && ( Hw_sub_rev == 'B') ) )
            {
                board_port_map_gs = board_port_map_gxw4216_v23;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4216_v23);
                board_hw_sw_ver = &board_ver_limit_gxw4216_v30b;
            }
            else
            {
               board_port_map_gs = board_port_map_gxw4216_v23;
               board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4216_v23);
               board_hw_sw_ver = &board_ver_limit_gxw4216_v23;
            }
        }
        else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4248_PNSTR) == 0)
        {
			if (Hw_rev >= 0x16 || Hw_rev == 0x14)
			{
				board_port_map_gs = board_port_map_gxw4248_v12;
				board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4248_v12);
				board_hw_sw_ver = &board_ver_limit_gxw4248_v14_16;
			}
            else if ( Hw_rev >= 0x12 )
            {
                board_port_map_gs = board_port_map_gxw4248_v12;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4248_v12);
                board_hw_sw_ver = &board_ver_limit_gxw4248_v12;
            }
            else
            {
                board_port_map_gs = board_port_map_gxw4248_v11;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4248_v11);
                board_hw_sw_ver = &board_ver_limit_gxw4248_v11;
            }
        }
        else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4296_PNSTR) == 0)
        {
            // Unsupported
        }
        else // Default to GXW4232
        {
            if ( ( Hw_rev > 0x30 ) || ( ( Hw_rev == 0x30 ) && ( Hw_sub_rev == 'B') ) )
            {
                board_port_map_gs = board_port_map_gxw4232_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v30b;
            }
            else if ( Hw_rev >= 0x24 )
            {
                board_port_map_gs = board_port_map_gxw4232_v24;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v24);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v24;
            }
            else
            {
                board_port_map_gs = board_port_map_gxw4232_v23;
                board_port_map_gs_count = GS_ARRAY_SIZE(board_port_map_gxw4232_v23);
                board_hw_sw_ver = &board_ver_limit_gxw4232_v23;
            }
        }
    }

    return 0;
}

static int
gs_board_factcfg_hwmodel_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    int             alias = (int)data;
    char            hw_model[128];
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

    len = 6;
    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        {
            if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4232_PNSTR) == 0)
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4232_NICKSTR);
                    memcpy(hw_model, HW_GXW4232_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4232_STR);
                    memcpy(hw_model, HW_GXW4232_STR, len);
                }
            }
            else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4224_PNSTR) == 0)
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4224_NICKSTR);
                    memcpy(hw_model, HW_GXW4224_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4224_STR);
                    memcpy(hw_model, HW_GXW4224_STR, len);
                }
            }
            else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4216_PNSTR) == 0)
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4216_NICKSTR);
                    memcpy(hw_model, HW_GXW4216_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4216_STR);
                    memcpy(hw_model, HW_GXW4216_STR, len);
                }
            }
            else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4296_PNSTR) == 0)
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4296_NICKSTR);
                    memcpy(hw_model, HW_GXW4296_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4296_STR);
                    memcpy(hw_model, HW_GXW4296_STR, len);
                }
            }
            else if (compare_pn_str((char *)&factPara->PartNumber, HW_GXW4248_PNSTR) == 0)
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4248_NICKSTR);
                    memcpy(hw_model, HW_GXW4248_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4248_STR);
                    memcpy(hw_model, HW_GXW4248_STR, len);
                }
            }
            else    //default to 4232
            {
                if (alias == 0)
                {
                    len = strlen(HW_GXW4232_NICKSTR);
                    memcpy(hw_model, HW_GXW4232_NICKSTR, len);
                }
                else
                {
                    len = strlen(HW_GXW4232_STR);
                    memcpy(hw_model, HW_GXW4232_STR, len);
                }
            }
        }
    }
    else
    {
        len = strlen(HW_UNKNOWN);
        memcpy(hw_model, HW_UNKNOWN, len);
    }

    if (alias == 0) //dev_alias
    {
        if (strlen(hw_feature_modelalias))
        {
            len = strlen(hw_feature_modelalias);
            memcpy(hw_model, hw_feature_modelalias, len);
        }
    }
    else if (alias == 2)    //dev_id
    {
        if (strlen(hw_feature_modelstr))
        {
            len = strlen(hw_feature_modelstr);
            memcpy(hw_model, hw_feature_modelstr, len);
        }
    }
    else    //dev_base
    {
        if (strlen(hw_feature_modelbase))
        {
            len = strlen(hw_feature_modelbase);
            memcpy(hw_model, hw_feature_modelbase, len);
        }
    }

    memcpy(page, hw_model, len);

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

int gs_board_factcfg_hwrev_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;
    char *hwrev = (char *)page;
    char revStr[128];

    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        {
            char *pn = (char *)factPara->PartNumber;
            sprintf(revStr, "V%c.%c%c", pn[8], pn[9], pn[10]);
        }
        len = strlen(revStr);
        memcpy(hwrev, revStr, len);
    }
    else
    {
        len = strlen(HW_UNKNOWN);
        memcpy(hwrev, HW_UNKNOWN, len);
    }

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_devmac_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;
    int             len = 0;
    char            *default_mac = "00:0B:82:FF:FF:FF";
    char *dev_mac = (char *)page;
    char revStr[128];

    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        sprintf(revStr, "%02X:%02X:%02X:%02X:%02X:%02X",
                (factPara->mac[0] >> 8) & 0xFF,
                factPara->mac[0] & 0xFF, 
                (factPara->mac[1] >> 8) & 0xFF,
                factPara->mac[1] & 0xFF, 
                (factPara->mac[2] >> 8) & 0xFF,
                factPara->mac[2] & 0xFF 
                );
        len = strlen(revStr);
        memcpy(dev_mac, revStr, len);
    }
    else
    {
        len = strlen(default_mac);
        memcpy(dev_mac, default_mac, len);
    }
    
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_pn_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;
    int             len = 0;
    char *dev_pn = (char *)page;

    if (factPara->checksum == fact_para_checksum_be(factPara))
    {
        char *pn = (char *)factPara->PartNumber;
        len = strlen(pn);
        memcpy(dev_pn, pn, len);
    }
    else
    {
        len = strlen(HW_UNKNOWN);
        memcpy(dev_pn, HW_UNKNOWN, len);
    }
    
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_fwkey_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char *dev_fwkey = (char *)page;

    if (strlen(hw_feature_fwkey))
    {
        len = strlen(hw_feature_fwkey);
        memcpy(dev_fwkey, hw_feature_fwkey, len);
    }
    else
    {
        len = strlen("nokey");
        memcpy(dev_fwkey, "nokey", len);
    }
    
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_vendor_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char *dev_vendor = (char *)page;
    int             get_fullname = (int)data;

    if (get_fullname)
    {
        if (strlen(hw_feature_vendor_fullname))
        {
            len = strlen(hw_feature_vendor_fullname);
            memcpy(dev_vendor, hw_feature_vendor_fullname, len);
        }
        else
        {
            len = strlen("Grandstream Networks, Inc.");
            memcpy(dev_vendor, "Grandstream Networks, Inc.", len);
        }
    }
    else
    {
        if (strlen(hw_feature_vendor))
        {
            len = strlen(hw_feature_vendor);
            memcpy(dev_vendor, hw_feature_vendor, len);
        }
        else
        {
            len = strlen("Grandstream");
            memcpy(dev_vendor, "Grandstream", len);
        }
    }
    
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_oemid_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    if (hw_feature_oem_id)
    {
        p += sprintf(p, "%d\n", hw_feature_oem_id);
    }
    else
        p += sprintf(p, "%d\n", 0);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

#if 0
static int
gs_board_dummy_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char *dev_pn = (char *)page;

    {
        len = 1;
        if (*((int*)data) == 1)
            memcpy(dev_pn, "1", len);
        else
            memcpy(dev_pn, "0", len);
    }
    
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}
#endif

static int
gs_board_timecheck_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    /* simple printk */
    printk("--------------->uptime\n");
    return 0;
}

static int
gs_board_port_map_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    if ( board_port_map_proc_entry == ENTRY_MAP_TABLE_INFO )
    {
        if ( board_port_map_gs )
        {
            p += sprintf(p, "%d,%d,%d,%d,%d\n", 
                board_port_map_gs[board_port_map_gs_index].vineticDevId,
                board_port_map_gs[board_port_map_gs_index].vineticDevCh,
                board_port_map_gs[board_port_map_gs_index].vineticPort,
                board_port_map_gs[board_port_map_gs_index].slicPort,
                board_port_map_gs[board_port_map_gs_index].svip16);
            board_port_map_gs_index++;
            if ( board_port_map_gs_index >= board_port_map_gs_count )
            {
                board_port_map_gs_index = 0;
            }
        }
        else
        {
            p += sprintf(p, "-1,-1,-1,-1,-1\n");
        }
    }
    else if ( board_port_map_proc_entry == ENTRY_MAP_TABLE_SIZE )
    {
        p += sprintf(p, "%d\n", board_port_map_gs_count);
    }
    else
    {
        p += sprintf(p, "Port map table index: %d\n", board_port_map_gs_index);
    }

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_port_map_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    char        *endp;
    unsigned long newValue = 0;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    newValue = simple_strtoul(state_string, &endp, 10);

    {
        *((unsigned long*)data) = newValue;
        board_port_map_gs_index = 0;
    }

    return (count);
}

static int
gs_board_hw_sw_ver_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    if ( board_hw_sw_ver )
    {
        p += sprintf(p, "%d.%d.%d.%d\n", 
         GS_SW_VER_HILONG_HIWORD(board_hw_sw_ver->suppored_sw_ver),
         GS_SW_VER_HILONG_LOWORD(board_hw_sw_ver->suppored_sw_ver),
         GS_SW_VER_LOLONG_HIWORD(board_hw_sw_ver->suppored_sw_ver),
         GS_SW_VER_LOLONG_LOWORD(board_hw_sw_ver->suppored_sw_ver)
         );
    }
    else
    {
        /* forbidden to downgrade. */
        p += sprintf(p, "255.255.255.255\n");
    }

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_nand_type_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", gs_nand_type);

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_require_pem_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", require_pem);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_require_sn_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", require_sn);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_require_admin_pwd_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", require_admin_pwd);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_require_xml_key_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", require_xml_key);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_factcfg_require_hash_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int             len = 0;
    char            *p = page;

    p += sprintf(p, "%d\n", require_hash);
    
    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}



static void
control_soft_pwm(int gpio_index, int on, int pwm_value)
{
    if (on)
    {
        board_gpio_gs[gpio_index].pwm.pwm_value = pwm_value;
        board_gpio_gs[gpio_index].pwm.tick = 0;
        board_gpio_gs[gpio_index].pwm.pwm_stop = 0;
        if (!soft_pwm_active)
        {
            mod_timer(&led_pattern_timer, jiffies + 500 * HZ / 1000);
            soft_pwm_active = 1;
        }
    }
    else
    {
        board_gpio_gs[gpio_index].pwm.pwm_value = 0;
        board_gpio_gs[gpio_index].pwm.pwm_stop = 1;
        /* If the pwm timer is running, we will be unable to stop the gpio here. */
        if (!soft_pwm_active)
            soft_pwm_tick();    /* turn it off */
    }
}

void start_half_led_blinking(void)
{
    int ii, index;
    for (ii = 0; ii < led_pattern_count; ii += 2)
    {
        index = board_gpio_gs_count - led_pattern_count + ii;
        control_soft_pwm(index, PWM_ON, 0xaaaaaaaa);
    }   
}

void stop_half_led_blinking(void)
{
    int ii, index;
    for (ii = 0; ii < led_pattern_count; ii += 2)
    {
        index = board_gpio_gs_count - led_pattern_count + ii;
        control_soft_pwm(index, PWM_OFF, 0);
    }   
}

int
soft_pwm_tick(void)
{
    register int ind, ii;
    int ret = 0;

    led_tickonoff = (led_tickonoff + 1) & 1;

    if (led_pattern)
    {
        /* led patten change */
        gxw_led_pattern_state_change();
        return 1;
    }
    else
    {
        /* prevent setting gpio other than LED */
        for (ind = 0; ind < led_pattern_count; ind ++)
        {
            ii = board_gpio_gs_count - led_pattern_count + ind; 
            if (board_gpio_gs[ii].gpio_dir == 0)     //input
                continue;

            if (board_gpio_gs[ii].pwm.pwm_stop)
            {
                if (!board_gpio_gs[ii].pwm.pwm_onoff)
                {
                   if (board_gpio_gs[ii].extended_gpio)
                       ifxmips_i2c_gpio_set(board_gpio_gs[ii].gpio_pin, 0);
                   else
                       gpio_direction_output(board_gpio_gs[ii].gpio_pin, 0);
                }
                board_gpio_gs[ii].pwm.pwm_value = 0;
                board_gpio_gs[ii].pwm.pwm_stop = 0;
                continue;
            }
            if (board_gpio_gs[ii].pwm.pwm_value)
            {
                int index = (board_gpio_gs[ii].pwm.tick + 1) & 0x1f;

                ret ++;

                /* 
                 * SXIE: we start on tick 1, but it is still okay we check tick 1 to make sure
                 * tick 1 is aligned on on/off state. 
                 * This is just to align first on tick.
                 */
                if (index == 1 && ((board_gpio_gs[ii].pwm.pwm_value >> 1 ) & 1) != led_tickonoff) 
                    continue; 

                board_gpio_gs[ii].pwm.tick = index;

                if (board_gpio_gs[ii].pwm.pwm_value & (1 << index))
                {
                    if (board_gpio_gs[ii].extended_gpio)
                        ifxmips_i2c_gpio_set(board_gpio_gs[ii].gpio_pin, 1);
                    else
                        gpio_direction_output(board_gpio_gs[ii].gpio_pin, 1);
                }
                else
                {
                    if (board_gpio_gs[ii].extended_gpio)
                        ifxmips_i2c_gpio_set(board_gpio_gs[ii].gpio_pin, 0);
                    else
                        gpio_direction_output(board_gpio_gs[ii].gpio_pin, 0);
                }

            }
        }
    }
    return ret;
}

static int
gs_board_gpio_init( unsigned int gpio_index)
{
    struct gsboard_gpio *p_gpio = &board_gpio_gs[gpio_index];
    unsigned int gpio_pin = (unsigned int)p_gpio->gpio_pin;

    if (p_gpio->skip_init || p_gpio->extended_gpio)
        return 0;

    /* set pin mux */
    if (gpio_request(gpio_pin, "gs_board_gpio") < 0) 
    {
        printk("gs_board_gpio_init, request gpio failed %d\n", gpio_pin);
        return -1;
    }

    /* set gpio direction */
    if (p_gpio->gpio_dir == 0) /* input */
        gpio_direction_input(gpio_pin);
    else
    {
        //printk("### gpio name is: %s, init_val: %d\n", p_gpio->gpio_name, p_gpio->init_val);
        gpio_direction_output(gpio_pin, p_gpio->init_val);
    }

    return 0;
}

static int
gs_board_gpio_proc_read (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    char        *p = page;
    int         len = 0;
    int         index;
    int         pin_val = 0;

    index = (int)data;
    pin_val = gpio_get_value(board_gpio_gs[index].gpio_pin);
    p += sprintf(p, "%x\n", pin_val? 1 : 0);

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int
gs_board_gpio_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data)
{
    char        state_string[12] = {'\0'};
    int         index;
    unsigned long newValue = 0;
    unsigned long newPWM = 0;
    int         pwm_changed = 0;

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';

    index = (int)data;

    if (state_string[0] == '1')
    {
        newValue = 1;
    }
    else
    {
        newValue = 0;
    }

    if ((state_string[1] == '.') && (strlen(state_string) > 2))
    {
        newPWM = simple_strtoul(&state_string[2], NULL, 16) & 0xffffffff;
        /* prevent setting the same pwm value twice */
        if (newPWM != board_gpio_gs[index].pwm.pwm_value)
        {
            board_gpio_gs[index].pwm.pwm_value = newPWM;
            pwm_changed = 1;
        }
    }
    else
    {
        if (board_gpio_gs[index].pwm.pwm_value)
        {
            board_gpio_gs[index].pwm.pwm_value = 0;
            pwm_changed = 1;
        }
    }
    board_gpio_gs[index].pwm.pwm_onoff = newValue;

    if (pwm_changed)
    {
        if (board_gpio_gs[index].pwm.pwm_value)
            control_soft_pwm(index, PWM_ON, board_gpio_gs[index].pwm.pwm_value);
        else
        {
            control_soft_pwm(index, PWM_OFF, board_gpio_gs[index].pwm.pwm_value);
            if (board_gpio_gs[index].pwm.pwm_onoff)
            {
               if (board_gpio_gs[index].extended_gpio)
                   ifxmips_i2c_gpio_set(board_gpio_gs[index].gpio_pin, newValue);
               else
                   gpio_set_value(board_gpio_gs[index].gpio_pin, newValue);
            }
        }
    }
    else if (newPWM == 0)
    {
        if (board_gpio_gs[index].extended_gpio)
            ifxmips_i2c_gpio_set(board_gpio_gs[index].gpio_pin, newValue);
        else
            gpio_set_value(board_gpio_gs[index].gpio_pin, newValue);
    }

    return (count);
}

/*
 * proc routine for mtd partitions
 */
static int
provision_mtd_proc_read_nand (
    char            *page,
    char            **start,
    off_t           off,
    int             count,
    int             *eof,
    void            *data)
{
    int         len = 0;
    int         ii = (int)data;
    int         ret = 0;
    int         isBootLoader = 0;
    int         addr;
    pImageHeader p = (pImageHeader)page;
    struct mtd_info * mtd;

    if ( memcmp(gs_nand_partition_info[ii].name, "bootloader", strlen("bootloader")) == 0 )
    {
        isBootLoader = 1;
    }

    if ( isBootLoader )
    {
        mtd = gs_mtd_256ecc;
    }
    else
    {
        mtd = gs_mtd;
    }

    len = sizeof(ImageHeader);
    addr = gs_nand_partition_info[ii].start;
    while ((*(mtd->block_isbad))(mtd, addr))
    {
        /* try the backup block */
        addr += mtd->erasesize;
        if (addr >= gs_nand_partition_info[ii].offset)
        {
            // no image header found
            ret = -1;
            break;
        }
    }

    if (addr < gs_nand_partition_info[ii].offset) {
        ret = gs_internal_mtd_read(mtd, addr, len, (char *)p);
    }

    if (gs_endian_conv32(p->Signature) != GS_PROD_SIGNATURE || ret != 0)
        memset(p, 0x0, len);

    //SXIE: do we need blank hw_model info to use same image for all the models?
    //p->reserved[0] = 0x0;    

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

/* mode indicates write to which copy of bootloader */
static int
provision_mtd_proc_write_boot (
    struct file     *file,
    const char      *buffer,
    unsigned long   count,
    void            *data)
{
    int         ii = (int)data;
    //int         len = 0;
    int         ret = 0;
    int         addr, end_addr;
    char*       image_buf;
    unsigned long   aligned_count;

    if (count > BOOT_SIZE)
    { 
        printk("Error, writing size can't be larger than boot partition"); 
        return (-EINVAL);
    }

    image_buf = vmalloc(BOOT_SIZE);

    if (image_buf == NULL)
        return (-ENOMEM);
    memset(image_buf, 0, BOOT_SIZE);
    addr = gs_nand_partition_info[ii].start + BOOT_SIZE;
    end_addr = addr + BOOT_SIZE;
    printk("write boot: %ld from 0x%08X to 0x%08X\n", count, addr, end_addr);
    while ((*(gs_mtd_256ecc->block_isbad))(gs_mtd_256ecc, addr))
    {
        /* try the backup block */
        addr += gs_mtd_256ecc->erasesize;
        if ((addr + count) >= end_addr)
        {
            // no space
            ret = -1;   
            break;
        }
    }

    if (ret != 0)
        return (-ENOSPC);

    if (copy_from_user(image_buf, buffer, count))
        return (-EFAULT);

    /* round count to block boundary, block size is 0x4000 */
    if (gs_mtd_256ecc->erasesize == 0x20000)
        aligned_count = (count + 0x1ffff) & ~0x1ffff;
    else if (gs_mtd_256ecc->erasesize == 0x4000)
        aligned_count = (count + 0x3fff) & ~0x3fff;
    else {
        printk("unsupported erasesize %x, check this\n", gs_mtd_256ecc->erasesize);
        aligned_count = (count + gs_mtd_256ecc->erasesize - 1) & ~(gs_mtd_256ecc->erasesize - 1);
    }

    printk("size aligned to 0x%08lX\n", aligned_count);
    gs_internal_mtd_erase(gs_mtd_256ecc, (u_int32_t)addr, aligned_count);

    /* Write Flash */
    gs_internal_mtd_write(gs_mtd_256ecc, (u_int32_t)addr, aligned_count, (char *)image_buf);
    vfree(image_buf);

    return (count);
}

static int
gsboard_kick_wdt_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long       count,
        void            *data)
{
    feed_wdt();
    return (count);
}


static int
provision_mtd_proc_write_nand (
    struct file     *file,
    const char      *buffer,
    unsigned long   count,
    void            *data)
{
    char        image_header[2048] = {'\0'};
    int         ii = (int)data;
    int         len = 0;
    int         ret = 0;
    int         addr;
    pImageHeader p = (pImageHeader)image_header;
    int         isBootLoader = 0;
    struct mtd_info * mtd;

    if ( memcmp(gs_nand_partition_info[ii].name, "bootloader", strlen("bootloader")) == 0 )
    {
        isBootLoader = 1;
    }

    if ( isBootLoader )
    {
        mtd = gs_mtd_256ecc;
    }
    else
    {
        mtd = gs_mtd;
    }
    printk("write nand: %s ", gs_nand_partition_info[ii].name);

    if (count != 2048)
    {
        if ( isBootLoader )
        {
            ret = provision_mtd_proc_write_boot(file, buffer, count, data);
            return ret;
        }
    }

    if ((count > 2048))
        return (-EINVAL);

    len = sizeof(ImageHeader);
    addr = gs_nand_partition_info[ii].start;
    while ((*(mtd->block_isbad))(mtd, addr))
    {
        /* try the backup block */
        addr += mtd->erasesize;
        if (addr >= gs_nand_partition_info[ii].offset)
        {
            // no image header found
            ret = -1;   
            break;
        }
    }

    if (ret != 0)
        return (-ENOSPC);

    printk("write: %ld from 0x%08X\n", count, addr);
    if (copy_from_user(p, buffer, count))
        return (-EFAULT);

    /* SXIE: not needed if header is not using the entire block */
    gs_internal_mtd_erase(mtd, (u_int32_t)addr, mtd->erasesize);

    /* Write Flash */
    gs_internal_mtd_write(mtd, (u_int32_t)addr, count, (char *)p);

    return (count);

}

// For security table item certificate and private key,
// their length may exceed one page, use seq_file

//=============================================================================
static void *
private_key_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    if ( sec_info.private_key.len == 0 ||
         *pos > 0 )
        return NULL;
    else
    	return (void *)( sec_info.private_key.data );

}

//=============================================================================
static void *
private_key_seq_next(
        struct seq_file *s, 
        void *v, 
        loff_t *pos
)
//=============================================================================
{
	(*pos)++;
	return private_key_seq_start(s, pos);
}

//=============================================================================
static void 
private_key_seq_stop(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	return;
}

//=============================================================================
static int private_key_seq_show(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	seq_printf(s, "%s", (char *) v);

	return 0;
}

static struct seq_operations private_key_seq_ops={
	.start = private_key_seq_start,
	.next = private_key_seq_next,
	.stop = private_key_seq_stop,
	.show = private_key_seq_show
};

//=============================================================================
static int 
private_key_proc_open(
        struct inode *inode, 
        struct file *file
)
//=============================================================================
{
	return seq_open(file, &private_key_seq_ops);
}

static struct file_operations private_key_proc_ops={
	.owner = THIS_MODULE,
	.open = private_key_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};

//=============================================================================
static void *
certificate_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    if ( sec_info.certificate.len == 0 ||
         *pos > 0 )
        return NULL;
    else
    	return (void *)( sec_info.certificate.data );

}

//=============================================================================
static void *
certificate_seq_next(
        struct seq_file *s, 
        void *v, 
        loff_t *pos
)
//=============================================================================
{
	(*pos)++;
	return certificate_seq_start(s, pos);
}

//=============================================================================
static void 
certificate_seq_stop(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	return;
}

//=============================================================================
static int certificate_seq_show(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	seq_printf(s, "%s", (char *) v);
	return 0;
}

static struct seq_operations certificate_seq_ops={
	.start = certificate_seq_start,
	.next = certificate_seq_next,
	.stop = certificate_seq_stop,
	.show = certificate_seq_show
};

//=============================================================================
static int 
certificate_proc_open(
        struct inode *inode, 
        struct file *file
)
//=============================================================================
{
	return seq_open(file, &certificate_seq_ops);
}

static struct file_operations certificate_proc_ops={
	.owner = THIS_MODULE,
	.open = certificate_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};

//=============================================================================
static int sec_feature_proc_read(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int len = 0;

    switch ( (int)data )
    {
        case SEC_FEATURE_SN:
            if ( sec_info.sn.len == 0 )
            {
                printk("serial number not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.sn.data );
            if( len > sec_info.sn.len )
                printk( "sn length is greater than the stored length\n" );
            memcpy( page, sec_info.sn.data, len );
            break;
        case SEC_FEATURE_PRIVATE_KEY:
            if ( sec_info.private_key.len == 0 )
            {
                printk("private key not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.private_key.data );
            if( len > sec_info.private_key.len )
                printk( "private key length is greater than the stored length\n" );
            memcpy( page, sec_info.private_key.data, len );
            break;
        case SEC_FEATURE_CERTIFICATE:
            if ( sec_info.certificate.len == 0 )
            {
                printk("certificate not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.certificate.data );
            if( len > sec_info.certificate.len )
                printk( "certificate length is greater than the stored length\n" );
            memcpy( page, sec_info.certificate.data, len );
            break;
        case SEC_FEATURE_CHECKSUM:
            if ( sec_info.checksum.len == 0 )
            {
                printk("checksum not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.checksum.data );
            if( len > sec_info.checksum.len )
                printk( "checksum length is greater than the stored length\n" );
            memcpy( page, sec_info.checksum.data, len );
            break;
        case SEC_FEATURE_VONAGE_HASH:
            if ( sec_info.hash.len == 0 )
            {
                printk("hash not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.hash.data );
            if( len > sec_info.hash.len )
                printk( "hash length is greater than the stored length\n" );
            memcpy( page, sec_info.hash.data, len );
            break;
        case SEC_FEATURE_VONAGE_PASSWORD:
            if ( sec_info.password.len == 0)
            {
                printk("password not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.password.data );
            if( len > sec_info.password.len )
                printk( "password length is greater than the stored length\n" );
            memcpy( page, sec_info.password.data, len );
            break;
        case SEC_FEATURE_VONAGE_XML_KEY:
            if ( sec_info.xml_key.len == 0)
            {
                printk("xml_key not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info.xml_key.data );
            if( len > sec_info.xml_key.len )
                printk( "xml key length is greater than the stored length\n" );
            memcpy( page, sec_info.xml_key.data, len );
            break;
        default:
            return -EINVAL;
    }

    if ( len <= off + count )
        *eof = 1;
    *start = page + off;
    len -= off;
    if ( len > count )
        len = count;
    if ( len < 0 )
        len = 0;

    return len;
}

//=============================================================================
static int
cfg_reload_write(
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data
)
//=============================================================================
{
    // use this to trigger factory reload after writing security table by factory tool 
    load_fact_para(gs_mtd, factParaBuf);
    return count;
}



static int mtdmap_install_proc_entry (void)
{
    struct proc_dir_entry *provision_root_dir;
    struct proc_dir_entry *gsboard_root_dir;
    struct proc_dir_entry   *ent;
    struct proc_dir_entry   *dent;
    int ii;

    /* install the proc entry */
    provision_root_dir = proc_mkdir( "provision", NULL);
    gsboard_root_dir = proc_mkdir( "gsboard", NULL);

    if (gsboard_root_dir != NULL)
    {
        /* create entry for wdt feeder */
        {
            ent = create_proc_entry("auto_wdt_feed", S_IFREG|S_IRUGO, gsboard_root_dir);
            if (!ent) {
                return (-1);
            }
            ent->read_proc  = gsboard_flag_proc_read;
            ent->write_proc = gsboard_flag_proc_write;
            ent->data       = (void *)&wdt_feed;
        }
        {
            ent = create_proc_entry("kick_wdt", S_IFREG|S_IRUGO, gsboard_root_dir);
            if (!ent) {
                return (-1);
            }
            ent->write_proc = gsboard_kick_wdt_proc_write;
        }

        ent = create_proc_entry("rtp_check_loss", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_control_proc_read;
            ent->write_proc = gsboard_control_proc_write;
            ent->data       = (void *)&rtp_port_inspect;
        }

        ent = create_proc_entry("seemem", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_seemem_proc_read;
            ent->write_proc  = gsboard_seemem_proc_write;
            ent->data       = (void *)0;
        }
        ent = create_proc_entry("network_link", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_flag_proc_read;
            ent->data       = (void *)&net_linked;
        }

#ifdef CONFIG_GXW42XX_ATE_BUILD
        ent = create_proc_entry("ATE_build", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_flag_proc_read;
            ent->data       = (void *)&ate_flag;
        }
#endif
#ifdef CONFIG_GXW42XX_RECOVERY_BUILD
        ent = create_proc_entry("RECOVERY_build", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_flag_proc_read;
            ent->data       = (void *)&recovery_flag;
        }
#endif
        printk("initialize gs board gpios\n");
        for (ii = 0; ii < board_gpio_gs_count; ii ++)
        {
            if (gs_board_gpio_init( ii ) == 0)
            {
                ent = create_proc_entry(board_gpio_gs[ii].gpio_name, S_IFREG|S_IRUGO, gsboard_root_dir);
                if (!ent) {
                    printk("create proc entry %s failed\n", board_gpio_gs[ii].gpio_name);
                    continue;
                }
                ent->read_proc  = gs_board_gpio_proc_read;
                ent->write_proc = gs_board_gpio_proc_write;
                ent->data       = (void*)ii;
            }
        }

#ifdef CONFIG_GXW42XX_BUILD_48_96
        /* this proc entry is for a single entrance for multiple slave chip reset with different gpios */
        ent = create_proc_entry("slave_reset", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->write_proc  = gs_board_slavereset_proc_write;
            ent->data       = NULL;
        }
#endif

        ent = create_proc_entry("uptime", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gs_board_timecheck_proc_read;
            ent->data       = NULL;
        }

        ent = create_proc_entry("port_map", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gs_board_port_map_proc_read;
            ent->write_proc = gs_board_port_map_proc_write;
            ent->data       = (void *)&board_port_map_proc_entry;
        }

        ent = create_proc_entry("ver_limit", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gs_board_hw_sw_ver_proc_read;
            ent->data       = NULL;
        }

        ent = create_proc_entry("nand_type", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gs_board_nand_type_proc_read;
            ent->data       = NULL;
        }

        ent = create_proc_entry("led_pattern", S_IFREG|S_IRUGO, gsboard_root_dir);
        if (ent) {
            ent->read_proc  = gsboard_flag_proc_read;
            ent->write_proc = gsboard_led_pattern_proc_write;
            ent->data       = (void *)&led_pattern;
        }

        dent = proc_mkdir("dev_info", gsboard_root_dir);

        /* entry for factory config data */
        ent = create_proc_entry("dev_data", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_proc_read;
            ent->write_proc = gs_board_factcfg_proc_write;
            ent->data       = NULL;
        }
        ent = create_proc_entry("dev_alias", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_hwmodel_proc_read;
            ent->data       = (void *)0;
        }
        ent = create_proc_entry("dev_base", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_hwmodel_proc_read;
            ent->data       = (void *)1;
        }
        ent = create_proc_entry("dev_id", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_hwmodel_proc_read;
            ent->data       = (void *)2;
        }
        ent = create_proc_entry("dev_rev", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_hwrev_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("fwkey", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_fwkey_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("vendor", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_vendor_proc_read;
            ent->data       = (void *)0;
        }
        ent = create_proc_entry("vendor_fullname", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_vendor_proc_read;
            ent->data       = (void *)1;
        }
        ent = create_proc_entry("oem_id", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_oemid_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("dev_mac", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_devmac_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("PN", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_pn_proc_read;
            ent->data       = NULL;
        }

        ent = create_proc_entry("require_pem", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_require_pem_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("require_sn", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_require_sn_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("require_admin_pwd", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_require_admin_pwd_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("require_xml_key", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_require_xml_key_proc_read;
            ent->data       = NULL;
        }
        ent = create_proc_entry("require_hash", S_IFREG|S_IRUGO, dent);
        if (ent) {
            ent->read_proc  = gs_board_factcfg_require_hash_proc_read;
            ent->data       = NULL;
        }

        ent = create_proc_entry( "cfg_reload", S_IFREG | S_IRUGO, gsboard_root_dir );
        if (ent) 
            ent->write_proc = cfg_reload_write;

        dent = proc_mkdir("security", dent);
        for ( ii = 0; ii < NUM_SEC_FEATURE_ENTRIES; ii ++ )
        {
            ent = create_proc_entry( sec_feature_entry[ ii ].name, S_IFREG | S_IRUGO, dent );
            if ( !ent )
            {
                printk("gs procfs sec_feature creation failure\n");
                return -1;
            }

            if ( sec_feature_entry[ ii ].value == SEC_FEATURE_PRIVATE_KEY )
            {
                ent->proc_fops = &private_key_proc_ops;
            }
            else if ( sec_feature_entry[ ii ].value == SEC_FEATURE_CERTIFICATE )
            {
                ent->proc_fops = &certificate_proc_ops;
            }
            else
            {
                ent->read_proc = sec_feature_proc_read;
                ent->write_proc = NULL;
                ent->data = (int*)( sec_feature_entry[ ii ].value );
            }
        }

#ifdef CONFIG_GSDSP_ACCEL
        dent = proc_mkdir("run_control", gsboard_root_dir);
        if (dent) {
            ent = create_proc_entry("accel_stat", S_IFREG|S_IRUGO, dent);
            if (ent) {
                ent->read_proc  = gsboard_control_proc_read;
                ent->write_proc = gsboard_control_proc_write;
                ent->data       = &do_accel_stat;
            }
            ent = create_proc_entry("accel_queue_size", S_IFREG|S_IRUGO, dent);
            if (ent) {
                ent->read_proc  = gsboard_control_proc_read;
                ent->write_proc = gsboard_control_proc_write;
                ent->data       = &accel_cmd_queue_size;
            }
        }
#endif
    }

    if (provision_root_dir != NULL)
    {
        struct proc_dir_entry   *dent;

        dent = proc_mkdir("partition", provision_root_dir);

        for (ii = 0; ii < gs_nand_partition_count; ii ++) 
        {
            if (gs_nand_partition_info[ii].start == gs_nand_partition_info[ii].offset)
            {
                /* not image partition */
                continue;
            }
            ent = create_proc_entry(gs_nand_partition_info[ii].name, S_IFREG|S_IRUGO, dent);
            if (!ent) {
                continue;
            }
            ent->read_proc  = provision_mtd_proc_read_nand;
            ent->write_proc  = provision_mtd_proc_write_nand;
            ent->data       = (void *)ii;
        }

    }
    return 0;
}

void gs_get_board_mac(char *mac)
{
    GS_FACTORYPARA*  factPara = (GS_FACTORYPARA*) factParaBuf;
    char *default_mac = ":=00:0b:82:ff:ff:ff";
    char *p = default_mac;
    int  tmp, j;

    for (j = 0; j < 6; j++)
    {
        p += 2;
        tmp = ascToInt(p);
        mac[j] = tmp << 4;
        p++;
        tmp = ascToInt(p);
        mac[j] |= tmp;
    }
    if (factPara->mac[0] == 0xffff && factPara->mac[1] == 0xffff)
    {
        printk ("No factory config data found, use default mac: 00:0b:82:ff:ff:ff.\n");
        return;
    }

    if (fact_para_checksum_be(factPara) == factPara->checksum &&
            !(factPara->mac[0] == 0 &&
              factPara->mac[1] == 0 &&
              factPara->mac[2] == 0))
    {
        // use the mac from fact para
        memcpy(mac, factPara->mac, 6);
    }
    else
    {
        printk("No valid MAC addr found in factory configuration, chksum 0x%x vs 0x%x\n", fact_para_checksum_be(factPara), factPara->checksum);
    }
    return;
}


void gs_init(struct mtd_info *mtd, struct mtd_info *mtd_256ecc)
{
    gs_mtd = mtd;
    gs_mtd_256ecc = mtd_256ecc;

    /* read out factory data */
    printk("Load factory configuration data\n");

    load_fact_para(mtd, factParaBuf);
    gs_board_factcfg_parse_feature();

    switch (led_gpioset)
    {
        case 0:
            board_gpio_gs = board_gpio_gxw4200_v21;
            board_gpio_gs_count = sizeof(board_gpio_gxw4200_v20) / sizeof(struct gsboard_gpio);
            break;
        case 1:
            board_gpio_gs = board_gpio_gxw4216_v22;
            board_gpio_gs_count = sizeof(board_gpio_gxw4216_v22) / sizeof(struct gsboard_gpio);
            led_pattern_count = 16;
            break;
        case 2:
            board_gpio_gs = board_gpio_gxw4224_v22;
            board_gpio_gs_count = sizeof(board_gpio_gxw4224_v22) / sizeof(struct gsboard_gpio);
            led_pattern_count = 24;
            break;
        case 3:
            board_gpio_gs = board_gpio_gxw4232_v22;
            board_gpio_gs_count = sizeof(board_gpio_gxw4232_v22) / sizeof(struct gsboard_gpio);
            led_pattern_count = 32;
            break;
        case 4:
            board_gpio_gs = board_gpio_gxw4248_v10;
            board_gpio_gs_count = sizeof(board_gpio_gxw4248_v10) / sizeof(struct gsboard_gpio);
            led_pattern_count = 48;
            break;
        case 5:
            board_gpio_gs = board_gpio_gxw4296_v10;
            board_gpio_gs_count = sizeof(board_gpio_gxw4296_v10) / sizeof(struct gsboard_gpio);
            led_pattern_count = 96;
            break;
        case 6:
        case 10:
            board_gpio_gs = board_gpio_gxw4216_v23;
            board_gpio_gs_count = sizeof(board_gpio_gxw4216_v23) / sizeof(struct gsboard_gpio);
            led_pattern_count = 16;
            break;
        case 7:
        case 11:
            board_gpio_gs = board_gpio_gxw4224_v23;
            board_gpio_gs_count = sizeof(board_gpio_gxw4224_v23) / sizeof(struct gsboard_gpio);
            led_pattern_count = 24;
            break;
        case 8:
        case 12:
            board_gpio_gs = board_gpio_gxw4232_v23;
            board_gpio_gs_count = sizeof(board_gpio_gxw4232_v23) / sizeof(struct gsboard_gpio);
            led_pattern_count = 32;
            break;
        case 9:
            board_gpio_gs = board_gpio_gxw4248_v11;
            board_gpio_gs_count = sizeof(board_gpio_gxw4248_v11) / sizeof(struct gsboard_gpio);
            led_pattern_count = 48;
            break;
		case 13:
            board_gpio_gs = board_gpio_gxw4248_v15;
            board_gpio_gs_count = sizeof(board_gpio_gxw4248_v15) / sizeof(struct gsboard_gpio);
            led_pattern_count = 48;	
			i2c_gpio_flag = 1;//0=tca6424 ;1=aw9523
			break;
        default:
            board_gpio_gs = board_gpio_gxw4200_v20;
            board_gpio_gs_count = sizeof(board_gpio_gxw4200_v21) / sizeof(struct gsboard_gpio);
            break;
    }

    {
        char board_mac[6];

        gs_get_board_mac(board_mac);
        ifxmips_replace_ethaddr(board_mac);
        switchip_set_mac_table_entry(board_mac);
    }

    /* register proc interface */
    mtdmap_install_proc_entry();

    setup_timer(&led_pattern_timer, gxw_led_pattern_timer, 0);
}

int gs_internal_read_appver(int part, unsigned int *ver)
{
    int addr, len;
    ImageHeader p;

    memset(&p, 0, sizeof(ImageHeader));
    len = sizeof(ImageHeader);
    addr = gs_nand_partition_info[part].start;
    while ((*(gs_mtd->block_isbad))(gs_mtd, addr))
    {
        /* try the backup block */
        addr += gs_mtd->erasesize;
        if (addr >= gs_nand_partition_info[part].offset)
        {
            // no image header found
            return -1;
        }
    }
    if(gs_internal_mtd_read(gs_mtd, addr, len, (char *)&p))
    {
        return -1;
    }
    if (gs_endian_conv32(p.Signature) != GS_PROD_SIGNATURE)
        return -1;
    *ver = p.Version;
    return 0;
}

int gs_internal_recreate_bbt(void)
{
    struct nand_chip *chip = gs_mtd->priv;
    /* Check, if we should skip the bad block table scan */
    if (chip->options & NAND_SKIP_BBTSCAN)
         return 0;
    /* Build bad block table */
    return chip->scan_bbt(gs_mtd);
}

int gs_erase_partition(char * name)
{
    int part = 0;
    int eb, i;
    u_int32_t eb_offs;
    struct mtd_partition *pPartition = NULL;
    struct mtd_info *mtd;

    while ( part < gs_nand_partition_count )
    {
        if ( 0 == strcmp(gs_nand_partition_info[part].name, name) )
        {
            pPartition = gs_nand_partition_info + part;
            break;
        }
        else
        {
            ++part;
        }
    }

    if ( !strcmp("bootloader", name) )
    {
        mtd = gs_mtd_256ecc;
    }
    else
    {
        mtd = gs_mtd;
    }
    if ( pPartition && mtd )
    {
        printk("Try to erase partition %s\n", pPartition->name);
        eb = 0;
        for ( i = 0; i < pPartition->size; i += mtd->erasesize, eb++ )
        {
            eb_offs = eb * mtd->erasesize;
            while ( gs_internal_mtd_checkbad( mtd, eb_offs ) != 0 )
            {
                eb++;
                eb_offs = eb * mtd->erasesize;
                if ( eb_offs >= pPartition->size )
                {
                    goto out;
                }
            }
            eb_offs += pPartition->start;
            gs_internal_mtd_erase( mtd, eb_offs, mtd->erasesize );
        }
    }
out:
    return 0;
}

EXPORT_SYMBOL(get_gs_fact_param_pointer);
EXPORT_SYMBOL(gs_get_board_mac);
EXPORT_SYMBOL(fact_para_checksum_be);

EXPORT_SYMBOL(gs_internal_mtd_erase);
EXPORT_SYMBOL(gs_internal_mtd_write);
EXPORT_SYMBOL(gs_internal_mtd_read);
EXPORT_SYMBOL(gs_internal_mtd_checkbad);
EXPORT_SYMBOL(gs_internal_partition_info);
EXPORT_SYMBOL(gs_internal_recreate_bbt);
EXPORT_SYMBOL(gs_erase_partition);

