/******************************************************************************

                               Copyright (c) 2009
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 81726 Munich, Germany

  THE DELIVERY OF THIS SOFTWARE AS WELL AS THE HEREBY GRANTED NON-EXCLUSIVE,
  WORLDWIDE LICENSE TO USE, COPY, MODIFY, DISTRIBUTE AND SUBLICENSE THIS
  SOFTWARE IS FREE OF CHARGE.

  THE LICENSED SOFTWARE IS PROVIDED "AS IS" AND INFINEON EXPRESSLY DISCLAIMS
  ALL REPRESENTATIONS AND WARRANTIES, WHETHER EXPRESS OR IMPLIED, INCLUDING
  WITHOUT LIMITATION, WARRANTIES OR REPRESENTATIONS OF WORKMANSHIP,
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, DURABILITY, THAT THE
  OPERATING OF THE LICENSED SOFTWARE WILL BE ERROR FREE OR FREE OF ANY THIRD
  PARTY CLAIMS, INCLUDING WITHOUT LIMITATION CLAIMS OF THIRD PARTY INTELLECTUAL
  PROPERTY INFRINGEMENT.

  EXCEPT FOR ANY LIABILITY DUE TO WILFUL ACTS OR GROSS NEGLIGENCE AND EXCEPT
  FOR ANY PERSONAL INJURY INFINEON SHALL IN NO EVENT BE LIABLE FOR ANY CLAIM
  OR DAMAGES OF ANY KIND, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.

*******************************************************************************
   Module      : ltq_svip_nat_pt.c

   Description : This file contains implementation of Lantiq's
                 VINETIC-SVIP NAT-Protocol Translation function
*******************************************************************************/
#ifdef CONFIG_LTQ_SVIP_NAT_PT

#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/if_vlan.h>
#include <linux/udp.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/in6.h> /* just to shut up a warning */
#include <linux/miscdevice.h>
#include <linux/udp_redirect.h>
#include <asm/checksum.h>
#include <linux/ltq_svip_nat_pt_io.h>
#include <linux/ifx_types.h>
#include <net/route.h>
MODULE_AUTHOR("Lantiq Deutschland GmbH");
MODULE_DESCRIPTION("SVIP NAT Protocol Translation module");
MODULE_LICENSE("GPL");

#define SVIP_NAT_PT_VERSION "0.1"

#define LTQ_SVIP_NAT_PT_INFO_STR "@(#)SVIP NAT PT, version "SVIP_NAT_PT_VERSION
#define MODNAME "SVIP NAT PT"

/* Define this in order to support same UDP port NATing. 
   This allows sockets to be bound the correct interface therefore 
   fixing the same UDP port ambiguity resulting in no audio */
#define SUPPORT_LOCAL_INTF_BIND

/* lookup port table for all possible ports */
#define SVIP_NAT_PORT_LOOKUP_TABLE_SIZE  (0x10000)

#define SVIP_NAT_RULES_TOTAL  (CONFIG_LTQ_SVIP_NAT_RULES_TOTAL)
#define SVIP_UDP_FROM         (CONFIG_LTQ_SVIP_NAT_UDP_PORT_BASE)
#define SVIP_UDP_TO           ((CONFIG_LTQ_SVIP_NAT_UDP_PORT_BASE)+(CONFIG_LTQ_SVIP_NAT_RULES_TOTAL)-1)
#define SVIP_NAT_INDEX_FROM   (0)
#define SVIP_NAT_INDEX_TO     ((0)+(CONFIG_LTQ_SVIP_NAT_RULES_TOTAL)-1)

#define SVIP_PORT_INRANGE(nPort) \
   ((nPort) >= (SVIP_UDP_FROM) && (nPort) <= (SVIP_UDP_TO))

#define SVIP_PORT_INDEX(nPort)   (nPort - SVIP_UDP_FROM)

#define SVIP_NAT_INDEX_INRANGE(nNatIndex) \
   ((nNatIndex) >= (SVIP_NAT_INDEX_FROM) && (nNatIndex) <= (SVIP_NAT_INDEX_TO))

#define SVIP_NET_DEV_ETH0_IDX       0
#define SVIP_NET_DEV_VETH0_IDX      1
#define SVIP_NET_DEV_LO_IDX         2

#define SVIP_NET_DEV_ETH0_NAME      "eth0"
#ifndef CONFIG_SVIP_INTLCC
/* SVIP used as External LCC */
#define SVIP_NET_DEV_VETH1_NAME     "eth1"
#else
/* SVIP used as Internal LCC */
#define SVIP_NET_DEV_VETH1_NAME     "veth0"
#endif
#define SVIP_NET_DEV_LO_NAME        "lo"

#define SVIP_NAT_STATS_LOC2REM   0
#define SVIP_NAT_STATS_REM2LOC   1
#define SVIP_NAT_STATS_TYPES     2

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
#define SVIP_NAT_FOR_EACH_NETDEV(d) for_each_netdev(&init_net, dev)
#define SVIP_NAT_IP_HDR(ethhdr) ip_hdr(ethhdr)
#else
#define SVIP_NAT_FOR_EACH_NETDEV(d) for(d=dev_base; dev; dev = dev->next)
#define SVIP_NAT_IP_HDR(ethhdr) (ethhdr)->nh.iph
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24) */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define SVIP_NAT_SKB_MAC_HEADER(ethhdr) (ethhdr)->mac.ethernet
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
#define SVIP_NAT_SKB_MAC_HEADER(ethhdr) (ethhdr)->mac.raw
#else
#define SVIP_NAT_SKB_MAC_HEADER(ethhdr) skb_mac_header(ethhdr)
#endif

typedef struct SVIP_NAT_PT_stats
{
   unsigned long           inPackets;
   unsigned long           outPackets;
   unsigned long           outErrors;
} SVIP_NAT_PT_stats_t;

typedef union
{
   struct sockaddr_in      addr;
   struct sockaddr_in6     addr6;
} u_sockaddr_in;

typedef struct SVIP_NAT_PT_sockdata
{
   struct socket           *sock;
   u_sockaddr_in           sockaddr;
   unsigned int            sockaddr_len;
   struct msghdr           msg;
   struct kvec             iov;
#define SOCK_STATE_CLOSED   (0x00000000)
#define SOCK_STATE_STARTING (0x00000001)
#define SOCK_STATE_RUNNING  (0x00000002)
#define SOCK_STATE_CLOSING  (0x00000003)
   unsigned int            sock_state;
} SVIP_NAT_PT_sockdata_t;

typedef struct SVIP_NAT_PT_table_entry
{
   int nat_index;
   int nat_entry_valid;
   SVIP_NAT_PT_IO_Rule_t   natRule;
   SVIP_NAT_PT_stats_t     natStats[SVIP_NAT_STATS_TYPES];
   SVIP_NAT_PT_sockdata_t  sockdata_loc;
   SVIP_NAT_PT_sockdata_t  sockdata_rem;
   IFX_uint8_t hookT38;
   IFX_uint8_t sendNoSignal;
   IFX_uint8_t udpErrCorrection;
   IFX_uint8_t updateT38SeqNum;
   IFX_uint8_t updateT38ErrRecovery;
   SVIP_NAT_PKT_STATS_t stats;
} SVIP_NAT_PT_table_entry_t;

typedef struct SVIP_UDPTL_PACKET
{
   IFX_uint16_t   seq_number;
   /* We do only need the sequence number here, 
     no need to define the following IFP structure. 
    */
   IFX_uint8_t    ifp;
} SVIP_UDPTL_PACKET_t, *PSVIP_UDPTL_PACKET_t;

/* pointer to the SVIP NAT PT table */
static SVIP_NAT_PT_table_entry_t *pNatPtTable = NULL;
static struct semaphore *sem_nat_tbl_access;
static int proc_read_in_progress = 0;
static int nDeviceOpen = 0;
/* saves the NAT table index between subsequent invocation */
static int nProcReadIdx = 0;
static SVIP_NAT_PT_table_entry_t * nat_port_table_ptr[SVIP_NAT_PORT_LOOKUP_TABLE_SIZE];
/* the map of nat rule index and local whole udp port */
static int svip_port_nat_rule_map[SVIP_NAT_PORT_LOOKUP_TABLE_SIZE];
/* Flag to indicate whether we have detected the far end no-signal. */
static IFX_uint8_t svip_no_signal_far_end[SVIP_NAT_PORT_LOOKUP_TABLE_SIZE];

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
/* Update current packet rate after PACKET_COUNT_RATE_UPDATE packets */
#define PACKET_COUNT_RATE_UPDATE (1000)
typedef struct SVIP_NAT_PT_performance_stats
{
   unsigned long packet_rate;
   unsigned long max_rate;
   unsigned long stats_total_packets;
   unsigned long stats_timestamp;
} SVIP_NAT_PT_performance_stats_t;

static SVIP_NAT_PT_performance_stats_t perfStats;
static void svip_nat_pt_stats_update(void);
#endif

static int svip_nat_pt_device_ioctl(struct inode *,struct file *,unsigned int ,unsigned long);
static int svip_nat_pt_device_release (struct inode *,struct file *);
static int svip_nat_pt_device_open    (struct inode *,struct file *);
static int svip_nat_pt_create_sock(SVIP_NAT_PT_table_entry_t *pNatPtEntry, SVIP_NAT_PT_IO_Rule_t *pNatPtRule, int bRem);
static void svip_nat_pt_release_sock(SVIP_NAT_PT_table_entry_t *pNatPtEntry, int bRem);
static int svip_nat_pt_create_ipv4_sock(struct socket **psock, SVIP_IPV4_ADDR_t *loc_ipv4, SVIP_UDP_PORT_t nUdp);
static int svip_nat_pt_create_ipv6_sock(struct socket **psock, SVIP_IPV6_ADDR_t *iff_ipv6, SVIP_UDP_PORT_t nUdp);
static void svip_nat_pt_prepare_sockaddr(SVIP_NAT_PT_table_entry_t *pNatPtEntry, SVIP_NAT_PT_IO_Rule_t *pNatPtRule, int bRem);
static int svip_nat_pt_sendmsg (SVIP_NAT_PT_sockdata_t *psockdata, void *data, unsigned int len);
static int svip_nat_pt_action(struct sock *sk,struct sk_buff *skb);
static int svip_nat_pt_tx_action(struct sk_buff *skb);
#if defined ( SUPPORT_LOCAL_INTF_BIND )
static int svip_nat_pt_lookup_sock_intf(struct socket *sock, struct sockaddr_in *locaddr, SVIP_IP_ADDR_UNION_t *pIP, SVIP_IP_ADDR_TYPE_t pTypeIP);
#endif
#if defined ( SUPPORT_SVIP_NAT_PT_DSCP )
static int svip_nat_pt_tos_config(struct socket *psock, char RTPipTOS);
#endif


/* This structure holds the interface functions supported by
   the SVIP NAT PT configuration device. */
struct file_operations svip_nat_pt_fops =
{
owner:	THIS_MODULE,
llseek:  NULL,                      /* seek */
read:    NULL,
write:   NULL,
readdir: NULL,                      /* readdir */
poll:    NULL,                      /* select */
ioctl:   svip_nat_pt_device_ioctl,  /* ioctl */
mmap:    NULL,                      /* mmap */
open:    svip_nat_pt_device_open,   /* open, */
flush:   NULL,                      /* flush */
release: svip_nat_pt_device_release /* close */
};

/** Structure holding MISC module operations */
static struct miscdevice svip_nat_pt_miscdev =
{
   minor:   MINOR_NUM_SVIP_NAT_PT,
   name:    SVIP_NAT_PT_DEVICE_NAME,
   fops:    &svip_nat_pt_fops
};

/******************************************************************************/
/**
   Function to read /proc/net/svip_nat/nat proc entry

\arguments
   page     - pointer to page buffer
   start    - pointer to start address pointer
   off      - offset
   count    - maximum data length to read
   eof      - end of file flag
   data     - proc read data (provided by the function
              pointed to by data)

\return
   length of read data

\remarks:
   Each call of this routine forces a copy_to_user of the data returned by
   'fn'. This routine will be called by the user until 'len = 0'.
****************************************************************************/
static int svip_nat_pt_proc_read (char *page, char **start, off_t off,
   int count, int *eof, void *data)
{
   unsigned long flags;
   int (*fn)(char *buf, int size);
   int len;

   /* If the NAT table index is negative, the reading has completed */
   if (nProcReadIdx < 0)
   {
      nProcReadIdx = 0;
      *eof = 1;
      proc_read_in_progress = 0;
      up(sem_nat_tbl_access);
      return 0;
   }

   local_irq_save(flags);
   if (!proc_read_in_progress)
   {
      proc_read_in_progress = 1;
      local_irq_restore(flags);
      /* we use this semaphore in order to ensure no other party(could be ioctl
         FIO_SVIP_NAT_PT_RULE_LIST), uses function SVIP_NAT_PT_proc_read_nat(), during
         the time read of the proc file takes place */
      down(sem_nat_tbl_access);
   }
   else
   {
      local_irq_restore(flags);
   }

   if (data != NULL)
   {
      fn = data;
      len = fn (page, count);
      /* In this setup each read of the proc entries returns the read data by
         'fn' to the user. The user keeps issuing read requests as long as the
         returned value of 'len' is greater than zero. */
      *eof = 1;
      *start = page;
   }
   else
   {
      len = 0;
   }

   return len;
}

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
/* Print statistincs for NAT performance */
static int svip_nat_pt_proc_read_stats(char *page, char **start, off_t off,
   int count, int *eof, void *data)
{
    unsigned long flags;
    char        *p = page;
    int         len = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
   local_irq_save(flags);
#else
   local_save_flags(flags);
#endif

    p += sprintf(p, "NAT Stats:\n----------------------------------------------------\n");
    p += sprintf(p, "%-30s: %8lu packets\n", "Total up/down link", perfStats.stats_total_packets );
    p += sprintf(p, "%-30s: %8lu packets/sec\n", "Current packet rate", perfStats.packet_rate );
    p += sprintf(p, "%-30s: %8lu packets/sec\n", "Max packet rate", perfStats.max_rate );

    /* Clear the curent packet rate. it will get updated if any packets are flowing in */
    perfStats.packet_rate = 0;

      /* update stat for highest packet rate */
    len = (p - page);
    if (len <= off+count)
         *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count)
         len = count;
    if (len<0)
         len = 0;

    local_irq_restore(flags);

   return len;
}
#endif

/******************************************************************************/
/**
   Function for setting up /proc/net/svip_nat read data

\arguments
   buf      - pointer to read buffer
   count    - size of read buffer

\return
   length of read data into buffer

\remarks:
   The global variable 'nProcReadIdx' is used to save the table index where
   the reading of the NAT table stopped. Reading is stopped when the end of
   the read buffer is approached. On the next itteration the reading continues
   from the saved index.
*******************************************************************************/
static int svip_nat_pt_proc_read_nat(char *buf, int count)
{
   int i, j;
   int len = 0;
   SVIP_NAT_PT_IO_Rule_t *pNatPtRule;

   if (nProcReadIdx == -1)
   {
      nProcReadIdx = 0;
      return 0;
   }

   if (nProcReadIdx == 0)
   {
      len = sprintf(buf+len,
                  " #  " /* 4 char */
                  "rem MAC            " /* 19 char */
                  "rem IP                                   " /* 41 char */
                  "rem UDP  " /* 9 char */
                  "loc MAC            " /* 19 char */
                  "loc IP                                   " /* 41 char */
                  "loc UDP  " /* 9 char */
                  "loc->rem(in/out/err)  " /* 22 char */
                  "rem->loc(in/out/err)\n\r"); /* 22 char = 182 total */

      len += sprintf(buf+len,
                  "--- " /* 19 char */
                  "-----------------  " /* 19 char */
                  "---------------------------------------  " /* 41 char */
                  "-------  " /* 9 char */
                  "-----------------  " /* 19 char */
                  "---------------------------------------  " /* 41 char */
                  "-------  " /* 9 char */
                  "--------------------  " /* 22 char */
                  "--------------------\n\r");
   }

   for (i = nProcReadIdx; i < CONFIG_LTQ_SVIP_NAT_RULES_TOTAL; i++)
   {
      int slen;

      pNatPtRule = &pNatPtTable[i].natRule;

      if (pNatPtRule->locUDP == 0)
         continue;

      /* make sure not to overwrite the buffer */
      if (count < len+182)
         break;

      /* remMAC */
      slen = 0;
      slen = sprintf(buf+len, "%3d ", i);

      for (j = 0; j < ETH_ALEN; j++)
      {
         slen += sprintf(buf+len+slen, "%02x%s",
                 pNatPtRule->remMAC[j], j < ETH_ALEN-1 ? ":" : "  ");
      }
      len += slen;
      for (j = 0; j < (19-slen); j++)
         len += sprintf(buf+len, " ");

      /* remIP */
      if (pNatPtRule->remTypeIP == SVIP_IPV4_ADDR_TYPE)
      {
         slen = sprintf(buf+len, "%d.%d.%d.%d",
                     (ntohl(pNatPtRule->remIP.v4) >> 24) & 0xff,
                     (ntohl(pNatPtRule->remIP.v4) >> 16) & 0xff,
                     (ntohl(pNatPtRule->remIP.v4) >> 8) & 0xff,
                     (ntohl(pNatPtRule->remIP.v4) >> 0) & 0xff);
         len += slen;
      }
      else
      {
         slen = sprintf(buf+len, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
                  ntohs(pNatPtRule->remIP.v6.ipv6_addr16[0]), ntohs(pNatPtRule->remIP.v6.ipv6_addr16[1]),
                  ntohs(pNatPtRule->remIP.v6.ipv6_addr16[2]), ntohs(pNatPtRule->remIP.v6.ipv6_addr16[3]),
                  ntohs(pNatPtRule->remIP.v6.ipv6_addr16[4]), ntohs(pNatPtRule->remIP.v6.ipv6_addr16[5]),
                  ntohs(pNatPtRule->remIP.v6.ipv6_addr16[6]), ntohs(pNatPtRule->remIP.v6.ipv6_addr16[7]));
         len += slen;
      }
      for (j = 0; j < (41-slen); j++)
         len += sprintf(buf+len, " ");

      /* remUDP */
      slen = sprintf(buf+len, "%d", pNatPtRule->remUDP);
      len += slen;
      for (j = 0; j < (9-slen); j++)
         len += sprintf(buf+len, " ");


      /* locMAC */
      slen = 0;
      for (j = 0; j < ETH_ALEN; j++)
      {
         slen += sprintf(buf+len+slen, "%02x%s",
                  pNatPtRule->locMAC[j], j < ETH_ALEN-1 ? ":" : " ");
      }
      len += slen;
      for (j = 0; j < (19-slen); j++)
         len += sprintf(buf+len, " ");

      /* locIP */
      if (pNatPtRule->locTypeIP == SVIP_IPV4_ADDR_TYPE)
      {
         slen = sprintf(buf+len, "%d.%d.%d.%d",
                     (ntohl(pNatPtRule->locIP.v4) >> 24) & 0xff,
                     (ntohl(pNatPtRule->locIP.v4) >> 16) & 0xff,
                     (ntohl(pNatPtRule->locIP.v4) >> 8) & 0xff,
                     (ntohl(pNatPtRule->locIP.v4) >> 0) & 0xff);
         len += slen;
      }
      else
      {
         slen = sprintf(buf+len, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
                  ntohs(pNatPtRule->locIP.v6.ipv6_addr16[0]), ntohs(pNatPtRule->locIP.v6.ipv6_addr16[1]),
                  ntohs(pNatPtRule->locIP.v6.ipv6_addr16[2]), ntohs(pNatPtRule->locIP.v6.ipv6_addr16[3]),
                  ntohs(pNatPtRule->locIP.v6.ipv6_addr16[4]), ntohs(pNatPtRule->locIP.v6.ipv6_addr16[5]),
                  ntohs(pNatPtRule->locIP.v6.ipv6_addr16[6]), ntohs(pNatPtRule->locIP.v6.ipv6_addr16[7]));
         len += slen;
      }
      for (j = 0; j < (41-slen); j++)
         len += sprintf(buf+len, " ");

      /* locUDP */
      slen = sprintf(buf+len, "%d", pNatPtRule->locUDP);
      len += slen;
      for (j = 0; j < (9-slen); j++)
         len += sprintf(buf+len, " ");

      /* NAT statistics, Local to Remote translation */
      slen = sprintf(buf+len, "(%ld/%ld/%ld)",
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_LOC2REM].inPackets,
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_LOC2REM].outPackets,
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_LOC2REM].outErrors);
      len += slen;
      for (j = 0; j < (22-slen); j++)
         len += sprintf(buf+len, " ");

      /* NAT statistics, Remote to Local translation */
      len += sprintf(buf+len, "(%ld/%ld/%ld)\n\r",
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_REM2LOC].inPackets,
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_REM2LOC].outPackets,
                  pNatPtTable[i].natStats[SVIP_NAT_STATS_REM2LOC].outErrors);
   }
   if (i == CONFIG_LTQ_SVIP_NAT_RULES_TOTAL)
      nProcReadIdx = -1;   /* reading completed */
   else
      nProcReadIdx = i;    /* reading still in process, buffer was full */

   return len;
}

/******************************************************************************/
/**
   Creates proc read/write entries

\return
   0 on success, -1 on error
*/
/******************************************************************************/
static int svip_nat_pt_proc_install(void)
{
   struct proc_dir_entry *pProcParentDir, *pProcDir;
   struct proc_dir_entry *pProcNode;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
   pProcParentDir = proc_net;
#else
   pProcParentDir = init_net.proc_net;
#endif
   pProcDir = proc_mkdir(SVIP_NAT_PT_DEVICE_NAME, pProcParentDir);
   if (pProcDir == NULL)
   {
      printk(KERN_ERR "SVIP NAT PT: cannot create proc dir %s/%s\n\r",
            pProcParentDir->name, SVIP_NAT_PT_DEVICE_NAME);
      return -1;
   }

   pProcNode = create_proc_read_entry("nat", S_IFREG|S_IRUGO, pProcDir,
         svip_nat_pt_proc_read, (void *)svip_nat_pt_proc_read_nat);
   if (pProcNode == NULL)
   {
      printk(KERN_ERR "SVIP NAT PT: cannot create proc entry %s/%s",
            pProcDir->name, "nat");
      return -1;
   }

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
   pProcNode = create_proc_read_entry("stats", S_IFREG|S_IRUGO, pProcDir,
         svip_nat_pt_proc_read_stats, (void *)0);
   if (pProcNode == NULL)
   {
      printk(KERN_ERR "SVIP NAT PT: cannot create proc entry %s/%s",
            pProcDir->name, "nat");
      return -1;
   }
#endif

   return 0;
}

/******************************************************************************/
/**
   No actions done here, simply a check is performed if an open has already
   been performed. Currently only a single open is allowed as it is a sufficient
   to have hat a single process configuring the SVIP NAT PT at one time.

\arguments
   inode       - pointer to disk file data
   file        - pointer to device file data

\return
   0 on success, else -1
*/
/******************************************************************************/
static int svip_nat_pt_device_open(struct inode *inode, struct file *file)
{
   unsigned long flags;
#if 0 /* debug RMi */
   struct in_device *in_dev;
   struct in_ifaddr *ifa;
#endif

   local_save_flags(flags);

   if (nDeviceOpen)
   {
      MOD_INC_USE_COUNT;
      local_irq_restore(flags);
      return 0;
   }

#if 0 /* debug RMi */
   /* find pointer to IP address of eth0 */
   if ((in_dev=in_dev_get(net_devs[SVIP_NET_DEV_ETH0_IDX])) != NULL)
   {
      for (ifa = in_dev->ifa_list; ifa != NULL; ifa = ifa->ifa_next)
      {
         if (!paddr_eth0 && ifa->ifa_address != 0)
         {
            paddr_eth0 = &ifa->ifa_address;
            continue;
         }
         if (paddr_eth0 && ifa->ifa_address != 0)
         {
            paddr_eth0_0 = &ifa->ifa_address;
            break;
         }
      }
      in_dev_put(in_dev);
   }
   if (paddr_eth0 == NULL || paddr_eth0_0 == NULL)
   {
      local_irq_restore(flags);
      return -ENODATA;
   }

   /* find pointer to IP address of veth0 */
   if ((in_dev=in_dev_get(net_devs[SVIP_NET_DEV_VETH0_IDX])) != NULL)
   {
      for (ifa = in_dev->ifa_list; ifa != NULL; ifa = ifa->ifa_next)
      {
         if (ifa->ifa_address != 0)
         {
            paddr_veth0 = &ifa->ifa_address;
            pmask_veth0 = &ifa->ifa_mask;
            break;
         }
      }
      in_dev_put(in_dev);
   }
   if (paddr_veth0 == NULL)
   {
      local_irq_restore(flags);
      return -ENODATA;
   }
#endif

   MOD_INC_USE_COUNT;
   nDeviceOpen++;
   local_irq_restore(flags);

   return 0;
}


/******************************************************************************/
/**
   This function is called when a process closes the SVIP NAT PT device file

\arguments
   inode       - pointer to disk file data
   file        - pointer to device file data

\return
   0 on success, else -1

*/
/******************************************************************************/
static int svip_nat_pt_device_release(struct inode *inode,
      struct file *file)
{
   unsigned long flags;

   local_save_flags(flags);

   MOD_DEC_USE_COUNT;

   local_irq_restore(flags);

   return 0;
}


/******************************************************************************/
/**
   This function is called when a process closes the SVIP NAT PT device file

\arguments
   inode       - pointer to disk file data
   file        - pointer to device file data
   ioctl_num   - ioctl number requested
   ioctl_param - pointer to data related to the ioctl number

\return
   0 on success, else -1

*/
/******************************************************************************/
int svip_nat_pt_device_ioctl (struct inode *inode, struct file *file,
      unsigned int ioctl_num, unsigned long ioctl_param)
{
   int err = 0;
   SVIP_NAT_PT_table_entry_t *pNatPtEntry;
   SVIP_NAT_PT_IO_Rule_t *pNatPtRule, *pNatPtRuleIn;
   SVIP_NAT_PKT_STATS_t *pNatPtStats;
   SVIP_NAT_PORT_MAP_Rule_t *pNatPortPtRuleIn;
   SVIP_NAT_PT_HOOK_t *pNatPtHookIn;
   SVIP_UDP_PORT_t nPort;
   int nNatIdx, nNatMapIdx;
   int bWrite = 0;
   int bRead = 0;
   unsigned char *pData = 0;
   int nSize;

   if (_IOC_DIR(ioctl_num) & _IOC_WRITE)
      bWrite = 1;
   if (_IOC_DIR(ioctl_num) & _IOC_READ)
      bRead = 1;
   nSize = _IOC_SIZE(ioctl_num);

   if (nSize > sizeof(int))
   {
      if (bRead || bWrite)
      {
         pData = kmalloc (nSize, GFP_KERNEL);
         if (bWrite)
         {
            if (copy_from_user ((void *)pData, (void *)ioctl_param, nSize) != 0)
            {
               printk(KERN_ERR "SVIP NAT PT: ioctl %x: copy_from_user() failed!\n", ioctl_num);
               err = -1;
               goto error;
            }
         }
      }
   }

   switch (ioctl_num)
   {
      case FIO_SVIP_NAT_PT_RULE_ADD:

         pNatPtRuleIn = (SVIP_NAT_PT_IO_Rule_t *)pData;

         /* check if the local UDP port is within range */
         nPort = ntohs(pNatPtRuleIn->locUDP);

         nNatIdx = ntohs(pNatPtRuleIn->nNatRuleIndex);
         /* check if the NatIndex is within range */
         if (!SVIP_NAT_INDEX_INRANGE(nNatIdx))
         {
            printk(KERN_ERR "%s: local UDP port %d, NAT index %d is out of range (%d..%d)\n", MODNAME,
                    nPort, nNatIdx, SVIP_UDP_FROM, SVIP_UDP_TO);
            err = -1;
            goto error;
         }

         pNatPtEntry = &pNatPtTable[nNatIdx];

         down(sem_nat_tbl_access);

         if ( ( pNatPtEntry->sockdata_loc.sock != NULL ) ||
              ( nat_port_table_ptr[nPort] != NULL ) )
         {
            printk(KERN_ERR "%s: NAT rule for local UDP port %d is already in use, NAT index %d is valid.\n", MODNAME,
                    nPort, nNatIdx);
            err = -1;
            up(sem_nat_tbl_access);
            goto error;
         }

//         printk("SVIP NAT: Adding nat rule for port %d on NAT index %d\n", nPort, nNatIdx );

         /* save the NAT rule index in the indexed NAT table */
         pNatPtEntry->nat_index = nNatIdx;
         pNatPtRule = &pNatPtEntry->natRule;

         if ((err = svip_nat_pt_create_sock(pNatPtEntry, pNatPtRuleIn, 0)) < 0)
         {
            memset((void *)pNatPtEntry, 0, sizeof(SVIP_NAT_PT_table_entry_t));
            up(sem_nat_tbl_access);
            goto error;
         }
         svip_nat_pt_prepare_sockaddr(pNatPtEntry, pNatPtRuleIn, 0);

         if ((err = svip_nat_pt_create_sock(pNatPtEntry, pNatPtRuleIn, 1)) < 0)
         {
            memset((void *)pNatPtEntry, 0, sizeof(SVIP_NAT_PT_table_entry_t));
            up(sem_nat_tbl_access);
            goto error;
         }
         svip_nat_pt_prepare_sockaddr(pNatPtEntry, pNatPtRuleIn, 1);

         /* add rule to the NAT table */

         /* copy remote peer address details */
         memcpy((void *)pNatPtRule->remMAC, (void *)pNatPtRuleIn->remMAC, ETH_ALEN);
         pNatPtRule->remTypeIP = pNatPtRuleIn->remTypeIP;
         if (pNatPtRule->remTypeIP == SVIP_IPV4_ADDR_TYPE)
         {
            pNatPtRule->remIP.v4 = pNatPtRuleIn->remIP.v4;
         }
         else
         {
            pNatPtRule->remIP.v6 = pNatPtRuleIn->remIP.v6;
         }
         pNatPtRule->remUDP = pNatPtRuleIn->remUDP;

         /* copy local peer address details */
         memcpy((void *)pNatPtRule->locMAC, (void *)pNatPtRuleIn->locMAC, ETH_ALEN);
         pNatPtRule->locTypeIP = pNatPtRuleIn->locTypeIP;
         if (pNatPtRule->locTypeIP == SVIP_IPV4_ADDR_TYPE)
         {
            pNatPtRule->locIP.v4 = pNatPtRuleIn->locIP.v4;
         }
         else
         {
            pNatPtRule->locIP.v6 = pNatPtRuleIn->locIP.v6;
         }
         pNatPtRule->locUDP = pNatPtRuleIn->locUDP;

         memset(pNatPtTable[nNatIdx].natStats, 0,
            sizeof(SVIP_NAT_PT_stats_t)*SVIP_NAT_STATS_TYPES);

         /* Added pointer to the NAT control structure in the udp port lookup table */
         nat_port_table_ptr[nPort] = pNatPtEntry;
         svip_port_nat_rule_map[nPort] = nNatIdx;
         pNatPtEntry->nat_entry_valid = 1;

         up(sem_nat_tbl_access);
         break;

      case FIO_SVIP_NAT_PT_RULE_REMOVE:

         pNatPtRuleIn = (SVIP_NAT_PT_IO_Rule_t *)pData;

         nNatIdx = ntohs(pNatPtRuleIn->nNatRuleIndex);
         nPort = pNatPtRuleIn->locUDP;

         /* check if the NatIndex is within range */
         if (!SVIP_NAT_INDEX_INRANGE(nNatIdx))
         {
             printk(KERN_ERR "%s: local UDP port %d, NAT index %d is out of range (%d..%d)\n", MODNAME,
                  nPort, nNatIdx, SVIP_UDP_FROM, SVIP_UDP_TO);
            err = -1;
             goto error;
         }

         pNatPtEntry = &pNatPtTable[nNatIdx];
         pNatPtRule = &pNatPtEntry->natRule;
         nPort = pNatPtRule->locUDP;

         /* remove rule from the NAT table */
         down(sem_nat_tbl_access);

         if ( !pNatPtEntry->nat_entry_valid ) //!nat_port_table_ptr[pNatPtRule->locUDP] )
         {
            printk(KERN_ERR "%s: NAT rule for local UDP port %d is already freed, NAT index %d is valid\n", MODNAME,
                    nPort, nNatIdx);
            err = -1;
            up(sem_nat_tbl_access);
            goto error;
         }

//         printk("SVIP NAT: Remove nat rule for port %d on NAT index %d\n", nPort, nNatIdx );

         svip_nat_pt_release_sock(pNatPtEntry, 0);
         svip_nat_pt_release_sock(pNatPtEntry, 1);

         nat_port_table_ptr[nPort] = NULL;
         svip_port_nat_rule_map[nPort] = -1;

         memset((void *)pNatPtEntry, 0, sizeof(SVIP_NAT_PT_table_entry_t));

         up(sem_nat_tbl_access);
         break;

      case FIO_SVIP_NAT_PT_RULE_LIST:
      {
         int len;
         char buf[256];

         down(sem_nat_tbl_access);
         while (nProcReadIdx != -1)
         {
            len = svip_nat_pt_proc_read_nat(buf, 256);
            if (len > 0)
               printk("%s", buf);
         }
         nProcReadIdx = 0;
         up(sem_nat_tbl_access);
         break;
      }
      case FIO_SVIP_NAT_PT_RULE_PREPARE:
         {
            pNatPortPtRuleIn = (SVIP_NAT_PORT_MAP_Rule_t *)pData;
            nPort = ntohs(pNatPortPtRuleIn->locUDP);
            nNatIdx = ntohs(pNatPortPtRuleIn->nNatRuleIndex);
            /* check if the NatIndex is within range */
            if (!SVIP_NAT_INDEX_INRANGE(nNatIdx))
            {
                printk(KERN_ERR "%s: try to map nat rule, local UDP port %d, NAT index %d is out of range (%d..%d)\n", MODNAME,
                     nPort, nNatIdx, SVIP_UDP_FROM, SVIP_UDP_TO);
               err = -1;
                goto error;
            }
            svip_port_nat_rule_map[nPort] = nNatIdx;
         }
         break;
      case FIO_SVIP_NAT_PT_RULE_GET:
         {
            pNatPtRuleIn = (SVIP_NAT_PT_IO_Rule_t *)pData;
            nPort = ntohs(pNatPtRuleIn->locUDP);
            nNatMapIdx = svip_port_nat_rule_map[nPort];
            nNatIdx = ntohs(pNatPtRuleIn->nNatRuleIndex);
            /* check if the NatIndex is within range */
            if (!SVIP_NAT_INDEX_INRANGE(nNatMapIdx) || nNatMapIdx != nNatIdx)
            {
                printk(KERN_ERR "%s: try to get nat rule, local UDP port %d, map NAT index %d Nat index %d is out of range (%d..%d)\n", MODNAME,
                     nPort, nNatMapIdx, nNatIdx, SVIP_UDP_FROM, SVIP_UDP_TO);
               err = -1;
                goto error;
            }

            down(sem_nat_tbl_access);
            /* fetch rule from the NAT table */
            memcpy(pData, &pNatPtTable[nNatIdx].natRule, sizeof(SVIP_NAT_PT_IO_Rule_t));
            up(sem_nat_tbl_access);
         }
         break;
      case FIO_SVIP_NAT_PT_STATS_GET:
         {
            pNatPtStats = (SVIP_NAT_PKT_STATS_t *)pData;
            nPort = ntohs(pNatPtStats->locUDP);
            nNatMapIdx = svip_port_nat_rule_map[nPort];
            /* check if the NatIndex is within range */
            if (!SVIP_NAT_INDEX_INRANGE(nNatMapIdx))
            {
                printk(KERN_ERR "%s: try to get nat stats, local UDP port %d, map NAT index %d is out of range (%d..%d)\n", MODNAME,
                     nPort, nNatMapIdx, SVIP_UDP_FROM, SVIP_UDP_TO);
               err = -1;
                goto error;
            }

            /* fetch stats from the NAT table */
            memcpy(pData, &pNatPtTable[nNatMapIdx].stats, sizeof(SVIP_NAT_PKT_STATS_t));
         }
         break;
      case FIO_SVIP_NAT_PT_HOOK_SET:
         {
            pNatPtHookIn = (SVIP_NAT_PT_HOOK_t *)pData;
            nNatIdx = ntohs(pNatPtHookIn->nNatRuleIndex);
            /* check if the NatIndex is within range */
            if (!SVIP_NAT_INDEX_INRANGE(nNatIdx))
            {
                printk(KERN_ERR "%s: try to set the hook rule, Nat index %d is out of range (%d..%d)\n", MODNAME,
                     nNatIdx, SVIP_NAT_INDEX_FROM, SVIP_NAT_INDEX_TO);
               err = -1;
                goto error;
            }

            down(sem_nat_tbl_access);
            pNatPtTable[nNatIdx].hookT38 = pNatPtHookIn->hookT38;
            pNatPtTable[nNatIdx].sendNoSignal = pNatPtHookIn->sendNoSignal;
            pNatPtTable[nNatIdx].udpErrCorrection = pNatPtHookIn->udpErrCorrection;
            switch ( pNatPtTable[nNatIdx].udpErrCorrection )
            {
            case IFX_SVIP_T38_FEC:
               {
                  if ( pNatPtTable[nNatIdx].sendNoSignal > 1 )
                  {
                     pNatPtTable[nNatIdx].sendNoSignal = 1;
                  }
                  /* no need to update error recovery packets under FEC */
                  pNatPtTable[nNatIdx].updateT38ErrRecovery = 0;
               }
               break;
            case IFX_SVIP_T38_RED:
            default:
               {
                  pNatPtTable[nNatIdx].updateT38ErrRecovery = 1;
               }
               break;
            }
            pNatPtHookIn->noSignalFarEnd = svip_no_signal_far_end[pNatPtTable[nNatIdx].natRule.locUDP];
            pNatPtHookIn->sendNoSignal = pNatPtTable[nNatIdx].sendNoSignal;
            if ( pNatPtTable[nNatIdx].sendNoSignal )
            {
               /* If far end no signal is detected, then we need to send out no signal at upper layer. */
               if (svip_no_signal_far_end[pNatPtTable[nNatIdx].natRule.locUDP])
               {
                  pNatPtTable[nNatIdx].sendNoSignal = 0;
               }
               pNatPtTable[nNatIdx].updateT38SeqNum = 1;
            }
            else
            {
               pNatPtTable[nNatIdx].updateT38SeqNum = 0;
               pNatPtTable[nNatIdx].updateT38ErrRecovery = 0;
               svip_no_signal_far_end[pNatPtTable[nNatIdx].natRule.locUDP] = 0;
            }
            up(sem_nat_tbl_access);
         }
         break;

      default:
         printk(KERN_ERR "SVIP NAT PT: unsupported ioctl (%x) command for device %s\n",
               ioctl_num, PATH_SVIP_NAT_PT_DEVICE_NAME);
         err = -1;
         goto error;
   }

   if (nSize > sizeof(int))
   {
      if (bRead)
      {
         if (copy_to_user ((void *)ioctl_param, (void *)pData, nSize) != 0)
         {
            printk(KERN_ERR "SVIP NAT PT: ioctl %x: copy_to_user() failed!\n", ioctl_num);
            err = -1;
            goto error;
         }
      }
   }

error:
   if (pData)
      kfree(pData);

   return err;
}

#if 0
void dump_msg(unsigned char *pData, unsigned int nLen)
{
   int i;

   for (i=0; i<nLen; i++)
   {
      if (!i || !(i%16))
         printk("\n    ");
      else if (i && !(i%4))
         printk(" ");
      printk("%02x", pData[i]);
   }
   if (--i%16)
      printk("\n");
}
#endif

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
/* Updates the socket statistics 
   MUST have sem_nat_tbl_access locked
*/
static void svip_nat_pt_stats_update(void)
{
    /* Initial timestamp */
    if ( perfStats.stats_total_packets == 0 )
        perfStats.stats_timestamp = jiffies;

    /* update the total number of packets processed since bootup */
    perfStats.stats_total_packets++;

    /* Update stats every PACKET_COUNT_RATE_UPDATE packets */
    if ( ( perfStats.stats_total_packets % PACKET_COUNT_RATE_UPDATE) == 0 )
    { 
        unsigned long temp_timestamp = (jiffies - perfStats.stats_timestamp) * 1000 / HZ;

        /* Reset timestamp for updating the current rate */
        perfStats.stats_timestamp = jiffies;

        /* update current packet rate */
        perfStats.packet_rate = ( PACKET_COUNT_RATE_UPDATE * 1000) / ( temp_timestamp + 1 );

        /* Record max packet rate*/
        if ( perfStats.max_rate < perfStats.packet_rate )
            perfStats.max_rate = perfStats.packet_rate;
    }
}
#endif

/**
	Creates UDP socket of the type depending the IP type of the respective peer.

	\param psock				pointer to socket pointer
   \param pNatPtRule			NAT PT rule data
	\param bRem					0 for local socket, else for remote socket

   \return
		on success, *psock contains a pointer to the allocated and successfully bound socket
						and 0 is returned
		on error, *psock is NULL and descriptive error code is returned
*/
static int svip_nat_pt_create_sock(SVIP_NAT_PT_table_entry_t *pNatPtEntry, SVIP_NAT_PT_IO_Rule_t *pNatPtRule, int bRem)
{
	int err;
	struct socket *sock;
	SVIP_IP_ADDR_TYPE_t *pTypeIP;
	SVIP_UDP_PORT_t *pUDP;
	unsigned int nMagic;
	int sk_data = 0;
#if defined ( SUPPORT_LOCAL_INTF_BIND )
	SVIP_IP_ADDR_UNION_t *pIP;
#endif

	if (bRem == 0)
	{
		/* IP address type of the local stream */
		pTypeIP = &pNatPtRule->locTypeIP;
		/* local stream is addressing the remote UDP port */
		pUDP = &pNatPtRule->remUDP;
#if defined ( SUPPORT_LOCAL_INTF_BIND )
		/* Wen setting up the socket for the remote IP copy we need to bind to 
		our local interface. Use the locIP to resolve the interface on which 
		we need to bind to*/
		pIP = &pNatPtRule->locIP;
#endif
		nMagic = UDP_REDIRECT_LOC_MAGIC;
	}
	else
	{
		/* IP address type of the remote stream */
		pTypeIP = &pNatPtRule->remTypeIP;
		/* remote stream is addressing the local UDP port */
		pUDP = &pNatPtRule->locUDP;
#if defined ( SUPPORT_LOCAL_INTF_BIND )
		/* Wen setting up the socket for the local IP copy we need to bind to 
		our local interface. Use the remIP to resolve the interface on which 
		we need to bind to*/
		pIP = &pNatPtRule->remIP;
#endif
		nMagic = UDP_REDIRECT_REM_MAGIC;
	}

	/*
		TODO: parameter ipv4loc and ipv6loc of svip_nat_pt_create_ipv4_sock() and
		svip_nat_pt_create_ipv6_sock(), should respectevely be set to the ipv4 and
		ipv6 address of the local Ethernet interface to which the socket is bound
		to. Setting NULL for these parameters, will bind the socket to all local
		Ethernet interfaces. With such setting, the bind process will fail when
		the locUDP and remUDP ports are identical.

		Best solution is to look into the routing table and find out over which
		interface the respective locIP and remIP hosts are reachable and thus use
		the address of the local interface to bind the socket to.
	*/
	if (*pTypeIP == SVIP_IPV4_ADDR_TYPE)
	{
		SVIP_IPV4_ADDR_t *ipv4loc = NULL;
#if defined ( SUPPORT_LOCAL_INTF_BIND )
		/* bug 170794
		 * Partial fix: #1
		 *    1, if remote UDP port and local UDP port are different, listen any address;
		 *    2, otherwise, listen specified address
		 */
		if ( pNatPtRule->remUDP == pNatPtRule->locUDP/* && !pNatPtRule->nMasterChip*/ ) {
			ipv4loc = &pIP->v4;
		}
#endif
		if ((err = svip_nat_pt_create_ipv4_sock(&sock, ipv4loc, *pUDP)) < 0)
			return err;
	}
	else
	{
		SVIP_IPV6_ADDR_t *ipv6loc = NULL;
#if defined ( SUPPORT_LOCAL_INTF_BIND )
		ipv6loc = &pIP->v6;
#endif
		if ((err = svip_nat_pt_create_ipv6_sock(&sock, ipv6loc, *pUDP)) < 0)
			return err;
	}

#if defined ( SUPPORT_SVIP_NAT_PT_DSCP )
	if ((err = svip_nat_pt_tos_config(sock, pNatPtRule->nDSCP)) < 0)
		return err;
#else
	/* increase the processing priority of the socket to maximum */
	sock->sk->sk_priority = 6;
#endif

	/* Place the NAT index in the MSByte of the sock user data ptr */
	sk_data = ( pNatPtEntry->nat_index << UDP_REDIRECT_INDEX_SHIFT );

	if (bRem == 0)
	{
		/* write the magic identifier in order the UDP packet hook identifies
		   our sockets, and at the same time to distinguish the local from
		   remote peer data, and vice-versa */
		sk_data |= UDP_REDIRECT_LOC_MAGIC;
		sock->sk->sk_user_data = (void *)sk_data;
		pNatPtEntry->sockdata_loc.sock = sock;
	}
	else
	{
		sk_data |= UDP_REDIRECT_REM_MAGIC;
		sock->sk->sk_user_data = (void *)sk_data;
		pNatPtEntry->sockdata_rem.sock = sock;
	}

	return 0;
}

static void svip_nat_pt_release_sock(SVIP_NAT_PT_table_entry_t *pNatPtEntry, int bRem)
{
   struct socket *sock = NULL;

   if (bRem == 0)
   {
      sock = pNatPtEntry->sockdata_loc.sock;
      pNatPtEntry->sockdata_loc.sock_state = SOCK_STATE_CLOSING;
   }
   else
   {
      sock = pNatPtEntry->sockdata_rem.sock;
      pNatPtEntry->sockdata_rem.sock_state = SOCK_STATE_CLOSING;
   }

   if (sock)
      sock_release(sock);
}


/**
   Creates UDP socket of IPv4 family and binds it to the specified local interface
	and UDP port number.

	\param psock			pointer to pointer of the socket structure
   \param ipv4loc       IPv4 address in network byte order of the local interface
								on which the socket will receive.
   \param nUdp				UDP port number to bind the socket to

   \return
	on success, *psock contains a pointer to the allocated and successfully bound socket
					and 0 is returned
	on error, *psock is NULL and descriptive error code is returned
*/
static int svip_nat_pt_create_ipv4_sock(struct socket **psock, SVIP_IPV4_ADDR_t *ipv4loc, SVIP_UDP_PORT_t nUdp)
{
    int err = 0;
    struct sockaddr_in locaddr;
    struct socket *sock;

    *psock = NULL;
    /* create IPv4 UDP socket */
    if ((err = sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &sock)) < 0)
    {
        printk("%s: error creating IPv4 UDP socket (%d)!\n", MODNAME, err);
        return err;
    }
    sock->sk->sk_reuse = 1;

    memset((void *)&locaddr, 0, sizeof(struct sockaddr_in));

#if defined ( SUPPORT_LOCAL_INTF_BIND )
    if ( ipv4loc ) {
        /* Resolve the local interface which needs to be used to send packets to the ipv4loc */
        err = svip_nat_pt_lookup_sock_intf(sock, &locaddr, (SVIP_IP_ADDR_UNION_t*)ipv4loc, SVIP_IPV4_ADDR_TYPE);
    } else {
        locaddr.sin_addr.s_addr = INADDR_ANY;
    }
    if ( err < 0 ) 
    {
        printk("%s: error finding routable interface (%d)!\n", MODNAME, err);
        return err;
    }
#else
    if (ipv4loc)
        locaddr.sin_addr.s_addr = *((__be32 *)ipv4loc);
    else
        locaddr.sin_addr.s_addr = INADDR_ANY;
#endif
    locaddr.sin_port = htons(nUdp);

//    printk("%s: Binding IPv4 UDP socket to address %08x/%d !\n", MODNAME, ntohl(locaddr.sin_addr.s_addr), nUdp);

	/* bind the socket */
	if((err = sock->ops->bind(sock, (struct sockaddr *)&locaddr,
			sizeof(struct sockaddr))) < 0)
	{
		printk("%s: error binding IPv4 UDP socket to address %08x/%d (%d)!\n", MODNAME,
		ntohl(locaddr.sin_addr.s_addr), nUdp, err);
		sock_release(sock);
		return err;
	}

	if(sock->sk == NULL)
	{
		printk("%s: error creating IPv4 UDP socket, sock->sk is NULL!\n", MODNAME);
		sock_release(sock);
		return -1;
	}

	*psock = sock;
	return 0;
}

/**
   Creates UDP socket of IPv6 family and binds it to the specified local interface
	and UDP port number.

	\param psock			pointer to pointer of the socket structure
   \param ipv6loc       IPv6 address of the local interface on which the socket
								will receive
   \param nUdp				UDP port number to bind the socket to

   \return
	on success, *psock contains a pointer to the allocated and successfully bound socket
					and 0 is returned
	on error, *psock is NULL and descriptive error code is returned
*/
int svip_nat_pt_create_ipv6_sock(struct socket **psock, SVIP_IPV6_ADDR_t *ipv6loc, SVIP_UDP_PORT_t nUdp)
{
#ifdef CONFIG_IPV6
	int err;
	struct sockaddr_in6 locaddr6;
	struct socket *sock;

	*psock = NULL;
	/* create IPv6 UDP socket */
	if ((err = sock_create(PF_INET6, SOCK_DGRAM, IPPROTO_UDP, &sock)) < 0)
	{
		printk("%s: error creating IPv6 UDP socket (%d)!\n", MODNAME, err);
		return err;
	}
	sock->sk->sk_reuse = 1;

	memset((void *)&locaddr6, 0, sizeof(struct sockaddr_in6));

	locaddr6.sin6_family = AF_INET6;
	if (ipv6loc)
		locaddr6.sin6_addr = *((struct in6_addr *)ipv6loc);
	else
		locaddr6.sin6_addr = in6addr_any;
	locaddr6.sin6_port = htons(nUdp);

	/* bind the socket */
	if((err = sock->ops->bind(sock, (struct sockaddr *)&locaddr6,
			sizeof(struct sockaddr_in6))) < 0)
	{
		printk("%s: error binding IPv6 UDP socket to address %04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x/%d (%d)!\n", MODNAME,
				locaddr6.sin6_addr.s6_addr16[7], locaddr6.sin6_addr.s6_addr16[6],
				locaddr6.sin6_addr.s6_addr16[5], locaddr6.sin6_addr.s6_addr16[4],
				locaddr6.sin6_addr.s6_addr16[3], locaddr6.sin6_addr.s6_addr16[2],
				locaddr6.sin6_addr.s6_addr16[1], locaddr6.sin6_addr.s6_addr16[0],
				nUdp, err);
		sock_release(sock);
		return err;
	}

	if(sock->sk == NULL)
	{
		printk("%s: error creating IPv6 UDP socket, sock->sk is NULL!\n", MODNAME);
		sock_release(sock);
		return -1;
	}


    psockdata->sock_state = SOCK_STATE_STARTING;
	*psock = sock;
	return 0;
#else
        printk("No IPv6 support at the moment\n");
        *psock = NULL;
        return -1;
#endif
}

#if defined ( SUPPORT_SVIP_NAT_PT_DSCP )
int svip_nat_pt_tos_config(struct socket *psock, char RTPipTOS)
{
	int err = 0;
	char rtp_tos_val = RTPipTOS;
    unsigned int rtp_eth_skb_prio = 6;
//	unsigned int rtp_qos_val = 5;
//        struct vlan_ioctl_args ifr;
//        char rtp_vlan_intf[10];

	if((err = kernel_setsockopt( psock, IPPROTO_IP, IP_TOS, &rtp_tos_val, sizeof( rtp_tos_val ) ) < 0) )
	{
		printk("%s: error setting socket DSCP val of %d (%d)!\n", MODNAME, rtp_tos_val, err);
		sock_release(psock);
		return err;
	}


#if 0
        /* Get the currently used VLAN interface for configuring the QoS */
        strcpy( rtp_vlan_intf, "eth0.24" );

        memset(&ifr, 0, sizeof(ifr));
        ifr.cmd = SET_VLAN_EGRESS_PRIORITY_CMD;
        sprintf( ifr.device1, rtp_vlan_intf);
        ifr.u.skb_priority = rtp_eth_skb_prio;
        ifr.vlan_qos = rtp_qos_val;

        /* Set the egress skb priority to be used for the RTP trafic */
        if ( (err = kernel_setsockopt( psock, SOL_SOCKET, SO_PRIORITY, &rtp_eth_skb_prio, sizeof( rtp_eth_skb_prio ) ) ) < 0 ) {
            printk("kernel_setsockopt: Failed to set RTP skb priority to %d (error %d)\n", rtp_eth_skb_prio, err);
            sock_release(psock);
            return err;
        }
        else {
            /* Set the QoS for SIP on the egress sbk buffer used SIP_VLAN_EGRESS_QUEUE on the curent VLAN interface */
            if ( (err = kernel_sock_ioctl( psock, SIOCSIFVLAN, (unsigned long )&ifr) ) < 0 ) {
                printk("QoS: Failed to set Layer 2 802.1p for RTP to %d (error %d)\n", rtp_qos_val, err);
                sock_release(psock);
                return err;
            }
            else {
                printk("QoS: Layer 2 802.1p for RTP set to %d @skb_prio %d\n", rtp_qos_val, rtp_eth_skb_prio);
            }
        }
#else
	psock->sk->sk_priority = rtp_eth_skb_prio;

#endif
	return 0;
}
#endif

#if defined ( SUPPORT_LOCAL_INTF_BIND )
static int svip_nat_pt_lookup_sock_intf(struct socket *sock, struct sockaddr_in *locaddr, SVIP_IP_ADDR_UNION_t *pIP, SVIP_IP_ADDR_TYPE_t pTypeIP)
{
    int no_route = 1;
    struct rtable   *rt = NULL;
    SVIP_IPV4_ADDR_t ipv4loc;

    if ( pTypeIP == SVIP_IPV6_ADDR_TYPE)
    {
        printk("SVIP_NAT_PT no support for ipv6 loc int lookup\n");
        return -1;
    }

    /* Setup a fail safe address */
    locaddr->sin_addr.s_addr = INADDR_ANY;
    locaddr->sin_family = AF_INET;

    if (!pIP)
    {
        printk("SVIP_NAT_PT no pIP support bind to INADDR_ANY\n");
        return -2;
    }

    ipv4loc = pIP->v4;

    /* Make sure we can route this packet. */
    rt = (struct rtable *)__sk_dst_check(sock->sk, 0);
    if (rt == NULL) 
    {
        struct sock *sk = sock->sk;
        struct inet_sock *inet = inet_sk(sk);
        struct ip_options *opt = inet->opt;

        __be32 daddr = ipv4loc;

        if(opt && opt->srr)
            daddr = opt->faddr;

        {
            struct flowi fl = { .oif = sk->sk_bound_dev_if,
                                .nl_u = { .ip4_u =
                                        { .daddr = daddr,
                                          .saddr = inet->saddr,
                                          .tos = RT_CONN_FLAGS(sk) } },
                                          .proto = sk->sk_protocol,
                                          .flags = inet_sk_flowi_flags(sk),
                                          .uli_u = { .ports =
                                                   { .sport = inet->sport,
                                                     .dport = inet->dport } } };

            /* If this fails, retransmit mechanism of transport layer will
             * keep trying until route appears or the connection times
             * itself out.
             */
            security_sk_classify_flow(sk, &fl);
            if (ip_route_output_flow(sock_net(sk), &rt, &fl, sk, 0))
            {
                no_route = 1;
                printk("ERROR NO ROUTE rt\n");
                goto done;
            }
        }
        sk_setup_caps(sk, &rt->u.dst);
        no_route = 0;        
    }
    else
    {
        no_route = 1;
        printk("SVIP NAT PT: NO ROUTE\n");
    }

done:
    if ( rt && !no_route )
    {
        locaddr->sin_addr.s_addr = ((__be32 )rt->rt_src);
//        printk("SRC IP = 0x%08X --> DST IP = 0x%08X\n", rt->rt_src, rt->rt_dst);
    }
    else
    {
        locaddr->sin_addr.s_addr = ipv4loc;
        printk("SRC IP = 0x%08X --> DST IP = ?\n", locaddr->sin_addr.s_addr);
    }
return 0;
}
#endif

/**
   Prepares the socket address and socket message data structures, i.e.
   struct sockaddr_in for IPv4 socket or, struct sockaddr_in6 for IPv6 for the
   direction specified by bRem. By preparing these details once during rule
   creation, any unnecessary repetition of the same step is omitted which
   minimizes the processing overhead during packet payload transfer
   from the incoming socket to the outgoing one.

   \param pNatPtEntry   pointer to the related NAT PT table entry
   \param pNatPtRule    pointer to the related NAT PT rule data
   \param bRem          0 for local socket, else for remote

   \return
      on success 0 is returned
	   on error, descriptive error code is returned
*/
static void svip_nat_pt_prepare_sockaddr(SVIP_NAT_PT_table_entry_t *pNatPtEntry, SVIP_NAT_PT_IO_Rule_t *pNatPtRule, int bRem)
{
    SVIP_IP_ADDR_TYPE_t *pTypeIP;
    SVIP_IP_ADDR_UNION_t *pIP;
    SVIP_UDP_PORT_t *pUDP;
    SVIP_NAT_PT_sockdata_t *psockdata;

    if (bRem == 0)
    {
        /* IP address type of the local stream */
        pTypeIP = &pNatPtRule->locTypeIP;
        /* IP address of the local stream */
        pIP = &pNatPtRule->locIP;
        /* local stream receiving UDP port */
        pUDP = &pNatPtRule->locUDP;
        psockdata = &pNatPtEntry->sockdata_loc;
    }
    else
    {
        /* IP address type of the remote stream */
        pTypeIP = &pNatPtRule->remTypeIP;
        /* IP address of the remote stream */
        pIP = &pNatPtRule->remIP;
        /* remote stream receiving UDP port */
        pUDP = &pNatPtRule->remUDP;
        psockdata = &pNatPtEntry->sockdata_rem;
    }

    psockdata->sock_state = SOCK_STATE_RUNNING;

    if (*pTypeIP == SVIP_IPV4_ADDR_TYPE)
    {
        struct sockaddr_in *psockaddr = &psockdata->sockaddr.addr;

        psockaddr->sin_family = AF_INET;
        psockaddr->sin_port = *pUDP;
        psockaddr->sin_addr = *((struct in_addr *)pIP);
        psockdata->sockaddr_len = sizeof(struct sockaddr_in);
        psockdata->msg.msg_name = (void *)psockaddr;
        psockdata->msg.msg_namelen = sizeof(struct sockaddr_in);
    }
    else
    {
        struct sockaddr_in6 *psockaddr6 = &psockdata->sockaddr.addr6;

        psockaddr6->sin6_family = AF_INET6;
        psockaddr6->sin6_port = *pUDP;
        psockaddr6->sin6_addr = *((struct in6_addr *)pIP);
        psockdata->sockaddr_len = sizeof(struct sockaddr_in6);
        psockdata->msg.msg_name = (void *)psockaddr6;
        psockdata->msg.msg_namelen = sizeof(struct sockaddr_in6);
    }

    psockdata->sock->sk->sk_allocation = GFP_ATOMIC;
    psockdata->msg.msg_flags = MSG_DONTWAIT | MSG_NOSIGNAL;
}

/**
   Send the UDP datagram data of length len and pointed to by data.

	\param psockdata     contains the socket and socket address details
   \param data          pointer to the UDP datagam data
   \param len           lenght of the datagramd data in bytes

   \return
      on success 0 is returned
	   on error, descriptive error code is returned
*/
static int svip_nat_pt_sendmsg (SVIP_NAT_PT_sockdata_t *psockdata, void *data, unsigned int len)
{
   int err;

   if ( !psockdata || !psockdata->sock)
   {
      printk("ERROR psockdata->sock = 0x%08X\n", (unsigned int)psockdata);
      err = -2;
      return err;
   }

   psockdata->iov.iov_base = (void *)data;
   psockdata->iov.iov_len = len;

   err = kernel_sendmsg(psockdata->sock, &psockdata->msg, &psockdata->iov, 1, len);

   return err;
}

static int svip_nat_pt_tx_action(struct sk_buff *skb)
{
   struct iphdr *iph;
   struct udphdr *udph;
   SVIP_NAT_PT_IO_Rule_t *pNatRule;
   SVIP_UDP_PORT_t locUDP, remoteUDP;
   int nNatIdx;
   struct ethhdr *ethh;

   /* consider only UDP packets. */
   iph = SVIP_NAT_IP_HDR(skb);
   if(iph->protocol != IPPROTO_UDP)
   {
      return 0;
   }

   udph = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
   locUDP = ntohs(udph->source);
   remoteUDP = ntohs(udph->dest);
   nNatIdx = svip_port_nat_rule_map[remoteUDP];
   if ( (locUDP == SVIP_PROBE_SOURCE_PORT)
      && (SVIP_NAT_INDEX_INRANGE(nNatIdx))
      )
   {
      ethh = (struct ethhdr *)SVIP_NAT_SKB_MAC_HEADER(skb);
      if ( !ethh )
         ethh = (struct ethhdr*)(skb->head + 2);
      /* process the packet if a respective NAT rule exists */
      pNatRule = &pNatPtTable[nNatIdx].natRule;
      if(ethh)
      {
         memcpy(pNatRule->remMAC, ethh->h_dest, ETH_ALEN);
      }
      return 0;
   }
   return 0;
}

/************************************************************************************************/
/************************ This part was ported from spandsp 0.0.6 ************************/
/************************************************************************************************/
static int decode_length(const uint8_t *buf, int limit, int *len, int *pvalue)
{
    if (*len >= limit)
        return -1;
    if ((buf[*len] & 0x80) == 0)
    {
        *pvalue = buf[(*len)++];
        return 0;
    }
    if ((buf[*len] & 0x40) == 0)
    {
        if (*len >= limit - 1)
            return -1;
        *pvalue = (buf[(*len)++] & 0x3F) << 8;
        *pvalue |= buf[(*len)++];
        return 0;
    }
    *pvalue = (buf[(*len)++] & 0x3F) << 14;
    /* Indicate we have a fragment */
    return 1;
}
/*- End of function --------------------------------------------------------*/

static int decode_open_type(const uint8_t *buf, int limit, int *len, const uint8_t **p_object, int *p_num_octets)
{
    int octet_cnt;
    int octet_idx;
    int stat;
    const uint8_t **pbuf;

    for (octet_idx = 0, *p_num_octets = 0;  ;  octet_idx += octet_cnt)
    {
        if ((stat = decode_length(buf, limit, len, &octet_cnt)) < 0)
            return -1;
        if (octet_cnt > 0)
        {
            *p_num_octets += octet_cnt;

            pbuf = &p_object[octet_idx];
            /* Make sure the buffer contains at least the number of bits requested */
            if ((*len + octet_cnt) > limit)
                return -1;

            *pbuf = &buf[*len];
            *len += octet_cnt;
        }
        if (stat == 0)
            break;
    }
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int encode_length(uint8_t *buf, int *len, int value)
{
    int multiplier;

    if (value < 0x80)
    {
        /* 1 octet */
        buf[(*len)++] = value;
        return value;
    }
    if (value < 0x4000)
    {
        /* 2 octets */
        /* Set the first bit of the first octet */
        buf[(*len)++] = ((0x8000 | value) >> 8) & 0xFF;
        buf[(*len)++] = value & 0xFF;
        return value;
    }
    /* Fragmentation */
    multiplier = (value < 0x10000)  ?  (value >> 14)  :  4;
    /* Set the first 2 bits of the octet */
    buf[(*len)++] = 0xC0 | multiplier;
    return multiplier << 14;
}
/*- End of function --------------------------------------------------------*/

static int __attribute__((unused)) encode_open_type(uint8_t *buf, int *len, const uint8_t *data, int num_octets)
{
    int enclen;
    int octet_idx;
    uint8_t zero_byte;

    /* If open type is of zero length, add a single zero byte (10.1) */
    if (num_octets == 0)
    {
        zero_byte = 0;
        data = &zero_byte;
        num_octets = 1;
    }
    /* Encode the open type */
    for (octet_idx = 0;  ;  num_octets -= enclen, octet_idx += enclen)
    {
        if ((enclen = encode_length(buf, len, num_octets)) < 0)
            return -1;
        if (enclen > 0)
        {
            memcpy(&buf[*len], &data[octet_idx], enclen);
            *len += enclen;
        }
        if (enclen >= num_octets)
            break;
    }

    return 0;
}
/*- End of function --------------------------------------------------------*/

static int t38_core_rx_ifp_stream(const uint8_t *buf, int len, uint8_t *ptype, int *pt30_indicator )
{
    int t30_indicator;
    int ptr;
    int pkt_len;
    int ret = 0;
    uint8_t type;
    uint8_t data_field_present;

    *ptype = -1;
    *pt30_indicator = -1;
    ptr = 0;
    pkt_len = len;
    if ((ptr + 1) > pkt_len)
        return ret;
    /*endif*/
    data_field_present = buf[ptr] & 0x80;
    type = (buf[ptr] >> 6) & 1;
    *ptype = type;
    switch (type)
    {
    case T38_TYPE_OF_MSG_T30_INDICATOR:
        /* Indicators should never have a data field */
        if (data_field_present)
        {
            return -1;
        }
        /*endif*/
        if ((buf[ptr] & 0x20))
        {
            /* Extension */
            if ((ptr + 2) > pkt_len)
                return ret;
            /*endif*/
            t30_indicator = T38_IND_V8_ANSAM + (((buf[ptr] << 2) & 0x3C) | ((buf[ptr + 1] >> 6) & 0x3));
            if (t30_indicator > T38_IND_V33_14400_TRAINING)
            {
                return -1;
            }
            /*endif*/
            ptr += 2;
        }
        else
        {
            t30_indicator = (buf[ptr] >> 1) & 0xF;
            ptr += 1;
        }
        *pt30_indicator = t30_indicator;
        /*endif*/
        break;
    case T38_TYPE_OF_MSG_T30_DATA:
        break;
    }
    /*endswitch*/
    if (ptr > pkt_len)
        return ret;
    /*endif*/
    return ptr;
}
/*- End of function --------------------------------------------------------*/

/************************************************************************************************/

/**
   This function wlll be called whenever UDP data is delivered
   to the socket previously created upon a NAT PT rule addition.
   Incoming packets' data on a UDP_REDIRECT_LOC_MAGIC socket will
   be written down to the pairing UDP_REDIRECT_REM_MAGIC socket,
   and vice-versa.

   \param sk            socket data
   \param skb           socket buffer containing the data

   \return
      on success 0 is returned
      on error, descriptive error code is returned
*/
static int svip_nat_pt_action(struct sock *sk, struct sk_buff *skb)
{
   int err = 0;
   SVIP_NAT_PT_table_entry_t *pNatPtEntry = NULL;
   SVIP_NAT_PKT_STATS_t *pNatPtStats = NULL;
   SVIP_NAT_PT_sockdata_t *psockdata = NULL, *psockdataRem = NULL;
   struct udphdr *udph = NULL;
   int offset, len, port, bRem = 0;
   int sk_data = 0;
   int nat_index = 0;
   PSVIP_UDPTL_PACKET_t pUdptl;
   IFX_uint16_t seq_num;
   char no_signal[] = {0x00, 0x00, 0x01, 0x00, 0x00, 0x00};
   char no_signal_ifp[] = {0x01, 0x00};
#define MAX_BUFFER_SIZE 1280
   unsigned char buffer[MAX_BUFFER_SIZE];
   unsigned char *pUdptlBuf;
   int fSendBuffer = 0;
   const uint8_t *msg;
   int msg_len;
   int stat, stat2, count, total_count = 0;
   int t30_indicator;
   int ptr = 0, err_recovery = 0;
   IFX_uint8_t type, updateT38SeqNum = 1;
   const IFX_uint8_t rtp_header_len = 12;

   /* Get the value assigned to the skb user data ptr. It si NOt a ptr but typecast int */
   sk_data = (int)sk->sk_user_data;
   /* MS byte holds the NAT table rule index */
   nat_index = (sk_data & UDP_REDIRECT_INDEX_MASK) >> UDP_REDIRECT_INDEX_SHIFT;

   udph = (struct udphdr *)skb->data;
   if ( udph == NULL )
   {
      printk("WARNING!!! SVIP NAT Incorrect skb buff ptr\n");
      return -1;
   }

   /* Handling packets coming from the LOCAL SVIPs destined for the outside world */
   if ((sk_data & UDP_REDIRECT_SOCK_MASK) == UDP_REDIRECT_LOC_MAGIC)
   {
      port = ntohs(udph->source);
      pNatPtEntry = nat_port_table_ptr[port];
      psockdata = &pNatPtEntry->sockdata_rem;
      bRem = 0;
   }
   /* Handling packets incoming from the outside world, destined for the local SVIPs */
   else if ((sk_data & UDP_REDIRECT_SOCK_MASK) == UDP_REDIRECT_REM_MAGIC)
   {
      port = ntohs(udph->dest);
      pNatPtEntry = nat_port_table_ptr[port];
      psockdata = &pNatPtEntry->sockdata_loc;
      bRem = 1;
      psockdataRem = &pNatPtEntry->sockdata_rem;
   }
   else
   {
      printk("WARNING!!! sk->sk_user_data = %p, rem = %d\n", sk->sk_user_data, bRem);
      return -1;
   }

   /* There is a smal chance for the NAT rule to be removed as a packet is veing processed. 
      In that case nat_port_table_ptr[port]; will be NULL so there is no access to the NAT rule
      that handles this socket. */
   if ( NULL == pNatPtEntry )
   {
//      printk("WARNING!!! Packet arrived on a closed socket. NAT rule for port %d has been removed.\n", port );
      return -1;
   }

//printk("sk->sk_user_data = 0x%08X port = %d, rem = %d\n", sk->sk_user_data, port, bRem);
   pNatPtEntry->natStats[bRem].inPackets++;
   pNatPtStats = &pNatPtEntry->stats;

   offset = sizeof(struct udphdr);
   if ((len = skb->len - offset) < 0)
      return -1;

   if ( psockdata->sock_state == SOCK_STATE_RUNNING )
   {
      if ( 0 == bRem )
      {
         ++pNatPtStats->preceived;
         pNatPtStats->oreceived += (( len > rtp_header_len ) ? ( len - rtp_header_len ) : 0);
         if ( pNatPtEntry->hookT38 )
         {
            pUdptlBuf = skb->data + offset;
            pUdptl = (PSVIP_UDPTL_PACKET_t)(pUdptlBuf);
            ptr += sizeof(pUdptl->seq_number);
            /* Break out the primary packet */
            if ((stat = decode_open_type(pUdptlBuf, len, &ptr, &msg, &msg_len)) != 0)
            {
                pNatPtEntry->hookT38 = 0;
                updateT38SeqNum = 0;
            }
            err_recovery = ptr;
            /* RED */
            if ( (pUdptlBuf[ptr++] & 0x80) == 0 )
            {
               total_count = 0;
               /* decode how many entries here. */
               do
               {
                  if ((stat2 = decode_length(pUdptlBuf, len, &ptr, &count)) < 0)
                  {
                     updateT38SeqNum = 0;
                     break;
                  }
                  total_count += count;
               }
               while ( stat2 > 0 );
            }
            /* FEC */
            else
            {
               /* No FEC is allowed under RED mode. */
               if ( pNatPtEntry->udpErrCorrection == IFX_SVIP_T38_RED )
               {
                  pNatPtEntry->hookT38 = 0;
                  updateT38SeqNum = 0;
               }
            }
            if ( pNatPtEntry->updateT38SeqNum && updateT38SeqNum )
            {
               while ( pNatPtEntry->sendNoSignal > 0 )
               {
                  svip_nat_pt_sendmsg (psockdata, no_signal, sizeof(no_signal));
                  pNatPtEntry->sendNoSignal--;
               }
               seq_num = ntohs(pUdptl->seq_number) + 1;
               pUdptl->seq_number = htons(seq_num);
               if ( pNatPtEntry->updateT38ErrRecovery )
               {
                  if ( pNatPtEntry->udpErrCorrection == IFX_SVIP_T38_RED )
                  {
                     total_count += 1;
                     /* max entries of secondary ifp packets on SVIP is 4. */
                     if ( ( total_count <= 4 )
                        && ( len < ( MAX_BUFFER_SIZE + sizeof(no_signal_ifp) ) )
                        )
                     {
                        err_recovery += 1;
                        encode_length(pUdptlBuf, &err_recovery, total_count);
                        memcpy(buffer, pUdptlBuf, len);
                        memcpy(buffer + len, no_signal_ifp, sizeof(no_signal_ifp));
                        fSendBuffer = 1;
                        len += sizeof(no_signal_ifp);
                     }
                     else
                     {
                        pNatPtEntry->updateT38ErrRecovery = 0;
                     }
                  }
               }
            }
         }
      }
      else
      {
         ++pNatPtStats->psent;
         pNatPtStats->osent += (( len > rtp_header_len ) ? ( len - rtp_header_len ) : 0);
         /* accelerate to send out the no-signal packets if we received the t30 indicator from far end. */
         if ( !svip_no_signal_far_end[pNatPtEntry->natRule.locUDP] )
         {
            pUdptlBuf = skb->data + offset;
            pUdptl = (PSVIP_UDPTL_PACKET_t)(pUdptlBuf);
            ptr += sizeof(pUdptl->seq_number);
            /* Break out the primary packet */
            if ((stat = decode_open_type(pUdptlBuf, len, &ptr, &msg, &msg_len)) == 0)
            {
               stat = t38_core_rx_ifp_stream(msg, msg_len, &type, &t30_indicator);
               if ( (stat >= 0)
                  && ( msg_len == 1 )
                  &&( type == T38_TYPE_OF_MSG_T30_INDICATOR )
                  && (t30_indicator == T38_IND_NO_SIGNAL )
                  )
               {
                  svip_no_signal_far_end[pNatPtEntry->natRule.locUDP] = 1;
               }
            }
         }
         if ( pNatPtEntry->sendNoSignal
            && pNatPtEntry->hookT38
            && svip_no_signal_far_end[pNatPtEntry->natRule.locUDP]
            )
         {
            while ( pNatPtEntry->sendNoSignal > 0 )
            {
               svip_nat_pt_sendmsg (psockdataRem, no_signal, sizeof(no_signal));
               pNatPtEntry->sendNoSignal--;
            }
         }
      }
      if ( fSendBuffer )
      {
         err = svip_nat_pt_sendmsg (psockdata, buffer, len);
      }
      else
      {
         err = svip_nat_pt_sendmsg (psockdata, skb->data + offset, len);
      }
   }
   else
   {
      psockdata->sock_state = SOCK_STATE_CLOSED;
      err = -1;
//      printk("WARNING!!! Packet arrived on a closed socket\n");
   }

   if (err < 0)
      pNatPtEntry->natStats[bRem].outErrors++;
   else
      pNatPtEntry->natStats[bRem].outPackets++;

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
   svip_nat_pt_stats_update();
#endif

   return err;
}

/******************************************************************************/
/**
   Function executed upon unloading of the SVIP NAT PT module. It unregisters the
   SVIP NAT PT configuration device and frees the memory used for the NAT table.

\remarks:
   Currently the SVIP NAT PT module is statically linked into the Linux kernel
   therefore this routine cannot be executed.
*******************************************************************************/
static int __init init(void)
{
#if 0 /* debug RMi */
   struct net_device *dev;
#endif

   if (misc_register(&svip_nat_pt_miscdev) != 0)
   {
      printk(KERN_ERR "%s: cannot register SVIP NAT PT device node.\n",
         svip_nat_pt_miscdev.name);
      return -EIO;
   }

   /* allocation of memory for NAT table */
   pNatPtTable = (SVIP_NAT_PT_table_entry_t *)kmalloc(
         sizeof(SVIP_NAT_PT_table_entry_t) * CONFIG_LTQ_SVIP_NAT_RULES_TOTAL, GFP_KERNEL);
   if (pNatPtTable == NULL)
   {
      printk (KERN_ERR "SVIP NAT PT: error allocating memory for NAT PT table\n");
      return -1;
   }

   /* clear the NAT table */
   memset((void *)pNatPtTable, 0, sizeof(SVIP_NAT_PT_table_entry_t) * CONFIG_LTQ_SVIP_NAT_RULES_TOTAL);

   memset(nat_port_table_ptr, 0, sizeof(nat_port_table_ptr)); //[SVIP_NAT_PORT_LOOKUP_TABLE_SIZE]
   memset(svip_port_nat_rule_map, -1, sizeof(svip_port_nat_rule_map));
   memset(svip_no_signal_far_end, 0, sizeof(svip_no_signal_far_end));
   printk("SVIP NAT: Initialize NAT udp port lookup table. Size %d\n", sizeof(nat_port_table_ptr));

#ifdef CONFIG_LTQ_SVIP_NAT_PT_PERFORMANCE_STATS
   /* clear the performance statistics table */
   memset((void *)&perfStats, 0, sizeof(SVIP_NAT_PT_performance_stats_t));
#endif

   if ((sem_nat_tbl_access = kmalloc(sizeof(struct semaphore), GFP_KERNEL)) == NULL)
   {
      printk (KERN_ERR "SVIP NAT PT: error allocating memory for semaphore\n");
      return -1;
   }
   sema_init(sem_nat_tbl_access, 1);

   svip_nat_pt_proc_install();

   /* set pointer to the UDP socket hook function */
   udp_do_redirect_fn = svip_nat_pt_action;
   udp_do_tx_redirect_fn = svip_nat_pt_tx_action;

#if 0
   /* find pointers to 'struct net_device' of eth0 and veth0, respectevely */
   read_lock(&dev_base_lock);
   SVIP_NAT_FOR_EACH_NETDEV(dev)
   {
      if (!strcmp(dev->name, SVIP_NET_DEV_ETH0_NAME))
      {
         net_devs[SVIP_NET_DEV_ETH0_IDX] = dev;
      }
      if (!strcmp(dev->name, SVIP_NET_DEV_VETH1_NAME))
      {
         net_devs[SVIP_NET_DEV_VETH0_IDX] = dev;
      }
   }
   read_unlock(&dev_base_lock);

   if (net_devs[SVIP_NET_DEV_ETH0_IDX] == NULL ||
       net_devs[SVIP_NET_DEV_VETH0_IDX] == NULL)
   {
      printk (KERN_ERR "SVIP NAT PT: Error, unable to locate eth0 and veth0 interfaces\n");
      return -1;
   }
#endif

   printk ("%s, (c) 2012, Lantiq Deutschland GmbH\n", &LTQ_SVIP_NAT_PT_INFO_STR[4]);

   return 0;
}


/******************************************************************************/
/**
   Function executed upon unloading of the SVIP NAT PT module. It unregisters the
   SVIP NAT PT configuration device and frees the memory used for the NAT table.

\remarks:
   Currently the SVIP NAT PT module is statically linked into the Linux kernel
   therefore this routine cannot be executed.
*******************************************************************************/
static void __exit fini(void)
{
   int i;
   SVIP_NAT_PT_table_entry_t *pNatPtEntry;
   unsigned long flags;

   down(sem_nat_tbl_access);

   /* unregister SVIP NAT PT configuration device */
   misc_deregister(&svip_nat_pt_miscdev);

   if (pNatPtTable == NULL)
   {
      up(sem_nat_tbl_access);
      return;
   }

   /* clear UDP REDIRECT callback function */
   udp_do_redirect_fn = NULL;

   /* go through all rules and release the sock data structure */
   for (i = 0; i < CONFIG_LTQ_SVIP_NAT_RULES_TOTAL; i++)
   {
      pNatPtEntry = &pNatPtTable[i];

      if (pNatPtEntry->sockdata_loc.sock == NULL)
         continue;

      svip_nat_pt_release_sock(pNatPtEntry, 0);
      svip_nat_pt_release_sock(pNatPtEntry, 1);
   }
   kfree (pNatPtTable);

   MOD_DEC_USE_COUNT;

   local_save_flags(flags);
   up(sem_nat_tbl_access);
   kfree(sem_nat_tbl_access);
   local_irq_restore(flags);
}

module_init(init);
module_exit(fini);

#endif
