#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include "msglib.h"
#include "pvalue_validation/pvalue_validation_lib.h"

int open_queue( key_t keyval )
{
    int qid;

    if ( ( qid = msgget( keyval, IPC_CREAT | 0660 ) ) == -1 )
    {
        //printf( "%s got error %d (%s)\n", __FUNCTION__, errno, strerror( errno ) );
        return ( -1 );
    }

    return ( qid );
}

int send_message( int qid, void *qbuf, int size )
{
    int result, length;

    /* The length is essentially the size of the structure minus sizeof(mtype) */
    length = size - sizeof( long );

    if ( ( result = msgsnd( qid, qbuf, length, IPC_NOWAIT ) ) == -1 )
    {
        //printf( "%s got error %d (%s)\n", __FUNCTION__, errno, strerror( errno ) );
        return ( -1 );
    }

    return ( result );
}

int send_message_retry( int qid, void *qbuf, int size, int retries, int* err )
{
    int result, length;
    int retrycount = 0;
    int errsv = 0;

    /* The length is essentially the size of the structure minus sizeof(mtype) */
    length = size - sizeof(long);

    while( 1 )
    {
        result = msgsnd( qid, qbuf, length, IPC_NOWAIT); 
        errsv = errno;
        if ( result == -1 && errsv == EAGAIN )
        {
            retrycount++;
            printf( "Cannot send message. Retry...\n" );
            if ( retries == -1 ) /*send infinitely*/
                continue;
            if( retrycount < retries )
            {
                continue;
            }
        }
        if ( err )
            *err = errsv;
        return result;
    }
    return -2; //Should not happen
}

int read_message( int qid, long type, void *qbuf, int size )
{
    int result, length;

    /* The length is essentially the size of the structure minus sizeof(mtype) */
    length = size - sizeof( long );

    if ( ( result = msgrcv( qid, qbuf, length, type, 0 ) ) == -1 )
    {
        //printf( "%s got error %d (%s)\n", __FUNCTION__, errno, strerror( errno ) );
        return ( -1 );
    }

    return ( result );
}

int peek_message( int qid, long type )
{
    int result;

    //printf( "Error: peek_message should not be used\n" );

    if ( ( result = msgrcv( qid, NULL, 0, type, IPC_NOWAIT ) ) == -1 )
    {
        if ( errno == E2BIG )
        {
            return ( 1 );
        }
    }

    return ( 0 );
}

int remove_queue( int qid )
{
    if ( msgctl( qid, IPC_RMID, 0 ) == -1 )
    {
        //printf( "%s got error %d (%s)\n", __FUNCTION__, errno, strerror( errno ) );
        return ( -1 );
    }

    return ( 0 );
}

int get_group( char* group, char* grp )
{
    long val;
    char *endptr = NULL;
    int retval = 0;

    if ( group )
    {
        val = strtol( group, &endptr, 10);

        if ( ( errno == ERANGE && ( val == LONG_MAX || val == LONG_MIN ) )
            ||( errno != 0 && val == 0 )
           )
        {
            retval = 1;
        }
        else if ( endptr == group )
        {
            retval = 1;
        }
        else
        {
            *grp = val;
        }
    }
    else
    {
        retval = -1;
    }

    return retval;
}

int scripts_signal( char *groups, char *data )
{
    int server_qid;
    key_t msgkey;
    admin_msg server_msg;
    int ret;

    server_msg.group = 0;
    get_group( groups, &( server_msg.group ) );

    /* Generate our IPC key value */
    msgkey = SERVER_MSGQ_KEY;

    /* Open/create the queue */
    if ( ( server_qid = open_queue( msgkey ) ) == -1 )
    {
        //perror("open_queue");
        return ( 1 );
    }

    server_msg.mtype = ADMMSG_MSG_DONE_SCRIPTS_EVT;
    server_msg.user_id = 0;
    memset( server_msg.mbody, 0, ADMIN_EVENT_MSG_LENGTH );

    if ( data && strlen( data ) )
    {
        strncpy( server_msg.mbody, data, ADMIN_EVENT_MSG_LENGTH );
    }

    if ( ( ret = send_message( server_qid, ( struct msgbuf * )&server_msg, sizeof( server_msg ) ) ) != 0 )
    {
        //printf(" send msg error, ret=%d\n", ret);
    }

    return 0;
}

