/*
 * NVRAM variable manipulation
 *
 * Copyright 2005, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram.h,v 1.4 2013/04/18 19:22:02 bmathew Exp $
 *
 * SCF - ADM5120 port - 6/30/2006
 *     - removed broadcom specific portions and OS indepent layer
 *     - removed MTD dependence, ported to raw flash access routines
 *     - added zlib support and checksumming for nvram file
 *
 */

#ifndef _bcmnvram_h_
#define _bcmnvram_h_

#define NVRAM_DEV_MAJOR 241
#define NVRAM_DEV_MINOR 0

#define MARK_TEMP         1
#define MARK_L1LOCKED     2
#define MARK_L2LOCKED     3

#define DO_NOT_NOTIFY_FLAG 1
#define DO_NOT_VALIDATE_FLAG 2
#define PROVISION_CHANGE_FLAG 4
#define TR069_CHANGE_FLAG 8
#define WEB_CHANGE_FLAG 0x10
#define SSH_CHANGE_FLAG 0x20
#define LOCK_FLAG_MASK 0xfffc

#define LINE_BUF            (2048*4) /* max. buffer allocated for one line, we need at lease support both two cert/key fields changed on Web */
#define MAX_KEYLEN          128
#define MAX_VALLEN          LINE_BUF    /* this is to allow the cert/key field */
#define MAX_STR_LEN         8192

#ifndef _LANGUAGE_ASSEMBLY

#include <string.h>
#include <linux/types.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * The definition of MAX_MSG_GROUP_INTEGER_ID is the bits number of __admin_msg.group
 */
#define MAX_MSG_GROUP_INTEGER_ID 8

typedef enum
{
    /* NOTE default group for broadcating to all listeners is 1 */
    MSG_GROUP_ID_BROADCAST = 0x1,
    MSG_GROUP_ID_01 = 0x2,
    MSG_GROUP_ID_02 = 0x4,
    MSG_GROUP_ID_03 = 0x8,
    MSG_GROUP_ID_GS_ATA = 0x10,
    MSG_GROUP_ID_05 = 0x20,
    MSG_GROUP_XML_PVALUE = 0x40,
    MSG_GROUP_ID_CPE = 0x80
} MSG_GROUP_ID_E;

struct nvram_header
{
    int magic;
    int len;
    int space_used;
    char reserved[10];
};

struct nvram_tuple
{
    char *name;
    char *value;
    struct nvram_tuple *next;
};

/*
 * we don't know how much data we'll get out of the compressed data stored in flash,
 * so we'll create a linked list element for each call to zlib_inflate with
 * the number of bytes available in the decompression buffer equal to the size of the
 * array pointed to by *data, then glue all the array elements together for processing
 */
struct nvram_inflate_stream
{
    char *data;
    struct nvram_inflate_stream *next;
};

/*
 * Initialize NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern int nvram_init( void* );

/*
 * Disable NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern void nvram_exit( void );

/*
 * Get the value of an NVRAM variable. The pointer returned may be
 * invalid after a set.
 * @param   name    name of variable to get
 * @return  value of variable or NULL if undefined
 */
extern char * nvram_get( const char *name );
extern char *nvram_l2get( const char *name );

/*
 * Race free version of nvram_get
 * @param name      name of variable to get
 * @param buffer    buffer to store result
 * @param length    length of buffer
 */
extern char * nvram_get_safe( const char * name, char * buffer, const unsigned int length );

/*
 * Get the value of an NVRAM variable.
 * @param   name    name of variable to get
 * @return  value of variable or NUL if undefined
 */
#define nvram_safe_get(name) (nvram_get(name) ? : "")

/*
 * Match an NVRAM variable.
 * @param   name    name of variable to match
 * @param   match   value to compare against value of variable
 * @return  TRUE if variable is defined and its value is string equal
 *      to match or FALSE otherwise
 */
static __inline__ int
nvram_match( char *name, char *match )
{
    const char *value = nvram_get( name );
    return ( value && !strcmp( value, match ) );
}

/*
 * Inversely match an NVRAM variable.
 * @param   name    name of variable to match
 * @param   match   value to compare against value of variable
 * @return  TRUE if variable is defined and its value is not string
 *      equal to invmatch or FALSE otherwise
 */
static __inline__ int
nvram_invmatch( char *name, char *invmatch )
{
    const char *value = nvram_get( name );
    return ( value && strcmp( value, invmatch ) );
}

/*
 * Set the value of an NVRAM variable. The name and value strings are
 * copied into private storage. Pointers to previously set values
 * may become invalid. The new value may be immediately
 * retrieved but will not be permanently stored until a commit.
 * @param   name    name of variable to set
 * @param   value   value of variable
 * @return  0 on success and errno on failure
 */
extern int nvram_set( const char *name, const char *value );
extern int nvram_l2set( const char *name, const char *value );
extern int nvram_set_unvalidated( const char *name, const char *value );
extern int nvram_set_skip_notify( const char *name, const char *value );
extern int nvram_set_provision( const char *name, const char *value );
extern int nvram_set_tr069( const char *name, const char *value );
extern int nvram_set_flag( const char *name, const char *value, unsigned int flag );

/*
 * Determines if a value is valid for a given pvalue
 * @param   name    name of variable to set
 * @param   value   value to verify
 * @return  1 if valid, 0 otherwise
 */
extern int nvram_is_valid( const char *name, const char *value );

/*
 * Unset an NVRAM variable. Pointers to previously set values
 * remain valid until a set.
 * @param   name    name of variable to unset
 * @return  0 on success and errno on failure
 * NOTE: use nvram_commit to commit this change to flash.
 */
extern int nvram_unset( const char *name );
extern int nvram_unset_tr069( const char *name );

/*
 * Erase all NVRAM variables
 * @return  0 on success and errno on failure
 */
extern int nvram_erase_all( void );

/*
 * Commit NVRAM variables to permanent storage. All pointers to values
 * may be invalid after a commit.
 * NVRAM values are undefined after a commit.
 * @return  0 on success and errno on failure
 */
extern int nvram_commit( void );
extern int nvram_commit_sync( void );

/*
 * Get all NVRAM variables (format name=value\0 ... \0\0).
 * @param   buf buffer to store variables
 * @param   count   size of buffer in bytes
 * @return  0 on success and errno on failure
 */
extern int nvram_getall( char *buf, int count );

/*
 * Reset NVRAM variables at a certain privilege level or lower 
 * @param	buf	buffer to store variables
 * @param	privilege_level	privilege level to erase
 * @param	mode 0 erase privilege level only, mode 1 erase privilege level and below
 */
extern void nvram_reset_level( char *buf, int privilege_level, int mode) ;
extern int nvram_erase_list( const char *filename );
extern int nvram_erase_not_list( const char *filename );
extern int nvram_space_used( void );
extern int nvram_ram_used( void );
extern int nvram_erase_all( void );
extern int nvram_get_ram_size( void );
extern char *nvram_alloc_ram_size_buffer( void );


/*
 * Get whether or not NVRAM variable belongs a specified group
 * @param name - the pvalue name or alias
 * @param group - the group we want to know wheter or not the pvalue blongs to
 * @return - 1 if belongs, 0 otherwise
 */
int nvram_has_notification_group( const char* name, const unsigned int group );

/*
 * Get help for pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_help( const char* name, char *buffer, unsigned int size );

/*
 * Check reboot flag
 * @attrib name - the pvalue name/alias
 * @return - the reboot flag is set (1) or not (0), -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int nvram_check_reboot( const char* name );

/*
 * Check pvalue check privilege flag
 * @param name - the pvalue name or alias
 * @return - the privilege flag on success, -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int nvram_check_privilege( const char* name );

/*
 * Check lock flag
 * @attrib name - the pvalue name/alias
 * @return - the lock flag > 0, -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int nvram_check_lock( const char* name );

/*
 * Check if pvalue is a password
 * @param name - the pvalue name or alias
 * @return - if pvalue is a password (1) or not (0), -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int nvram_check_password( const char* name );

/*
 * Check if pvalue matches datatype
 * @param name - the pvalue name or alias
 * @param type - the pvalue category type to match
 * @return - if pvalue has the same dataype (1) or not(0) , -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int nvram_match_datatype( const char* name, const char* type );

/*
 * Get proper name of pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_name( const char* name, char *buffer, unsigned int size );

/*
 * Get default of pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_default( const char* name, char *buffer, unsigned int size );

/*
 * Get alias of pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_alias( const char* name, char *buffer, unsigned int size );

/*
 * Get scripts of pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_scripts( const char* name, char *buffer, unsigned int size );

/*
 * Get limits of pvalue
 * @param name - the pvalue name or alias
 * @param buffer - buffer to store result
 * @param size - size of the buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int nvram_get_limits( const char* name, char *buffer, unsigned int size );

#ifdef __cplusplus
}
#endif

#endif /* _LANGUAGE_ASSEMBLY */

#define NVRAM_MAGIC           0x48534C46/* 'FLSH' */
#define NVRAM_SPACE_USED      0x00000001
#define NVRAM_RAM_USED        0x00000002
#define NVRAM_COMMIT_SYNC     0x00000004
#define NVRAM_ERASE_ALL       0x00000010
#define NVRAM_RAM_SIZE        0x00000020

#endif /* _bcmnvram_h_ */
