/*
 * NVRAM variable manipulation (Linux user mode half)
 *
 * Copyright 2004, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram_linux.c,v 1.5.2.2 2015/09/30 16:17:16 bmathew Exp $
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <error.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/syslog.h>

#include "nvram.h"
#include "msg_notify.h"
#include "nvram_linux.h"
#include "pvalue_validation/pvalue_validation_lib.h"

#define PATH_DEV_NVRAM "/dev/nvram"
#define PATH_DEV_NVRAM_SAFE "/dev/nvram_safe"
#define PROPER_NAME_LEN (128)

/* Globals */
static int nvram_fd = -1;
static int nvram_safe_fd = -1;
static int kernel_nvram_ram_size = -1;
static char *nvram_buf = NULL;
int check_action( void );
int file_to_buf( char *path, char *buf, int len );
static int dummy_fd0 = -1;  /*fake stdin*/
static int dummy_fd1 = -1;  /*fake stdout*/
static int dummy_fd2 = -1;  /*fake stderr*/

inline void
print_err( char *errmsg )
{
#if 0   //SXIE

    if ( errno )
    {
        perror( errmsg );
    }

#else
    //if (errno)
    //perror(NULL);
#endif
}

int
nvram_init( void *unused )
{
    /*
     * Open three file handles before opening to /dev/nvram
     * to prevent /dev/nvram being inadvertently
     * used as stdin, stdout or stderr.
     */
    if ( ( dummy_fd0 = open( "/dev/null", O_RDONLY ) ) < 0 )
    {
        goto err;
    }

    if ( ( dummy_fd1 = open( "/dev/null", O_WRONLY ) ) < 0 )
    {
        goto err;
    }

    if ( ( dummy_fd2 = open( "/dev/null", O_RDWR ) ) < 0 )
    {
        goto err;
    }

    if ( ( nvram_fd = open( PATH_DEV_NVRAM, O_RDWR ) ) < 0 )
    {
        goto err;
    }

    if ( ( nvram_safe_fd = open( PATH_DEV_NVRAM_SAFE, O_RDWR ) ) < 0 )
    {
        goto err;
    }

    if ( ( kernel_nvram_ram_size = nvram_get_ram_size() ) <= 0 )
    {
        printf( "Kernel does not support NVRAM_RAM_SIZE ioctl\n" );
        goto err;
    }

    /* Map kernel string buffer into user space */
    if ( ( nvram_buf = mmap( NULL, kernel_nvram_ram_size, PROT_READ, MAP_SHARED, nvram_fd, 0 ) ) == MAP_FAILED )
    {
        close( nvram_fd );
        nvram_fd = -1;
        goto err;
    }

    return 0;

err:
    print_err( PATH_DEV_NVRAM );
    return errno;
}

char *
nvram_get_safe( const char *name, char * buffer, const unsigned int length )
{
    size_t count = strnlen( name, MAX_VALLEN ) + 1;
    char * ret = buffer;
    char * kbuf = buffer;
    char * target;
    char propername[PROPER_NAME_LEN] = {0};

    if ( count == (MAX_VALLEN + 1) )
    {
        return NULL;
    }

    if ( nvram_fd < 0 )
    {
        if ( nvram_init( NULL ) )
        {
            return NULL;
        }
    }

    if ( ( name[0] == MARK_TEMP ) ||
            ( name[0] == MARK_L1LOCKED ) ||
            ( name[0] == MARK_L2LOCKED )
       )
    {
        propername[0] = name[0];
        nvram_get_name( &name[1], &propername[1], ( PROPER_NAME_LEN - 1 ) );

        if ( strlen( &propername[1] ) == 0 )
        {
            target = ( char * )name;
        }
        else
        {
            target = propername;
        }
    }
    else
    {
        nvram_get_name( name, propername, PROPER_NAME_LEN );

        if ( strlen( propername ) == 0 )
        {
            target = ( char * )name;
        }
        else
        {
            target = propername;
        }
    }

    count = strlen( target ) + 1;

    if ( count > length )
    {
        if ( !( kbuf = malloc( count ) ) )
        {
            return NULL;
        }

        memset( kbuf, 0, count );
    }
    else
    {
        count = length;
    }

    /* Get offset into mmap() space */
    strcpy( ( char * )kbuf, target );

    count = read( nvram_safe_fd, kbuf, count );

    if ( count <= 0 )
    {
        memset( buffer, 0, length );
        ret = NULL;
    }
    else if ( count <= length )
    {
        if ( kbuf != buffer )
        {
            memcpy( buffer, kbuf, count );
        }
    }
    else
    {
        memset( buffer, 0, length );
        ret = NULL;
    }

    if ( count < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( kbuf != buffer )
    {
        free( kbuf );
    }

    return ret;
}

char *
nvram_get( const char *name )
{
    size_t count = 0;
    char tmp[MAX_VALLEN], *value; //SXIE 041607, increase the buf size to MAX_VALLEN from 100
    unsigned long *off = ( unsigned long * ) tmp;
    char propername[PROPER_NAME_LEN] = {0};
    char * target;

    if ( nvram_fd < 0 )
        if ( nvram_init( NULL ) )
        {
            return NULL;
        }

    nvram_get_name( name, propername, PROPER_NAME_LEN );

    if ( strlen( propername ) == 0 )
    {
        target = ( char * )name;
    }
    else
    {
        target = propername;
    }

    count = strlen( target ) + 1;

    if ( count > sizeof( tmp ) )
    {
        if ( !( off = malloc( count ) ) )
        {
            return NULL;
        }

        memset( off, 0, count );
    }

    /* Get offset into mmap() space */
    strcpy( ( char * ) off, target );

    count = read( nvram_fd, off, count );

    if ( count == sizeof( unsigned long ) )
    {
        value = &nvram_buf[*off];
    }
    else
    {
        value = NULL;
    }

    if ( count < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( off != ( unsigned long * ) tmp )
    {
        free( off );
    }

    return value;
}

char *
nvram_l2get_safe( const char *name, char * buffer, unsigned int length )
{
    size_t count = strnlen( name, MAX_VALLEN ) + 2;
    char * kbuf = buffer;
    char * ret = buffer;
    char propername[PROPER_NAME_LEN] = {0};
    char * target;

    if ( count == ( MAX_VALLEN + 2 ) )
    {
        return NULL;
    }

    if ( nvram_fd < 0 )
        if ( nvram_init( NULL ) )
        {
            return NULL;
        }

    nvram_get_name( name, propername, PROPER_NAME_LEN );

    if ( strlen( propername ) == 0 )
    {
        target = ( char * )name;
    }
    else
    {
        target = propername;
    }

    count = strlen( target ) + 2;

    if ( count > length )
    {
        if ( !( kbuf = malloc( count ) ) )
        {
            return NULL;
        }

        memset( kbuf, 0, count );
    }
    else
    {
        count = length;
    }

    /* Get offset into mmap() space */
    sprintf( ( char * )kbuf, " %s", name );
    kbuf[0] = MARK_L2LOCKED; // mark the first char

    count = read( nvram_safe_fd, kbuf, count );

    if ( count <= 0 )
    {
        ret = NULL;
    }
    else if ( count <= length )
    {
        memcpy( buffer, kbuf, count );
    }
    else
    {
        ret = NULL;
    }

    if ( count < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( kbuf != buffer )
    {
        free( kbuf );
    }

    return ret;
}


char *
nvram_l2get( const char *name )
{
    size_t count = strlen( name ) + 2;
    char tmp[MAX_VALLEN], *value; //SXIE 041607, increase the buf size to MAX_VALLEN from 100 for some ssl key fields
    unsigned long *off = ( unsigned long * ) tmp;
    char propername[PROPER_NAME_LEN] = {0};
    char * target;

    if ( nvram_fd < 0 )
        if ( nvram_init( NULL ) )
        {
            return NULL;
        }

    nvram_get_name( name, propername, PROPER_NAME_LEN );

    if ( strlen( propername ) == 0 )
    {
        target = ( char * )name;
    }
    else
    {
        target = propername;
    }

    count = strlen( target ) + 2;

    if ( count > sizeof( tmp ) )
    {
        if ( !( off = malloc( count ) ) )
        {
            return NULL;
        }

        memset( off, 0, count );
    }

    /* Get offset into mmap() space */
    sprintf( ( char * ) tmp, " %s", name );
    tmp[0] = MARK_L2LOCKED; // mark the first char

    count = read( nvram_fd, off, count );

    if ( count == sizeof( unsigned long ) )
    {
        value = &nvram_buf[*off];
    }
    else
    {
        value = NULL;
    }

    if ( count < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( off != ( unsigned long * ) tmp )
    {
        free( off );
    }

    return value;
}

int
nvram_getall( char *buf, int count )
{
    int ret;
    int read_len = count;

    if ( nvram_fd < 0 )
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return ret;
        }

    if ( read_len == 0 )
    {
        read_len = kernel_nvram_ram_size;
    }

    /* Get all variables */
    *buf = '\0';

    ret = read( nvram_fd, buf, read_len );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    return ( ret == read_len ) ? 0 : ret;
}

//note: this function must always support utf-8
static int
dos2unix( const char * input_string, const size_t size_in, char * output_string, const size_t size_out )
{
    int in_pos = 0;
    int out_pos = 0;

    if ( input_string == NULL ||  output_string == NULL || size_in == 0 || size_out == 0 )
    {
        return 1;
    }

    while ( ( in_pos < size_in ) && ( out_pos < size_out ) )
    {
        //we need to remove the \r from Windows stype \r\n line endings
        //in some cases the calling function removes the \n at end of line (see load_setting for example)
        //but leaves the \r, as \r is always invalid we should strip out every \r whether it is paired or not
        if ( input_string[in_pos] != '\r' )
        {
            output_string[out_pos] = input_string[in_pos];

            if ( input_string[in_pos] == '\0' )
            {
                return 0;
            }

            out_pos++;
        }

        in_pos++;
    }

    if ( out_pos < size_out ) //handle input not being null-terminated (or strlen used for size_in)
    {
        output_string[ out_pos ] = '\0';
        return 0;
    }
    else if ( output_string[ size_out - 1 ] != '\0' ) //converted input longer than output buffer
    {
        output_string[ size_out - 1 ] = '\0';
    }

    return 1; //return failure by default
}

#if 0 //ascii characters do not occur in utf-8 characters so we do not need this for now
static int
dos2unix( const char * input_string, const size_t size_in, char * output_string, const size_t size_out )
{
    int in_pos = 0;
    int out_pos = 0;
    int count;

    if ( input_string == NULL ||  output_string == NULL || size_in == 0 || size_out == 0 )
    {
        return 1;
    }

    while ( ( in_pos < size_in ) && ( out_pos < size_out ) )
    {
        /*byte value is 0xxx xxxx - ascii*/
        if ( !( input_string[in_pos] & 0x80 ) )
        {
            //we need to remove the \r from Windows stype \r\n line endings
            //in some cases the calling function removes the \n at end of line (see load_setting for example)
            //but leaves the \r, as \r is always invalid we should strip out every \r whether it is paired or not
            if ( input_string[in_pos] != '\r' )
            {
                output_string[out_pos] = input_string[in_pos];

                if ( input_string[in_pos] == '\0' )
                {
                    return 0;
                }

                out_pos++;
            }

            in_pos++;
        }
        else //if utf-8 character
        {
            /*byte value is 1111 0xxx*/
            if ( ( input_string[in_pos] & 0xf8 ) == 0xf0 )
            {
                count = 4;
            }
            /*byte value is 1110 xxxx*/
            else if ( ( input_string[in_pos] & 0xf0 ) ==  0xe0 )
            {
                count = 3;
            }
            /*byte value is 110x xxxx*/
            else if ( ( input_string[in_pos] & 0xe0 ) == 0xc0 )
            {
                count = 2;
            }
            else
            {
                //printf( "ERROR: Encountered invalid UTF-8 character\n" );
                output_string[ out_pos ] = '\0';
                return 1;
            }

            if ( ( out_pos + count ) < size_out )
            {
                memcpy( &output_string[ out_pos ], &input_string[ in_pos ], count );
            }
            else
            {
                output_string[ out_pos ] = '\0';
                return 1;
            }

            in_pos += count;
            out_pos += count;
        }
    }

    if ( out_pos < size_out ) //handle input not being null-terminated
    {
        output_string[ out_pos ] = '\0';
        return 0;
    }
    else if ( output_string[ size_out - 1 ] != '\0' ) //converted input longer than output buffer
    {
        output_string[ size_out - 1 ] = '\0';
    }

    return 1; //return failure by default
}
#endif

static const char* 
check_pvalue_flags( const char *name, const char *value_to_set, int arg )
{
    int locked = nvram_check_lock( name );
    int locked_flag = arg & LOCK_FLAG_MASK; /*separate locked flags from validate and notify flags*/

    if ( 0 != locked )
    {
        if( 0 == locked_flag )
        {
            //printf( "WARNING!!! Locked - pvalue %s\n", name );
            return NULL;
        }
        if( 0 != locked_flag && ( ( locked_flag & locked ) == 0 ) )
        {
            //printf( "WARNING!!! Locked - pvalue %s\n", name );
            return NULL;
        }
    }
    if ( DO_NOT_VALIDATE_FLAG != ( arg & DO_NOT_VALIDATE_FLAG ) )
    {
        if( 0 < validate_pvalue( name, value_to_set ) )
        {
            //printf( "WARNING!!! Attempted to set invalid input for pvalue %s\n", name );
            return NULL;
        }
    }
    return name;
}

static int
_nvram_set( const char *name, const char *value, unsigned int arg )
{
    size_t count = 0;
    char tmp[100], *buf = tmp;
    char propername[PROPER_NAME_LEN] = {0};
    char converted_value[MAX_VALLEN] = {0};
    const char * value_to_set;
    unsigned int groups = 0;
    char * target = NULL;
    int ret;
    //int validate_ret = 0;

    if ( ( value != NULL ) && ( strlen( value ) > 0 ) )
    {
        if ( 1 == dos2unix( value, strlen( value ), converted_value, sizeof( converted_value ) ) )
        {
            printf( "WARNING!!! non-UTF-8/value too long - pvalue %s\n", name );
            return -1;
        }

        value_to_set = converted_value;
    }
    else
    {
        value_to_set = value;
    }

    /*
    * validate value using validation server
    * return if invalid value
    * Continue if name was not found or
    * if validation server was not running.
    */
    if ( ( name[0] == MARK_TEMP ) ||
            ( name[0] == MARK_L1LOCKED ) ||
            ( name[0] == MARK_L2LOCKED )
       )
    {
        propername[0] = name[0];
        nvram_get_name( &name[1], &propername[1], ( PROPER_NAME_LEN - 1 ) );

        if ( strlen( &propername[1] ) == 0 )
        {
            target = ( char* )check_pvalue_flags( &name[1], value_to_set, arg | DO_NOT_VALIDATE_FLAG );
            if ( NULL != target )
            {
                target = ( char* )name;
            }
        }
        else
        {   
            target = ( char* )check_pvalue_flags( &propername[1], value_to_set, arg );
            if ( NULL != target )
            {
                target = propername;
            }
        }
    }
    else
    {
        nvram_get_name( name, propername, PROPER_NAME_LEN );

        if ( strlen( propername ) == 0 )
        {
            target = ( char* )check_pvalue_flags( name, value_to_set, arg | DO_NOT_VALIDATE_FLAG );
        }
        else
        {
            target = ( char* )check_pvalue_flags( propername, value_to_set, arg );
        }
    }
    if ( NULL == target )
    {
        return -1;
    }

    count = strlen( target ) + 1;

    /*if ( ( validate_ret > 0 ) )
    {
        printf( "WARNING!!! Attempted to set invalid input for pvalue %s\n", target );
        return -1;
    }*/

    if ( nvram_fd < 0 )
    {
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return ret;
        }
    }

    /* Unset if value is NULL */
    if ( value_to_set )
    {
        count += strlen( value_to_set ) + 1;
    }

    if ( count > sizeof( tmp ) )
    {

        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    if ( value_to_set )
    {
        sprintf( buf, "%s=%s", target, value_to_set );
    }
    else
    {
        strcpy( buf, target );
    }

    errno = 0;
    ret = write( nvram_fd, buf, count );

    if ( ret < 0 && errno > 0 )
    {
        ret = -errno;
    }

    if ( buf != tmp )
    {
        free( buf );
    }

//printf( "check notification\n" );
    if ( DO_NOT_NOTIFY_FLAG != ( arg & DO_NOT_NOTIFY_FLAG ) )
    {
        if ( ret == count )
        {
            //printf( "check notification\n" );
            get_pvalue_notification_group_mask( target, &groups );
            notify_setting_change( target, groups );
        }
    }

    return ( ret == count || ret == -EEXIST ) ? 0 : ret;
}


//returns 1 if valid, 0 otherwise
int
nvram_is_valid( const char *name, const char *value )
{
    char propername[PROPER_NAME_LEN] = {0};
    int validate_ret = 0;
    char converted_value[MAX_VALLEN] = {0};

    if ( NULL == name || NULL == value )
    {
        return 0;
    }

    if ( 1 == dos2unix( value, strlen( value ), converted_value, sizeof( converted_value ) ) )
    {
        return 0;
    }

    /*
    * validate value using validation server
    * return if invalid value
    */
    if ( ( name[0] == MARK_TEMP ) ||
            ( name[0] == MARK_L1LOCKED ) ||
            ( name[0] == MARK_L2LOCKED )
       )
    {
        propername[0] = name[0];
        nvram_get_name( &name[1], &propername[1], ( PROPER_NAME_LEN - 1 ) );

        if ( strlen( &propername[1] ) == 0 )
        {
            validate_ret = validate_pvalue( &name[1], converted_value );
        }
        else
        {
            validate_ret = validate_pvalue( &propername[1], converted_value );
        }
    }
    else
    {
        nvram_get_name( name, propername, PROPER_NAME_LEN );

        if ( strlen( propername ) == 0 )
        {
            validate_ret = validate_pvalue( name, converted_value );
        }
        else
        {
            validate_ret = validate_pvalue( propername, converted_value );
        }
    }

    return ( 0 == validate_ret );
}

int
nvram_set( const char *name, const char *value )
{
    int ret;

    ret = _nvram_set( name, value, 0 );

    return ret;
}

int
nvram_set_unvalidated( const char *name, const char *value )
{
    int ret;

    ret = _nvram_set( name, value, DO_NOT_VALIDATE_FLAG );

    return ret;
}

int
nvram_set_skip_notify( const char *name, const char *value )
{
    int ret;

    ret = _nvram_set( name, value, DO_NOT_NOTIFY_FLAG );

    return ret;
}

int 
nvram_set_flag( const char *name, const char *value, unsigned int flag )
{
    int ret;

    ret = _nvram_set( name, value, flag );

    return ret;
}

int
nvram_set_provision( const char *name, const char *value )
{
    int ret;

    ret = _nvram_set( name, value, PROVISION_CHANGE_FLAG );

    return ret;
}

int
nvram_set_tr069( const char *name, const char *value )
{
    int ret;

    ret = _nvram_set( name, value, TR069_CHANGE_FLAG );

    return ret;
}

int
nvram_unset( const char *name )
{
    return _nvram_set( name, NULL, 0 );
}

int
nvram_unset_flag( const char *name, unsigned int flag )
{
    return _nvram_set( name, NULL, flag );
}

int
nvram_unset_tr069( const char *name )
{
    return _nvram_set( name, NULL, TR069_CHANGE_FLAG );
}

int
nvram_settemp( const char *name, const char *value )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( value == NULL )
    {
        return -EINVAL;
    }

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_TEMP; // mark the first char

    ret = _nvram_set( buf, value, 0 );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;

}

int
nvram_unsettemp( const char *name )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_TEMP; // mark the first char

    ret = _nvram_set( buf, NULL, 0 );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;

}


int
nvram_l1lock( const char *name )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_L1LOCKED; // mark the first char

    ret = _nvram_set( buf, NULL, nvram_check_lock( name ) | DO_NOT_VALIDATE_FLAG );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;
}

int
nvram_l1set( const char *name, const char *value )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( value == NULL )
    {
        return -EINVAL;
    }

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_L1LOCKED; // mark the first char

    ret = _nvram_set( buf, value, nvram_check_lock( name ) );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;
}

int
nvram_l2lock( const char *name )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_L2LOCKED; // mark the first char

    ret = _nvram_set( buf, NULL, nvram_check_lock( name ) | DO_NOT_VALIDATE_FLAG );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;
}

int
nvram_l2set( const char *name, const char *value )
{
    size_t count = strlen( name ) + 2;
    char tmp[100], *buf = tmp;
    int ret;

    if ( value == NULL )
    {
        return -EINVAL;
    }

    if ( count > sizeof( tmp ) )
    {
        if ( !( buf = malloc( count ) ) )
        {
            return -ENOMEM;
        }

        memset( buf, 0, count );
    }

    sprintf( buf, " %s", name );
    buf[0] = MARK_L2LOCKED; // mark the first char

    ret = _nvram_set( buf, value, nvram_check_lock( name ) );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    if ( buf != tmp )
    {
        free( buf );
    }

    return ret;
}

int
nvram_commit( void )
{
    int ret;

    {
        if ( nvram_fd < 0 )
            if ( ( ret = nvram_init( NULL ) ) )
            {
                return ret;
            }

        ret = ioctl( nvram_fd, NVRAM_MAGIC, NULL );

        if ( ret < 0 )
        {
            print_err( PATH_DEV_NVRAM );
        }
        else
        {
            notify_setting_change( "nvram_commit", MSG_GROUP_ID_BROADCAST ); //notify all groups
        }
    }

    return ret;
}

int nvram_commit_sync( void )
{
    int ret;

    if ( nvram_fd < 0 )
    {
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return ret;
        }
    }

    ret = ioctl( nvram_fd, NVRAM_COMMIT_SYNC, NULL );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    return ret;
}

int
nvram_space_used( void )
{
    int ret;

    {
        if ( nvram_fd < 0 )
            if ( ( ret = nvram_init( NULL ) ) )
            {
                return ret;
            }

        ret = ioctl( nvram_fd, NVRAM_SPACE_USED, NULL );

        if ( ret < 0 )
        {
            print_err( PATH_DEV_NVRAM );
        }

    }

    return ret;
}

//=======================================================================
int
nvram_get_ram_size(
    void
)
//=======================================================================
{
    int ret = 0;

    if ( nvram_fd < 0 )
    {
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return ret;
        }
    }

    ret = ioctl( nvram_fd, NVRAM_RAM_SIZE, NULL );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    return ret;
}

//=======================================================================
char *
nvram_alloc_ram_size_buffer(
    void
)
//Description: Allocate space as big as the kernel NVRAM deflated buffer
//=======================================================================
{
    int ret = 0;

    if ( nvram_fd < 0 )
    {
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return NULL;
        }
    }

    /* Allocate space as big as the kernel NVRAM deflated buffer */
    return malloc( kernel_nvram_ram_size );
}


int
nvram_ram_used( void )
{
    int ret;

    if ( nvram_fd < 0 )
    {
        if ( ( ret = nvram_init( NULL ) ) )
        {
            return ret;
        }
    }

    ret = ioctl( nvram_fd, NVRAM_RAM_USED, NULL );

    if ( ret < 0 )
    {
        print_err( PATH_DEV_NVRAM );
    }

    return ret;
}

int
nvram_erase_all( void )
{
    int ret;

    {
        if ( nvram_fd < 0 )
            if ( ( ret = nvram_init( NULL ) ) )
            {
                return ret;
            }

        ret = ioctl( nvram_fd, NVRAM_ERASE_ALL, NULL );

        if ( ret < 0 )
        {
            print_err( PATH_DEV_NVRAM );
        }

    }

    return ret;
}

int file2nvram( char *filename, char *varname )
{
    FILE *fp;
    int c, count;
    int i = 0, j = 0;
    char mem[10000], buf[30000];

    if ( !( fp = fopen( filename, "rb" ) ) )
    {
        return 0;
    }

    count = fread( mem, 1, sizeof( mem ), fp );
    fclose( fp );

    for ( j = 0; j < count; j++ )
    {
        if ( i > sizeof( buf ) - 3 )
        {
            break;
        }

        c = mem[j];

        if ( c >= 32 && c <= 126 && c != '\\' && c != '~' )
        {
            buf[i++] = ( unsigned char ) c;
        }
        else if ( c == 0 )
        {
            buf[i++] = '~';
        }
        else
        {
            buf[i++] = '\\';
            sprintf( buf + i, "%02X", c );
            i += 2;
        }
    }

    if ( i == 0 )
    {
        return 0;
    }

    buf[i] = 0;
    //fprintf(stderr,"================ > file2nvram %s = [%s] \n",varname,buf);
    nvram_set( varname, buf );
    nvram_commit(); //Barry adds for test
    return 0;
}

int nvram2file( char *varname, char *filename )
{
    FILE *fp;
    int c, tmp;
    int i = 0, j = 0;
    char *buf;
    char mem[10000];

    if ( !( fp = fopen( filename, "wb" ) ) )
    {
        return 0;
    }

    buf = strdup( nvram_safe_get( varname ) );

    //fprintf(stderr,"=================> nvram2file %s = [%s] \n",varname,buf);
    while ( buf[i] && j < sizeof( mem ) - 3 )
    {
        if ( buf[i] == '\\' )
        {
            i++;
            tmp = buf[i + 2];
            buf[i + 2] = 0;
            sscanf( buf + i, "%02X", &c );
            buf[i + 2] = tmp;
            i += 2;
            mem[j] = c;
            j++;
        }
        else if ( buf[i] == '~' )
        {
            mem[j] = 0;
            j++;
            i++;
        }
        else
        {
            mem[j] = buf[i];
            j++;
            i++;
        }
    }

    if ( j <= 0 )
    {
        return j;
    }

    j = fwrite( mem, 1, j, fp );
    fclose( fp );
    free( buf );
    return j;
}

int
file_to_buf( char *path, char *buf, int len )
{
    FILE *fp;

    memset( buf, 0, len );

    if ( ( fp = fopen( path, "r" ) ) )
    {
        fgets( buf, len, fp );
        fclose( fp );
        return 1;
    }

    return 0;
}

/*
*Erase a list of pvalues leaving the rest untouched.
*/
int nvram_erase_list( const char *filename )
{
    FILE *fp;
    int count, total_count;
    char buf[MAX_VALLEN];
    char *delimiter, *start;

    if ( !( fp = fopen( filename, "rb" ) ) )
    {
        return -1;
    }

    delimiter = buf;

    while ( delimiter )
    {
        start = buf;
        total_count = 0;

        while ( 1 )
        {
            count = fread( delimiter, 1, 1, fp );

            if ( !count )
            {
                fclose( fp );
                return -1;
            }

            total_count += count;

            if ( !( *delimiter ) )
            {
                fclose( fp );
                return 0;
            }

            if ( *delimiter == ' ' )
            {
                break;
            }

            delimiter++;

            if ( total_count > 2047 )
            {
                fclose( fp );
                return -1;
            }
        }

        *delimiter = 0;
        _nvram_set( start, NULL, DO_NOT_VALIDATE_FLAG );
        delimiter = buf;
    }

    fclose( fp );
    return 0;
}

/*
* Data structures for nvram_erase_not_list
*/
struct nv_pair
{
    char *name;
    char *value;
    struct nv_pair *next;
};
struct nv_pair *list_head = NULL, *list_tail = NULL;

/*
* Helper functions for nvram_erase_not_list
*/
int store_pairs( char *name, char *value )
{
    unsigned int len1 = 0, len2 = 0;
    struct nv_pair *ptr;

    if ( !name )
    {
        return -1;
    }

    if ( !value )
    {
        return 0;
    }

    ptr = malloc( sizeof( struct nv_pair ) );
    memset( ptr, 0, sizeof( struct nv_pair ) );

    if ( !list_head )
    {
        list_head = ptr;
    }

    if ( list_tail )
    {
        list_tail->next = ptr;
    }

    list_tail = ptr;

    len1 = strnlen( name, MAX_VALLEN );
    ptr->name = malloc( sizeof( char ) * (len1 + 1) );
    memset( ptr->name, 0, sizeof( char ) * (len1 + 1) );
    strncpy( ptr->name, name, len1 );

    len2 = strnlen( value, MAX_VALLEN );
    ptr->value = malloc( sizeof( char ) * (len2 + 1) );
    memset( ptr->value, 0, sizeof( char ) * (len2 + 1) );
    strncpy( ptr->value, value, len2 );

	ptr->next = NULL ;

	return 0 ;
}

/*
* Helper functions for nvram_erase_not_list
*/
int restore_pairs()
{
    struct nv_pair *list_curr = list_head;
    struct nv_pair *prev;
    list_head = NULL;
    list_tail = NULL;

    while ( list_curr )
    {
        _nvram_set( list_curr->name, list_curr->value, nvram_check_lock( list_curr->name ) );
        free( list_curr->name );
        free( list_curr->value );
        prev = list_curr;
        list_curr = list_curr->next;
        free( prev );
    }

    return 0;
}

/*
*Erase pvalues not on the list leaving the list untouched.
*/
int nvram_erase_not_list( const char *filename )
{
    FILE *fp;
    int count, total_count;
    char buf[MAX_VALLEN], buf1[MAX_VALLEN];
    char *delimiter, *start;
    char *value;

    if ( !( fp = fopen( filename, "rb" ) ) )
    {
        return -1;
    }

    delimiter = buf;

    while ( delimiter )
    {
        start = buf;
        total_count = 0;

        while ( 1 )
        {
            count = fread( delimiter, 1, 1, fp );

            if ( !count )
            {
                delimiter = 0;
                break;
            }

            total_count += count;

            if ( !( *delimiter ) )
            {
                delimiter = 0;
                break;
            }

            if ( *delimiter == ' ' )
            {
                break;
            }

            delimiter++;

            if ( total_count > (MAX_VALLEN - 1) )
            {
                delimiter = 0;
                break;
            }
        }

        if ( !delimiter )
        {
            continue;
        }

        *delimiter = 0;

        if ( *start )
        {
            value = nvram_get_safe( start, buf1, MAX_VALLEN );
            store_pairs( start, value );
        }

        delimiter = buf;
    }

    fclose( fp );

    nvram_erase_all();
    restore_pairs();
    return 0;
}

/*
 * reset nvram values at a certain privilege level or
 * at and below a certain privilege level
 */
void
nvram_reset_level( char *buf, int privilege_level, int mode )
{
    int privilege;
    char *name, *tmp_str, *tmp_name;

    for ( name = buf; *name; )
    {
        tmp_name = name + strlen( name ) + 1;
        tmp_str = strchr( name, '=' );

        if ( tmp_str != NULL )
        {
            *tmp_str = '\0';
        }

        privilege = check_pvalue_privilege_flag( name );

        if ( privilege_level == privilege )
        {
            _nvram_set( name, NULL, DO_NOT_VALIDATE_FLAG );
        }
        else if ( privilege_level > privilege && mode )
        {
            _nvram_set( name, NULL, DO_NOT_VALIDATE_FLAG );
        }

        name = tmp_name;
    }
}

/*
Determine if a pvalue has a notificatiion group
*/
int nvram_has_notification_group( const char* name, const unsigned int group )
{
    if ( group < MAX_MSG_GROUP_INTEGER_ID )
    {
        return pvalue_has_notification_group( name, group );
    }
    else
    {
        printf( "Error: %s input error\n", __FUNCTION__ );
        return 0;
    }
}

/*
Determine if a pvalue has a notificatiion group
*/
int nvram_has_notification_group_str( const char* name, const char * group )
{
    unsigned int group_num = strtoul( group, NULL, 0 );

    return nvram_has_notification_group( name, group_num );
}

/*
Add pvalue change notification group by number
*/
int nvram_add_notification_group( const char* name, const char * group )
{
    unsigned int group_num = strtoul( group, NULL, 0 );

    if ( group_num < MAX_MSG_GROUP_INTEGER_ID )
    {
        return add_pvalue_notification_group( name, group_num );
    }

    printf( "Error: %s input error\n", __FUNCTION__ );
    return -1;
}

/*
Remove pvalue change notification group by number
*/
int nvram_remove_notification_group( const char* name, const char * group )
{
    unsigned int group_num = strtoul( group, NULL, 0 );

    if ( group_num < MAX_MSG_GROUP_INTEGER_ID )
    {
        return remove_pvalue_notification_group( name, group_num );
    }

    printf( "Error: %s input error\n", __FUNCTION__ );
    return -1;
}

/*
Add pvalue change notification groups by mask
*/
int nvram_set_notification_group_bitmask( const char* name, const char* groups )
{
    unsigned int groups_mask;

    if ( 0 == str_to_group_mask( groups, &groups_mask ) )
    {
        return set_pvalue_notification_group_bitmask( name, groups_mask );
    }

    printf( "Error: %s input error\n", __FUNCTION__ );
    return -1;
}

/*
Remove pvalue change notification groups by mask
*/
int nvram_clear_notification_group_bitmask( const char* name, const char* groups )
{
    unsigned int groups_mask;

    if ( 0 == str_to_group_mask( groups, &groups_mask ) )
    {
        return clear_pvalue_notification_group_bitmask( name, groups_mask );
    }

    printf( "Error: %s input error\n", __FUNCTION__ );
    return -1;
}

/*
Set pvalue change notification flag
*/
int nvram_set_notification( const char* name )
{
    return set_pvalue_notification_group_bitmask( name, MSG_GROUP_ID_GS_ATA | MSG_GROUP_ID_CPE );
}

/*
Unset pvalue change notification flag
*/
int nvram_unset_notification( const char* name )
{
    return clear_pvalue_notification_group_bitmask( name, MSG_GROUP_ID_GS_ATA | MSG_GROUP_ID_CPE );
}

/*
Check pvalue change notification flag
*/
int nvram_check_notification( const char* name )
{
    unsigned int mask = 0;
    get_pvalue_notification_group_mask( name, &mask );
    return mask;
}

/*
Set pvalue change privilege flag
*/
int nvram_set_privilege( const char* name, int val )
{
    return set_pvalue_privilege_flag( name, val );
}

/*
Check pvalue check privilege flag
*/
int nvram_check_privilege( const char* name )
{
    return check_pvalue_privilege_flag( name );
}

/*
Check if pvalue is a password
*/
int nvram_check_password( const char* name )
{
    if ( ( check_datatype( name ) != PASSWORD )
            && ( check_datatype( name ) != CERT )
            && ( check_datatype( name ) != DEVICEPASSWORD )
        )
    {
        return 0;
    }

    return 1;
}

/*
Check reboot flag
*/
int nvram_check_reboot( const char* name )
{
    return check_pvalue_reboot_flag( name );
}

/*
Check lock flag
*/
int nvram_check_lock( const char* name )
{
    int ret = check_pvalue_lock_flag( name );
    if ( ret < 0 )
    {
        return 0;
    }
    return ret;
}

/*
Check if pvalue matches datatype
*/
int nvram_match_datatype( const char* name, const char* type )
{
    int datatype = check_datatype( name );
    const char *str = get_str_datatype( datatype );

    if ( str == NULL )
    {
        return 0;
    }

    if ( strcmp( str, type ) )
    {
        return 0;
    }

    return 1;
}

/*
Get the datatype of pvalue
*/
int nvram_check_datatype( const char* name )
{
    return check_datatype( name );
}

/*
Get proper name of pvalue
*/
int nvram_get_name( const char* name, char *buffer, unsigned int size )
{
    return pvalue_name( name, buffer, size );
}

/*
Get default of pvalue
*/
int nvram_get_default( const char* name, char *buffer, unsigned int size )
{
    return pvalue_default( name, buffer, size );
}

/*
Get alias of pvalue
*/
int nvram_get_alias( const char* name, char *buffer, unsigned int size )
{
    return pvalue_alias( name, buffer, size );
}

/*
Get help for pvalue
*/
int nvram_get_help( const char* name, char *buffer, unsigned int size )
{
    return pvalue_help( name, buffer, size );
}

/*
Get scripts of pvalue
*/
int nvram_get_scripts( const char* name, char *buffer, unsigned int size )
{
    return pvalue_scripts( name, buffer, size );
}


/*
Get limits of pvalue
*/
int nvram_get_limits( const char* name, char *buffer, unsigned int size )
{
    return pvalue_desc( name, buffer, size );
}
