#ifndef PVALUE_VALIDATION_LIB_H
#define PVALUE_VALIDATION_LIB_H

#ifdef __cplusplus
extern "C" {
#endif

#include "datatypes.h"

#define NOTFOUND -1
#define SOCKETERROR -2
#define SERVERERROR -3

/*
 * validate_pvalue
 * validates value for given pvalue using pvalue_validation_server
 * @attrib value - the value to be validated
 * @attrib name - the pvalue name/alias that the value is for
 * @return - 1 if invalid, 0 if valid, -1 if pvalue not found, -2 on socket error, -3 on server error
 */
int validate_pvalue( const char *name, const char *value );

/*
 * pvalue_name
 * returns the proper name (non-alias) for a given pvalue  (used to translate alias into name)
 * @attrib name - the pvalue name/alias
 * @attrib ret_str - a buffer to write the pvalue name to (will be empty string if not found)
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int pvalue_name( const char *name, char *ret_str, unsigned int size );

/*
 * pvalue_default
 * returns the default value for a given pvalue
 * @attrib name - the pvalue name/alias
 * @attrib ret_str - a buffer to write the result to
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int pvalue_default( const char *name, char *ret_str, unsigned int size );

/*
 * pvalue_alias
 * returns the alias for a given pvalue (used to translate name into alias)
 * @attrib name - the pvalue name/alias
 * @attrib ret_str - a buffer to write the pvalue alias to
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int pvalue_alias( const char *name, char *ret_str, unsigned int size );

/*
 * pvalue_scripts
 * gets script list for pvalue from pvalue_validation_server
 * @attrib name - the pvalue name/alias
 * @attrib ret_str - a buffer to write the result to
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int pvalue_scripts( const char *name, char *ret_str, unsigned int size );

/*
 * pvalue_desc
 * gets description of pvalue from pvalue_validation_server
 * @attrib name - the pvalue name/alias
 * @attrib ret_str - a buffer to write the result to
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int pvalue_desc( const char *name, char *ret_str, unsigned int size );

/*
 * pvalue_desc
 * gets base of numrical pvalue
 * @attrib name - the pvalue name/alias
 * @return - the base of the pvalue, -1 on failure
 */
int pvalue_base( const char *name );

/*
 * set_pvalue_notification_group
 * Adds notification groups for a pvalue usning a mask
 * if pvalue is not in the list then add it to the list.
 * @attrib name - the pvalue name/alias
 * @attrib group_mask - A mask of the groups that should be added or removed
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int set_pvalue_notification_group_bitmask( const char *name, unsigned int group_mask );

/*
 * clear_pvalue_notification_group
 * removes notification groups for a pvalue using a mask
 * if pvalue is not in the list then add it to the list.
 * @attrib name - the pvalue name/alias
 * @attrib group_mask - A mask of the groups that should be added or removed
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int clear_pvalue_notification_group_bitmask( const char *name, unsigned int group_mask );

/*
 * add_pvalue_notification_group
 * Adds a notification group to a pvalue's list of groups to notify
 * @attrib name - the name/alias for the pvalue
 * @attrib group - A notification group to add to the pvalue
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int add_pvalue_notification_group( const char *name, unsigned int group );

/*
 * remove_pvalue_notification_groups
 * Removes a notification group from a pvalue's list of groups to notify
 * @attrib name - the name/alias for the pvalue
 * @attrib group - A notification group to add to the pvalue
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int remove_pvalue_notification_group( const char *name, unsigned int group );

/*
 * get_pvalue_notification_groups
 * Get the notification group mask for a given pvalue
 * @attrib name - the pvalue name/alias
 * @attrib group_mask - a pointer to an unsigned integer to hold the output mask
 * @return - 0 on success, -1 on invalid input, -2 on socket error, -3 on server error
 */
int get_pvalue_notification_group_mask( const char *name, unsigned int *group_mask );

/*
 * pvalue_has_notification_group
 * Gets a boolean value for whether or not a pvalue has a given notification group.
 * @attrib name - the pvalue name/alias
 * @attrib group - the notification group to check the pvalue for
 * @Return - 1 if pvalue exists and has group, 0 otherwise
 */
int pvalue_has_notification_group( const char *name, unsigned int group );

/*
 * check_pvalue_reboot_flag
 * Check the reboot flag of pvalue
 * @attrib name - the pvalue name/alias
 * @return - the reboot flag on success (0 or 1), -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int check_pvalue_reboot_flag( const char *name );

/*
 * check_pvalue_lock_flag
 * Check the lock flag of pvalue
 * @attrib name - the pvalue name/alias
 * @return - the lock flag on success (0 or greater than 0), -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int check_pvalue_lock_flag( const char *name );

/*
 * set_pvalue_privilege_flag
 * sets privilege flag of pvalue
 * @attrib name - the pvalue name/alias
 * @attrib value - the value to set the flag to (0 or 1)
 * @return - 0 on success, -1 if not found and could not add pvalue,
 *           -2 on socket error, -3 on server error
 */
int set_pvalue_privilege_flag( const char *name, int value );

/*
 * check_pvalue_privilege_flag
 * Check the privilege flag of pvalue
 * @attrib name - the pvalue name/alias
 * @return - the script flag on success (0 or 1), -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int check_pvalue_privilege_flag( const char *name );

/*
 * check_datatype
 * Check the data type of pvalue
 * @attrib name - the pvalue name/alias
 * @return - catagory enumeration on success, -1 on invalid input,
 *           -2 on socket error, -3 on server error
 */
int check_datatype( const char *name );

/*
 * check_datatype
 * return string version of datatype
 * @attrib name - the pvalue name/alias
 * @return - pointer to datatype string on success, NULL on failure
 */
const char* get_str_datatype( enum category );

/*
 * return catagory enumeration for datatype string
 * @attrib str - datatype string
 * @return: the enumeration:
 * "number"        -> NUMBER (0)
 * "utf8string"    -> UTF8_STRING (1)
 * "string"        -> STRING (2)
 * "password"      -> PASSWORD (3)
 * "address"       -> ADDRESS (4)
 * "ipaddress"     -> IPADDRESS (5)
 * "serveraddress" -> SERVERADDRESS (6)
 * "netmask"       -> NETMASK (7)
 * "url"           -> URL (8)
 * "cert"          -> CERTIFICATE (9)
 * "blank"         -> BLANK (10)
 * "device_password"-> DEVICEPASSWORD (11)
 * returns -1 for invalid strings
 */
enum category get_cat_datatype( char *str );

/*
 * pvalue_help
 * Fetches a help string giving information about the requested pvalue
 * @attrib name - the name/alias for the pvalue
 * @attrib ret_str - a buffer to save the help string to
 * @attrib size - the size of the provided buffer
 * @return - 0 on success, -2 on socket error
 */
int pvalue_help( const char *name, char *ret_str, unsigned int size );

/*
 * str_to_group_mask
 * Converts a base 10 digit string into a group mask
 * @attrib groups - a base 10 string containing a group mask
 * @attrib group_mask - a pointer to an unsigned integer to hold the output mask
 * @return - 0 on success, -1 on input error
 */
int str_to_group_mask( const char* groups_str, unsigned int* groups );

#ifdef __cplusplus
}
#endif

#endif
