#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include "datatypes.h"
#include "xml_pvalues.h"
#include "xml_parser.h"

/*from kernel nvram code*/
#define MARK_TEMP 1
#define MARK_L1LOCKED 2
#define MARK_L2LOCKED 3

int sock;
const char *socketname;
pvalue_xml *list_head = NULL ;
int loop = 1;

void process_query( int );

void clean_up( int sig, siginfo_t *siginfo, void *context )
{
    printf( "Termination Signal Received From: PID: %ld, UID: %ld\n", ( long )siginfo->si_pid, ( long )siginfo->si_uid );
    loop = 0;
    close( sock );
    unlink( socketname );
    pvalue_xml_tree_free( &list_head );
    exit( 0 );
}

char *strip_temp_char( char *str )
{
    if ( str == NULL )
    {
        printf( "Error: strip_temp_char called with NULL pvalue name\n" );
        return NULL;
    }

    if ( str[0] == ':' || str[0] == MARK_TEMP ||
            str[0] == MARK_L1LOCKED || str[0] == MARK_L2LOCKED )
    {
        str = &( str[1] );
    }

    return str;
}

int main( int argc, char **argv )
{
    int msgsock;
    struct sockaddr_un server;
    struct sigaction act;
    const char *filename;

    if ( argc < 2 )
    {
        socketname = "/tmp/.pvalue_validation";
    }
    else
    {
        socketname = argv[1];
    }

    if ( argc < 3 )
    {
        filename = "/etc/pvalues.xml";
    }
    else
    {
        filename = argv[2];
    }

    memset( &act, 0, sizeof( act ) );
    act.sa_sigaction = &clean_up;
    act.sa_flags = SA_SIGINFO;

    if ( sigaction( SIGINT, &act, NULL ) < 0 )
    {
        perror( "sigaction" );
        exit( 1 );
    }

    if ( sigaction( SIGTERM, &act, NULL ) < 0 )
    {
        perror( "sigaction" );
        exit( 1 );
    }

    signal(SIGPIPE, SIG_IGN);

    calculate_int_limits();
    parse_xml_limits_file( &list_head, filename );
    sock = socket( AF_UNIX, SOCK_STREAM, 0 );

    if ( sock < 0 )
    {
        perror( "opening stream socket" );
        exit( 1 );
    }

    server.sun_family = AF_UNIX;
    strcpy( server.sun_path, socketname );

    if ( bind( sock, ( struct sockaddr * ) &server, sizeof( struct sockaddr_un ) ) )
    {
        perror( "binding stream socket" );
        exit( 1 );
    }

    printf( "Pvalue validation server started. Socket name is %s\n", server.sun_path );
    listen( sock, 5 );

    while ( loop )
    {
        msgsock = accept( sock, 0, 0 );

        if ( msgsock == -1 )
        {
            perror( "accept connection" );
        }
        else
        {
            process_query( msgsock );
        }
    }

    close( sock );
    unlink( socketname );
    pvalue_xml_tree_free( &list_head );
    return 0;
}

void process_query( int msgsock )
{
    int rval, ret_val;
    char buf[BUFSIZE], *command, *name, *value,
    ret_str[BUFSIZE];
    char *pSave;

    memset( buf, 0x0, sizeof( buf ) );
    memset( ret_str, 0x0, sizeof( ret_str ) );

    if ( ( rval = read( msgsock, buf, BUFSIZE ) ) < 0 )
    {
        perror( "reading stream message" );
    }
    else if ( rval > 0 )
    {
        command = strtok_r( buf, " ", &pSave );
        name = strtok_r( NULL, " ", &pSave );
        value = strtok_r( NULL, "\0", &pSave );

        if ( !command )
        {
            sprintf( ret_str, "ERROR" );
        }
        else if ( !name )
        {
            if ( !strncmp( command, "DEFAULT", 8 ) ||
                    !strncmp( command, "DESC", 5 ) ||
                    !strncmp( command, "HELP", 5 ) ||
                    !strncmp( command, "ALIAS", 6 ) ||
                    !strncmp( command, "PROPERNAME", 11 )
					)
            {
                ret_str[0] = 0;
            }
            else
            {
                sprintf( ret_str, "-1" );
            }
        }
        else if ( !strncmp( command, "DEFAULT", 8 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            char *tmp_str = pvalue_xml_tree_default( &list_head, atoi( name ) );
#else
            char *tmp_str = pvalue_xml_tree_default( &list_head, hash( name ) );
#endif

            if ( tmp_str )
            {
                if ( strlen( tmp_str ) )
                {
                    sprintf( ret_str, "%s", tmp_str );
                }
                else
                {
                    sprintf( ret_str, "NOTFOUND" );
                }
            }
            else
            {
                sprintf( ret_str, "NOTFOUND" );
            }
        }
        else if ( !strncmp( command, "DESC", 5 ) )
        {
            size_t len = 0;
            char *tmp_str = ret_str;
            char tmp_buf[128] = {0};
            int number = 0;

            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            number = pvalue_xml_tree_propername( &list_head, atoi( name ), tmp_buf, 128 );
#else
            number = pvalue_xml_tree_propername( &list_head, hash( name ), tmp_buf, 128 );
#endif

            if ( number > 0 && strlen( tmp_buf ) )
            {
                name = tmp_buf;
            }

            len = sprintf( ret_str, "P%s ", name );
            tmp_str += len;
#ifdef DESKTOP_DEBUG
            pvalue_xml_tree_description( &list_head, atoi( name ), tmp_str, BUFSIZE - len );
#else
            pvalue_xml_tree_description( &list_head, hash( name ), tmp_str, BUFSIZE - len );
#endif

            if ( strlen( ret_str ) == len )
            {
                ret_str[0] = 0;
            }
        }
        else if ( !strncmp( command, "HELP", 5 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            char *tmp_str = ( char * )pvalue_xml_tree_help( &list_head, atoi( name ) );
#else
            char *tmp_str = ( char * )pvalue_xml_tree_help( &list_head, hash( name ) );
#endif

            if ( tmp_str )
            {
                if ( strlen( tmp_str ) )
                {
                    sprintf( ret_str, "%s", tmp_str );
                }
            }
        }
        else if ( !strncmp( command, "ALIAS", 6 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            char *tmp_str = pvalue_xml_tree_alias( &list_head, atoi( name ) );
#else
            char *tmp_str = pvalue_xml_tree_alias( &list_head, hash( name ) );
#endif

            if ( tmp_str )
            {
                if ( strlen( tmp_str ) )
                {
                    sprintf( ret_str, "%s", tmp_str );
                }
            }
        }
        else if ( !strncmp( command, "BASE", 5 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            int base = pvalue_xml_tree_get_base( &list_head, atoi( name ) );
#else
            int base = pvalue_xml_tree_get_base( &list_head, hash( name ) );
#endif
            sprintf( ret_str, "%i", base );
        }
        else if ( !strncmp( command, "GROUPS", 7 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            unsigned int tempnum = pvalue_xml_tree_groups( &list_head, atoi( name ) );
#else
            unsigned int tempnum = pvalue_xml_tree_groups( &list_head, hash( name ) );
#endif
            sprintf( ret_str, "%u", tempnum );
        }
        else if ( !strncmp( command, "PROPERNAME", 11 ) )
        {
            name = strip_temp_char( name );
            char tmp_buf[128] = {0};
            int number = 0;
#ifdef DESKTOP_DEBUG
            number = pvalue_xml_tree_propername( &list_head, atoi( name ), tmp_buf, 128 );
#else
            number = pvalue_xml_tree_propername( &list_head, hash( name ), tmp_buf, 128 );
#endif

            if ( number > 0 )
            {
                if ( strlen( tmp_buf ) )
                {
                    sprintf( ret_str, "%s", tmp_buf );
                }
            }
        }
        else if ( !strncmp( command, "SCRIPTS", 8 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            pvalue_xml_tree_scripts( &list_head, atoi( name ), ret_str, BUFSIZE );
#else
            pvalue_xml_tree_scripts( &list_head, hash( name ), ret_str, BUFSIZE );
#endif
        }
        else  if ( !strncmp( command, "VALIDATE", 9 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            ret_val = pvalue_xml_tree_validate( &list_head, atoi( name ), value );
#else
            ret_val = pvalue_xml_tree_validate( &list_head, hash( name ), value );
#endif
            sprintf( ret_str, "%d", ret_val );
        }
        else  if ( !strncmp( command, "ADDORREMGROUPS", 15 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            unsigned int groups = pvalue_xml_tree_groups( &list_head, atoi( name ) );
#else
            unsigned int groups = pvalue_xml_tree_groups( &list_head, hash( name ) );
#endif
            unsigned long long tmp_long;
            unsigned int groups_to_change = 0;
            int add = 0;
            char * endptr;
            char * tmpptr;

            if ( !value )
            {
                sprintf( ret_str, "-1" );
            }
            else
            {
                tmp_long = strtoull( value, &endptr, 0 );

                if ( endptr[0] == '\0' || tmp_long > UINT_MAX || ( endptr[0] >= '0' && endptr[0] <= '9' ) )
                {
                    sprintf( ret_str, "-1" );
                }
                else
                {

                    groups_to_change = tmp_long;
                    tmpptr = endptr;

                    tmp_long = strtoul( endptr, &endptr, 0 );

                    if ( tmp_long > 1 || endptr == tmpptr )
                    {
                        sprintf( ret_str, "-1" );
                    }
                    else
                    {

                        add = tmp_long;

                        if ( add )
                        {
                            groups |= groups_to_change;
                        }
                        else
                        {
                            groups &= ~( groups_to_change );
                        }

#ifdef DESKTOP_DEBUG
                        add = pvalue_set_notification_groups( &list_head, atoi( name ), groups );
#else
                        add = pvalue_set_notification_groups( &list_head, hash( name ), groups );
#endif

                        sprintf( ret_str, "%d", add );
                    }
                }
            }
        }
        else  if ( !strncmp( command, "SETPRIVILEGEFLAG", 17 ) )
        {
            name = strip_temp_char( name );

            if ( !value )
            {
                ret_val = 0;
            }
            else
            {
                ret_val = atoi( value );
#ifdef DESKTOP_DEBUG
                printf( "%d name %s %d\n", __LINE__, name, atoi( name ) );
                sprintf( ret_str, "%d", pvalue_set_privilege_flag( &list_head, atoi( name ), ret_val ) );
#else
                sprintf( ret_str, "%d", pvalue_set_privilege_flag( &list_head, hash( name ), ret_val ) );
#endif
            }
        }
        else  if ( !strncmp( command, "PRIVILEGEFLAG", 14 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            printf( "%d name %s %d\n", __LINE__, name, atoi( name ) );
            sprintf( ret_str, "%d", pvalue_check_privilege_flag( &list_head, atoi( name ) ) );
#else
            sprintf( ret_str, "%d", pvalue_check_privilege_flag( &list_head, hash( name ) ) );
#endif
        }
        else  if ( !strncmp( command, "REBOOTFLAG", 11 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            printf( "%d name %s %d\n", __LINE__, name, atoi( name ) );
            sprintf( ret_str, "%d", pvalue_check_reboot_flag( &list_head, atoi( name ) ) );
#else
            sprintf( ret_str, "%d", pvalue_check_reboot_flag( &list_head, hash( name ) ) );
#endif
        }
        else  if ( !strncmp( command, "CATEGORY", 9 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            printf( "%d name %s %d\n", __LINE__, name, atoi( name ) );
            sprintf( ret_str, "%d", pvalue_check_category( &list_head, atoi( name ) ) );
#else
            sprintf( ret_str, "%d", pvalue_check_category( &list_head, hash( name ) ) );
#endif
        }
        else  if ( !strncmp( command, "LOCKFLAG", 9 ) )
        {
            name = strip_temp_char( name );
#ifdef DESKTOP_DEBUG
            printf( "%d name %s %d\n", __LINE__, name, atoi( name ) );
            sprintf( ret_str, "%d", pvalue_check_lock_flag( &list_head, atoi( name ) ) );
#else
            sprintf( ret_str, "%d", pvalue_check_lock_flag( &list_head, hash( name ) ) );
#endif
        }
        else
        {
            sprintf( ret_str, "ERROR" );
        }

        write( msgsock, ret_str, strlen( ret_str ) );

    }

    close( msgsock );
}
