#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>  /* Add IPv6 support */

#include "validate_functions.h"


/*
*check if string contains numeric values
*/
int is_numeric( const char *value )
{
    int i = 1;

    if ( value[0] != '-' && ( value[0] < 0x30 || value[0] > 0x39 ) )
    {
        return 1;
    }

    while ( value[i] )
    {
        if ( value[i] < 0x30 || value[i] > 0x39 )
        {
            return 1;
        }

        i++;
    }

    return 0;
}

/*
 * RFC 3696#section-2
     2.  Restrictions on domain (DNS) names

     The authoritative definitions of the format and syntax of domain
     names appear in RFCs 1035 [RFC1035], 1123 [RFC1123], and 2181
     [RFC2181].

     Any characters, or combination of bits (as octets), are permitted in
     DNS names.  However, there is a preferred form that is required by
     most applications.  This preferred form has been the only one
     permitted in the names of top-level domains, or TLDs.  In general, it
     is also the only form permitted in most second-level names registered
     in TLDs, although some names that are normally not seen by users obey
     other rules.  It derives from the original ARPANET rules for the
     naming of hosts (i.e., the "hostname" rule) and is perhaps better
     described as the "LDH rule", after the characters that it permits.
     The LDH rule, as updated, provides that the labels (words or strings
     separated by periods) that make up a domain name must consist of only
     the ASCII [ASCII] alphabetic and numeric characters, plus the hyphen.
     No other symbols or punctuation characters are permitted, nor is
     blank space.  If the hyphen is used, it is not permitted to appear at
     either the beginning or end of a label.  There is an additional rule
     that essentially requires that top-level domain names not be all-
     numeric.          

     A DNS label may be no more than 63 octets long.

     A complete, fully-qualified, domain name must not exceed 255 octets.

 * From hostname manual,
     Each element of the hostname must be from 1 to 63 characters long
     and the entire hostname, including the dots, can be at most 253
     characters long.  Valid characters for hostnames are ASCII(7)
     letters from a to z, the digits from 0 to 9, and the hyphen (-).
     A hostname may not start with a hyphen.
 */
int check_label( const char *str ) 
{
    int i = 0;

    if ( '\0' == str[0] )
    {
        return -1;
    }
    while( '\0' != str[i] )
    {
        if ( 0 == i && '-' == str[0] ) /*cannot start with hyphen*/
        {
            return -1;
        }
        if ( ( ( str[i] < '0' ) || ( str[i] > '9' ) ) &&
             ( ( str[i] < 'A' ) || ( str[i] > 'Z' ) ) &&
             ( ( str[i] < 'a' ) || ( str[i] > 'z' ) ) &&
             ( str[i] != '-' ) &&
             ( str[i] != '_' )    /* bug 93511 */
        )
        {
            return -1;
        }
        if ( 62 < i ) /*max allowable length is 63*/
        {
            return -1;
        }
        i++;
    }
    if ( '-' == str[i-1] ) /*cannot end with hyphen*/
    {
        return -1;
    }
    return 0;
}
/*
Checks if the address is valid
*/
int check_address( char *address_str )
{
    char temp_addr[BUFFER_LENGTH] = { 0 }, *addr1 = NULL, *addr2 = NULL, *addr3 = NULL, *addr4 = NULL;
    int i = 0;
    char *pSave = NULL;
    size_t str_len = 0;

    if ( address_str[0] == TERMINAL_CHAR )
    {
        return INVALID_ADDRESS ;    /*return invalid address*/
    }

    /*copy to temporary buffer */
    strncpy( temp_addr, address_str, BUFFER_LENGTH - 1 );
    temp_addr[BUFFER_LENGTH - 1] = '\0';
    /* take out four tokens */
    addr1 = strtok_r( temp_addr, ".", &pSave );
    addr2 = strtok_r( NULL, ".", &pSave );
    addr3 = strtok_r( NULL, ".", &pSave );
    addr4 = strtok_r( NULL, ".", &pSave );

    /* if all four tokens exists and they are all numeric*/
    /*must be an IP address*/
    if ( addr1 && addr2 && addr3 && addr4 &&
            !is_numeric( addr1 ) && !is_numeric( addr2 ) &&
            !is_numeric( addr3 ) && !is_numeric( addr4 )
       )
    {
        /*if all values are between 0 and 255 except the first one */
        if ( ( atoi( addr1 ) < 255 ) && ( atoi( addr1 ) > -1 ) &&
                ( atoi( addr2 ) <= 255 ) && ( atoi( addr2 ) > -1 ) &&
                ( atoi( addr3 ) <= 255 ) && ( atoi( addr3 ) > -1 ) &&
                ( atoi( addr4 ) <= 255 ) && ( atoi( addr4 ) > -1 ) /*Bug 29294
            sip server and outbound proxy can not config the host address as like x.x.x.0*/
           )
        {
            return 0 ;    /*return OK*/
        }

        return INVALID_ADDRESS ; /*else return invalid address*/
    }

    /* strtok_r modified temp_addr. so we need to copy it again*/
    strncpy( temp_addr, address_str, BUFFER_LENGTH - 1 );
    str_len = strnlen( temp_addr, BUFFER_LENGTH );
    if ( '.' == temp_addr[str_len-1] )/*can have an optional dot at the end*/
    {
        temp_addr[str_len-1] = '\0';
        str_len--;
    }
    if ( 253 < str_len )/*max allowable length is 253*/
    {
        return INVALID_ADDRESS;
    }
    addr1 = temp_addr;
    addr2 = strsep( &addr1, "." );
    i = 0;
    while( NULL != addr2 )
    {
        i++;
        if ( 0 != check_label( addr2 ) )
        {
            return INVALID_ADDRESS;
        }

        addr3 = addr2;
        addr2 = strsep( &addr1, "." );
    }

    if ( 0 == i )
    {
        return INVALID_ADDRESS;
    }
    i = 0;
    while ( '\0' != addr3[i] ) /*last label cannot be only numeric*/
    {
        if ( ( addr3[i] < '0' ) || ( addr3[i]  > '9' ) ) 
        {
            return 0;
        }
        i++;
    }
    return INVALID_ADDRESS;
}

/*
Checks if the ip address is valid
*/
int check_ip_address( char *address_str )
{
    char temp_addr[256], *addr1, *addr2, *addr3, *addr4;
    char *pSave;

    if ( address_str[0] == '\0' )
    {
        return INVALID_ADDRESS ;    /*return invalid address*/
    }

    /*copy to temporary buffer */
    strncpy( temp_addr, address_str, 255 );
    temp_addr[255] = '\0';
    /* take out four tokens */
    addr1 = strtok_r( temp_addr, ".", &pSave );
    addr2 = strtok_r( NULL, ".", &pSave );
    addr3 = strtok_r( NULL, ".", &pSave );
    addr4 = strtok_r( NULL, ".", &pSave );

    /* if all four tokens exists and they are all numeric*/
    if ( addr1 && addr2 && addr3 && addr4 &&
            !is_numeric( addr1 ) && !is_numeric( addr2 ) &&
            !is_numeric( addr3 ) && !is_numeric( addr4 )
       )
    {
        /*if all values are between 0 and 255*/
        if ( ( atoi( addr1 ) < 255 ) && ( atoi( addr1 ) > -1 ) &&
                ( atoi( addr2 ) <= 255 ) && ( atoi( addr2 ) > -1 ) &&
                ( atoi( addr3 ) <= 255 ) && ( atoi( addr3 ) > -1 ) &&
                ( atoi( addr4 ) <= 255 ) && ( atoi( addr4 ) > -1 ) /*Bug 29294
            sip server and outbound proxy can not config the host address as like x.x.x.0*/
           )
        {
            return 0 ;    /*return OK*/
        }
    }
    return INVALID_ADDRESS ; /*return invalid address*/
}

/*
extracts server address into server_addr
returns INVALID_ADDRESS if invalid address detected
returns position of end of server_address in input_str
*/
int get_server_address( char *input_str, char *server_addr )
{
    int cnt = 0, j = 0;
    int slash_flg = 0;

    /*while input_str does not end and count is less than buffer length*/
    while ( ( input_str[j] != TERMINAL_CHAR ) &&
            ( cnt < ( BUFFER_LENGTH - 1 ) )
          )
    {
        /*if input_str has ended break loop*/
        if ( input_str[j] == TERMINAL_CHAR )
        {
            break;
        }

        /*if there are leading slashes ignore them and continue*/
        if ( ( input_str[j] == '/' ) && !slash_flg )
        {
            j++;
            continue;
        }
        /*if a colon is found server address has ended break loop*/
        else if ( input_str[j] == ':' )
        {
            break;
        }
        /*if a slash is found server address has ended break loop*/
        else if ( ( input_str[j] == '/' ) && slash_flg )
        {
            break;
        }

        slash_flg++;
        server_addr[cnt] = input_str[j];
        cnt++;
        j++;
    }

    /*set terminating null character */
    server_addr[cnt] = TERMINAL_CHAR;

    /*if the address is a invalid address*/
    if ( check_address( server_addr ) == INVALID_ADDRESS )
    {
        return INVALID_ADDRESS ; /*return invalid */
    }

    return j ; /*return present position */
}

/*extracts port number if it exists into port
returns INVALID_PORT if port number invalid
returns index value of the end of port number in input_str
*/
int get_port( char *input_str, char *port, int position )
{
    int cnt = 0;
    int colon_flg = position ; /*store position */
    int colon_position = position;

    /*while input_str does not end and count is less than the port buffer
                                                                    length*/
    while ( ( input_str[position] != TERMINAL_CHAR ) &&
            ( cnt < ( PORT_BUFFER_LENGTH - 1 ) )
          )
    {
        /*if input_str has ended break loop*/
        if ( input_str[position] == TERMINAL_CHAR )
        {
            break;
        }

        /*if colon has not been found ignore characters and continue*/
        if ( ( input_str[position] != ':' ) &&
                ( colon_flg != -1 )
           )
        {
            position++;
            continue;
        }
        /* if colon has been found set flag and proceed*/
        else if ( ( input_str[position] == ':' ) &&
                  ( colon_flg != -1 )
                )
        {
            colon_flg = -1;
            position++;
        }

        /*if non numeric character has been found port has ended break loop*/
        if ( ( input_str[position] < 48 ) ||
                ( input_str[position] > 57 )
           )
        {
            break;
        }

        port[cnt] = input_str[position];
        cnt++;
        position++;
    }

    /*set terminating null character */
    port[cnt] = TERMINAL_CHAR;

    /*if colon was found port exists*/
    if ( colon_flg == -1 )
    {
        /* port number cannot be greater than MAX_PORT_NUMBER*/
        if ( atoi( port ) > MAX_PORT_NUMBER )
        {
            return INVALID_PORT ;    /*return invalid port */
        }

        /*port cannot be empty with a colon*/
        if ( position == ( colon_position + 1 ) )
        {
            return INVALID_PORT;
        }

        return position ; /*return present position */
    }

    return colon_flg ;    /*else return old position */
}

/*
extracts path from input_str into path
*/
int get_path( char *input_str, char *path, int position )
{
    int cnt = 0;
    int slash_flg = 0;

    /*while input_str does not end and count is less than the buffer length*/
    while ( ( input_str[position] != TERMINAL_CHAR ) &&
            ( cnt < ( BUFFER_LENGTH - 1 ) )
          )
    {
        /*if input_str has ended break loop*/
        if ( input_str[position] == TERMINAL_CHAR )
        {
            break;
        }

        /*if there are leading slashes ignore them and continue*/
        if ( ( input_str[position] != '/' ) && !slash_flg )
        {
            position++;
            continue;
        }

        /*if there are more than one slash in succession ignore them and
                                                               continue*/
        if ( ( input_str[position] == '/' ) &&
                ( path[cnt - 1] == '/' )
           )
        {
            position++;
            continue;
        }

        slash_flg++;
        path[cnt] = input_str[position];
        cnt++;
        position++;
    }

    /*set terminating null character */
    path[cnt] = TERMINAL_CHAR;

    if ( path[0] == TERMINAL_CHAR )
    {
        return INVALID_PATH;
    }

    return SUCCESS;
}

int check_server_address( char *input_str )
{
    char server_addr[256], port[6];
    int return_value;

    /* get the server address */
    return_value = get_server_address( input_str, server_addr );

    /* checking if the server address is valid */
    if ( return_value < 0 )
    {
        return return_value;
    }

    /* get the port if it exists and is valid */
    return_value = get_port( input_str, port, return_value );

    if ( return_value < 0 )
    {
        return return_value;
    }

    /*Added check to prevent non-ascii characters*/
    while ( return_value < 256 && input_str[return_value] != TERMINAL_CHAR )
    {
        if ( input_str[return_value] < 32 )
        {
            return INVALID_ADDRESS;
        }

        return_value++;
    }

    return SUCCESS;
}

/*
Checks if the netmask is valid
*/
int check_netmask( char *address_str )
{
    char temp_addr[256], *addr1, *addr2, *addr3, *addr4;
    char *pSave;

    if ( address_str[0] == '\0' )
    {
        return -1 ;    /*return invalid address*/
    }

    /*copy to temporary buffer */
    strncpy( temp_addr, address_str, 255 );
    temp_addr[255] = '\0';
    /* take out four tokens */
    addr1 = strtok_r( temp_addr, ".", &pSave );
    addr2 = strtok_r( NULL, ".", &pSave );
    addr3 = strtok_r( NULL, ".", &pSave );
    addr4 = strtok_r( NULL, ".", &pSave );

    /* if all four tokens exists and they are all numeric*/
    if ( addr1 && addr2 && addr3 && addr4 &&
            !is_numeric( addr1 ) && !is_numeric( addr2 ) &&
            !is_numeric( addr3 ) && !is_numeric( addr4 )
       )
    {
        /*if all values are between 0 and 255*/
        if ( ( atoi( addr1 ) <= 255 ) && ( atoi( addr1 ) > -1 ) &&
                ( atoi( addr2 ) <= 255 ) && ( atoi( addr2 ) > -1 ) &&
                ( atoi( addr3 ) <= 255 ) && ( atoi( addr3 ) > -1 ) &&
                ( atoi( addr4 ) <= 255 ) && ( atoi( addr4 ) > -1 )
           )
        {
            return 0 ;    /*return OK*/
        }
    }

    return INVALID_ADDRESS ; /*return invalid address*/
}

/*
checks if a url is valid(assumed http or https)
*/
int check_url( char *input_str )
{
    int cnt = 0, colon_flg = 0;

    while ( input_str[cnt] != TERMINAL_CHAR &&
            cnt < ( BUFFER_LENGTH - 1 ) )
    {
        if ( input_str[cnt] == ':' && !colon_flg )
        {
            colon_flg = cnt;

            if ( !cnt && colon_flg )
            {
                return INVALID_ADDRESS;
            }

            break;
        }

        if ( ( ( input_str[cnt] >= '0' ) && ( input_str[cnt]  <= '9' ) ) ||
                ( ( input_str[cnt] >= 'A' ) && ( input_str[cnt]  <= 'Z' ) ) ||
                ( ( input_str[cnt] >= 'a' ) && ( input_str[cnt]  <= 'z' ) ) ||
                ( input_str[cnt] == '-' ) ||
                ( input_str[cnt] == '.' ) ||
                ( input_str[cnt] == '+' )
           )
        {
            cnt++;
            continue;
        }

        return INVALID_ADDRESS;
    }

    if ( !colon_flg )
    {
        return INVALID_ADDRESS;
    }

    return check_server_address( input_str + colon_flg + 1 );
}

/*
checks if string is utf8
*/
int validate_utf8( const unsigned char *input_str, int debug )
{
    int i = 0, count = 0, cnt;
    unsigned long int val;

    if ( !input_str[0] )
    {
        return 0;
    }

    while ( input_str[i] )
    {
        /*byte value is 0xxx xxxx - ascii*/
        if ( !( input_str[i] & 0x80 ) )
        {
            i++;
            continue;
        }

        val = 0;

        if ( debug )
        {
            printf( "%02x ", input_str[i] );
        }

        /*The following values can never appear in an utf8 string - RFC 3629*/
        if ( input_str[i] == 0xc0 || input_str[i] == 0xc1 ||
                ( input_str[i] >= 0xf5 ) ) //&& input_str[i] <= 0xff ) )
        {
            if ( debug )
            {
                printf( "invalid values( RFC 3629 )\n" );
            }

            return 1;
        }

        /*byte value is 1111 0xxx*/
        if ( ( input_str[i] & 0xf8 ) == 0xf0 )
        {
            count = 4;
            val = input_str[i] & 0x7;
        }
        /*byte value is 1110 xxxx*/
        else if ( ( input_str[i] & 0xf0 ) ==  0xe0 )
        {
            count = 3;
            val = input_str[i] & 0xf;
        }
        /*byte value is 110x xxxx*/
        else if ( ( input_str[i] & 0xe0 ) == 0xc0 )
        {
            count = 2;
            val = input_str[i] & 0x1f;
        }

        if ( !count )
        {
            if ( debug )
            {
                printf( "no count\n" );
            }

            return 1;
        }

        cnt = 1;

        while ( cnt < count )
        {
            val <<= 6;

            if ( debug )
            {
                printf( "|%02x ", input_str[i + cnt] );
            }

            /*string cannot be over*/
            if ( !input_str[i + cnt] )
            {
                if ( debug )
                {
                    printf( "invalid length\n" );
                }

                return 1;
            }

            /*byte value should be 10xx xxxx*/
            if ( ( ( input_str[i + cnt] & 0xc0 ) != 0x80 ) )
            {
                if ( debug )
                {
                    printf( "invalid byte value\n" );
                }

                return 1;
            }

            val |= ( input_str[i + cnt] & 0x3f );
            cnt++;
        }

        if ( debug )
        {
            printf( " val %lx ", val );
        }

        if ( count == 4 )
        {
            /*val must be between 0x10000 and 0x10ffff for four byte encoding*/
            if ( val < 0x10000 || val > 0x10ffff )
            {
                if ( debug )
                {
                    printf( "invalid val for 4 bytes\n" );
                }

                return 1;
            }
        }
        else if ( count == 3 )
        {
            /*val must be between 0x800 and 0xffff for three byte encoding*/
            if ( val < 0x800 || val > 0xffff )
            {
                if ( debug )
                {
                    printf( "invalid val for 3 bytes\n" );
                }

                return 1;
            }
        }
        else if ( count == 2 )
        {
            /*val must be between 0x80 and 0x7ff for two byte encoding*/
            if ( val < 0x80 || val > 0x7ff )
            {
                if ( debug )
                {
                    printf( "invalid val for 2 bytes\n" );
                }

                return 1;
            }
        }

        /*utf16 surrogate halves and byte ordering used by utf16 invalid*/
        if ( ( val >= 0xd800 && val <= 0xdfff ) ||
                val == 0xfffe || val == 0xffff )
        {
            if ( debug )
            {
                printf( "invalid val ( utf16 )\n" );
            }

            return 1;
        }

        i += count;
    }

    if ( debug )
    {
        printf( "\n" );
    }

    return 0;
}

int strlen_utf8( const unsigned char *input_str, int debug )
{
    int i = 0, count = 0, length = 0;

    if ( validate_utf8( input_str, debug ) )
    {
        return -1;
    }

    if ( !input_str[i] )
    {
        return -1;
    }

    while ( input_str[i] )
    {
        /*byte value is 0xxx xxxx - ascii*/
        if ( !( input_str[i] & 0x80 ) )
        {
            i++;
            length++;
            continue;
        }

        if ( debug )
        {
            printf( "%02x ", input_str[i] );
        }

        /*byte value is 1111 0xxx*/
        if ( ( input_str[i] & 0xf8 ) == 0xf0 )
        {
            count = 4;
        }
        /*byte value is 1110 xxxx*/
        else if ( ( input_str[i] & 0xf0 ) ==  0xe0 )
        {
            count = 3;
        }
        /*byte value is 110x xxxx*/
        else if ( ( input_str[i] & 0xe0 ) == 0xc0 )
        {
            count = 2;
        }

        if ( debug )
        {
            printf( "count %d\n", count );
        }

        i += count;
        length++;
    }

    if ( debug )
    {
        printf( "\n" );
    }

    return length;
}

long get_code_point_utf8( const unsigned char *input_str, unsigned char **ret_ptr, int debug )
{
    int i = 0, count = 0, cnt;
    unsigned long int val;

    if ( validate_utf8( input_str, debug ) )
    {
        return -1;
    }

    if ( !input_str[0] )
    {
        return -1;
    }

    /*byte value is 0xxx xxxx - ascii*/
    if ( !( input_str[i] & 0x80 ) )
    {
        i++;
        return input_str[i];
    }

    val = 0;

    /*byte value is 1111 0xxx*/
    if ( ( input_str[i] & 0xf8 ) == 0xf0 )
    {
        count = 4;
        val = input_str[i] & 0x7;
    }
    /*byte value is 1110 xxxx*/
    else if ( ( input_str[i] & 0xf0 ) ==  0xe0 )
    {
        count = 3;
        val = input_str[i] & 0xf;
    }
    /*byte value is 110x xxxx*/
    else if ( ( input_str[i] & 0xe0 ) == 0xc0 )
    {
        count = 2;
        val = input_str[i] & 0x1f;
    }

    if ( debug )
    {
        printf( "count %d\n", count );
    }

    cnt = 1;

    while ( cnt < count )
    {
        val <<= 6;

        if ( debug )
        {
            printf( "|%02x ", input_str[i + cnt] );
        }

        val |= ( input_str[i + cnt] & 0x3f );

        if ( debug )
        {
            printf( "\\ %lx \\", val );
        }

        cnt++;
    }

    if ( debug )
    {
        printf( " val %lx ", val );
    }

    if ( debug )
    {
        printf( "\n" );
    }

    *ret_ptr = ( unsigned char * )( input_str + i + cnt );
    return val;
}

int validate_certificate( char* value, char is_unpkcs8 )
{
    char begin_str[] = "-----BEGIN";
    char end_str[] = "-----END";
    char pkcs8_str[] = "ENCRYPTED PRIVATE KEY";
    if ( is_unpkcs8 && NULL != strstr( value, pkcs8_str ) )
    {
        return 1;
    }
    if ( NULL != strstr( value, begin_str ) && NULL != strstr( value, end_str ) )
    {
        unsigned int begin = 0;
        unsigned int end = 0;
        size_t len = strlen( value );
        if ( 8172 < len )
        {
            return 1;
        }

        char* tmp = strstr( value, begin_str );
        if ( NULL != tmp )
        {
            begin = tmp - value;
            tmp = &value[begin] + strlen( begin_str );
            /* commented for P8472
            if ( NULL != strstr( tmp, begin_str ) )
            {
                return 1;
            }
            */
        }
        tmp = strstr( value, end_str );
        if ( NULL != tmp )
        {
            end = tmp - value;
            tmp = &value[end] + strlen( end_str );
            /* commented for P8472
            if ( NULL != strstr( tmp, end_str ) )
            {
                return 1;
            }
            */
        }
        if ( begin + strlen( begin_str ) >= end )
        {
            return 1;
        }
    }
    else
    {
        return 1;
    }
    return 0;
}

