#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "datatypes.h"
#include "xml_pvalues.h"
#include "xml_parser.h"
#include "validate_functions.h"
#include "../nvtool/nvram.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#ifdef BUILD_PRE_PARSER
void* preparse_malloc( size_t size );
void preparse_free( void* p );
char* preparse_strdup( char* p );
#define MALLOC(x) preparse_malloc(x)
#define FREE(x) preparse_free(x)
#define STRDUP(x) preparse_strdup(x)
#else
#define MALLOC(x) malloc(x)
#define FREE(x) free(x)
#define STRDUP(x) strdup(x)
#endif

#define MAX_TAG_LEN     50
#define MAX_BUFFER_LEN  2048
#define MAX_VAL_LEN     90

/*
 * copy contents of in_buffer to out_buffer upto size. 
 * Return number of copied char.
 */
int get_tag_buffer( char *in_buffer, char *out_buffer, int size )
{
    int i = 0;

    while ( i < size )
    {
        out_buffer[i] = in_buffer[i];
        if ( in_buffer[i] == '>' )
        {
            break;
        }
        i++;
    }
    if ( in_buffer[i] != '>' )
    {
        out_buffer[0] = '\0';
        i = 0;
    }
    else
    {
        out_buffer[i] = '\0';
        i++;
    }

    return i;
}

/*
 * copy contents of FILE pointer fp into buffer upto size
 * Return number of copied char
 */
int get_tag_file( FILE *fp, char *buffer, int size )
{
    int i = 0;

    while ( !feof( fp ) && i < size )
    {
        fread( buffer + i, sizeof(char), 1, fp );
        i++;
        if ( buffer[i - 1] == '>' )
        {
            break;
        }
    }
    if ( buffer[i - 1] != '>' )
    {
        buffer[0] = '\0';
    }
    else
    {
        buffer[i - 1] = '\0';
    }

    return i * sizeof( char );
}

/*
 * Start reading from file from a start tag to an end tag and store in buffer.
 * Return number of copied char
 */
int get_value_file( FILE *fp, char *buffer, int size, char *start_tag, int tag_size )
{
    int i = 0, read_count, total_count = 0;
    char end_tag[tag_size];
    read_count = get_tag_file( fp, start_tag, tag_size );

    if ( !start_tag[0] )
    {
        fprintf( stderr, "%s: Tag size is too large, read_count=%d\n", __FUNCTION__, read_count );
        fclose( fp );
        exit( 1 );
    }
    if ( start_tag[0] == '/' )
    {
        fprintf( stderr, "%s: Tag is an end tag\n", __FUNCTION__ );
        fclose( fp );
        exit( 1 );
    }

    i = 0;
    total_count += read_count;
    while ( !feof( fp ) && i < size )
    {
        fread( buffer + i, sizeof(char), 1, fp );
        total_count++;
        if ( buffer[i] == '<' )
        {
            read_count = get_tag_file( fp, end_tag, tag_size );
            if ( end_tag[0] != '/' )
            {
                fseek( fp, -read_count, SEEK_CUR );
                i++;
                continue;
            }
            else
            {
                if ( strncmp( start_tag, end_tag + 1, tag_size ) )
                {
                    fseek( fp, -read_count, SEEK_CUR );
                    i++;
                    continue;
                }
                else
                {
                    total_count += read_count;
                    break;
                }
            }
        }
        i++;
    }
    if ( strncmp( start_tag, end_tag + 1, tag_size ) )
    {
        buffer[0] = '\0';
    }
    else
    {
        buffer[i] = '\0';
    }

    return total_count * sizeof( char );
}

/*
 * Read in from buffer from a start tag to an end tag and store in another buffer.
 * Return number of copied char
 */
int get_value_buffer( char *in_buffer, char *out_buffer, int size, char *start_tag, int tag_size, int newlineflag )
{
    int i = 0, read_count, total_count = 0;
    char end_tag[tag_size];

    if ( in_buffer[0] != '<' )
    {
        fprintf( stderr, "Incorrect starting point\n" );
        exit( 1 );
    }
    read_count = get_tag_buffer( in_buffer + 1, start_tag, tag_size );

    if ( !start_tag[0] )
    {
        fprintf( stderr, "get_value_buffer(): Tag size is too large\n" );
        exit( 1 );
    }
    if ( start_tag[0] == '/' )
    {
        fprintf( stderr, "%d:Tag is an end tag %s\n", __LINE__, start_tag );
        exit( 1 );
    }
    //printf( "%d %s\n", start_tag );

    i = 0;
    total_count += read_count + 1;
    while ( i < size )
    {
        if ( in_buffer[total_count + i] == '\n' && !newlineflag )
        {
            i++;
            continue;
        }
        out_buffer[i] = in_buffer[total_count + i];
        if ( in_buffer[total_count + i] == '<' )
        {
            i++;
            read_count = get_tag_buffer( in_buffer + total_count + i, end_tag, tag_size );
            if ( !read_count )
            {
                break;
            }
            i += read_count;
            if ( end_tag[0] != '/' )
            {
                i -= read_count;
                continue;
            }
            else
            {
                if ( strncmp( start_tag, end_tag + 1, tag_size ) )
                {
                    i -= read_count;
                    continue;
                }
                else
                {
                    break;
                }
            }
        }
        i++;
    }
    total_count +=  i;
    if ( strncmp( start_tag, end_tag + 1, tag_size ) )
    {
        total_count = 0;
        out_buffer[0] = '\0';
    }
    else
    {
        i -= ( read_count + 1 );
        out_buffer[i] = '\0';
    }
    return total_count;
}

/*
 * Find contents between <tag> and </tag> inside the buffer and store in value.
 */
void find_store_tag( char *in_buffer, char *tag, char *value, int value_size, int tag_size, int buffer_size )
{
    int i = 0, count = 0;
    char buffer[buffer_size], start_tag[tag_size];
    while ( i < buffer_size )
    {
        if ( in_buffer[i] == '\0' )
            break;
        if ( in_buffer[i] == '<' )
        {
            if ( strnlen( tag, 8 ) == 7 && !strncmp( tag, "default", 7 ) )
            {
                count = get_value_buffer( in_buffer + i, buffer, value_size, start_tag, tag_size, 1 );
            }
            else
            {
                count = get_value_buffer( in_buffer + i, buffer, value_size, start_tag, tag_size, 0 );
            }
            if ( !count )
            {
                break;
            }
        }
        else
        {
            i++;
            continue;
        }
        if ( strncmp( tag, start_tag, tag_size ) )
        {
            i += count;
            continue;
        }
        else
        {
            strncpy( value, buffer, value_size - 1 );
            value[value_size - 1] = '\0';
            return;
        }
        i++;
    }
    value[0] = '\0';
}

/*
 * Find contents between <tagi> and </tagi> inside the buffer and store in value.
 */
int find_store_tags( char *in_buffer, char *tag, char *value, int buffer_size, int suffix, int tag_size, int value_size )
{
    int i = 0, count = 0;
    char buffer[buffer_size], start_tag[tag_size];
    while ( i < buffer_size )
    {
        if ( in_buffer[i] == '\0' )
        {
            break;
        }
        if ( !in_buffer[i] )
        {
            break;
        }
        if ( in_buffer[i] == '<' )
        {
            count = get_value_buffer( in_buffer + i, buffer, value_size, start_tag, tag_size, 0 );
            if ( !count )
            {
                break;
            }
        }
        else
        {
            i++;
            continue;
        }
        if ( !strncmp( tag, start_tag, strnlen( tag, tag_size ) ) && 
            atoi( start_tag + strnlen( tag, tag_size ) ) == suffix && 
            ( start_tag + strnlen( tag, tag_size ) )[0] >= 0x30 && 
            ( start_tag + strnlen( tag, tag_size ) )[0] <= 0x39 )
        {
            strncpy( value, buffer, value_size - 1 );
            value[value_size - 1] = '\0';
            return 0;
        }
        else
        {
            i += count;
            continue;
        }
        i++;
    }
    value[0] = '\0';
    return 1;
}

//parse and store number category
void store_number_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN], value1[MAX_VAL_LEN], value2[MAX_VAL_LEN], *endptr = NULL;
    long long int val;

    find_store_tag( buffer, "base", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).number.base = atoi( value );
    }
    else
    {
        ( pvalue -> t ).number.base = 10;
    }
    if ( ( pvalue -> t ).number.base < 2 || ( pvalue -> t ).number.base > 36 )
    {
        ( pvalue -> t ).number.base = 10;
    }

    find_store_tag( buffer, "min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).number.min ), val );
    }
    find_store_tag( buffer, "max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).number.max ), val );
    }
    find_store_tag( buffer, "not_min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).number.not_min ), val );
    }
    find_store_tag( buffer, "not_max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).number.not_max ), val );
    }
    i = 0;
    while ( !find_store_tags( buffer, "range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( !is_numeric( value1 ) && !is_numeric( value2 ) )
        {
            rlist_add( &( ( pvalue -> t ).number.range ), atoll( value1 ), atoll( value2 ) );
        }
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( !is_numeric( value1 ) && !is_numeric( value2 ) )
        {
            rlist_add( &( ( pvalue -> t ).number.not_range ), atoll( value1 ), atoll( value2 ) );
        }
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "mult", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        endptr = NULL;
        val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
        if ( !( *endptr ) && value[0] )
        {
            llist_add( &( ( pvalue -> t ).number.mult ), val );
        }
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_mult", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        endptr = NULL;
        val = strtoll( value, &endptr, ( pvalue -> t ).number.base );
        if ( !( *endptr ) && value[0] )
        {
            llist_add( &( ( pvalue -> t ).number.not_mult ), val );
        }
        i++;
    }
}

//parse and store utf8 string category
void store_utf8_string_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN], value1[MAX_VAL_LEN], value2[MAX_VAL_LEN], *endptr = NULL;
    long int val;

    find_store_tag( buffer, "min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, 16 );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).utf8_str.min ), val );
    }
    find_store_tag( buffer, "max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, 16 );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).utf8_str.max ), val );
    }
    find_store_tag( buffer, "not_min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, 16 );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).utf8_str.not_min ), val );
    }
    find_store_tag( buffer, "not_max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    val = strtoll( value, &endptr, 16 );
    if ( !( *endptr ) && value[0] )
    {
        llist_add( &( ( pvalue -> t ).utf8_str.not_max ), val );
    }
    i = 0;
    while ( !find_store_tags( buffer, "range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( !is_numeric( value1 ) && !is_numeric( value2 ) )
        {
            rlist_add( &( ( pvalue -> t ).utf8_str.range ), atoll( value1 ), atoll( value2 ) );
        }
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( !is_numeric( value1 ) && !is_numeric( value2 ) )
        {
            rlist_add( &( ( pvalue -> t ).utf8_str.not_range ), atoll( value1 ), atoll( value2 ) );
        }
        i++;
    }
}

//parse and store string category
void store_string_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN], value1[MAX_VAL_LEN], value2[MAX_VAL_LEN];

    find_store_tag( buffer, "minlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).str.minlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).str.minlength = -1;
    }
    find_store_tag( buffer, "maxlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).str.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).str.maxlength = -1;
    }
    find_store_tag( buffer, "min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( value[0] ) 
    {
        if ( strnlen( value, 90 ) == 1 )
        {
            cllist_add( &( ( pvalue -> t ).str.min ), value[0] );
        }
        else
        {
            cllist_add( &( ( pvalue -> t ).str.min ), strtol( value, NULL, 10 ) );
        }
    }
    find_store_tag( buffer, "max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( value[0] ) 
    {
        if ( strnlen( value, 90 ) == 1 )
        {
            cllist_add( &( ( pvalue -> t ).str.max ), value[0] );
        }
        else
        {
            cllist_add( &( ( pvalue -> t ).str.max ), strtol( value, NULL, 10 ) );
        }
    }
    find_store_tag( buffer, "not_min", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( value[0] ) 
    {
        if ( strnlen( value, 90 ) == 1 )
            cllist_add( &( ( pvalue -> t ).str.not_min ), value[0] );
        else
        {
            cllist_add( &( ( pvalue -> t ).str.not_min ), strtol( value, NULL, 10 ) );
        }
    }
    find_store_tag( buffer, "not_max", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( value[0] ) 
    {
        if ( strnlen( value, 90 ) == 1 )
        {
            cllist_add( &( ( pvalue -> t ).str.not_max ), value[0] );
        }
        else
        {
            cllist_add( &( ( pvalue -> t ).str.not_max ), strtol( value, NULL, 10 ) );
        }
    }
    i = 0;
    while ( !find_store_tags( buffer, "range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( strlen( value1 ) == 1 && strlen( value2 ) == 1 )
        {
            crlist_add( &( ( pvalue -> t ).str.range ), value1[0], value2[0] );
        }
        else if ( strlen( value1 ) > 1 && strlen( value2 ) > 1 )
        {
            crlist_add( &( ( pvalue -> t ).str.range ), strtol( value1, NULL, 10 ), strtol( value2, NULL, 10 ) );
        }
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_range", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        find_store_tag( value, "min", value1, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        find_store_tag( value, "max", value2, MAX_VAL_LEN, 4, MAX_VAL_LEN );
        if ( strlen( value1 ) == 1 && strlen( value2 ) == 1 )
        {
            crlist_add( &( ( pvalue -> t ).str.not_range ), value1[0], value2[0] );
        }
        else if ( strlen( value1 ) > 1 && strlen( value2 ) > 1 )
        {
            crlist_add( &( ( pvalue -> t ).str.not_range ), strtol( value1, NULL, 10 ), strtol( value2, NULL, 10 ) );
        }
        i++;
    }
    find_store_tag( buffer, "ascii", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    ( pvalue -> t ).str.is_ascii = atoi( value ) ? 1:0;
}

//parse and store device password category
void store_device_password_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];

    find_store_tag( buffer, "minlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).device_passwd.minlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).device_passwd.minlength = -1;
    }
    find_store_tag( buffer, "maxlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).device_passwd.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).device_passwd.maxlength = -1;
    }
    i = 0;
    ( pvalue -> t ).device_passwd.small = 0;
    ( pvalue -> t ).device_passwd.large = 0;
    ( pvalue -> t ).device_passwd.num = 0;
    ( pvalue -> t ).device_passwd.special = 0;
    ( pvalue -> t ).device_passwd.required = 0;
    while ( !find_store_tags( buffer, "class", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        if ( 0 == strcmp( value, "[a-z]" ) )
        {
            ( pvalue -> t ).device_passwd.small = 1;
        }
        else if ( 0 == strcmp( value, "[A-Z]" ) )
        {
            ( pvalue -> t ).device_passwd.large = 1;
        }
        else if ( 0 == strcmp( value, "[0-9]" ) )
        {
            ( pvalue -> t ).device_passwd.num = 1;
        }
        else if ( 0 == strcmp( value, "special" ) )
        {
            ( pvalue -> t ).device_passwd.special = 1;
        }
        i++;
    }
    find_store_tag( buffer, "required", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    ( pvalue -> t ).device_passwd.required = atoi( value );
    if ( ( pvalue -> t ).device_passwd.required > ( pvalue -> t ).device_passwd.small + ( pvalue -> t ).device_passwd.large + ( pvalue -> t ).device_passwd.num + ( pvalue -> t ).device_passwd.special )
    {
        ( pvalue -> t ).device_passwd.required = ( pvalue -> t ).device_passwd.small + ( pvalue -> t ).device_passwd.large + ( pvalue -> t ).device_passwd.num + ( pvalue -> t ).device_passwd.special;
    }
    find_store_tag( buffer, "ascii", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    ( pvalue -> t ).device_passwd.is_ascii = atoi( value ) ? 1:0;
}
//parse and store address category
void store_address_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];
    while ( !find_store_tags( buffer, "allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).address.allowed_address ), value );
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).address.not_allowed_address ), value );
        i++;
    }
    find_store_tag( buffer, "maxlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).address.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).address.maxlength = -1;
    }
}

//parse and store ipaddress category
void store_ipaddress_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];
    while ( !find_store_tags( buffer, "allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).ipaddress.allowed_address ), value );
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).ipaddress.not_allowed_address ), value );
        i++;
    }
}

//parse and store serveraddress category
void store_serveraddress_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];
    while ( !find_store_tags( buffer, "allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).serveraddress.allowed_address ), value );
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_allowed_address", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).serveraddress.not_allowed_address ), value );
        i++;
    }
    find_store_tag( buffer, "maxlength", value, 90, 50, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).serveraddress.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).serveraddress.maxlength = -1;
    }
}

//parse and store netmask category
void store_netmask_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];
    while ( !find_store_tags( buffer, "allowed_netmask", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).netmask.allowed_mask ), value );
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_allowed_netmask", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).netmask.not_allowed_mask ), value );
        i++;
    }
}

//parse and store url category
void store_url_cat( char *buffer, pvalue_xml *pvalue )
{
    int i = 0;
    char value[MAX_VAL_LEN];
    while ( !find_store_tags( buffer, "allowed_url", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).url.allowed_url ), value );
        i++;
    }
    i = 0;
    while ( !find_store_tags( buffer, "not_allowed_url", value, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
    {
        str_list_add( &( ( pvalue -> t ).url.not_allowed_url ), value );
        i++;
    }
    find_store_tag( buffer, "maxlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).url.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).url.maxlength = -1;
    }
}

//parse and store certificate category
void store_cert_cat( char *buffer, pvalue_xml *pvalue )
{
    char value[MAX_VAL_LEN];
    find_store_tag( buffer, "minlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).cert.minlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).cert.minlength = -1;
    }
    find_store_tag( buffer, "maxlength", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    if ( !is_numeric( value ) )
    {
        ( pvalue -> t ).cert.maxlength = atoi( value );
    }
    else
    {
        ( pvalue -> t ).cert.maxlength = -1;
    }
    find_store_tag( buffer, "ascii", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    ( pvalue -> t ).cert.is_ascii = atoi( value ) ? 1:0;
    find_store_tag( buffer, "unpkcs8", value, MAX_VAL_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
    ( pvalue -> t ).cert.is_unpkcs8 = atoi( value ) ? 1:0;
}

//parse the xml limits file and store the obtained values
int parse_xml_limits_file( pvalue_xml **list_head, const char *filename )
{
    FILE *fp;
    char buffer[MAX_BUFFER_LEN], ch, start_tag[MAX_TAG_LEN], out_buffer[MAX_BUFFER_LEN];
    int i = 0;
    pvalue_xml *lh = *list_head;
    int notify_flag;
    char *endptr = NULL;
    long long int val;

    fp = fopen( filename, "r" );
    if ( !fp )
    {
        fprintf( stderr, "Failed to open file\n" );
        return -1;
    }
    while ( !feof( fp ) )
    {
        fread( &ch, sizeof( char ) , 1, fp );
        if ( ch == '<' )
        {
            memset( buffer, 0x0, MAX_BUFFER_LEN );
            get_value_file( fp, buffer, MAX_BUFFER_LEN, start_tag, MAX_TAG_LEN );
            if ( !strncmp( start_tag, "Pvalue", 6 ) || !strncmp( start_tag, "pvalue", 6 ) )
            {
                pvalue_xml pvalue;
                memset( &pvalue, 0x0, sizeof( pvalue_xml ) );
                memset(out_buffer, 0, MAX_BUFFER_LEN);
                find_store_tag( buffer, "name", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                if ( !strlen( out_buffer ) )
                {
                    continue;
                }
                pvalue.name = STRDUP( out_buffer );
#ifdef DESKTOP_DEBUG
                pvalue.hash = atoi( out_buffer );
#else
                pvalue.hash = hash( out_buffer );
#endif
                find_store_tag( buffer, "default", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                if ( strlen( out_buffer ) )
                {
                    pvalue.defval = STRDUP( out_buffer );
                }
                find_store_tag( buffer, "alias", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                if ( strlen( out_buffer ) )
                {
                    pvalue.alias = STRDUP( out_buffer );
                }
                find_store_tag( buffer, "mandatory", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                if ( !strlen( out_buffer ) )
                {
                    pvalue.mandatory = 0;
                }
                else
                {
                    pvalue.mandatory = atoi( out_buffer );
                }
                find_store_tag( buffer, "lock", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                pvalue.lock_flag = atoi( out_buffer ) ? 1:0;
                find_store_tag( buffer, "notify", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                notify_flag = atoi( out_buffer ) ? 1:0;
                find_store_tag( buffer, "group", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                pvalue.notification_groups = MSG_GROUP_ID_BROADCAST | MSG_GROUP_XML_PVALUE;
                if ( strlen( out_buffer ) )
                {
                    val = strtoll( out_buffer, &endptr, 10 );
                    if ( !( *endptr ) && out_buffer[0] && ( val >= 0 && val < MAX_MSG_GROUP_INTEGER_ID ) )
                    {
                        pvalue.notification_groups |= 1 << val;
                    }
                }
                else if ( notify_flag )
                {
                    pvalue.notification_groups |= MSG_GROUP_ID_GS_ATA;
                }
                find_store_tag( buffer, "privilege", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                pvalue.privilege_flag = atoi( out_buffer );
                find_store_tag( buffer, "reboot", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                pvalue.reboot_flag = atoi( out_buffer ) ? 1:0;
                find_store_tag( buffer, "script", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );
                if ( strlen( out_buffer ) )
                {
                    val = strtoll( out_buffer, &endptr, 10 );
                    if ( !( *endptr ) && out_buffer[0] )
                    {
                        llist_add( &( pvalue.scripts ), val );
                    }
                }
                i = 0;
                while ( !find_store_tags( buffer, "script", out_buffer, MAX_BUFFER_LEN, i, MAX_TAG_LEN, MAX_VAL_LEN ) )
                {
                    val = strtoll( out_buffer, &endptr, 10 );
                    if ( !( *endptr ) && out_buffer[0] )
                    {
                        llist_add( &( pvalue.scripts ), val );
                    }
                    i++;
                }
                find_store_tag( buffer, "category", out_buffer, MAX_BUFFER_LEN, MAX_TAG_LEN, MAX_BUFFER_LEN );

                if ( !strlen( out_buffer ) || !strncmp( out_buffer, "number", 7 ) )
                {
                    pvalue.cat = NUMBER;
                    store_number_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "utf8string", 11 ) )
                {
                    pvalue.cat = UTF8_STRING;
                    store_utf8_string_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "string", 7 ) )
                {
                    pvalue.cat = STRING;
                    store_string_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "password", 9 ) )
                {
                    pvalue.cat = PASSWORD;
                    store_string_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "device_password", 9 ) )
                {
                    pvalue.cat = DEVICEPASSWORD;
                    store_device_password_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "address", 8 ) ) 
                {
                    pvalue.cat = ADDRESS;
                    store_address_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "ipaddress", 10 ) ) 
                {
                    pvalue.cat = IPADDRESS;
                    store_ipaddress_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "serveraddress", 14 ) ) 
                {
                    pvalue.cat = SERVERADDRESS;
                    store_serveraddress_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "netmask", 8 ) )
                {
                    pvalue.cat = NETMASK;
                    store_netmask_cat( buffer, &pvalue );
                }
                else if ( !strncmp( out_buffer, "url", 4 ) )
                {
                    pvalue.cat = URL;
                    store_url_cat( buffer, &pvalue );
                }
                else if( !strncmp( out_buffer, "tr069", 6 ) )
                {
                    pvalue.cat = TR069;
                }
                else if ( !strncmp( out_buffer, "certificate", 12 ) )
                {
                    pvalue.cat = CERT;
                    store_cert_cat( buffer, &pvalue );
                }
                else
                {
                    fprintf( stderr, "Invalid category\n" );
                    continue;
                }
                pvalue_xml_tree_add( &lh, &pvalue );
#ifdef DESKTOP_DEBUG
                printf( "NAME = %llu Hash = %llu Default value = %s Mandatory = %d Category = %d\n",
                    pvalue.name, pvalue.hash, pvalue.defval, pvalue.mandatory, pvalue.cat );
#endif
            }
        }
    }
    fclose( fp );
    *list_head = lh;
#ifdef DESKTOP_DEBUG
    pvalue_xml_tree_print( &lh );
#endif    
    return 0;
}

