#define _GNU_SOURCE
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <getopt.h>

#include <asm/types.h>
#include "mtd/mtd-user.h"
#include "provision.h"

#define PROGRAM "scratch_load"
#define VERSION "$Revision: 1.1 $"

#define MAX_PAGE_SIZE	2048
#define MAX_OOB_SIZE	64

#define SQUASH_OFFSET_POS       10

/*
 * Buffer array used for writing data
 */
unsigned char writebuf[MAX_PAGE_SIZE];
unsigned char oobbuf[MAX_OOB_SIZE];
unsigned char oobreadbuf[MAX_OOB_SIZE];

// oob layouts to pass into the kernel as default
struct nand_oobinfo none_oobinfo = {
	.useecc = MTD_NANDECC_OFF,
};

struct nand_oobinfo autoplace_oobinfo = {
	.useecc = MTD_NANDECC_AUTOPLACE,
    .eccbytes = 6,
    .eccpos = {0, 1, 2, 3, 6, 7},
    .oobfree = { {8, 8} }
};

void display_help (void)
{
	printf("Usage: scratch_load [OPTION] MTD_DEVICE IMAGE_FILE\n"
	       "Writes to the specified MTD device.\n"
	       "\n"
	       "  -s addr, --start=addr set start address (default is 0)\n"
	       "  -e addr, --end=addr set end address\n"
	       "  -q, --quiet    	don't display progress messages\n"
	       "  -S bytes,--skip=bytes 	skip the specified number of bytes on input\n"
	       "  -C bytes,--count=bytes	write specified bytes only\n"
	       "      --help     	display this help and exit\n"
	       "      --version  	output version information and exit\n");
	exit(0);
}

void display_version (void)
{
	printf(PROGRAM " " VERSION "\n");
	exit(0);
}

char 	*mtd_device, *img;
int 	mtdoffset = 0;
int 	mtdend = 0;
int     skipinput = 0;
int     outcount = 0;
int 	quiet = 0;

void process_options (int argc, char *argv[])
{
	int error = 0;

	for (;;) {
		int option_index = 0;
		static const char *short_options = "pqs:e:";
		static const struct option long_options[] = {
			{"help", no_argument, 0, 0},
			{"version", no_argument, 0, 0},
			{"quiet", no_argument, 0, 'q'},
			{"start", required_argument, 0, 's'},
			{"end", required_argument, 0, 'e'},
			{"skip", required_argument, 0, 'S'},
			{"count", required_argument, 0, 'C'},
			{0, 0, 0, 0},
		};

		int c = getopt_long(argc, argv, short_options,
				    long_options, &option_index);
		if (c == EOF) {
			break;
		}

		switch (c) {
		case 0:
			switch (option_index) {
			case 0:
				display_help();
				break;
			case 1:
				display_version();
				break;
			}
			break;
		case 'q':
			quiet = 1;
			break;
		case 's':
			mtdoffset = strtol (optarg, NULL, 16);
			break;
		case 'e':
			mtdend = strtol (optarg, NULL, 16);
			break;
		case 'S':
			skipinput = strtol (optarg, NULL, 10);
			break;
		case 'C':
			outcount = strtol (optarg, NULL, 10);
			break;
		case '?':
			error = 1;
			break;
		}
	}

	if ((argc - optind) != 2 || error)
		display_help ();

	mtd_device = argv[optind++];
	img = argv[optind];
}

/*
 * Main program
 */
int main(int argc, char **argv)
{
	int cnt, fd, ifd, imglen = 0, pagelen, baderaseblock, blockstart = -1;
	struct mtd_info_user meminfo;
	struct erase_info_user mtd_erase_info;
	struct mtd_oob_buf oob;
	loff_t offs;
	int ret, readlen;
	int oobinfochanged = 0;
	struct nand_oobinfo old_oobinfo;
    int saved_mtd_offset;
    int bad_blocks_skipped = 0;

	process_options(argc, argv);

	memset(oobbuf, 0xff, sizeof(oobbuf));

	/* Open the device */
	if ((fd = open(mtd_device, O_RDWR)) == -1) {
		perror("open flash");
		exit(1);
	}

	/* Fill in MTD device capability structure */
	if (ioctl(fd, MEMGETINFO, &meminfo) != 0) {
		perror("MEMGETINFO");
		close(fd);
		exit(1);
	}
    if (mtdend == 0 || mtdend > meminfo.size) 
        mtdend = meminfo.size;

	/* Make sure device page sizes are valid */
	if (!(meminfo.oobsize == 16 && meminfo.oobblock == 512) &&
	    !(meminfo.oobsize == 8 && meminfo.oobblock == 256) &&
	    !(meminfo.oobsize == 64 && meminfo.oobblock == 2048)) {
		fprintf(stderr, "Unknown flash (not normal NAND)\n");
		close(fd);
		exit(1);
	}

	/* Read the current oob info */
	if (ioctl (fd, MEMGETOOBSEL, &old_oobinfo) != 0) {
		perror ("MEMGETOOBSEL");
		close (fd);
		exit (1);
	}

	// autoplace ECC ?
	if ((old_oobinfo.useecc != MTD_NANDECC_AUTOPLACE)) {

		if (ioctl (fd, MEMSETOOBSEL, &autoplace_oobinfo) != 0) {
			perror ("MEMSETOOBSEL");
			close (fd);
			exit (1);
		}
		oobinfochanged = 1;
	}

	oob.length = meminfo.oobsize;
	oob.ptr = oobbuf;

	/* Open the input file */
	if ((ifd = open(img, O_RDONLY)) == -1) {
		perror("open input file");
		goto restoreoob;
	}

	// get image length
    if (outcount > 0)
        imglen = outcount;
    else
   	    imglen = lseek(ifd, 0, SEEK_END);
    if (skipinput > 0 && skipinput < imglen)
    {
        imglen -= skipinput;
        lseek (ifd, skipinput, SEEK_SET);
    }
    else
        lseek (ifd, 0, SEEK_SET);

	pagelen = meminfo.oobblock;

	// Check, if length fits into device
	if ( ((imglen / pagelen) * meminfo.oobblock) > (meminfo.size - mtdoffset)) {
		fprintf (stderr, "Image %d bytes, NAND page %d bytes, OOB area %u bytes, device size %u bytes\n",
				imglen, pagelen, meminfo.oobblock, meminfo.size);
		perror ("Input file does not fit into device");
		goto closeall;
	}

    /* reserve the first two blocks for image header */
    saved_mtd_offset = mtdoffset;
#if 0
    mtdoffset += 2 * meminfo.erasesize;
#endif
	/* Get data from input and write to the device */
	//while (imglen && (mtdoffset < meminfo.size)) 
    if (!quiet)
        fprintf(stdout, "start offset=0x%x\n", mtdoffset);
	while (imglen && (mtdoffset < mtdend)) 
    {
		// new eraseblock , check for bad block(s)
		// Stay in the loop to be sure if the mtdoffset changes because
		// of a bad block, that the next block that will be written to
		// is also checked. Thus avoiding errors if the block(s) after the
		// skipped block(s) is also bad (number of blocks depending on
		// the blockalign
		while (blockstart != (mtdoffset & (~meminfo.erasesize + 1))) {
			blockstart = mtdoffset & (~meminfo.erasesize + 1);
			offs = blockstart;
	        baderaseblock = 0;
			if (!quiet)
				fprintf (stdout, "\rWrite block %x", blockstart);

		    /* Check all the blocks in an erase block for bad blocks */
			do {
			   	if ((ret = ioctl(fd, MEMGETBADBLOCK, &offs)) < 0) {
					perror("ioctl(MEMGETBADBLOCK)");
					goto closeall;
				}
				if (ret == 1) {
					baderaseblock = 1;
				   	if (!quiet)
						fprintf (stderr, "Bad block at %x block(s) from %x will be skipped\n", (int) offs, blockstart);
				}

				if (baderaseblock) {
					mtdoffset = blockstart + meminfo.erasesize;
                    bad_blocks_skipped ++;
				}
#if 1
                else
                {
                    /* SXIE: erase the block */
	                mtd_erase_info.length = meminfo.erasesize;
	                mtd_erase_info.start = blockstart;
		            if(ioctl(fd, MEMERASE, &mtd_erase_info)) {
					    perror("ioctl(MEMERASE)");
					    goto closeall;
                    }
                }
#endif
			    offs +=  meminfo.erasesize;
			} while ( offs < blockstart + meminfo.erasesize );

		}

		readlen = meminfo.oobblock;
		if ((imglen < readlen))
		{
			readlen = imglen;
			memset(writebuf + readlen, 0xff, meminfo.oobblock - readlen);
		}

		/* Read Page Data from input file */
		if ((cnt = read(ifd, writebuf, readlen)) != readlen) {
			if (cnt == 0)	// EOF
				break;
			perror ("File I/O error on input file");
			goto closeall;
		}

        oob.start = mtdoffset;
        oobbuf[SQUASH_OFFSET_POS - 1] =  bad_blocks_skipped & 0xff;
        if (ioctl(fd, MEMWRITEOOB, &oob) != 0) {
            perror ("ioctl(MEMWRITEOOB)");
            goto closeall;
        }

		/* Write out the Page data */
		if (pwrite(fd, writebuf, meminfo.oobblock, mtdoffset) != meminfo.oobblock) {
			perror ("pwrite");
			goto closeall;
		}
		imglen -= readlen;
		mtdoffset += meminfo.oobblock;
	}
    if (!quiet)
        fprintf(stdout, "\n");

 closeall:
	close(ifd);

 restoreoob:
	if (oobinfochanged) {
		if (ioctl (fd, MEMSETOOBSEL, &old_oobinfo) != 0) {
			perror ("MEMSETOOBSEL");
			close (fd);
			exit (1);
		}
	}

	close(fd);

	if (imglen > 0) {
		printf ("failed to write into mtd device, imglen=%d\n", imglen);
		exit (1);
	}

	/* Return happy */
	return 0;
}
