/* With code from OpenSSL (BSD License) */
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <openssl/err.h>
#include <openssl/asn1.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/ssl.h>

#define FORMAT_PEM 3
BIO *bio_err = NULL;
char * pass = NULL;

int password_callback( char * buf, int bufsiz, int verify )
{
  int res;
  
  if ( !pass ) return 0;
  
  res = strlen( pass );
  if ( res > bufsiz )
    res = bufsiz;
  memcpy( buf, pass, res );
  return res;
}

X509 *load_cert(BIO *err, const char *file, int format,
    const char *pass, ENGINE *e, const char *cert_descrip)
    {
    BUF_MEM *buf=NULL;
    X509 *x=NULL;
    BIO *cert;

    if ((cert=BIO_new(BIO_s_file())) == NULL)
        {
        ERR_print_errors(err);
        goto end;
        }

    if (file == NULL)
        {
        setvbuf(stdin, NULL, _IONBF, 0);
        BIO_set_fp(cert,stdin,BIO_NOCLOSE);
        }
    else
        {
        if (BIO_read_filename(cert,file) <= 0)
            {
            BIO_printf(err, "Error opening %s %s\n",
                cert_descrip, file);
            ERR_print_errors(err);
            goto end;
            }
        }

    if  (format == FORMAT_PEM)
	{
         x=PEM_read_bio_X509_AUX(cert,NULL,
            (pem_password_cb *)password_callback, NULL);
	}
    else    {
        BIO_printf(err,"bad input format specified for %s\n",
            cert_descrip);
        goto end;
        }
end:
    if (x == NULL)
        {
        BIO_printf(err,"unable to load certificate\n");
        ERR_print_errors(err);
        }
    if (cert != NULL) BIO_free(cert);
    if (buf != NULL) BUF_MEM_free(buf);
    return(x);
}

int main_cert_verify( int argc, char ** argv )
{
    SSL_CTX *ctx;
    char * infile;

    if ( argc != 2 ) {
        printf("SSL Error: unsupported argc: %d\n", argc);
        return -1;
    }
    infile = argv[1];

    SSL_load_error_strings();
    SSL_library_init();
    OpenSSL_add_all_algorithms();

    ctx = SSL_CTX_new(SSLv23_server_method());
    if ( ctx == NULL ) {
        printf("SSL Error: %s\n", ERR_error_string(ERR_get_error(), NULL));
        return -1;
    }

    if ( SSL_CTX_use_certificate_file(ctx, infile, SSL_FILETYPE_PEM) < 0 )
    {
        printf("SSL Error: %s\n", ERR_error_string(ERR_get_error(), NULL));
        return -1;
    }

    if (SSL_CTX_use_PrivateKey_file(ctx, infile, SSL_FILETYPE_PEM) < 0 )
    {
        printf("SSL Error: %s\n", ERR_error_string(ERR_get_error(), NULL));
        return -1;
    }

    if ( SSL_CTX_check_private_key(ctx) != 1 )
    {
        printf("SSL Error: %s\n", ERR_error_string(ERR_get_error(), NULL));
        return -1;
    }

    return 0;
}

int main( int argc, char ** argv )
{
    ENGINE *e = NULL;
    X509 *x;
    BIO *STDout;
    int informat;
    int ret = -1;
    char * infile;
    char *name = argv[0];
    char* p = NULL;

    if (strchr (name, '/') != NULL)
    {
        do
        {
            /* move behind next / */
            name = strchr (name, '/') + 1;
        }
        while (name != NULL && strlen(name) > 0 && strchr (name, '/') != NULL);
    }
    if (name == NULL) return -1;

    if ( !strcmp("cert_verify", name) )
    {
        return main_cert_verify(argc, argv);
    }

    if ( argc != 3 )
      return -1;
    
    infile = argv[1];
    pass = argv[2];
    
    signal(SIGPIPE, SIG_IGN);
    CRYPTO_malloc_init();
    ERR_load_crypto_strings();
    OpenSSL_add_all_algorithms();
    
    STDout = BIO_new_fp(stdout, BIO_NOCLOSE);
    informat = FORMAT_PEM;

    x=load_cert( bio_err, infile, informat, NULL, e, "Certificate" );
    if ( x == NULL )
    {
        goto err_out;
    }

#if 0
    BIO_puts(STDout, "notBefore=");
    ASN1_TIME_print(STDout, X509_get_notBefore(x));
    BIO_puts(STDout, "\n");
#endif
    {
        ASN1_TIME *tm = X509_get_notBefore(x);

        if ( tm->type == V_ASN1_UTCTIME )
        {
            char *v;
            int gmt=0;
            int i;
            int y=0,M=0,d=0,h=0,m=0,s=0;

            i=tm->length;
            v=(char *)tm->data;

            if (i < 10) goto err_out;
            if (v[i-1] == 'Z') gmt=1;
            for (i=0; i<10; i++)
                if ((v[i] > '9') || (v[i] < '0')) goto err_out;
            y= (v[0]-'0')*10+(v[1]-'0');
            if (y < 50) y+=100;
            M= (v[2]-'0')*10+(v[3]-'0');
            if ((M > 12) || (M < 1)) goto err_out;
            d= (v[4]-'0')*10+(v[5]-'0');
            h= (v[6]-'0')*10+(v[7]-'0');
            m=  (v[8]-'0')*10+(v[9]-'0');
            if (tm->length >=12 &&
                (v[10] >= '0') && (v[10] <= '9') &&
                (v[11] >= '0') && (v[11] <= '9'))
                s=  (v[10]-'0')*10+(v[11]-'0');

            // YYYYMMDDHHMM.SS
            printf("%d%02d%02d2359.59\n", y+1900, M, d);  
        }
        else if ( tm->type == V_ASN1_GENERALIZEDTIME )
        {
            char *v;
            int gmt=0;
            int i;
            int y=0,M=0,d=0,h=0,m=0,s=0;

            i=tm->length;
            v=(char *)tm->data;

            if (i < 12) goto err_out;
            if (v[i-1] == 'Z') gmt=1;
            for (i=0; i<12; i++)
                if ((v[i] > '9') || (v[i] < '0')) goto err_out;
            y= (v[0]-'0')*1000+(v[1]-'0')*100 + (v[2]-'0')*10+(v[3]-'0');
            M= (v[4]-'0')*10+(v[5]-'0');
            if ((M > 12) || (M < 1)) goto err_out;
            d= (v[6]-'0')*10+(v[7]-'0');
            h= (v[8]-'0')*10+(v[9]-'0');
            m=  (v[10]-'0')*10+(v[11]-'0');
            if (tm->length >= 14 &&
                (v[12] >= '0') && (v[12] <= '9') &&
                (v[13] >= '0') && (v[13] <= '9'))
                s=  (v[12]-'0')*10+(v[13]-'0');
          
            // YYYYMMDDHHMM.SS
            printf("%d%02d%02d2359.59\n", y, M, d);           
        }
    }

    X509_free(x);

    return 0;

err_out:
    return ret;
}
